#!/usr/bin/env python

import sys
sys.path.append('/usr/share/inkscape/extensions')

import math
import inkex
from simplestyle import *
import random

inxGen = 0

"""
This extension creates visual noise of various kinds, using
3 ingredients, scratches, chips, and specks.

Extensions -> Generate -> Noise -> ChipScratches

The 3 ingredients go into 3 Inkscape groups. Where they are in
the tree is controlled by the object selected at the time the
extension is called.  They become sisters of the selected object,
under the same parent.  If there is no selection, they go to
the root level.

In the text below, each * indicates a settable parameter in
the pop-up box.

The extension fills a rectangle, which defaults* to the page
size, or is width* by height*.  The overall size* and number*
of marks is set.  However, for each of the 3, a multiplier
changes the size* and number*.  To get the overall value,
just set the multiplier to 1.0

Also, each of the 3 may be separately enabled* or disabled.

For each ingredient, it is possible to set the size to grow*
with y.  For this case, it is possible to set the size to
a negative multiplier, so that it, when growing, will pass
through zero somewhere in the middle of the box.

For scratches, it is possible to change their orientation 
from vertical to angular*, radiating from the center of the box.
(If so, growth is measured from the center.)  The initial
angle* and the change with distance* are set.

For chips and specks, there are similar controls, except the
angular ones, which are omitted.

Last, there is the option to use a density gradient*. In the
simple case, the probability of making a mark is the same
everywhere.  With a gradient, it changes from 0 to 1 across
the box from left to right.

--

Some of the features above may be of little use, while some
omitted ones might be important.  This is a beta version, so
I await user feedback.

I notice that the extension might take 5 seconds to run.  Is
this important?  A bit of clean-up might help this.

Two features that might slow it down considerably are clusters
and collisions.  Collisions, if implemented, would keep any 2
marks from landing on top of each other.  Slow, because each
mark must search all the previous marks.

Clusters is some method of putting new marks near old ones.
I think this would be especially useful for specks, which
could be clustered around the end of scratches.  Right now,
they're uniform, and so fade into the background.

Again, feedback?

Technical note:  This file may also be used to generate the inx
file. (The bundled inx file was generated this way.)  Set the
global /inxGen/ =1 and run the file stand-alone.

- AK Sept 2018
"""

# Define passed parameters
parms = """
#tabs#Overall
n Nmain   "page" , Current index tab
b pgsizep True , Default rectangle to page size?
i rx      1000 max="10000" Width
i ry      1000 max="10000" Height
f mainSize 1.0 max="100.0" Size of objects
i mainNum 200 max="5000" Number of objects
#tab#Scratches
b honp    True , Enable scratches
f hsize   2.0 max="100.0",min="-100.0" Size of scratches
f hgrow   0.0 max="100.0",min="-100.0" Grow scratches with distance
f hnum    0.2 max="100.0" Number of scratches
b hrad    False , Angle scratches toward center
f hang    90. max="180.0",min="-180.0" Angle from radius
f hcurve  0.0 max="100.0",min="-100.0" Change angle with distance
b hgrad   False , Use density gradient
#tab#Chips
b conp    True , Enable chips
f csize   1.0 max="100.0",min="-100.0" Size of chips
f cgrow   0.0 max="100.0",min="-100.0" Grow chips with distance
f cnum    1.0 max="100.0" Number of chips
b cgrad   False , Use density gradient
#tab#Specks
b sonp    True , Enable specks
f ssize   1.0 max="100.0",min="-100.0" Size of specks
f sgrow   0.0 max="100.0",min="-100.0" Grow specks with distance
f snum    10.0 max="100.0" Number of specks
b sgrad   False , Use density gradient
#tabe#

        """
tdic = {
    "i" : "int",
    "f" : "float",
    "s" : "str",
    "n" : "str", #For the index tab name
    "b" : "inkbool",
}

tdic2 = {
    "i" : "int",
    "f" : "float",
    "s" : "str",
    "n" : "str", #For the index tab name
    "b" : "bool",
}


class noiseEffect(inkex.Effect):
    """
Creates visual noise.  3 kinds: Scratches, chips, specks.
50, 100, 100 instances are predefined.  These are chosen
randomly and scattered randomly across the page. Controls
for number and size, as well as some specific to each type.
    """
    def __init__(self):
        """
        Constructor.
        Defines the parms option of a script.
        """
        # Call the base class constructor.
        inkex.Effect.__init__(self)

        def callOP( p, self ) :
            call = ("self.OptionParser.add_option('', '--%s', action = 'store',"
                    + "type = '%s', dest = '%s', default = %s, help = '%s')")
            w = p.split(None, 4)
            eval(call % (w[1], tdic[w[0]], w[1], w[2], w[4] ))

        for parm in parms.split('\n') :
            if mt(parm) :
                continue
            if parm[0] == '#' :
                continue
            callOP( parm, self )



    def effect(self):
        """
        Effect behaviour.
        Overrides base class' method.
        """

        for parm in parms.split('\n') :
            if mt(parm) :
                continue
            if parm[0] == '#' :
                continue
            w = parm.split(None, 4)
            exec('%s = %s( self.options.%s)' % (w[1], tdic2[w[0]], w[1]) )

        # Get access to main SVG document element and get its dimensions.
        svg = self.document.getroot()

        if pgsizep :
            try :
                width  = inkex.unittouu(svg.get('width'))
                height = inkex.unittouu(svg.attrib['height'])
            except AttributeError :
                width  = self.unittouu(svg.get('width'))
                height = self.unittouu(svg.attrib['height'])
            rx = width
            ry = height
            #inkex.errormsg("Page size %d %d" % (width, height))

        svg = self.document.getroot()
        
        seld = self.selected
        if len(seld) :
            id = seld.keys()[0]
            sel = seld[id]
            parent = self.getParentNode(sel)
        else :
            parent = svg

#Create scratches


        pdic = {
            'rx' :   rx,
            'ry' :   ry,
            'size' : mainSize * hsize,
            'enable' : honp,
            'num' :  int( mainNum * hnum ),
            'grow' : hgrow,
            'radial' : hrad,
            'ang' : hang,
            'curve' : hcurve,
            'grad' : hgrad,
            }
        draw( scratches, parent, pdic)

#Create chips


        pdic = {
            'rx' :   rx,
            'ry' :   ry,
            'size' : mainSize * csize,
            'enable' : conp,
            'num' :  int( mainNum * cnum ),
            'grow' : cgrow,
            'radial' : False,
            'ang' : False,
            'curve' : False,
            'grad' : cgrad,
            }
        draw( chips, parent, pdic)

#Create specks


        pdic = {
            'rx' :   rx,
            'ry' :   ry,
            'size' : mainSize * ssize,
            'enable' : sonp,
            'num' :  int( mainNum * snum ),
            'grow' : sgrow,
            'radial' : False,
            'ang' : False,
            'curve' : False,
            'grad' : sgrad,
            }
        draw( specks, parent, pdic)

def draw( obj, parent, pdic) :
    if not pdic['enable'] :
        return

    group = inkex.etree.SubElement(parent, 'g')

    if pdic['grad'] :
        f = lambda x: math.sqrt(x)
    else :
        f = lambda x: x

    rx = pdic['rx']
    ry = pdic['ry']
    fmstyle = formatStyle({"fill" : "#000000"})
    mrx = rx // 2
    mry = ry // 2
    mrmax = mrx if mrx > mry else mry
    #Curves 90 going from the center to the edge (long dim.) if parm = 1.0
    curve = pdic['curve'] * 0.5 * math.pi / mrmax

    ctrs = [(rx * f(random.random()) , ry * random.random())
            for i in range(pdic['num'])]
    zp = 2 + 3 * int(random.random()*len(obj) / 3 )
    for ctr in ctrs :
        path = inkex.etree.Element(inkex.addNS('path','svg'))
        path.set('style', fmstyle)
        zp = (zp + 3) % len(obj)
        path.set('d', obj[zp])
        trstring = "translate (%g %g)" % ctr
        if pdic['radial'] :
            ang = math.atan2(ctr[1]-mry , ctr[0]-mrx) + curve * math.sqrt(
                (ctr[1]-mry)**2 + (ctr[0]-mrx)**2)
            trstring += " rotate (%g)" % (ang * 180 / math.pi + pdic['ang'])
        if pdic['size'] != 1.0 or pdic['grow'] != 0.0 :
            if pdic['radial'] :
                size = math.sqrt((ctr[1]-mry)**2 + (ctr[0]-mrx)**2) * (
                    pdic['grow'] / mrmax ) + pdic['size']
            else :
                size = ctr[1] * pdic['grow'] / mrmax + pdic['size']
            trstring += " scale (%g)" % size
        path.set('transform', trstring)
        group.append(path)


chips=(
4.17065, 3.83403,
"M -2.98542 -3.8338  c 0.34234,0 0.74533,0.14835 0.89583,0.32966 0.15043,0.18131 0.19014,0.32965 0.088,0.32965 -0.10212,0 0.0341,0.15422 0.30304,0.34254 0.26887,0.18832 0.56984,0.29206 0.66874,0.23093 0.0989,-0.0612 0.44473,-0.1589 0.76833,-0.21763 0.48389,-0.0877 0.6122,-0.0113 0.72327,0.43138 0.0745,0.29608 0.0843,0.73458 0.0215,0.97438 -0.0648,0.24806 -0.005,0.43833 0.13992,0.44082 0.13978,0.003 0.49312,0.18902 0.7855,0.41379 0.40627,0.31226 0.68285,0.3654 1.1727,0.22492 0.66407,-0.19047 1.37707,0.0827 1.57744,0.60479 0.14702,0.38319 -1.10951,0.94947 -2.1067,0.94947 -0.77639,0 -0.84516,0.0496 -0.84516,0.60652 0,0.36844 -0.17253,0.71404 -0.43955,0.88079 -0.24174,0.15096 -0.43953,0.47139 -0.43953,0.71211 0,0.56109 -0.49051,0.55224 -1.73627,-0.0314 -0.53191,-0.24917 -0.87804,-0.51738 -0.7692,-0.59622 0.10884,-0.0788 0.21711,-0.44315 0.24037,-0.80954 0.0361,-0.56828 -0.10116,-0.75694 -0.93402,-1.28385 -0.53699,-0.33972 -0.97558,-0.78166 -0.97308,-0.98208 0.003,-0.31564 0.0318,-0.32027 0.21635,-0.0349 0.14197,0.21974 0.21384,0.23816 0.21631,0.0549 0.003,-0.15109 -0.19671,-0.47488 -0.44125,-0.7194 -0.31961,-0.31961 -0.38937,-0.55126 -0.24852,-0.82414 0.10773,-0.20875 0.26287,-0.75031 0.34467,-1.20358 0.13057,-0.72333 0.22505,-0.82414 0.77135,-0.82414 z",
4.13855, 4.24985,
"M -1.85248 -4.24978  c 0.27589,0.003 0.47037,0.0646 0.48312,0.20271 0.0137,0.1478 0.0399,0.47805 0.0582,0.73402 0.026,0.36225 0.21413,0.47395 0.84811,0.50359 0.44794,0.0209 0.94806,0.0473 1.11126,0.0582 0.16314,0.011 0.39599,0.28179 0.51767,0.601744 0.12163,0.31997 0.37346,0.58176 0.55949,0.58176 0.18595,0 0.48549,0.14734 0.66583,0.3277 0.18035,0.18035 0.57257,0.36361 0.87127,0.40723 0.35897,0.0524 0.62021,0.31158 0.77039,0.76447 0.20387,0.61474 0.15184,0.74842 -0.50541,1.29714 -0.40292,0.33638 -0.73264,0.79074 -0.73264,1.0099 0,0.21915 -0.16606,0.65189 -0.36907,0.96172 -0.20296,0.30985 -0.30165,0.67266 -0.21907,0.80628 0.0894,0.1446 -0.091,0.24316 -0.44539,0.24316 -0.32756,0 -0.5954,-0.0917 -0.5954,-0.20362 0,-0.11191 0.15163,-0.63243 0.33677,-1.15715 0.18525,-0.52472 0.2637,-0.99731 0.17453,-1.04989 -0.0892,-0.0526 -0.63351,-0.17283 -1.20942,-0.26725 -0.57596,-0.0944 -1.10945,-0.26987 -1.18578,-0.38997 -0.0763,-0.12012 -0.53312,-0.25519 -1.01492,-0.29996 -1.10287,-0.10248 -1.29329,-0.29762 -1.33349,-1.3685 -0.0191,-0.50625 -0.13647,-0.82688 -0.28496,-0.77674 -0.76171,0.25701 -1.09657,-1.581564 -0.42769,-2.348844 0.31962,-0.36665 1.3196,-0.64382 1.92663,-0.63767 z",
4.22195, 4.21947,
"M 1.69537 -4.21932  c -0.0345,-8.8e-4 -0.0688,0.002 -0.1027,0.008 -0.18069,0.0328 -0.35094,0.1712 -0.50284,0.46983 -0.0104,1.09687 -0.6782,2.01812 -0.47774,3.15318 0.24531,0.74578 -0.54886,0.45228 -0.99871,0.50283 -1.3297,0.0104 -0.14408,-1.07574 -1.55709,-0.976 -0.42439,-0.67608 -0.98245,-2.34963 -1.81896,-0.91876 -0.17688,0.51739 -0.9194,1.1638 -0.0326,1.53546 0.37737,0.014 0.60982,0.25307 0.77516,0.58044 -0.0269,0.19027 -0.0344,0.50819 -0.0112,0.88577 0.0503,0.81728 -0.0383,1.14531 -0.37549,1.39186 -0.48057,0.35139 -0.36025,0.73074 0.3532,1.11256 0.18114,0.0969 0.28904,0.14046 0.33832,0.12594 0.0493,-0.0146 0.0402,-0.0871 -0.0116,-0.22214 -0.0798,-0.20803 0.0612,-0.40175 0.36204,-0.49725 0.33273,-0.1056 0.47599,-0.33503 0.4903,-0.78121 0.13921,-0.0295 0.30389,-0.11064 0.50376,-0.26119 1.46647,-0.57469 1.19941,1.53559 2.36963,1.92306 0.98055,1.45246 1.30915,-1.39831 1.46716,-2.1424 0.10424,-1.37891 -0.4829,-1.40293 0.43107,-1.92205 0.91202,0.23598 1.36087,-0.55942 1.32261,-1.38814 -0.0299,-0.64451 -0.35397,-1.30896 -0.98476,-1.52384 -0.44029,-0.27942 -1.02203,-1.0427 -1.53965,-1.05586 z",
2.29614, 3.86905,
"M -0.15038 -3.86873  c -0.32838,-0.009 -0.63053,0.17109 -0.7121,0.59531 -0.44133,-0.069 -1.15128,-1.08751 -1.3529,-0.1297 0.0799,0.51796 -0.18084,1.09953 -0.0362,1.560632 0.0859,0.54933 0.0879,1.28864 0.86402,1.13739 0.41931,0.12913 0.19391,1.48787 0.86352,1.32705 -0.27268,0.12418 -0.56414,0.20057 -0.90951,0.15607 -0.33358,0.0733 -0.0995,0.50263 0.10232,0.59686 0.69781,0.49054 1.25008,-1.24221 1.59784,0.0651 -0.10673,0.74125 -1.42009,1.11534 -1.22938,1.91513 0.45471,0.77949 1.52934,0.52133 2.19882,0.20722 0.426,-0.25368 0.96231,-0.36258 0.60153,-1.01906 -0.16613,-0.63544 -0.25215,-1.31953 -0.28578,-1.93476 0.19172,-0.28339 0.73189,-0.81591 0.74414,-1.18391 0.004,-0.12267 -0.0506,-0.22674 -0.19637,-0.29765 -0.12346,0.2062 -0.2481,0.38506 -0.39325,0.52399 -0.0738,-0.30337 -0.26755,-0.59016 -0.57981,-0.70486 -0.97504,-0.90992 0.71916,-1.624862 1.05005,-2.274802 v -5.1e-4 c -0.35914,-0.76136 -1.12554,0.174 -1.4578,-0.003 -0.18611,-0.32959 -0.54083,-0.52743 -0.86918,-0.5364 z",
3.31745, 5.20184,
"M -3.04057 -3.12517  c -0.12013,0.49557 0.22954,0.992902 -0.0125,1.481942 -0.3042,0.4355 -0.16661,0.96485 -0.24617,1.47138 -0.13051,0.54661 0.48075,0.72787 0.5505,1.18424 0.34005,0.28842 -0.34128,0.56491 -0.0684,0.64532 -0.0254,0.006 -0.0507,0.009 -0.076,0.0172 -0.12397,0.005 -0.0329,0.12902 0.0254,0.13283 0.0916,0.0214 0.17487,0.0566 0.25509,0.0975 -0.0618,0.0963 -0.13661,0.20956 -0.0942,0.3182 -0.0579,0.17996 -0.64502,0.46539 -0.41368,0.58716 0.27677,0.0877 0.30423,0.26778 0.27821,0.46656 -0.10304,0.0342 -0.16783,0.13651 -0.0334,0.1966 -0.0441,0.20892 -0.0836,0.41478 0.0836,0.54124 0.26204,0.0239 0.56304,-0.0127 0.85778,-0.0113 0.15595,0.053 0.3112,0.10898 0.45235,0.19132 0.44943,0.28858 0.77251,0.76143 1.30321,0.92948 0.251,0.13085 0.47062,0.0872 0.47613,-0.0674 0.003,-0.0703 -0.0394,-0.1635 -0.14241,-0.2736 -0.32012,-0.35787 -0.80262,-0.50472 -1.20241,-0.70974 0.26829,0.001 0.53112,0.2429 0.72923,-0.0975 -0.0774,-0.49202 -0.79569,-0.59976 -1.12805,-0.94766 -0.38855,-0.23155 -1.03547,-0.33263 -0.82509,-0.89511 0.19784,0.16083 0.38233,0.34262 0.59643,0.46985 0.2225,0.39004 0.96906,0.0149 1.31442,0.31886 0.30953,-0.35321 0.59538,0.79256 0.80227,0.11169 0.0337,-0.0757 0.0554,-0.16401 0.0621,-0.24848 v 0 c 0.0205,-0.25339 -0.0915,-0.47343 -0.40312,-0.22337 -0.19188,0.0244 -0.12003,-0.54699 -0.50655,-0.40708 -0.4008,-0.0492 -0.88145,0.32259 -1.01738,-0.24187 -0.22216,-0.0951 -0.4652,-0.19534 -0.71735,-0.25079 0.0546,-0.37763 0.0308,-0.78884 0.12161,-1.10494 -0.53717,-0.3656 -0.0495,-0.98532 -0.21974,-1.39571 -0.0611,-0.41652 0.91697,0.0243 0.77055,0.42327 0.0859,0.35784 0.84999,1.14566 0.90668,0.36743 0.24362,-0.49009 -0.0688,-1.08376 -0.55477,-1.27742 -0.23421,-0.13565 -0.50438,-0.44811 -0.0324,-0.37734 0.54993,-0.25032 0.8983,0.51127 1.39901,0.21378 0.31014,0.12139 0.34301,0.5908 0.78477,0.50158 0.32429,0.42527 0.37542,0.96608 0.52008,1.43504 -0.0457,0.40839 0.12259,0.77239 0.32746,1.1089 -0.0386,0.0876 -0.0483,0.19899 -0.0139,0.33208 -0.033,0.23478 0.0471,0.26766 0.266,0.26566 0.0265,0.0178 0.0539,0.0311 0.0816,0.0423 0.0757,0.35193 0.0297,0.752 0.33572,0.99755 0.38603,0.3712 0.90219,0.0857 0.7319,-0.43054 -0.0466,-0.30798 -0.29681,-0.49153 -0.49234,-0.70745 0.0877,-0.0846 0.14598,-0.18839 0.14571,-0.28813 0,-0.12995 -0.0965,-0.25111 -0.34364,-0.30729 -0.0113,-0.01 -0.0233,-0.0175 -0.035,-0.0261 0,-0.008 0,-0.0144 0.002,-0.0221 -0.0783,-0.58378 -0.29364,-1.04867 -0.57824,-1.54836 -0.003,-0.54764 -0.26269,-1.00095 -0.42559,-1.5107 -0.16414,-0.40027 -0.18097,-0.931632 -0.68992,-1.070242 -0.35382,-0.0791 -0.24553,0.52958 -0.32482,0.740812 -0.19094,0.0194 -0.35815,-0.327932 -0.61756,-0.303992 -0.0498,-0.45204 -0.10633,-0.96193 -0.59179,-1.08313 -0.21979,-0.35661 -0.7287,-0.40785 -0.73915,-0.88256 -1.36004,-2.36534 -1.95836,0.29957 -1.63423,1.12013 z",
2.98467, 4.66325,
"M 0.614136 -2.29735  c -0.0974,-0.40618 -0.7125,-0.83436 -0.90598,-1.29221 -0.16432,-0.38885 0.094,-0.83201 -0.0209,-0.95324 -0.29875,-0.3152 -0.35677,0.40541 -0.62004,0.28161 -0.5706,-0.2683 -0.96784,-0.4991 -1.56386,-0.36057 -0.0124,0.2114 0.004,0.42436 -0.0171,0.63525 -0.31209,0.73101 0.0309,1.55931 -0.31858,2.0946 0.0848,0.12448 0.14997,0.25099 0.20148,0.37838 0.0156,0.21967 0.0503,0.44298 0.0725,0.67113 0.0238,0.0613 0.0395,0.12615 0.0585,0.18949 0.50136,1.84607 -0.0172,1.42527 -0.43442,1.35738 -0.24015,0.81789 0.45195,1.85386 0.52381,2.81106 0.18281,2.0407 0.89697,0.86601 2.0265,0.22853 10e-4,-0.91029 -0.0244,-0.95992 0.52581,-1.46817 0.74242,-0.0378 1.71531,-0.0436 1.76344,-0.97207 0.0521,-0.0462 0.0914,-0.0908 0.12716,-0.13473 0.0493,0.0552 0.15625,0.0926 0.36013,0.10319 0.40274,0.0431 0.56401,-0.18544 0.58867,-0.47407 0.0542,-0.63499 -0.55312,-1.55926 -0.70765,-0.51248 0.0324,0.14484 -0.0426,0.29511 -0.1234,0.43883 -0.16329,-0.54242 -1.37396,-0.96101 -1.47446,-1.59914 -0.1211,-0.0939 -0.22103,-0.19865 -0.29781,-0.30976 0.11513,-0.10721 0.23173,-0.21303 0.34755,-0.31918 0.004,-0.24309 -0.0645,-0.52303 -0.11148,-0.7939 z",
2.04389, 5.29494,
"M -1.85309 -5.04915  c 0.23783,0.26043 0.21099,0.66518 0.22712,1.04875 -0.4069,0.38004 0.1825,0.8748 0.0781,1.30474 -0.0389,1.015465 -0.55187,0.49378 -0.49101,1.508525 0.13462,0.44679 0.16841,0.72546 0.81526,1.43512 0.29219,0.0781 -0.22258,1.36221 -0.28588,2.05254 -0.0792,0.86353 0.71412,1.04806 0.51067,1.67394 l 1.06331,1.32048 c -0.0314,-0.1812 -0.0448,-0.35515 0.0627,-0.47402 0.28009,0.15408 0.64109,0.57435 0.77582,0.0169 0.14718,-0.48429 0.42232,-0.0327 0.35542,0.2494 0.26229,0.10277 0.4328,-0.33043 0.71684,-0.13395 0.23715,-0.46565 -0.19546,-0.38058 -0.45959,-0.49191 -0.009,-0.3719 0.26324,-0.82167 -0.25285,-0.92357 -0.0662,-0.19471 0.006,-0.46625 -0.0348,-0.68452 -0.17436,-0.24732 -0.63358,-0.4809 -0.23403,-0.8166 0.25025,-0.42226 -0.85867,-0.85207 -0.48249,-1.2856 0.22844,0.18931 0.44217,0.179 0.35732,-0.17819 -0.13207,-0.6247 0.35102,-0.79202 -0.0534,-1.09863 -0.38584,-0.3369 -0.93599,-0.41779 -1.28967,-0.79966 -0.019,-0.23696 -0.17121,-0.4219 -0.2883,-0.61236 0.0439,-0.11202 0.10964,-0.220495 0.20235,-0.323135 0.29428,-0.26016 0.28836,-0.24769 0.10101,-0.59292 0.21808,-0.22172 -0.30092,-0.49114 -0.22305,-0.80405 -0.27966,-0.40328 -0.31858,-0.96801 -0.26602,-1.46128 0.16807,-0.43289 -0.90489,0.07 -0.90489,0.07 z",
5.1605, 3.51819,
"M 0.197111 -3.51809  c -0.25831,-0.008 -0.64247,0.48592 -1.65703,1.8944 -0.30917,0.4292 -1.11412,0.75117 -2.18011,0.87133 -1.27846,0.1441 -1.64072,0.31871 -1.48755,0.71783 0.11143,0.29022 0.20229,0.59651 0.20229,0.68056 0,0.084 0.36946,0.14096 0.82122,0.12688 0.7781,-0.0243 0.78325,-0.0409 0.0914,-0.31499 -0.4578,-0.18138 0.0596,-0.25533 1.38729,-0.19875 1.16464,0.0496 2.21889,0.19137 2.34248,0.31499 0.12366,0.12358 -0.0554,0.22449 -0.39706,0.22449 -0.50809,0 -0.54984,0.1332 -0.23026,0.73025 0.17922,0.33488 0.49805,0.6159 0.76974,0.70096 0.0157,0.0639 0.0352,0.12709 0.063,0.18811 0,0.0641 0.1163,0.0991 0.11048,0.1646 0.28267,0.39771 0.73888,1.40678 1.24267,0.67834 0.003,-0.26195 -0.48278,-1.34025 0.0102,-1.11978 0.13677,0.65215 0.88947,0.55265 1.32385,0.40816 0.52365,0.38992 1.46911,0.38685 1.70496,-0.35447 0.0158,-0.0853 0.0252,-0.17953 0.0359,-0.27196 0.0967,0.31684 0.14402,0.24344 0.17214,-0.30257 0.007,-0.13895 2.7e-4,-0.25575 -0.0191,-0.35714 0.0953,-0.19421 0.25677,-0.34478 0.56255,-0.38997 0.0741,-0.0998 0.10036,-0.18124 0.0932,-0.24534 -0.0312,-0.27771 -0.69621,-0.22987 -0.87843,0.0617 -0.18949,0.0256 -0.34877,-0.047 -0.49421,-0.15972 0.68347,-0.14164 0.77742,-0.40214 0.7724,-2.03813 -0.004,-1.25111 -0.11592,-1.71698 -0.33141,-1.37666 -0.17911,0.28293 -0.24087,0.73482 -0.13754,1.00399 0.10335,0.26916 0.005,0.48979 -0.21607,0.48979 -0.22226,0 -0.41684,-0.20557 -0.43256,-0.45651 -0.023,-0.36431 -0.0665,-0.3644 -0.21338,-4.5e-4 -0.14313,0.35465 -0.34793,0.34082 -0.92502,-0.0634 -0.40789,-0.28557 -0.99692,-0.42235 -1.30788,-0.30302 -0.31722,0.12184 -0.46945,0.062 -0.34693,-0.1362 0.1201,-0.19428 0.0622,-0.59178 -0.12911,-0.88331 -0.11326,-0.17268 -0.20468,-0.27964 -0.3221,-0.28394 z m -0.32296,4.90546 c -0.003,0.0156 -0.004,0.031 -0.007,0.0466 -0.0389,-0.0151 -0.0395,-0.0307 0.007,-0.0466 z",
3.88374, 3.51074,
"M 0.152516 -3.50963  c -0.0671,-0.0035 -0.13954,0.0012 -0.21726,0.01501 0.0239,0.38383 -0.009,0.664347 -0.0834,0.870756 -0.5491,-0.372894 -1.21382,-0.623817 -1.99453,-0.680445 0.004,0.330283 0.0274,0.672343 0.0583,1.019123 -0.60216,-0.112963 -1.22906,-0.191918 -1.71237,0.117895 l -0.087,0.185898 c 1.00994,1.039798 2.17682,1.967379 2.19764,3.587233 0.17881,0.262722 0.36799,0.463849 0.56477,0.61642 0.13818,0.425596 0.28373,0.847187 0.44201,1.256679 1.15883,0.174004 1.9076,-0.383239 2.23915,-1.246964 0.44076,-0.188455 0.8768,-0.414276 1.29111,-0.606262 0.84054,-0.06761 1.08622,-0.328197 1.02353,-0.67868 -0.11969,-0.669085 -1.36303,-1.666368 -1.734,-2.275359 -0.58197,-0.72387 -0.79212,-0.277084 -0.95819,0.185455 -0.11392,-0.207857 -0.24645,-0.407202 -0.39388,-0.596988 0.20718,-0.755268 0.0844,-1.73265 -0.63582,-1.769775 z",
3.17107, 3.48989,
"M -0.695755 -3.48975  c -0.14458,-0.002 -0.31898,0.0184 -0.53872,0.0514 -0.48796,0.0735 -0.9685,0.32099 -1.33449,0.68684 -0.48142,0.48132 -0.57636,0.733 -0.59963,1.58863 -0.0182,0.66578 0.0642,1.0927 0.23826,1.23725 0.14629,0.12144 0.26581,0.37135 0.26581,0.5554 0,0.44729 0.44932,0.79765 0.83375,0.65015 0.1706,-0.0654 0.31045,-0.035 0.31045,0.0672 0,0.10242 0.35026,0.62619 0.77866,1.16423 0.44864,0.56348 0.93335,0.97854 1.14294,0.97854 0.35308,0 1.38754,-1.22023 1.59865,-1.8853 0.056,-0.17641 0.36274,-0.47456 0.68184,-0.66306 0.45927,-0.27126 0.55359,-0.42928 0.45275,-0.75654 -0.23387,-0.759 -0.79899,-1.36205 -1.72673,-1.84317 -0.54539,-0.28283 -1.03708,-0.70924 -1.20888,-1.04864 -0.29712,-0.58699 -0.46072,-0.77923 -0.89466,-0.78284 z",
4.76911, 3.80814,
"M 1.68749 -3.80769  c -0.0899,-0.004 -0.20175,0.019 -0.3478,0.0548 -0.2754,0.068 -0.80502,0.1779 -1.295,0.27337 -0.002,-5.7e-4 -0.003,-0.003 -0.005,-0.004 v 5.1e-4 c -0.0214,-0.004 -0.0479,0.004 -0.0775,0.0196 -0.0805,0.0155 -0.16738,0.0341 -0.24289,0.0481 -0.55451,0.10268 -1.25868,0.42197 -1.56475,0.70952 l -0.55708,0.52297 -0.1452,-0.57826 c -0.22357,-0.89072 -0.83344,-0.9771 -1.61438,-0.22893 -1.22476,1.1734 -0.44024,2.73403 0.96946,1.92856 0.38743,-0.22135 0.94538,-0.33453 1.30844,-0.2651 0.50575,0.0967 0.65979,0.0172 0.7674,-0.39377 0.0749,-0.28637 0.25625,-0.4356 0.4563,-0.46716 0.007,0.37514 0.15968,0.7233 0.44339,0.89866 0.15237,0.0942 0.1253,0.23529 -0.091,0.47439 -0.32686,0.36118 -0.77208,1.06994 -1.10431,1.69498 -0.72099,0.39409 -1.14366,1.04266 -0.95602,1.53169 0.0709,0.18476 0.27588,0.33538 0.45527,0.33538 0.35557,0 1.2275,-0.60376 1.86139,-1.20819 0.12891,0.3163 0.20815,0.7041 0.18085,1.048 -0.0504,0.63594 -0.005,0.73571 0.32609,0.71675 0.21154,-0.012 0.52486,-0.0817 0.69607,-0.15503 0.21968,-0.0938 0.34301,-0.0121 0.4191,0.27905 0.17931,0.68569 0.76809,0.39235 0.94774,-0.47232 0.086,-0.41422 0.22873,-0.84663 0.31782,-0.96067 0.089,-0.11402 0.22021,-0.45146 0.29146,-0.7493 0.0898,-0.37587 0.33261,-0.6263 0.79272,-0.81856 0.42373,-0.17705 0.70601,-0.44922 0.7829,-0.75551 0.19283,-0.76824 -0.0249,-1.0573 -0.79841,-1.0573 -0.39962,0 -0.93906,-0.17198 -1.26298,-0.40256 -0.12655,-0.0901 -0.23128,-0.15678 -0.32761,-0.20515 -0.0225,-0.18343 -0.0534,-0.37777 -0.0956,-0.58808 -0.19008,-0.94639 -0.26088,-1.21336 -0.5307,-1.2268 z",
2.54464, 4.86321,
"M 2.11268 -4.86312  -2.4e-4,3.4e-4 c -0.42259,-0.005 -0.71865,0.16899 -1.01028,0.58895 -0.2341,0.33711 -0.33434,0.40738 -0.61019,0.42851 -0.19518,0.0149 -0.37421,0.089 -0.43826,0.18143 -0.0182,0.0262 -0.0368,0.0486 -0.0555,0.0679 -0.009,-10e-6 -0.0175,-0.002 -0.0259,-6.4e-4 -0.071,0.004 -0.13477,0.0243 -0.18907,0.0571 -0.0242,-0.0244 -0.0391,-0.59467 -0.0391,-0.65359 0,-0.36718 -0.95879,-0.22794 -1.02811,0.14931 -0.0212,0.11589 0.0311,0.75145 0.13054,0.79443 0.23059,0.0996 0.20734,0.48697 -0.037,0.66412 -0.0408,-0.002 -0.0836,-0.005 -0.13238,-0.0129 -0.2574,-0.0398 -0.36727,-0.008 -0.58024,0.1696 -0.28227,0.235 -0.2933,0.288473 -0.26634,1.268603 0.0133,0.48463 0.0593,0.66991 0.18495,0.74461 0.11022,0.0655 0.20912,0.0728 0.28966,0.0212 0.0568,-0.0364 0.11754,-0.12674 0.17124,-0.23402 0.107,0.0658 0.25755,0.11689 0.41792,0.14507 -0.0242,0.0438 -0.0493,0.0775 -0.0743,0.0935 -0.22968,0.14714 -0.18073,0.29031 0.16346,0.47951 0.16459,0.0905 0.24667,0.16091 0.28464,0.26651 -0.22328,0.31999 -0.54905,0.56105 -0.79857,0.56105 -0.26805,0 -0.38363,0.13101 -0.38363,0.43504 0,0.099 -0.14109,0.28428 -0.31344,0.41175 -0.29557,0.21858 -0.40894,0.4781 -0.23503,0.53776 0.0431,0.0148 0.21951,0.23939 0.39186,0.49911 0.29454,0.44378 0.49384,0.61168 1.14242,0.96274 0.2234,0.12093 0.26771,0.2085 0.26771,0.52841 0,0.21082 0.0564,0.42682 0.1253,0.48004 0.17164,0.13252 0.36279,0.12141 0.72039,-0.0417 0.2969,-0.13538 0.35429,-0.23188 0.31892,-0.53634 -0.0158,-0.13554 0.343,-0.51902 0.92317,-0.98692 0.28788,-0.23216 0.26618,-0.30644 -0.12894,-0.44069 -0.37244,-0.12655 -0.37109,-0.11805 -0.22497,-1.22413 0.0601,-0.45428 0.10536,-0.49778 0.48606,-0.46787 0.10405,0.008 0.18291,-0.0399 0.1829,-0.11171 0,-0.16011 -0.092,-0.15842 -0.53246,0.01 -0.43953,0.16783 -0.58282,0.30447 -0.72314,0.68954 -0.0603,0.16556 -0.19888,0.33774 -0.30795,0.3828 -0.10906,0.045 -0.22827,0.15462 -0.26475,0.24337 -0.0547,0.1331 -0.11029,0.17837 -0.15911,0.14437 0.0191,-0.0362 0.0267,-0.0666 0.0146,-0.0817 -0.0268,-0.0333 -0.007,-0.17466 0.0454,-0.31397 0.0701,-0.18872 0.056,-0.28296 -0.0553,-0.36886 -0.17292,-0.13348 -0.20148,-0.65681 -0.04,-0.73384 0.0536,-0.0256 0.0859,-0.25521 0.0912,-0.59759 0.0308,-0.0283 0.0572,-0.054 0.0821,-0.0789 0.15372,-0.0696 0.2735,-0.27581 0.20759,-0.4084 -0.0286,-0.0576 0.009,-0.313 0.0811,-0.59847 0.0866,0.0779 0.22739,0.11682 0.37448,0.10077 0.33605,-0.0365 0.41939,-0.41959 0.11843,-0.54393 -0.1668,-0.0689 -0.21038,-0.19846 -0.25309,-0.75131 -0.0284,-0.367453 -0.0136,-0.746523 0.0327,-0.842563 0.0464,-0.0961 0.0457,-0.2254 0.0137,-0.35209 0.24966,-0.16135 0.77006,-0.17218 0.9474,-0.007 0.3168,0.29467 0.52241,0.0917 0.52241,-0.51569 0,-0.31975 0.0512,-0.41964 0.31732,-0.61647 0.40678,-0.30086 0.38252,-0.55289 -0.0562,-0.58224 -0.0293,-0.002 -0.058,-0.004 -0.0862,-0.004 z",
2.20356, 4.42723,
"M -0.779833 -4.42535  c -0.11252,0.01 -0.20766,0.0645 -0.31936,0.15707 -0.19303,0.1602 -0.30553,0.46533 -0.24987,0.67821 0.0557,0.21288 -0.031,0.75799 -0.19276,1.21128 -0.16176,0.45332 -0.29873,0.73139 -0.30366,0.61778 -0.006,-0.11371 -0.12301,-0.0935 -0.26176,0.0452 -0.19127,0.19127 -0.10817,0.47025 0.34411,1.1537 0.45766,0.69158 0.55864,1.03463 0.43407,1.47638 -0.0892,0.31659 -0.13322,0.67331 -0.0981,0.79292 0.0352,0.11957 0.15515,0.65783 0.267,1.19605 0.11176,0.53827 0.24306,1.10126 0.29128,1.25078 0.1727,0.53503 0.7492,0.25702 1.23653,-0.59635 0.27267,-0.47751 0.6625,-0.86921 0.86619,-0.87004 0.20765,-9.2e-4 0.5269,-0.33172 0.72679,-0.75294 0.29035,-0.6119 0.31116,-0.88889 0.11183,-1.49304 -0.13459,-0.40798 -0.22205,-1.12174 -0.19417,-1.58632 0.11331,-1.88242 -0.006,-2.47587 -0.49595,-2.47587 -0.25281,0 -0.83601,-0.21102 -1.29552,-0.46928 -0.44311,-0.24903 -0.67905,-0.35297 -0.86669,-0.33554 z",
2.98889, 3.67435,
"M 0.0402496 -3.67435  c -0.32831,0 -1.18189,0.92749 -1.79096,1.94648 -0.3584,0.59962 -0.44172,0.96993 -0.33994,1.5125 0.10659,0.56847 0.0207,0.89088 -0.41379,1.54736 -0.30817,0.46568 -0.51868,1.04377 -0.47984,1.31799 0.0652,0.45978 0.15767,0.48811 1.5781,0.48811 0.9272,0 1.60942,0.10087 1.77078,0.26195 0.68945,0.68812 2.06575,-0.0134 2.06575,-1.05284 0,-0.21065 0.15263,-0.67888 0.33994,-1.04091 0.34011,-0.65768 0.33957,-0.65885 -0.40829,-1.25881 -0.58894,-0.47247 -0.77875,-0.80242 -0.89043,-1.54736 -0.14276,-0.95204 -0.94727,-2.17447 -1.43132,-2.17447 z",
3.40151, 5.32865,
"M 2.71045 -5.32831  v 3.7e-4 c -0.67587,-0.012 -1.14971,0.34982 -1.61607,1.21978 -0.37425,0.69821 -0.53418,0.84374 -0.97521,0.8875 -0.312,0.0309 -0.59827,0.1844 -0.70069,0.37577 -0.2226,0.41586 -0.49528,0.41624 -0.49528,7e-4 0,-0.76048 -1.53295,-0.4721 -1.64379,0.30924 -0.0339,0.24004 0.0495,0.46036 0.20873,0.54939 0.37193,0.2081 0.33168,1.02339 -0.0686,1.38571 -0.49393,0.44696 -1.00322,2.96143 -0.75706,3.73717 0.13368,0.42094 0.97103,0.75376 1.58712,0.63054 0.60075,-0.12019 0.73147,0.051 0.50116,0.65685 -0.0961,0.2532 -0.13673,0.56011 -0.0898,0.68207 0.0467,0.12199 0.19735,0.22188 0.33479,0.22188 0.31413,0 0.6262,-0.56153 0.49789,-0.89591 -0.053,-0.13863 0.0311,-0.82668 0.18716,-1.52901 0.31845,-1.4316 0.19566,-1.84197 -0.4343,-1.44858 -0.48884,0.30529 -0.65563,0.11304 -0.66886,-0.77128 -0.007,-0.43091 -0.0729,-0.93369 -0.14734,-1.11741 -0.18115,-0.44714 0.24354,-0.91857 0.82717,-0.91857 0.24714,0 0.58532,-0.15039 0.75154,-0.33411 0.36082,-0.39872 1.28482,-0.44747 1.58678,-0.0837 0.50646,0.6103 0.83488,0.19 0.83488,-1.06807 0,-0.66226 0.0822,-0.86909 0.50773,-1.27679 0.65033,-0.6231 0.61157,-1.14509 -0.09,-1.20588 -0.0467,-0.005 -0.093,-0.008 -0.13814,-0.008 z",
3.13042, 4.93641,
"M 0.240783 -4.93553  c -0.15688,-0.0178 -0.29819,0.23397 -0.49049,0.81373 -0.2825,0.85157 -0.93362,1.59307 -1.39943,1.59307 -0.39149,0 -0.56027,0.24786 -0.56027,0.82307 0,0.18722 -0.20606,0.53783 -0.45779,0.77901 -0.4317,0.41355 -0.59724,0.90453 -0.34324,1.01741 0.0626,0.028 0.32056,0.45291 0.57229,0.94429 0.43015,0.8396 0.72124,1.15726 1.66856,1.82145 0.32624,0.22878 0.39098,0.39448 0.39098,0.99971 0,0.39888 0.0823,0.80753 0.18298,0.90823 0.25071,0.2507 0.52985,0.2297 1.05213,-0.0788 0.43363,-0.25616 0.51743,-0.4387 0.4658,-1.01474 -0.0233,-0.25645 0.50065,-0.98195 1.348,-1.8672 0.42045,-0.43923 0.38879,-0.57978 -0.18835,-0.83377 -0.54393,-0.23942 -0.54167,-0.22334 -0.32824,-2.31596 0.0877,-0.8595 0.15391,-0.94177 0.70992,-0.88519 0.15197,0.0155 0.26679,-0.0756 0.26679,-0.21137 0,-0.30291 -0.13407,-0.30005 -0.77732,0.018 -0.64195,0.31753 -0.85123,0.57637 -1.05619,1.3049 -0.0882,0.31325 -0.29044,0.63899 -0.44972,0.72425 -0.15934,0.0852 -0.33341,0.29254 -0.38672,0.46045 -0.16715,0.5267 -0.33974,0.3349 -0.41637,-0.46279 -0.0465,-0.4866 -0.17065,-0.83911 -0.33826,-0.96165 -0.24082,-0.17607 -0.24289,-0.21137 -0.022,-0.39467 0.14751,-0.1224 0.24275,-0.43967 0.24275,-0.80907 0,-0.42563 0.11425,-0.74365 0.38136,-1.06115 0.43886,-0.52158 0.46645,-0.71391 0.16257,-1.12559 -0.0841,-0.11391 -0.15837,-0.17762 -0.2297,-0.18565 z",
3.36677, 6.16908,
"M 2.46494 -6.16587  c -0.0541,-0.01 -0.12003,0.003 -0.18482,0.0437 -0.14889,0.092 -0.21185,0.45187 -0.17387,0.99402 l 0.0595,0.84965 -0.8909,0.25437 c -0.85575,0.24425 -0.91404,0.30547 -1.49387,1.561 -0.68138,1.47535 -0.8842,1.39519 -0.95601,-0.37731 -0.0677,-1.67347 -0.53514,-1.57118 -0.54675,0.1197 -0.008,1.20372 -0.14815,1.49162 -0.91678,1.88573 -0.33093,0.1697 -0.55445,0.43025 -0.55445,0.64665 0,0.31078 0.1713,0.37963 1.20109,0.48124 0.9293,0.0916 1.27169,0.056 1.51288,-0.15692 0.17144,-0.15173 0.40028,-0.22045 0.50832,-0.15367 0.34578,0.21373 0.21149,0.66082 -0.35303,1.17641 -0.30209,0.2759 -0.64427,0.77409 -0.76029,1.10686 -0.28842,0.82736 -0.63087,0.63547 -0.63087,-0.35346 0,-0.78244 -0.0622,-0.90627 -0.45738,-0.90627 -0.31541,0 -0.7871,0.90286 -1.04012,1.99129 -0.23161,0.9964 -0.22259,1.08589 0.16459,1.6532 0.22515,0.32984 0.51912,0.64232 0.65352,0.69436 0.40194,0.15571 0.95683,-0.25666 0.95683,-0.71094 0,-0.48104 0.73081,-0.96973 1.18811,-0.79425 0.41463,0.1592 0.35885,0.40487 -0.17185,0.7526 -0.495,0.32433 -0.61311,0.99321 -0.24064,1.36566 0.25886,0.25887 0.69051,0.28229 1.0462,0.0571 0.1824,-0.11553 0.21563,-0.36728 0.12213,-0.92083 -0.13213,-0.78273 -0.0775,-0.84677 0.66079,-0.77645 0.16235,0.0149 0.40663,-0.10649 0.54313,-0.27095 0.19821,-0.23888 0.20282,-0.37194 0.0222,-0.6612 -0.12434,-0.19914 -0.17751,-0.44051 -0.11809,-0.53664 0.0594,-0.0961 0.0229,-0.48362 -0.0813,-0.86098 -0.23192,-0.83997 0.12462,-1.35152 0.94265,-1.35152 0.63152,0 0.89091,-0.51453 0.89091,-1.76764 0,-0.91589 -0.27505,-1.06452 -0.83144,-0.4497 -0.45014,0.49738 -0.77029,0.46537 -0.87594,-0.0874 -0.091,-0.47571 0.45137,-1.57555 0.77684,-1.57555 0.18146,0 0.93054,-1.3103 0.93054,-1.62773 0,-0.0901 -0.16195,-0.3501 -0.35992,-0.57829 -0.19792,-0.22822 -0.3899,-0.50528 -0.42664,-0.61551 -0.0186,-0.0556 -0.061,-0.0898 -0.11526,-0.0999 z",
3.36642, 4.71204,
"M 2.14302 -4.71204  c -0.63965,0.002 -0.47682,2.13844 -1.24978,1.55409 -0.92613,-1.50861 -3.40173,-0.29544 -3.14116,1.46395 0.51593,1.36936 -1.77871,0.91049 -0.92686,1.85562 1.40951,0.21819 -0.44125,1.4145 0.82327,1.66419 -0.57592,1.3499 0.6215,1.87772 1.29993,2.57621 l 0.0604,0.24159 c 1.21026,0.51184 1.30627,-1.98878 1.77943,-2.80526 0.61441,-1.01698 0.0716,-4.43446 2.03773,-3.03058 1.29395,-0.57329 -0.14993,-2.07687 -0.0896,-3.11886 -0.25448,-0.28697 -0.44513,-0.40085 -0.59283,-0.40095 z",
2.67932, 7.19187,
"M 1.11365 -7.19037  c -0.3679,0.0246 -0.63802,0.36228 -0.63802,0.85647 0,0.30813 -0.20548,0.69239 -0.54292,1.01568 -0.4374,0.41905 -0.65076,0.49812 -1.10192,0.40788 -0.45379,-0.0908 -0.64747,-0.0176 -1.02293,0.38733 -0.49761,0.53665 -0.51748,0.65878 -0.46996,2.8971 0.0235,1.10675 0.10496,1.52987 0.32646,1.70046 0.19433,0.14956 0.36827,0.16624 0.51027,0.0483 0.29856,-0.24777 0.67015,-1.55748 0.53886,-1.89956 -0.13302,-0.34677 0.56593,-0.78978 0.75734,-0.48002 0.19788,0.32019 -0.0997,2.15035 -0.38855,2.39006 -0.4049,0.33604 -0.31862,0.66299 0.28818,1.09508 0.4482,0.31914 0.55259,0.52604 0.55259,1.09468 0,0.38551 -0.1213,1.04395 -0.26923,1.46305 -0.32207,0.91276 -0.13816,3.08081 0.28414,3.35093 0.18196,0.11631 0.35857,0.0495 0.5824,-0.21965 0.17673,-0.21254 0.28298,-0.44885 0.23579,-0.52518 -0.0472,-0.0763 -0.0118,-0.39886 0.0802,-0.71701 0.12359,-0.43098 0.0987,-0.64618 -0.0975,-0.84237 -0.30484,-0.30484 -0.35517,-1.49996 -0.0705,-1.67587 0.1173,-0.0725 0.18049,-0.84131 0.16121,-1.95881 -0.0334,-1.94431 0.16284,-2.55058 0.62232,-1.92213 0.14856,0.20308 0.41417,0.30595 0.69123,0.26682 0.59243,-0.0836 0.73976,-0.95822 0.20916,-1.24219 -0.29405,-0.15734 -0.37126,-0.4532 -0.44656,-1.71577 -0.05,-0.83913 -0.0241,-1.70487 0.0576,-1.92415 0.21457,-0.57599 -0.13731,-1.67514 -0.5816,-1.81613 -0.0932,-0.0296 -0.18312,-0.0407 -0.26802,-0.0351 z",
3.567, 4.28391,
"M 2.1794 -4.28301  c -0.79452,0.0388 -2.13823,1.33278 -2.9006,1.67225 -1.91886,1.17374 2.21816,1.25105 0.10284,2.06499 -0.58361,-0.83118 -3.01823,-1.30637 -2.94711,0.17002 3.14717,0.20852 0.16743,2.80766 0.98237,4.53047 l 0.2031,0.12919 c 1.35802,-1.01667 2.60945,-2.23253 4.49275,-2.01435 2.35172,-1.19271 1.28122,-3.42569 0.70123,-5.26324 0.0473,-0.98574 -0.21839,-1.30966 -0.63458,-1.28933 z",
2.56174, 3.97214,
"M 1.23387 -3.97214  c -0.18031,0 -0.50408,0.11364 -0.92088,0.35086 -0.75763,0.43141 -0.89736,0.64381 -0.99891,1.519285 -0.10988,0.94689 -0.17227,1.02365 -0.88521,1.09244 -0.89236,0.0859 -1.14888,0.68747 -0.90125,2.11048 0.0846,0.48667 0.1492,1.23005 0.14367,1.65208 -0.008,0.60558 0.0866,0.7674 0.44804,0.7674 0.25202,0 0.52438,0.10634 0.6046,0.23617 0.21463,0.34726 0.57153,0.29583 2.12908,-0.30748 1.35943,-0.52657 1.40925,-0.57732 1.58906,-1.59422 0.28866,-1.63263 0.097,-2.0538 -1.0387,-2.2784 -1.22331,-0.24191 -1.64524,-0.80341 -1.07591,-1.43248 0.2218,-0.24509 0.40307,-0.615415 0.40307,-0.823195 0,-0.20788 0.17214,-0.54991 0.3824,-0.76018 0.3479,-0.34787 0.35277,-0.53337 0.12094,-0.53276 z",
1.40175, 2.35464,
"M 0.715347 1.73249  c 0.5213,-0.7219 1.00154,-1.43136 0.42262,-1.81986 -0.0242,-0.002 -0.0403,-0.0133 -0.0667,-0.0124 -0.2124,-0.35703 -0.18293,-1.55466 -0.66093,-1.50121 -0.38391,0.0613 -0.36171,-0.99746 -1.04282,-0.70124 -0.84233,0.26952 -1.03878,1.15204 -0.35968,1.72237 -0.28818,0.80849 -0.72983,1.85273 -0.017,2.57194 0.11293,0.13641 0.31123,0.26109 0.54043,0.35233 0.20746,0.0826 0.99861,-0.35508 1.18408,-0.61193 z",
3.58152, 4.34829,
"M -2.44195 -4.34714  c -0.75319,0.0362 -0.7762,0.92587 -0.55537,1.58591 -0.55027,0.26441 -0.90264,0.93541 -0.1653,1.39825 0.67743,0.40826 1.43356,1.01809 1.58319,1.83852 0.64834,0.73248 -0.1527,1.61798 -0.33278,2.42552 l 0.54118,0.63011 c 0.0142,-0.52312 0.48191,-0.59895 0.82051,-0.61429 -0.003,0.37955 0.1004,0.74248 0.40424,1.04909 0.11924,0.14402 0.32868,0.2752 0.57065,0.37153 0.21902,0.0872 1.05405,-0.37478 1.24987,-0.64594 0.55034,-0.76211 1.05742,-1.51075 0.44627,-1.92088 -0.0255,-0.002 -0.0426,-0.014 -0.0704,-0.0131 -0.0106,-0.0172 -0.0174,-0.0448 -0.0267,-0.0654 0.21024,-0.3016 0.38758,-0.56962 0.74632,-0.0954 0.65582,0.69676 1.24636,-0.49914 0.37097,-0.65739 -0.4622,0.24406 -0.37168,-0.72499 -0.69993,-0.96236 -0.10039,-0.91415 -0.4508,-1.68175 -0.81563,-2.50845 -0.28239,-0.46623 -0.83252,-1.22127 -1.4135,-1.3415 -0.0742,-0.0449 -0.15322,-0.0625 -0.2357,-0.0556 -0.32052,0.0264 -0.68498,0.39222 -0.99017,0.71794 -0.2134,-0.59389 -0.55451,-1.1417 -1.22475,-1.1271 -0.0728,-0.01 -0.14044,-0.0118 -0.20295,-0.01 z",
3.66123, 3.92165,
"M -2.89919 -0.28264  c 0.5248,0.64771 0.38465,1.62029 -0.46302,1.88928 -0.85786,0.6722 0.38319,1.31277 0.57724,1.97921 0.11171,0.25581 0.51808,0.48818 0.72707,0.20878 0.35103,-0.2002 0.83823,0.2502 1.11623,-0.3514 0.28826,-0.50439 0.39449,-1.18406 1.07021,-1.30949 0.38793,-0.46627 -0.45389,-0.99265 0.11266,-1.47639 0.35536,-0.69249 0.56184,-1.84188 1.34721,-2.07791 0.63362,0.38352 1.48161,0.65582 1.72651,-0.31936 0.39002,-0.4159 0.59822,-1.04678 -0.19431,-1.0914 -0.59473,-0.216 -1.75681,-0.9684 -1.89285,-1.06904 -0.17327,-0.12817 -0.40811,0.35444 -0.60881,0.59258 -0.2757,0.48096 -1.0685,0.56942 -0.82373,1.34875 0.16859,0.56791 -0.21712,0.74465 -0.57515,0.30903 -0.73642,-0.18936 -0.47345,1.25664 -1.18184,0.50643 -0.15757,-0.16278 -0.39659,0.0541 -0.64211,-0.053 -0.23726,-0.1035 0.0848,-0.3456 -0.27012,-0.34843 -0.37484,-0.003 -0.30986,0.87936 -0.0252,1.26233 z",
4.66444, 5.18369,
"M 2.11104 -5.18252  c -0.17844,-0.0365 -0.59098,0.79076 -0.87075,0.86971 -0.28109,-0.28971 -0.64463,-0.14558 -0.67074,0.31678 0.17296,0.35303 -0.32848,0.92855 -0.38034,0.37982 -0.48361,-0.88691 -0.84545,0.5953 -1.32914,0.67386 -0.34102,0.46556 0.0161,0.50242 0.25064,0.78445 -0.56409,0.44035 -0.67167,0.83211 -0.3612,1.14774 -0.006,0.0911 0.005,0.19332 0.0424,0.31316 0.16859,0.56791 -0.21712,0.74465 -0.57516,0.30903 -0.73641,-0.18936 -0.47344,1.25663 -1.18184,0.50642 -0.15758,-0.16277 -0.39629,0.0539 -0.64182,-0.0532 -0.23725,-0.1035 0.0846,-0.34547 -0.27027,-0.3483 -0.37484,-0.003 -0.30999,0.87948 -0.0253,1.26246 0.52481,0.64771 0.38466,1.6203 -0.46302,1.88928 -0.85785,0.6722 0.3837,1.31277 0.57775,1.97921 0.1117,0.25581 0.51758,0.48817 0.72657,0.20878 0.35102,-0.20021 0.83873,0.25019 1.11673,-0.3514 0.28826,-0.5044 0.39399,-1.18405 1.06971,-1.30949 0.38793,-0.46627 -0.4539,-0.99265 0.11266,-1.47639 0.35533,-0.69249 0.56234,-1.84188 1.34771,-2.07791 0.30376,0.18387 0.65616,0.34025 0.96842,0.34778 0.33359,0.3812 0.84312,0.50562 1.2237,-0.11369 -0.30948,-1.21831 1.01658,-2.14727 1.88669,-2.2686 -0.41399,-0.6498 -0.0859,-1.85265 -0.94827,-2.09135 -0.29061,-0.0875 -1.51122,0.0898 -1.51876,-0.51108 0.006,-0.26471 -0.0266,-0.37397 -0.0858,-0.38705 z",
3.05335, 5.71572,
"M 2.29249 -5.71375  c -0.44087,0.0257 -0.84558,0.29637 -0.4935,0.77566 0.043,0.5726 -0.66244,1.13031 -1.01182,1.61954 -0.59536,0.0643 -0.67884,0.58579 -0.62891,1.04283 -0.0328,0.0299 -0.0632,0.056 -0.0997,0.092 -0.45413,0.43121 -0.7044,1.02376 -0.9338,1.62522 -0.0229,-0.0163 -0.0443,-0.0282 -0.0687,-0.0486 -0.64201,-0.37339 -0.381,-1.68417 -1.16064,-1.73633 -0.89157,0.53168 -0.63784,1.71779 -0.94877,2.58124 0.41748,0.0224 0.78081,-0.11303 0.77824,-0.56637 -0.029,-0.81387 0.44852,-0.14497 0.38963,0.22686 0.1398,0.26873 0.29702,0.57106 0.45476,0.83974 -0.22764,0.40185 -0.52107,0.74432 -0.9617,0.95911 -0.36584,-0.3203 -0.60315,0.95868 -0.12867,0.68575 0.24482,-0.0374 0.49482,-0.0125 0.74567,0.0408 -0.13684,0.27374 -0.30318,0.52271 -0.54673,0.70848 -0.72324,1.44096 1.7649,0.37106 0.93432,1.48157 -1.67179,-0.43188 -1.37967,2.24445 0.49197,0.51159 0.93927,-0.75809 1.02925,-2.0752 2.374,-2.39416 0.0427,-0.19658 0.0496,-0.37691 0.0326,-0.54519 0.007,-0.013 0.0142,-0.0244 0.0212,-0.0377 -0.0109,-0.006 -0.0181,-0.0127 -0.0284,-0.0186 -0.0376,-0.26876 -0.14378,-0.50259 -0.29869,-0.70797 0.0675,-0.50253 0.39688,-1.12995 0.27337,-1.65519 0.0776,-0.12843 0.15978,-0.26383 0.2372,-0.3974 0.51546,-0.0608 0.96877,-0.52849 0.25114,-0.86041 -0.99645,-0.23462 -0.10779,-1.19683 -0.005,-1.79162 -0.13147,-0.84084 0.53896,-1.24912 1.0418,-1.73478 0.20727,-0.50166 -0.26911,-0.72137 -0.71004,-0.69608 z",
3.25546, 5.30144,
"M 1.7045 -3.73289  c -0.59536,0.0643 -0.67884,0.58579 -0.62891,1.04283 -0.0328,0.0299 -0.0632,0.056 -0.0997,0.092 -0.11954,0.11351 -0.22424,0.23899 -0.31885,0.37207 -0.16346,-0.31767 -0.41725,-0.59374 -0.693,-0.86196 -0.0135,-0.57623 -0.52533,-0.68908 -0.93429,-0.86196 -0.3143,-0.11364 -0.30232,-0.22096 -0.26924,-0.50178 10e-4,-0.35694 -0.19302,-0.64802 -0.63614,-0.70849 -0.52435,-0.42985 -0.56304,0.22719 -0.74414,0.56328 0.27,0.2644 0.29136,0.75232 -0.0961,1.02939 -0.41132,0.0707 -0.85182,1.03037 -0.23047,1.04387 0.0806,0.4722 0.45971,1.10991 0.004,1.55132 -0.11771,0.33019 0.0386,0.9166 0.39997,1.19631 -0.16174,0.3523 -0.42196,0.8126 -0.35452,1.31207 0.12679,0.12631 0.14632,0.23085 0.13179,0.31729 l 0.17828,0.42323 1.16996,0.47026 c -0.67056,1.39319 1.77057,0.35041 0.9462,1.45262 -1.67179,-0.43188 -1.37969,2.24446 0.49194,0.5116 0.93927,-0.7581 1.02929,-2.0752 2.37403,-2.39417 0.0427,-0.19657 0.0496,-0.37691 0.0325,-0.54519 0.007,-0.013 0.0142,-0.0244 0.0212,-0.0377 -0.0109,-0.006 -0.0181,-0.0127 -0.0284,-0.0186 -0.0375,-0.26875 -0.14377,-0.50258 -0.29869,-0.70796 0.0675,-0.50253 0.39688,-1.12996 0.27337,-1.6552 0.0776,-0.12843 0.15978,-0.26383 0.2372,-0.39739 0.51546,-0.0608 0.96877,-0.52849 0.25114,-0.86041 -0.99645,-0.23463 -0.10795,-1.19683 -0.005,-1.79163 -0.13147,-0.84083 -0.82474,-0.52493 -1.17413,-0.0357 z",
2.78553, 3.0102,
"M 1.13709 -3.01011  c -0.11999,0.003 -0.27165,0.0974 -0.47025,0.32039 0.066,1.27176 -1.98331,0.51776 -2.42723,1.40147 -0.0487,-0.008 -0.0958,-0.0135 -0.13954,-0.0155 v 0 c -0.1647,-0.007 -0.38814,1.2022 -0.42942,1.30522 -0.044,0.10842 -0.0601,0.18666 0.12544,0.44857 -0.29803,0.73252 -0.15986,0.47603 -0.30581,1.26773 -0.70138,0.52342 0.0894,1.15559 0.6594,0.67903 0.0741,-0.0339 0.14901,-0.0598 0.22426,-0.0801 0.006,0.0576 0.0156,0.14035 0.019,0.18707 0.3897,0.22008 0.6114,0.17646 0.72191,-0.0186 l 0.27597,0.52503 c 0.45595,-0.0258 0.61214,-0.2684 0.60356,-0.57309 0.61931,0.11342 1.21306,0.13704 1.71154,-0.50953 1.48952,-0.17693 -0.82934,-2.42653 0.68212,-1.97766 0.62791,0.54053 0.50184,-1.28519 -0.30697,-0.9617 -0.50501,-0.32149 -0.42357,-2.01307 -0.94361,-1.99832 z",
3.40832, 4.84469,
"M 2.62836 -4.84469  c -0.55086,0.67388 -0.63529,1.54387 -0.99475,2.24637 -0.0561,-0.0679 -0.1206,-0.11028 -0.20103,-0.108 h -5.3e-4 c -0.11999,0.003 -2.65788,1.24415 -2.89783,1.72186 -0.0487,-0.009 -0.0958,-0.0135 -0.13954,-0.0155 -0.0819,-0.003 -0.17807,0.29411 -0.2594,0.60668 -0.20463,-0.0148 -0.40645,-0.0353 -0.59997,-0.0636 -0.21733,0.51377 -0.40796,0.43238 -0.59894,0.21445 -0.86921,0.67457 0.0811,1.38892 0.91467,1.8433 -0.0122,0.0981 -0.0303,0.22652 -0.0661,0.42065 -0.70135,0.52342 0.0894,1.15558 0.6594,0.67903 0.0741,-0.0339 0.14904,-0.0598 0.22428,-0.0801 0.006,0.0576 0.0157,0.14035 0.0191,0.18707 0.3897,0.22008 0.61142,0.17645 0.72194,-0.0186 l 0.27593,0.52503 c 0.0851,-0.005 0.15658,-0.0193 0.2217,-0.0382 -0.54399,0.66639 2.71703,2.46603 2.19988,1.0232 -0.0139,-0.61741 -0.002,-1.35716 -0.17984,-1.98593 0.0244,-0.028 0.0494,-0.0506 0.0734,-0.0816 0.0125,-0.001 0.0226,-0.004 0.0346,-0.006 0.0456,0.024 0.0912,0.0457 0.13642,0.0744 0.28681,-0.0617 0.92583,0.13145 0.99012,-0.27905 -0.20907,-0.18446 -0.45444,-0.38854 -0.68937,-0.61029 -8e-5,-9e-5 1.1e-4,-4.4e-4 0,-5.2e-4 -0.18285,-0.66281 -0.73432,-1.43655 0.21035,-1.156 0.62791,0.54053 0.50183,-1.28519 -0.30697,-0.9617 -0.16449,-0.10471 -0.26461,-0.35892 -0.34571,-0.64802 0.40217,-0.89611 1.62198,-1.24553 1.16322,-2.50115 0.10718,-0.40833 0.32133,-0.44801 0.15195,-0.57722 -0.14476,-0.11041 -0.66517,-0.33109 -0.71729,-0.41031 z",
2.91331, 4.2784,
"M 2.48505 -2.50743  -1.1297,-0.69932 c -0.26081,0.34403 -0.73919,0.6277 -1.09585,0.98835 -0.24912,0.29904 -0.87887,0.48065 -0.71385,-0.14805 0.12544,-0.65962 -0.50944,-0.96834 -0.73168,-1.47416 -0.346,0.0114 -0.71232,0.27983 -0.97235,-0.16999 -0.28809,-0.52063 -1.0299,-0.21119 -0.64799,0.32288 0.0949,0.43869 -0.36202,1.16319 0.1949,1.57459 0.20779,0.3961 0.46321,0.88104 0.35823,1.27627 -0.32873,0.54046 0.48828,0.91234 0.59778,1.40794 -0.21794,0.44035 -0.88762,0.52868 -0.68223,1.17211 -0.14326,0.65293 0.62787,0.80659 0.88904,0.25258 0.25359,-0.22465 0.61647,-0.63708 0.95934,-0.57843 0.36095,0.23638 0.80362,0.33863 0.64427,0.90689 -0.14441,0.70198 1.27406,0.83383 0.55797,1.45519 -0.0887,0.73566 0.84516,-0.20652 0.74062,0.41996 0.0243,0.13225 0.1875,0.0766 0.22429,-0.008 0.52299,-0.10663 0.83538,-0.53091 0.87788,-1.00509 -0.28215,-0.43305 -0.9743,-0.71636 -0.8522,-1.3157 0.1777,-0.15854 0.72403,0.14848 0.86856,-0.2217 0.0626,-0.53635 0.50097,-0.95686 0.279,-1.47081 -0.26406,-0.25211 -0.7397,-0.041 -0.99059,-0.11792 -0.1503,-0.59096 0.62362,-0.44277 0.78004,-0.89386 0.23735,-0.31162 -0.0744,-0.59806 -0.34594,-0.66771 0.35082,-0.27779 0.25851,-0.66602 0.19046,-1.00583 z",
3.95524, 4.27828,
"M -1.82143 -4.27815  c -0.25411,0.0105 -0.45471,0.25635 -0.21601,0.59015 0.0949,0.43869 -0.3616,1.16318 0.19532,1.57458 0.2078,0.3961 0.4631,0.88118 0.35814,1.27641 -0.0519,0.0854 -0.0729,0.16631 -0.0749,0.24443 -0.30033,-0.55817 -0.63892,-1.39159 -1.45005,-0.93483 -0.68429,0.018 -1.4623,0.76829 -0.48681,1.12758 0.43926,0.0751 -0.26699,0.60334 0.20516,0.87695 -0.0159,0.0235 -0.0497,0.0421 -0.0672,0.0651 l 1.56943,0.97152 c -0.003,-0.009 0,-0.0182 -0.003,-0.0269 0.0739,-0.017 0.13893,-0.0512 0.20052,-0.0894 -0.0235,0.0959 -0.022,0.20757 0.0222,0.34571 -0.14325,0.65293 0.6282,0.80671 0.88937,0.2527 0.0456,-0.0404 0.0984,-0.0907 0.15036,-0.13952 0.28739,0.18954 0.63963,0.18337 0.97256,-0.0904 0.0949,0.088 0.29792,0.16289 0.50125,0.2434 0.0167,0.087 0.0148,0.18794 -0.0206,0.3147 -0.0412,0.20018 0.049,0.35167 0.17622,0.4837 -0.60791,0.27447 1.22695,0.76516 1.12241,1.39164 0.0243,0.13225 0.18749,0.0763 0.22429,-0.008 0.52297,-0.10663 0.83547,-0.53093 0.87797,-1.00511 -0.18696,-0.28696 -0.5497,-0.50948 -0.74052,-0.80047 0.3039,-0.33202 0.6305,-0.62391 1.08003,-0.69298 0.3443,-0.30868 -0.18545,-0.98462 0.27493,-1.39681 -0.003,-0.027 0.0145,-0.0359 0.0156,-0.0605 v -5.3e-4 c -0.15378,-0.23531 -0.0824,-0.95349 -0.51263,-0.59635 -0.008,0.0349 -0.0352,0.0725 -0.0403,0.10697 -0.23939,0.0364 -0.4218,0.16976 -0.58653,0.33073 -0.0687,0.002 -0.13319,1.9e-4 -0.18605,-0.016 -0.15028,-0.59096 0.62391,-0.44292 0.78031,-0.894 0.23736,-0.31162 -0.0747,-0.59801 -0.34621,-0.66766 0.35081,-0.2778 0.25874,-0.66582 0.19069,-1.00563 v -5.3e-4 l -1.12967,-0.69918 c -0.2608,0.34403 -0.73938,0.62792 -1.09606,0.98857 -0.24911,0.29904 -0.87866,0.48039 -0.71364,-0.14831 0.12544,-0.65962 -0.5095,-0.96851 -0.73173,-1.47433 -0.34602,0.0114 -0.71252,0.2798 -0.97255,-0.17002 -0.10803,-0.19523 -0.27956,-0.27346 -0.43202,-0.26716 z",
4.74428, 5.06938,
"M -3.5022 -4.25506  c 0.0377,1.35963 1.56777,-0.33227 2.03738,0.80243 0.10379,0.0956 0.208,0.18438 0.31094,0.26506 -0.0406,0.39354 -0.0321,0.76444 0.0226,0.99137 -0.0937,0.13068 -0.17696,0.30235 -0.24523,0.49401 l -0.016,0.006 c 0.002,0.008 0.003,0.018 0.005,0.0266 -0.0996,0.28994 -0.16492,0.62239 -0.17565,0.93429 -0.0765,0.0524 -0.1308,0.11334 -0.15194,0.17898 l -7.4e-4,1e-4 c -0.0328,0.10411 0.0199,0.22082 0.19471,0.33612 0.0223,0.0965 0.0512,0.18645 0.0924,0.26307 -0.0649,0.0713 -0.11275,0.14849 -0.14391,0.22932 l -4e-4,-3.2e-4 c -0.0213,0.0536 -0.0342,0.10888 -0.042,0.16478 -0.07,-0.0196 -0.14851,-0.0772 -0.2347,-0.1906 -0.4414,-0.50596 -1.07643,-0.19724 -1.61154,-0.33475 -0.20468,0.2792 -0.21946,0.73356 -0.73384,0.66033 -0.58741,-0.0949 -0.8024,0.67948 -0.14649,0.70885 0.40367,0.19623 0.69169,1.00265 1.35917,0.8186 0.43988,0.0811 0.97908,0.17983 1.22514,0.50646 0.22215,0.59228 1.0194,0.17931 1.47686,0.39919 0.0351,0.0735 0.0519,0.1516 0.0635,0.23048 -0.003,0.23736 -0.0256,0.46963 -0.1023,0.6877 0.73066,0.46689 1.85021,2.14597 1.91273,0.29922 l -0.0399,-0.38017 c 0.17772,-0.45619 0.21761,-0.97787 0.11319,-1.43135 0.18902,-0.0344 0.36991,0.007 0.54937,0.25255 0.46301,0.54704 1.44266,-0.4874 1.48929,0.45953 0.5239,0.52402 0.35947,-0.79213 0.7877,-0.32307 0.11904,0.0625 0.17579,-0.10037 0.1322,-0.18137 0.23905,-0.4772 0.0983,-0.98501 -0.24849,-1.31123 -0.51486,-0.0455 -1.16504,0.32403 -1.56109,-0.14209 -0.015,-0.23767 0.56363,-0.47753 0.36169,-0.8198 -0.38324,-0.38042 -0.44322,-0.98487 -0.98457,-1.1276 -0.36135,0.052 -0.48894,0.55663 -0.70435,0.70647 -0.55766,-0.24666 0.0368,-0.76423 -0.22141,-1.1658 -0.0503,-0.19364 -0.18405,-0.2692 -0.33237,-0.2754 -0.0404,-0.3067 -0.57848,-0.81058 -0.65503,-0.97077 -0.40971,-0.63902 0.92366,-1.3558 0.28285,-1.97222 -0.0522,-0.16312 -0.21544,-0.27501 -0.49717,-0.33243 -0.0587,-0.0278 -0.11515,-0.0458 -0.1692,-0.0573 -0.0289,-0.0113 -0.057,-0.0238 -0.0853,-0.0361 -1.03981,-0.53291 -2.51397,0.22465 -3.31269,0.63055 z",
3.95515, 2.62566,
"M 3.95505 -0.686011  c -0.15408,-0.23494 -0.0822,-0.95418 -0.51263,-0.59686 -0.008,0.0349 -0.0352,0.0725 -0.0403,0.10697 -0.83055,0.12629 -1.00177,1.38378 -1.97093,1.12551 -0.48091,0.005 -0.94795,-0.32656 -1.45987,-0.33538 -0.37338,-0.52472 -0.57999,-1.00169 -0.26509,-1.59473 0.0936,-0.67218 -1.16382,-0.98056 -0.65939,-0.11679 0.26345,0.50822 0.18968,1.40264 -0.508,0.75758 -0.32954,-0.53085 -0.64802,-1.61561 -1.54771,-1.10898 -0.68429,0.018 -1.46227,0.76829 -0.48678,1.12758 0.43926,0.0751 -0.26701,0.60334 0.20516,0.87695 -0.0159,0.0235 -0.0497,0.0426 -0.0672,0.0656 l 1.56941,0.97152 c -0.002,-0.009 -10e-4,-0.0182 -0.003,-0.0269 0.39897,-0.0917 0.59063,-0.63554 0.87437,-0.0842 0.29559,0.58427 0.85214,0.78099 1.36064,0.36277 0.29091,0.26967 1.59176,0.4144 0.83405,0.98495 -0.96634,0.19424 0.0436,1.30965 0.4563,0.51211 0.65617,-0.49446 1.06992,-1.43769 1.93064,-1.56993 0.3443,-0.30868 -0.18545,-0.98462 0.27493,-1.39681 -0.002,-0.0272 0.0148,-0.0361 0.0155,-0.061 z",
3.39269, 3.51493,
"M 0.0750514 -3.51461  c -0.19846,-0.006 -0.39854,0.0731 -0.58965,0.27905 -0.3796,0.57495 -1.2206,0.87046 -1.87431,0.68161 -0.71429,-0.005 -1.49539,0.87511 -0.60666,1.43299 0.6417,0.38671 1.3579,0.96436 1.49963,1.7415 0.61413,0.69383 -0.14464,1.5326 -0.31522,2.29753 l 0.51263,0.59686 c 0.0171,-0.63284 0.7448,-0.57701 1.00822,-0.58962 0.0858,-1.15842 1.23714,0.33687 1.55236,-0.74983 0.69887,0.1792 0.74695,-1.30546 1.36168,-0.49299 0.62122,0.65999 1.1806,-0.4728 0.35139,-0.62271 -0.4378,0.23119 -0.35205,-0.68673 -0.66301,-0.91157 -0.0951,-0.86591 -0.42699,-1.59301 -0.77256,-2.37608 -0.28776,-0.47507 -0.86916,-1.26771 -1.4645,-1.28674 z",
2.56824, 2.063,
"M -1.65563 -2.06134  c -1.25863,0.0605 -1.00656,1.90514 -0.4826,2.70738 1.27622,-0.20063 1.59649,2.35287 2.91245,1.0399 0.34188,-0.16365 0.38346,-0.31718 0.30757,-0.47967 0.0248,-0.0232 0.0505,-0.0458 0.0754,-0.0689 0.78294,-0.61042 2.07913,-1.20867 0.99887,-2.30315 -0.1618,-0.25964 -0.35631,-0.35126 -0.56606,-0.33394 -0.4303,0.0354 -0.91952,0.5265 -1.3292,0.96373 -0.28645,-0.79722 -0.7443,-1.53257 -1.64403,-1.51297 -0.0978,-0.0128 -0.18852,-0.0162 -0.27243,-0.0128 z",
2.69804, 3.17552,
"M -0.907231 -3.17534  c -0.24929,0.007 -0.46483,0.2028 -0.59435,0.71789 -0.1246,0.75535 -1.2685,1.31142 -0.66913,2.14191 -1.32502,1.15191 0.1607,2.97757 1.02977,3.49106 l 0.1303,-0.11175 0.15128,-0.13697 c 0.8895,-0.21167 1.24411,-0.76409 1.26851,-1.69723 0.19571,-0.87034 1.02336,-1.14446 1.34665,-0.12867 0.20799,-0.56415 1.74212,-1.10319 0.39257,-1.55429 -0.99736,-0.1586 -0.23361,-1.73999 -1.4904,-1.42319 -0.30638,-0.417 -1.01682,-1.31425 -1.56523,-1.29875 z",
2.85984, 5.28311,
"M 2.73793 -5.28207  c -0.015,-0.002 -0.0319,-10e-4 -0.0501,0.002 l -0.34004,0.31678 c -0.17511,0.027 -0.38357,0.16702 -0.5669,0.42116 -0.73686,-0.0649 -1.46563,0.13236 -1.66706,0.13746 -0.67315,0.017 -0.52935,1.98311 -0.49609,2.62775 0.0175,0.11077 0.0314,0.16367 0.0424,0.19378 -0.0163,0.0123 -0.0413,0.0577 -0.0791,0.18656 -0.0357,0.12199 -0.10261,0.35732 -0.14883,0.51521 -0.17169,-0.11412 -0.34298,-0.19051 -0.50125,-0.18604 -0.24929,0.007 -0.46477,0.2027 -0.59428,0.71779 -0.12462,0.75535 -1.2686,1.3115 -0.66921,2.14199 -1.32504,1.15191 0.16084,2.97725 1.02989,3.49074 l 0.13022,-0.11162 0.1509,-0.13694 c 0.8895,-0.21167 1.24425,-0.76392 1.26867,-1.69706 0.19572,-0.87034 1.02339,-1.14446 1.34668,-0.12867 0.20799,-0.56415 1.74228,-1.10333 0.39275,-1.55443 -0.98558,-0.15672 -0.25485,-1.69952 -1.45005,-1.4304 0.0337,-0.19455 0.0813,-0.37702 0.15608,-0.53589 0.30318,-0.62037 1.08029,-1.8405 1.08415,-2.40243 0.005,-0.76342 0.83722,-0.66127 0.91932,-1.45315 0.13685,-0.53732 0.26681,-1.08517 0.0419,-1.11414 z",
2.25069, 2.15736,
"M -1.31572 -2.15659  c -0.1647,-0.007 -0.28795,0.0329 -0.32919,0.13591 -0.0487,0.11989 0.0156,0.32374 0.25373,0.63355 -0.18714,0.21042 -0.36899,0.44338 -0.58703,0.64182 -0.85828,0.33362 0.60076,1.16166 0.42995,1.80145 0.20592,0.53574 0.73787,1.3365 1.35496,0.78031 l 0.16846,0.32091 c 0.59552,-0.0337 1.38631,-0.20816 2.03242,-1.02836 l -0.19119,-0.8568 c -0.137,-0.23584 -0.2603,-0.40707 -0.37311,-0.53743 0.37941,-0.26434 1.13768,-0.0887 0.64751,-0.82579 -0.0527,-0.0367 -0.10821,-0.0708 -0.1633,-0.10542 0.49546,-0.53719 -0.1007,-0.64494 -0.7457,-0.38034 -0.43736,-0.17954 -0.90406,-0.30825 -1.35702,-0.38189 -0.0559,-0.008 -0.10962,-0.0114 -0.16124,-0.0119 -0.0978,-10e-4 -0.18828,0.009 -0.27284,0.0295 -0.24662,-0.12465 -0.50609,-0.20615 -0.70591,-0.21549 z",
3.67275, 3.21569,
"M -0.0460498 -3.21533  c -0.11996,0.001 -0.23781,0.009 -0.32142,0.0129 -0.29093,0.0121 -0.53765,0.0695 -0.3235,0.53536 -0.22627,0.0914 -0.90234,0.55329 -1.04024,-0.0388 -0.31879,-0.0384 -0.77856,0.15926 -0.93379,0.43718 -0.43863,-0.21291 -0.84127,0.042 -0.99891,0.44856 -0.043,0.32893 0.0751,0.60333 0.26046,0.84749 -0.009,0.3245 0.1084,0.47004 0.30022,0.49558 l -0.38343,0.45217 c 0.29165,0.62004 0.78181,0.43344 1.16014,0.0506 0.0781,0.0553 0.15775,0.10901 0.23977,0.16071 -0.53198,0.40496 -0.7605,0.97563 0.31369,1.28933 l 0.72294,-0.49765 c 0.0642,-0.0825 0.11237,-0.15383 0.16278,-0.22686 0.0327,0.0105 0.0654,0.0181 0.0982,0.0279 0.16068,0.39255 0.77761,0.90608 0.66302,1.33532 0.20593,0.53574 0.73787,1.3365 1.35496,0.78032 l 0.16846,0.32091 c 0.59552,-0.0337 1.38581,-0.20816 2.03192,-1.02837 l -0.19069,-0.85679 c -0.137,-0.23584 -0.2608,-0.40707 -0.37361,-0.53743 0.37941,-0.26434 1.13818,-0.0887 0.64801,-0.82579 -0.0527,-0.0367 -0.10871,-0.0708 -0.16383,-0.10542 0.49549,-0.5372 -0.10067,-0.64494 -0.74567,-0.38034 -0.43736,-0.17954 -0.90406,-0.30825 -1.35702,-0.38189 -0.0559,-0.008 -0.10962,-0.0114 -0.16124,-0.0119 -0.0978,-0.001 -0.18828,0.009 -0.27284,0.0294 -0.0795,-0.0402 -0.15954,-0.0736 -0.23926,-0.10387 0.11316,-0.10188 0.2712,-0.17912 0.47128,-0.21652 0.0483,-0.28576 0.18888,-1.14252 -0.003,-1.27021 -0.32433,-0.21564 -0.36465,-0.44676 -0.55502,-0.44183 -0.10408,0.003 -0.0796,-0.25731 -0.19894,-0.27234 -0.0916,-0.0237 -0.21338,-0.0292 -0.33332,-0.0279 z",
4.37172, 2.82177,
"M 4.15107 -2.76363  c -0.12588,-0.061 -0.31491,-0.0778 -0.59006,-0.0321 -0.39183,1.50176 -2.22417,-0.002 -2.85969,1.25239 -0.22059,0.57212 0.36731,0.82629 -0.0102,1.2242 -0.47996,0.50597 -1.83247,1.95136 -2.83873,1.05509 -1.37002,-0.71385 -1.45547,1.3313 -2.22411,1.81568 0.29352,0.31468 0.80913,0.29068 1.2014,0.23962 1.08757,-0.8648 2.32743,0.27342 3.58569,-0.0204 0.84516,-0.67246 1.49671,-1.57107 2.62485,-1.94719 0.46688,-0.61992 1.62754,-0.4097 1.06677,-1.52838 -0.0528,-0.66878 0.58863,-1.79182 0.0441,-2.05885 z",
4.37157, 4.1273,
"M 0.620322 -4.12594  c -0.55557,0.0573 -1.03581,1.84758 -1.54823,2.21485 0.3837,0.6876 0.283,1.30919 0.0419,1.9172 -0.42523,0.16787 -0.86971,0.17674 -1.26193,-0.1726 -1.37001,-0.71385 -1.45499,1.33153 -2.22363,1.81591 0.29352,0.31468 0.8087,0.29032 1.20097,0.23926 0.48778,-0.38788 1.00761,-0.36748 1.54461,-0.25941 -0.14566,0.3988 -0.2277,0.81069 -0.15658,1.24953 0.36957,0.23628 0.50355,0.43109 0.52038,0.60462 l 0.91723,0.64388 c 0.68612,-0.27271 2.29931,-0.23369 1.41335,-1.51153 0.38381,-0.70654 1.05535,-1.29742 0.97981,-2.16938 0.29503,-0.21386 0.61597,-0.39961 0.99218,-0.52503 0.46689,-0.61992 1.62738,-0.40991 1.06659,-1.52859 -0.0528,-0.66878 0.58846,-1.79176 0.0439,-2.05879 -0.12586,-0.061 -0.31498,-0.0777 -0.59012,-0.032 -0.31404,1.2036 -1.55213,0.47891 -2.35903,0.78858 -0.16962,-0.30286 -0.26535,-0.61458 -0.18346,-0.94051 -0.13758,-0.21015 -0.26969,-0.28917 -0.39791,-0.27595 z",
2.40615, 2.34567,
"M 1.22057 -2.34541  c -0.11996,0.001 -0.23783,0.009 -0.32144,0.0129 -0.29094,0.0121 -0.53763,0.0695 -0.32348,0.53537 -0.22627,0.0914 -0.90236,0.55323 -1.04026,-0.0388 -0.3188,-0.0384 -0.77856,0.15925 -0.9338,0.43718 -0.43862,-0.21292 -0.84126,0.042 -0.9989,0.44855 -0.043,0.32893 0.0751,0.60334 0.26045,0.8475 -0.009,0.3245 0.10843,0.47003 0.30025,0.49557 l -0.38343,0.45217 c 0.29165,0.62004 0.78179,0.4334 1.16012,0.0506 0.0781,0.0553 0.15777,0.10901 0.23979,0.16071 -0.53197,0.40497 -0.76052,0.97563 0.31366,1.28933 l 0.72295,-0.49764 c 0.0642,-0.0825 0.11239,-0.15383 0.1628,-0.22686 0.21412,0.0689 0.42973,0.123 0.64233,0.15606 l 0.0419,-0.0403 c 0.16619,-0.0829 0.52523,-0.15309 0.76946,-0.27698 0.23728,0.032 0.30295,-0.10345 0.24339,-0.30644 -0.006,-0.0574 -0.03,-0.12095 -0.0847,-0.19327 -2.9e-4,-2.4e-4 -7.2e-4,-2.8e-4 -10e-4,-5.2e-4 -0.0729,-0.13224 -0.17968,-0.27385 -0.3147,-0.40928 -0.1447,-0.38814 0.092,-0.78214 0.63511,-0.88366 0.0483,-0.28576 0.18899,-1.14252 -0.003,-1.27021 -0.32432,-0.21564 -0.36462,-0.44676 -0.55499,-0.44184 -0.10408,0.003 -0.0796,-0.25731 -0.19896,-0.27233 -0.0916,-0.0237 -0.21336,-0.0292 -0.3333,-0.0279 z",
3.50783, 3.58322,
"M 2.96957 -3.58221  c -0.054,-0.004 -0.10962,0.004 -0.16682,0.0204 -0.50287,0.14531 -1.11843,0.96574 -1.59686,0.75172 -0.10443,0.2409 -0.0893,0.42169 -0.004,0.56879 -0.16015,0.0374 -0.31561,0.12717 -0.45888,0.27949 -0.18095,-0.16288 -0.36634,-0.32735 -0.52012,-0.46987 -0.20437,-0.0568 -0.39359,-0.0485 -0.5582,7.6e-4 -0.3902,0.11676 -0.63677,0.46713 -0.60175,0.74583 -0.82773,-0.51376 -1.62373,-1.30541 -2.57037,-1.06536 h -4e-4 c 0.26652,0.79551 1.77909,1.50405 1.31935,2.59942 -0.29573,1.43857 0.96759,1.36306 1.9262,0.86162 -0.19051,0.27421 -0.33048,0.55606 -0.0695,0.87617 0.2907,0.72136 0.51069,1.5495 1.19922,1.99646 0.20081,-0.39583 0.52596,-0.80888 0.35365,-1.27772 -0.24696,-0.37352 -0.82161,-0.82316 -0.85651,-1.19804 0.67455,0.0369 1.28244,0.50499 1.61609,-0.44907 0.0895,-0.17713 0.12714,-0.35251 0.13189,-0.52562 0.5227,-0.002 0.92832,-0.16026 0.80786,-0.83572 -0.02,-0.11225 -0.0568,-0.22688 -0.10279,-0.34112 0.29762,0.56825 0.4526,0.29332 0.43494,-0.27949 0.56932,-0.62085 -0.0145,-1.20088 0.14248,-1.85712 -0.11901,-0.28003 -0.26387,-0.39114 -0.4255,-0.40157 z m -3.15799,2.26732 c -0.008,0.02 -0.0114,0.0391 -0.0165,0.0585 -0.0297,-0.0163 -0.0596,-0.0336 -0.0906,-0.0557 0.0354,-6.9e-4 0.0686,8.4e-4 0.10709,-0.003 z",
3.7615, 3.60448,
"M 2.71615 -3.60346  c -0.054,-0.004 -0.10972,0.004 -0.1669,0.0202 -0.32652,0.0944 -0.70003,0.46997 -1.05317,0.6656 -0.0392,-0.0864 -0.0913,-0.17253 -0.16949,-0.2589 -0.13989,0.11483 -0.28059,0.23472 -0.44546,0.30385 -0.37044,0.0914 -0.55509,0.29917 -0.72657,0.58653 -0.0611,-0.0552 -0.13033,-0.11623 -0.185,-0.16691 -0.20439,-0.0568 -0.39351,-0.0488 -0.55811,5.1e-4 -0.35949,0.10757 -0.59423,0.413107 -0.60203,0.678517 -0.0136,0.0138 -0.0277,0.0274 -0.0398,0.0413 -0.81341,-0.513967 -1.59923,-1.275407 -2.53061,-1.039227 h -5e-4 c 0.26651,0.795517 1.77903,1.503967 1.31929,2.599337 -0.29573,1.43857 0.9679,1.3634 1.92651,0.86196 -0.19053,0.27421 -0.33076,0.5558 -0.0698,0.87592 0.012,0.0298 0.0233,0.0613 0.0351,0.0915 -0.14874,0.19134 -0.58581,0.55204 -0.12969,0.59428 0.55962,-0.15196 0.50268,0.68369 0.64232,1.02681 0.27718,0.29651 0.87279,0.45861 1.25006,0.19069 0.25513,-0.26582 0.48609,-6.5e-4 0.76999,-0.005 0.16695,-0.18991 0.32821,-0.0961 0.49556,-0.0584 0.71837,-0.71386 0.97589,-1.56297 1.28365,-2.4846 -0.70366,-0.43516 -0.773,-0.65102 -0.81441,-1.42007 l -0.0104,-0.33796 c 0.0471,-0.0983 0.0696,-0.27697 0.0625,-0.50798 0.56933,-0.62085 -0.0149,-1.200497 0.14211,-1.856747 -0.11901,-0.28003 -0.26366,-0.39109 -0.42529,-0.40152 z",
2.73778, 2.64431,
"M -1.85179 2.64296  c 0.2123,0.0405 0.70405,-0.84096 1.03766,-0.92501 0.33497,0.30812 0.76833,0.15488 0.79944,-0.33693 -0.20608,-0.37546 0.3914,-0.98771 0.45321,-0.40411 0.57615,0.94326 1.00706,-0.63268 1.58337,-0.71623 0.18095,0.0986 0.3791,0.08 0.57309,0.10232 0.38124,-1.55226 -0.0574,-1.82885 -0.65008,-1.86604 0.0889,-0.23201 0.18206,-0.4544 0.29609,-0.63924 -0.83934,-0.59743 -2.12622,-0.83961 -2.63705,0.29611 -0.0335,0.0842 -0.0746,0.16672 -0.12041,0.24856 -0.21743,0.0386 -0.42992,0.0931 -0.60772,0.19482 -0.90374,-0.25609 -0.4962,0.82188 -0.69969,1.33377 -0.68622,0.72594 -1.23518,1.47159 -0.69607,2.40967 0.21852,-0.0632 0.39881,-0.18586 0.55965,-0.33848 0.002,0.0764 0.006,0.15365 0.006,0.22945 -0.008,0.28213 0.0321,0.39784 0.10282,0.41134 z",
2.73778, 2.64431,
"M -1.85179 2.64296  c 0.2123,0.0405 0.70405,-0.84096 1.03766,-0.92501 0.33497,0.30812 0.76833,0.15488 0.79944,-0.33693 -0.20608,-0.37546 0.3914,-0.98771 0.45321,-0.40411 0.57615,0.94326 1.00706,-0.63268 1.58337,-0.71623 0.18095,0.0986 0.3791,0.08 0.57309,0.10232 0.38124,-1.55226 -0.0574,-1.82885 -0.65008,-1.86604 0.0889,-0.23201 0.18206,-0.4544 0.29609,-0.63924 -0.83934,-0.59743 -2.12622,-0.83961 -2.63705,0.29611 -0.0335,0.0842 -0.0746,0.16672 -0.12041,0.24856 -0.21743,0.0386 -0.42992,0.0931 -0.60772,0.19482 -0.90374,-0.25609 -0.4962,0.82188 -0.69969,1.33377 -0.68622,0.72594 -1.23518,1.47159 -0.69607,2.40967 0.21852,-0.0632 0.39881,-0.18586 0.55965,-0.33848 0.002,0.0764 0.006,0.15365 0.006,0.22945 -0.008,0.28213 0.0321,0.39784 0.10282,0.41134 z",
2.72104, 3.39079,
"M 0.286178 -3.39079  c -0.14007,0.11497 -0.28086,0.23514 -0.44596,0.30437 -0.43748,0.10788 -0.61692,0.37041 -0.8201,0.74518 -0.20775,0.47047 -0.48958,-0.22623 -0.70953,-0.14625 -0.17844,0.25772 -0.66421,0.48419 -0.68522,0.74259 0.48728,-0.0488 0.29403,0.53326 0.005,0.63149 -0.19034,0.18451 -0.59938,0.72542 -0.14417,0.90072 0.44725,-0.12109 0.48747,0.26394 0.52038,0.58136 -0.074,-0.44637 0.62326,-0.57247 0.80926,-0.27647 -0.19566,0.38226 -0.23154,0.87394 -0.36896,1.30586 -0.1065,0.17514 -0.65649,0.59395 -0.16746,0.63924 0.55962,-0.15196 0.50271,0.68369 0.64236,1.02681 0.27718,0.29651 0.87278,0.45861 1.25005,0.19069 0.25514,-0.26582 0.48606,-0.001 0.76996,-0.005 0.16695,-0.18991 0.32824,-0.0961 0.49559,-0.0584 0.71837,-0.71386 0.97589,-1.56297 1.28366,-2.4846 -0.70366,-0.43516 -0.77306,-0.65079 -0.81447,-1.41984 l -0.0209,-0.66544 c -0.14473,-0.20288 -0.29131,-0.40206 -0.49768,-0.48162 -0.51565,0.12706 -0.62662,-0.15055 -0.84926,-0.31646 -0.0172,-0.40469 0.11379,-0.80938 -0.25281,-1.21407 z",
2.54738, 3.09564,
"M -2.54738 -1.21382  c 0.12657,0.31725 -0.0579,0.48889 0.29639,1.02693 0.28022,0.42555 1.10094,1.23068 1.25829,1.46812 0.39733,0.5996 0.15592,0.35314 -0.13001,1.1135 l 0.41975,0.48888 c 0.0197,-0.0589 0.63189,-0.11044 0.65209,-0.17001 0.0222,0.0431 0.0455,0.0859 0.0661,0.12919 0.0877,0.19747 0.15333,0.26483 0.2067,0.25115 0.16015,-0.0429 0.20981,-0.81511 0.41291,-0.98341 0.33369,0.10267 0.58274,-0.14635 0.44182,-0.49713 -0.2667,-0.19187 -0.0552,-0.81306 0.18034,-0.42943 0.71052,0.46274 0.48797,-0.77085 0.8594,-1.01906 0.15785,0.008 0.28929,-0.0697 0.43098,-0.11834 -0.24882,-1.20063 -0.644,-1.24748 -1.06662,-1.07745 -0.0152,-0.19 -0.0245,-0.3749 -0.006,-0.54054 -0.15658,-0.0274 -0.32065,-0.0374 -0.48318,-0.0305 l 0.0367,-0.18293 c -0.41522,-0.73638 -1.14845,-2.2421 -2.38538,-0.50798 -0.386,0.22163 0.0136,0.52023 -0.16174,0.80202 -0.29782,0.28407 -0.20672,-0.38153 -0.40463,-0.14883 -0.23569,0.10251 -0.42783,0.26499 -0.62375,0.42581 z",
4.16689, 3.81427,
"M -1.95926 -3.81051  c -0.29384,0.0315 -0.63098,0.25701 -1.01751,0.79892 -0.386,0.22163 0.0136,0.52023 -0.16173,0.80202 -0.29785,0.28407 -0.20672,-0.38153 -0.40463,-0.14883 -0.23569,0.10251 -0.42783,0.265 -0.62376,0.42582 0.12658,0.31725 -0.0582,0.48929 0.29612,1.02733 0.0583,0.0886 0.14341,0.19507 0.23771,0.30799 -0.0747,9e-4 -0.14412,0.0272 -0.19894,0.0687 h -5.3e-4 c -0.17727,0.13413 -0.20177,0.41817 0.24545,0.45268 -0.39825,0.48749 0.43336,0.66351 0.60359,1.13223 -0.0301,0.59814 0.20492,1.21839 -0.0532,1.81385 -0.21561,0.33237 0.23479,0.51094 0.38809,0.76015 0.0564,0.0501 0.0678,0.1023 0.10335,0.154 h 5.3e-4 c 0.4419,0.12956 0.9933,-0.19327 1.50844,-0.36587 0.76589,-0.53181 1.53929,-1.25257 2.29753,-1.59473 0.2199,0.13691 0.43069,0.36556 0.6263,0.4749 0.45749,0.20053 0.51117,0.84887 1.06042,0.88005 0.3851,0.004 0.70237,-0.18814 0.91467,-0.51573 -0.35343,-0.19801 -0.27543,-0.7836 0.0424,-1.07745 0.47445,-0.42372 0.22762,-1.0411 -0.13642,-1.4304 -0.40209,-0.32194 0.51559,-0.78393 -0.0935,-0.79737 -0.15158,-0.14276 -0.26191,-0.30716 -0.3855,-0.45837 l -0.23204,0.0605 c -0.52811,0.011 -1.14326,-0.0144 -1.2087,0.19172 l -0.31109,0.56431 c -0.38256,0.0826 -0.41923,0.31447 -1.16274,0.24132 -0.0683,0.071 -0.12409,0.1551 -0.17259,0.24754 l -0.003,-0.004 c 0.0758,-0.27877 0.11591,-0.63338 0.29712,-0.75448 0.15785,0.008 0.2893,-0.0697 0.43098,-0.11834 -0.24881,-1.20062 -0.64397,-1.24748 -1.06658,-1.07745 -0.0152,-0.19 -0.0242,-0.37489 -0.006,-0.54053 -0.15658,-0.0274 -0.32065,-0.0374 -0.48319,-0.0305 l 0.0367,-0.18294 c -0.28546,-0.50626 -0.72144,-1.37616 -1.36787,-1.3069 z",
3.3826, 3.20591,
"M 0.917115 -3.20545  c -0.0393,0.67856 -0.8055,0.38483 -1.10476,0.76992 0.21062,0.50964 -0.42472,0.8769 -0.74454,1.18986 -0.0806,-0.39166 -0.0452,-0.8526 -0.44329,-0.91195 -0.65614,0.0726 -0.81839,-0.6789 -1.28317,-0.87078 -0.2656,0.13569 -0.49393,0.3231 -0.72396,0.51189 0.16808,0.42128 -0.0773,0.64903 0.39322,1.36349 0.3721,0.56508 1.46193,1.63419 1.67087,1.94948 0.52763,0.79619 0.20746,0.4691 -0.17223,1.47876 l 0.55718,0.64914 c 0.0262,-0.0782 0.83916,-0.14666 0.86601,-0.22576 0.0295,0.0572 0.0604,0.11406 0.0878,0.17155 0.11646,0.26222 0.20361,0.35166 0.27446,0.33349 0.21267,-0.057 0.27862,-1.08235 0.54829,-1.30583 0.44311,0.13633 0.77382,-0.19433 0.5867,-0.66013 -0.35415,-0.25477 -0.0733,-1.07964 0.23947,-0.57022 0.9435,0.61446 0.64796,-1.0236 1.14116,-1.35319 0.2096,0.0106 0.38415,-0.0925 0.57228,-0.15714 -0.33038,-1.59428 -0.85513,-1.65649 -1.41632,-1.43072 -0.017,-0.21332 -0.024,-0.41843 -0.012,-0.61071 -0.22056,0.0324 -0.42912,0.0689 -0.60935,0.12495 -0.11459,-0.34405 -0.26659,-0.45533 -0.42751,-0.44603 z",
3.6319, 4.78495,
"M 0.896993 -4.78491  c -1.04618,0.86919 1.22964,1.02801 0.62706,2.09804 -0.0111,0.14066 -0.0168,0.27728 -0.0167,0.40806 -0.33465,0.21103 -0.62095,0.44688 -0.76564,0.63005 -0.40971,0.0176 -0.98566,0.24842 -1.41538,0.56247 -0.0885,-0.0278 -0.17012,-0.0331 -0.23481,-0.009 l -4.4e-4,-4.3e-4 c -0.10214,0.0385 -0.1614,0.15187 -0.14414,0.36058 -0.0621,0.0771 -0.11487,0.15557 -0.14971,0.23526 -0.0962,-0.007 -0.18664,0.003 -0.26946,0.0282 v -4.3e-4 c -0.51314,0.15257 -0.75518,0.85767 -0.51544,1.44917 0.0468,0.13673 0.0411,0.38968 0.0235,0.67624 -0.563,0.31662 -0.99981,0.90929 -1.60573,1.11511 0.0839,0.86303 -0.5451,2.78068 0.94615,1.68955 l 0.27419,-0.26605 c 0.59919,-0.18144 1.14669,-0.60514 1.40596,-1.11211 0.70056,-0.69037 0.4631,-1.79139 0.86788,-2.69387 0.38804,-0.19262 0.71359,-0.58299 1.22675,-0.0295 0.503,0.48469 0.58214,-0.85407 0.69164,-1.07533 0.24966,-0.71685 1.63714,-0.11067 1.72634,-0.99534 0.0961,-0.14176 0.083,-0.33963 -0.0458,-0.59669 -0.0144,-0.0633 -0.0351,-0.11846 -0.0594,-0.1681 -0.009,-0.0297 -0.0165,-0.0595 -0.0243,-0.0894 -0.22286,-1.14696 -1.72872,-1.83888 -2.54117,-2.21652 z m -1.97015,4.93649 c 0.004,0.004 0.009,0.008 0.0133,0.0124 -0.0133,0.047 -0.0288,0.0941 -0.044,0.14115 0.009,-0.0514 0.0193,-0.10257 0.0308,-0.15356 z",
2.20024, 4.20283,
"M 0.680372 -4.20263  c -0.17134,-0.003 -0.35314,0.0281 -0.51107,0.0956 h -0.002 c -0.37912,0.16215 -0.6214,0.53304 -0.25527,1.13533 0.0428,0.30239 0.0412,0.55368 -5.3e-4,0.76378 -0.20661,0.20366 -0.36107,0.43947 -0.42891,0.67541 -0.25368,0.17815 -0.60508,0.27842 -1.04386,0.33693 -0.81735,0.77798 1.48963,1.90805 0.56173,2.45463 -0.32718,0.0973 -0.67947,0.11365 -0.9953,0.0207 -0.64842,0.99087 0.36237,2.72148 1.30794,2.69131 l 0.25683,-0.16433 0.13385,-0.19999 c 0.0466,-0.0443 0.0923,-0.0929 0.13795,-0.14159 0.17233,0.17619 0.32211,0.35812 0.41807,0.55397 0.56925,-0.35075 1.47415,0.57288 1.80713,-0.0207 -1.12112,-0.65447 -1.73331,-1.46106 -0.64026,-2.46807 -0.1388,-0.15461 -0.33557,-0.2819 -0.49663,-0.40411 0.0189,-0.22272 0.0125,-0.44358 -0.0258,-0.65784 0.0824,-0.0552 0.18539,-0.11327 0.31935,-0.17622 0.92186,-0.47154 0.76654,-2.17728 0.97665,-2.5032 -0.36297,-0.32872 -0.97897,0.11256 -1.34102,-0.54364 -0.0805,0.02 -0.16004,0.0479 -0.23873,0.0796 0.0343,-0.20274 0.15489,-0.40552 0.40256,-0.60616 0.59182,-0.58976 0.17092,-0.91141 -0.34314,-0.92139 z",
2.88699, 5.39383,
"M 1.3671 -5.39363  c -0.17122,-0.003 -0.3528,0.0282 -0.51057,0.0956 h -0.003 c -0.37912,0.16215 -0.62142,0.53304 -0.25529,1.13533 0.0428,0.30239 0.0412,0.55368 0,0.76378 -0.20662,0.20366 -0.36108,0.43947 -0.42892,0.67541 -0.25368,0.17815 -0.6051,0.27842 -1.04386,0.33693 -0.81735,0.77798 1.4896,1.90805 0.56171,2.45463 -0.32718,0.0973 -0.67945,0.11362 -0.99528,0.0207 -0.14761,0.22555 -0.20556,0.49003 -0.20154,0.76275 -0.051,0.0146 -0.10085,0.0324 -0.14729,0.0579 -0.23484,0.6226 -0.50334,1.23094 -0.7865,1.83296 -0.19645,0.25916 -0.14105,0.63084 -0.33951,0.83613 -0.11512,0.138 -0.11041,0.31176 -0.0977,0.48472 l 1.21801,1.32705 c 0.11266,-0.0498 0.24308,-0.0702 0.37931,-0.0165 0.29117,0.0276 0.68913,0.0748 0.86971,-0.23874 0.30268,-0.33688 0.72999,0.11684 1.03198,-0.17725 0.082,-0.27859 0.31242,-0.40815 0.41085,-0.68988 0.33589,-0.752 0.92244,-0.0445 1.15649,-0.42471 0.35317,-0.13097 0.0367,-1.3743 -0.16174,-1.55863 0.0322,-0.0731 0.0626,-0.14742 0.0972,-0.21911 0.2109,0.004 0.28993,-0.14967 0.11834,-0.38292 -0.0568,-0.22605 0.26795,-0.75271 -0.0594,-0.80564 -0.2096,0.24628 -0.25599,0.65901 -0.5855,0.79892 -0.1397,-0.40155 -0.0145,-0.84498 0.51882,-1.33635 -0.1388,-0.15462 -0.33554,-0.2819 -0.49659,-0.40412 0.0188,-0.22271 0.0124,-0.44358 -0.0259,-0.65784 0.0824,-0.0552 0.18539,-0.11326 0.31938,-0.17621 0.92183,-0.47154 0.76655,-2.17729 0.97668,-2.50321 -0.36298,-0.32872 -0.97896,0.11256 -1.34101,-0.54364 -0.0805,0.02 -0.16005,0.0479 -0.23874,0.0796 0.0343,-0.20274 0.15489,-0.40552 0.40254,-0.60616 0.59163,-0.58956 0.17121,-0.9112 -0.34261,-0.92139 z",
2.2926, 5.13722,
"M 0.802274 -5.13667  c -0.15648,0.008 -0.31396,0.11235 -0.39378,0.36329 0.17158,0.46045 -0.71247,0.26413 -0.98288,0.53433 -0.18795,-0.0953 -0.37721,-0.18365 -0.56276,-0.25735 -0.30393,-0.35352 -0.57627,-0.46456 -0.78031,-0.41599 h -0.001 c -0.34007,0.0809 -0.48919,0.60594 -0.27337,1.19424 1.00303,1.38655 -0.11896,2.78412 0.99219,3.28301 0.18788,-0.0184 0.35102,-0.008 0.49866,0.0202 0.0272,0.0107 0.055,0.0168 0.0842,0.017 1.09791,0.28456 1.07577,1.83267 1.1281,2.82101 1.29392,0.57385 -0.73445,2.82023 1.25884,2.61276 0.17748,0.17897 0.47238,0.11556 0.51522,-0.14728 0.081,-1.48014 -0.80518,-2.85605 -0.76896,-4.31808 0.16182,-0.1731 0.22826,-0.33544 0.2403,-0.49299 0.17579,-0.13499 0.36803,-0.25536 0.53588,-0.29662 -0.45833,-0.34903 -0.47617,-0.81306 -0.45474,-1.07591 -1.13773,-0.23427 0.49343,-1.18353 -0.54364,-1.63607 0.26771,-0.66998 -1.19856,-0.90889 -0.31369,-1.41387 0.50499,-0.3651 0.16598,-0.81 -0.17827,-0.79168 z",
4.57828, 3.5071,
"M 2.91251 -3.50687  c -0.4486,0.0106 0.45675,0.54041 0.18913,0.76585 -0.86641,-0.29492 -1.26553,1.09463 -1.96578,0.93793 -0.86302,0.0543 -1.79475,0.32429 -2.49597,0.83405 -0.39846,-0.0495 -0.7705,-0.009 -1.1157,0.12454 -0.006,0.0507 0.007,0.10957 -0.003,0.15813 0.0966,0.98514 -1.0613,0.70185 -1.47535,0.23151 -0.37552,-0.57285 -1.03049,0.23261 -0.27905,0.28888 -0.31933,0.39262 0.19955,0.57156 0.46199,0.88831 -0.002,0.0106 -0.005,0.0172 -0.006,0.0284 0.0115,0.11487 0.0542,0.20711 0.11732,0.28474 0.007,0.52372 0.17613,1.06164 -0.0476,1.58079 -0.20304,0.31453 0.22357,0.4826 0.36896,0.71778 0.0533,0.0472 0.064,0.0964 0.0977,0.14521 h 5e-4 c 0.41788,0.12151 0.93829,-0.18412 1.42473,-0.3483 0.40148,-0.2801 0.80497,-0.61424 1.20716,-0.9126 0.50226,-0.12855 1.03791,-0.70742 1.45571,-0.97152 -0.39362,-0.88195 1.08802,-0.72117 1.19787,-1.3348 -0.12419,-0.6673 -0.7335,-1.91974 0.42995,-1.96474 0.86471,-0.32419 0.78534,1.59558 1.21698,0.39687 -0.25257,-0.73375 0.38536,-0.64026 0.82942,-0.33848 0.28241,-0.45436 -0.55153,-1.06703 -1.10898,-1.37252 v 5.1e-4 c -0.0792,-0.0443 -0.15266,-0.0811 -0.21706,-0.1111 -0.12671,-0.0221 -0.21858,-0.031 -0.28266,-0.0295 z",
3.74688, 2.41634,
"M 2.72883 -2.41634  -0.35677,0.016 c -0.33291,0.36885 -0.63684,0.43941 -0.88624,0.44063 -0.0292,-0.0173 -0.0593,-0.0354 -0.0868,-0.0506 v 3.3e-4 c -0.051,-0.0285 -0.0983,-0.0522 -0.13977,-0.0716 -0.0816,-0.0142 -0.14078,-0.02 -0.18205,-0.019 -0.10932,0.003 -0.0935,0.0532 -0.0413,0.1228 -0.0466,0.007 -0.0874,0.0211 -0.1198,0.0519 -0.0676,0.10644 -0.089,0.22694 -0.0958,0.35209 -0.32395,0.17106 -0.55217,0.6454 -0.88722,0.57042 -0.5558,0.035 -1.15583,0.20884 -1.6074,0.53712 -0.25661,-0.0319 -0.49621,-0.006 -0.7185,0.0802 -0.004,0.0326 0.005,0.0706 -0.002,0.10184 0.0622,0.63443 -0.68347,0.45199 -0.95012,0.14909 -0.24183,-0.36891 -0.66363,0.14979 -0.17971,0.18604 -0.20564,0.25284 0.12851,0.36808 0.29752,0.57207 -0.001,0.007 -0.003,0.0111 -0.004,0.0183 0.007,0.074 0.0349,0.13337 0.0755,0.18337 0.005,0.33727 0.1135,0.68368 -0.0306,1.01801 -0.13076,0.20256 0.14398,0.3108 0.23761,0.46226 0.0343,0.0304 0.0412,0.0621 0.0629,0.0935 h 3.2e-4 c 0.26912,0.0782 0.60427,-0.11857 0.91752,-0.2243 0.25856,-0.18038 0.5184,-0.39557 0.77741,-0.58771 0.32345,-0.0828 0.66843,-0.45558 0.93749,-0.62566 -0.18482,-0.41411 0.27077,-0.47225 0.55943,-0.62599 0.042,0.26188 -0.16431,0.57971 -0.0466,0.81602 -0.0361,0.3214 0.2697,0.38876 0.29186,0.65727 -0.0155,0.41805 0.31928,0.10309 0.49155,0.006 0.24347,0.16794 0.51027,0.0323 0.71949,-0.1644 0.15749,-0.44026 0.53019,-0.15638 0.81901,-0.0952 0.098,-0.20391 0.35342,-0.32496 0.4526,-0.4992 -0.0531,-0.28513 0.23897,-0.57539 0.41732,-0.72149 v -3.4e-4 c 0.3295,-0.64851 0.11566,-1.00699 -0.21897,-1.51055 l 0.28918,-0.44994 0.22499,-0.29984 c -0.18304,-0.14669 -0.3661,-0.12842 -0.54912,-0.0343 l -0.38038,0.11848 c 0.28124,-0.43579 0.14347,-0.51087 -0.0885,-0.57341 z",
1.5426, 3.75375,
"M -0.18215 -3.74938  c -0.43006,0.0954 -0.0238,1.66304 -0.29143,2.06565 -1.23283,0.67467 0.70648,1.73509 -0.66571,2.29879 -1.27771,1.27642 0.87232,1.5933 1.05811,1.42669 -1.72666,0.27299 0.54523,3.0984 0.51438,0.84809 0.0778,-1.31836 1.51452,-1.87061 0.99737,-3.31344 -0.0269,-1.19813 0.0509,-2.86032 -1.37276,-3.26853 -0.10038,-0.0537 -0.17819,-0.0707 -0.23966,-0.0572 z",
2.6102, 2.40582,
"M -2.60391 1.07532  1.21829,1.32714 c 0.11265,-0.0498 0.24287,-0.0704 0.37911,-0.0167 0.29115,0.0276 0.68939,0.0748 0.86997,-0.23871 0.30266,-0.33688 0.72982,0.11701 1.03183,-0.17708 0.082,-0.27859 0.31221,-0.40824 0.41064,-0.68997 0.0829,-0.35912 0.3014,-0.69002 0.49691,-1.00001 0.25988,-0.36573 0.40154,-0.79981 0.5954,-1.20228 0.21088,0.004 0.29,-0.14987 0.11841,-0.38313 -0.0568,-0.22605 0.26786,-0.75294 -0.0595,-0.80587 -0.21044,0.24728 -0.25622,0.66273 -0.58937,0.80098 -0.0881,0.24739 -0.32811,0.21877 -0.5183,0.099 -0.43381,-0.0984 -0.16894,-0.56282 -0.0503,-0.81595 -0.18037,-0.10226 -0.47473,0.10933 -0.66436,0.16038 0.14099,0.42689 -0.26992,0.60149 -0.59482,0.72943 -0.10138,0.24534 -0.36733,-0.0479 -0.17905,-0.12239 0.19862,-0.22162 0.39983,-0.47805 0.66131,-0.63217 0.23857,-0.24684 -0.1743,-0.27544 -0.34069,-0.2571 -0.28559,0.11508 -0.52696,7.5e-4 -0.74386,-0.19182 -0.1358,-0.11552 -0.22022,-0.0715 -0.27824,0.0899 -0.15588,0.0841 -0.36986,0.0793 -0.5395,0.17231 -0.23485,0.6226 -0.50305,1.23123 -0.78622,1.83325 -0.19645,0.25916 -0.14117,0.63063 -0.33965,0.83592 -0.11511,0.138 -0.11073,0.31194 -0.098,0.4849 z",
3.83978, 2.30066,
"M 1.89988 -2.28596  c -0.46473,-0.0533 -0.93451,0.0301 -1.2983,0.38777 -0.61395,-0.10651 -1.14946,0.60908 -1.8877,0.45323 -0.65836,-0.53157 -1.19607,0.74513 -1.84876,0.2517 -0.22541,0.36322 -0.81294,0.89409 -0.68754,1.46714 0.20541,0.12631 0.2374,0.23071 0.21384,0.31715 l 0.28852,0.42359 5.18412,1.28604 c -0.12156,-0.38789 0.57265,-0.65546 1.06815,-0.66503 0.3002,-0.42158 1.17579,-0.37128 0.82666,-1.04765 -0.64248,0.006 -0.4043,-0.75074 -0.0849,-0.96787 z",
2.48198, 2.67999,
"M -0.947181 2.42314  c 0.18893,-0.12837 0.39015,-0.1443 0.5237,0.25685 0.51028,-0.15853 1.30811,-0.11944 1.84667,-0.42769 1.22161,0.34616 0.47447,-1.31052 0.96133,-1.80427 0.18748,-0.84884 0.0458,-1.7551 0.0512,-2.62474 0.0372,-1.37538 -0.84814,0.48939 -1.39027,0.62596 -0.40821,-0.37552 -0.93592,-0.18917 -0.97385,0.41022 0.25116,0.45758 -0.477,1.20377 -0.55234,0.49252 -0.7022,-1.14958 -1.22788,0.77102 -1.93025,0.87284 -0.42731,0.88954 1.20329,1.9184 1.46381,2.19831 z",
4.26056, 3.08976,
"M -2.80517 -3.08919  c -0.26757,0.53645 -1.24706,-0.0831 -1.33583,0.28009 0.54354,0.28297 0.25361,0.49804 -0.0894,0.49764 -0.15417,0.30689 0.31539,0.50952 0.66921,0.73174 -0.43301,0.39009 -0.69014,0.69733 -0.58912,0.83612 0.13288,0.3976 0.66614,0.90708 0.14624,0.88781 -0.3356,0.1386 0.28826,0.9763 0.62944,1.30069 0.0448,-0.0159 0.0277,0.18537 0.2382,0.28009 0.59182,-0.12048 0.63855,0.74505 1.24386,0.67799 0.29986,0.0447 0.27308,0.39182 0.33383,0.68678 0.2408,-0.23569 0.71932,-0.51227 0.5607,-0.89607 0.24093,-0.31002 0.88424,-0.0359 0.83043,-0.6966 0.27241,-0.35349 0.87897,-0.0301 1.26195,-0.22117 0.25652,0.75949 1.11287,0.11487 0.97616,-0.46044 0.16131,-0.52891 0.74101,-0.64995 0.6723,-0.0481 0.77777,0.57404 1.08543,-0.60114 0.85577,-1.1653 -0.30509,-0.30498 -0.47604,-0.87464 0.16433,-0.83974 0.20368,0.0876 0.36298,0.0215 0.49766,-0.10077 -0.0896,-0.39051 -1.13962,0.24844 -1.35134,-0.41238 -0.39409,-0.50595 -0.71183,-0.003 -0.74157,0.42736 -0.27477,0.54361 -1.36311,1.08291 -1.44074,0.11938 0.49909,-0.34336 0.0274,-0.70102 -0.47749,-0.47543 -0.34422,0.48082 -1.39303,0.25803 -0.98856,-0.43408 -0.39558,-0.16344 -0.54981,-0.83329 -0.99528,-0.80667 -0.46308,0.47435 -0.67871,-0.19078 -1.07075,-0.16898 z",
2.85227, 3.44376,
"M 2.39399 0.825598  c 0.51165,-1.00701 0.17934,-1.56385 -0.34027,-2.34578 l 0.44899,-0.69865 0.34956,-0.46536 c -0.28419,-0.22778 -0.56839,-0.19965 -0.85258,-0.0535 l -0.59069,0.18407 c 0.43669,-0.6767 0.22244,-0.79303 -0.13779,-0.89014 l -0.55392,0.0248 c -1.05242,1.16603 -1.92058,0.41374 -2.26059,0.73685 -0.31726,0.4997 -0.002,1.19184 -0.27419,1.71558 -0.0601,0.59507 -1.33712,0.72874 -0.96788,1.37654 0.68403,0.42661 -0.0301,1.17727 0.22283,1.68512 -0.056,0.49908 0.41904,0.60361 0.45347,1.02055 -0.0241,0.64915 0.49545,0.16052 0.76293,0.009 0.37808,0.26076 0.79269,0.0499 1.11759,-0.25546 0.24455,-0.68364 0.82318,-0.24258 1.27166,-0.14754 0.15219,-0.31662 0.54842,-0.50454 0.70243,-0.77509 -0.0825,-0.44275 0.37149,-0.89378 0.64845,-1.12065 z",
3.408, 3.57197,
"M 2.62753 -3.57195  c -0.64407,0.7879 -0.64716,1.84506 -1.20222,2.58759 -1.34442,-0.27745 -1.52748,0.69891 -1.31733,1.75739 -0.78212,0.16845 -1.74702,0.16312 -2.57312,0.0427 -0.21732,0.51378 -0.40761,0.43217 -0.59858,0.21424 -1.0749,0.8342 0.63602,1.73677 1.46282,2.1222 1.23346,0.6503 2.55782,-0.35092 3.77246,0.4198 0.2868,-0.0617 0.92584,0.13183 0.99013,-0.27867 -0.83826,-0.73953 -2.26236,-1.79647 -1.23986,-3.02911 0.18907,-1.17518 1.78492,-1.4423 1.271,-2.84888 0.10716,-0.40834 0.32162,-0.44801 0.15222,-0.57722 -0.14475,-0.11041 -0.6654,-0.33083 -0.71752,-0.41005 z",
2.16401, 4.29525,
"M 1.65199 -3.9757  c -0.1751,0.027 -0.38331,0.16713 -0.56662,0.42127 -0.73687,-0.0649 -1.46581,0.13196 -1.66725,0.13706 -0.67315,0.017 -0.52923,1.98326 -0.49599,2.6279 0.0831,0.52678 0.10231,-0.0945 -0.0368,0.38051 -0.25456,0.86925 -0.82603,2.79733 -0.99814,3.68266 -0.0219,0.0245 -0.0307,0.0543 -0.0512,0.0793 0.26405,0.40175 0.43883,0.65708 0.62561,0.94225 1.81312,-0.12134 0.99017,-2.46123 1.53551,-3.61943 0.3032,-0.62038 1.0801,-1.84071 1.08397,-2.40264 0.005,-0.76342 0.83708,-0.66154 0.9192,-1.45341 0.14595,-0.57315 0.28406,-1.1579 -0.008,-1.11255 z",
2.42152, 3.19498,
"M 1.79275 3.19496  c 0.0358,-0.35442 0.0143,-0.94751 0.14977,-1.20163 0.29216,-0.54826 0.76509,-0.61761 0.24914,-1.04756 l -0.46191,-0.29639 c -0.0199,-0.009 -0.0339,-0.008 -0.0548,-0.0191 0.008,-0.67733 -0.40627,-1.15393 -0.87696,-1.61178 -0.0135,-0.57623 -0.52506,-0.68936 -0.93404,-0.86224 -0.3143,-0.11364 -0.30235,-0.2211 -0.26928,-0.50192 0.001,-0.35694 -0.19325,-0.6476 -0.63636,-0.70807 -0.52435,-0.42985 -0.56303,0.22718 -0.74414,0.56327 0.27001,0.2644 0.29136,0.75233 -0.0961,1.0294 -0.41132,0.0707 -0.85183,1.03036 -0.23048,1.04386 0.0806,0.4722 0.45958,1.10992 0.004,1.55133 -0.1628,0.45669 0.19683,1.40675 0.90123,1.35186 0.45163,-0.12672 0.99602,0.14704 1.99196,-0.16774 0.14912,0.4712 0.54383,0.84975 1.00786,0.87673 z",
2.60961, 2.6311,
"M 2.50705 -1.19561  -0.25339,-0.1347 c -0.4738,-0.41913 -0.99447,-0.94138 -1.21945,-0.8131 l -0.73175,0.24406 c -0.40381,-0.23794 -0.62439,-0.0636 -1.21942,-0.73175 -0.45864,0.0258 -0.91727,0.38107 -1.37591,0.59171 -0.0117,0.11781 -0.0183,0.23633 -0.0649,0.35298 -0.27654,0.35904 -0.46239,0.91911 0.20227,0.9358 0.65658,0.0504 0.51399,0.93494 0.66863,1.40794 0.23986,0.54801 -0.23986,1.16298 0.21828,1.63643 0.33587,0.31601 0.77101,0.40448 1.22387,0.28852 -0.1503,-0.46126 0.39437,-0.91344 0.91269,-0.91403 0.76165,0.0123 1.04546,-0.73144 1.04116,-1.36969 -0.0925,-0.60988 1.09071,-0.27154 0.56549,-0.77799 -0.0175,-0.24875 0.0184,-0.48273 0.0324,-0.71618 z",
2.91238, 3.10557,
"M -2.9122 1.73814  c 0.28378,0.47289 1.03118,0.63668 1.62475,0.90306 1.10867,0.1542 2.37832,0.14746 3.32429,0.46437 0.29646,-0.1176 0.26342,-0.96007 0.29166,-1.66776 l 0.0964,-0.71277 c 0.0983,-0.21454 0.10887,-0.66087 0.22079,-0.8799 0.101,-0.19778 0.31208,-0.17115 0.25795,-0.31721 -0.37435,-0.55516 -0.83678,-0.93415 -1.38048,-1.15063 -0.0489,0.0422 -0.0876,0.10835 -0.13886,0.14474 -0.75872,0.99895 -1.59127,-0.26181 -1.57102,-1.05567 0.14438,-0.85593 -1.15818,-0.67142 -0.50818,0.0282 -0.74639,0.10568 -0.15692,0.9358 -0.38769,1.48657 -0.51224,0.50199 -0.80922,1.23824 -1.51997,1.55271 -0.45966,0.11748 -0.20829,0.6411 -0.27572,0.98493 0.009,0.0899 -0.0234,0.14503 -0.0341,0.21939 z",
3.04347, 2.84609,
"M 0.490388 -2.84481  c -0.17818,-0.0381 -0.59112,0.79085 -0.87113,0.86987 -0.28109,-0.28971 -0.64432,-0.14548 -0.67045,0.31688 0.17296,0.35303 -0.32861,0.92862 -0.38047,0.37989 -0.4836,-0.88691 -0.84563,0.59496 -1.32932,0.67352 -0.34102,0.46556 0.0163,0.50226 0.25082,0.78429 -0.88963,0.69447 -0.64769,1.26964 0.58717,1.61261 0.64827,-0.33916 1.89403,-1.4075 1.52897,0.052 0.13212,0.76935 0.98486,1.48997 1.55068,0.56922 -0.30948,-1.21831 1.0167,-2.14737 1.88681,-2.26869 -0.414,-0.6498 -0.0862,-1.85279 -0.94851,-2.09149 -0.29062,-0.0875 -1.51076,0.0901 -1.51829,-0.51083 0.006,-0.26528 -0.0269,-0.37452 -0.0863,-0.38722 z",
2.60354, 3.87863,
"M -2.58741 2.42622  5e-5,-5.1e-4 c 0.0545,-0.28255 0.24695,-0.66339 0.55812,-1.04225 0.58545,-0.71275 0.69899,-1.23193 0.40168,-1.83633 -0.12573,-0.25554 -0.10279,-0.32366 0.0655,-0.19436 0.14242,0.10941 0.35573,0.007 0.47396,-0.22648 0.13004,-0.25726 0.0644,-0.53105 -0.16725,-0.69336 -0.61079,-0.42848 -0.34025,-1.8077 0.39236,-1.99985 0.39389,-0.1033 0.67136,-0.0486 0.77194,0.15245 0.10567,0.21111 0.29186,0.16355 0.57011,-0.14555 0.38478,-0.42741 0.42246,-0.42673 0.54193,0.0137 0.0706,0.26033 0.052,0.57305 -0.0418,0.69491 -0.0936,0.12183 0.004,0.1741 0.21775,0.11617 0.29389,-0.0797 0.38668,0.21619 0.38089,1.21691 -0.004,0.7281 0.16873,1.53845 0.38543,1.80433 0.21632,0.26545 0.31499,0.58433 0.21915,0.70905 -0.0958,0.12468 -0.0624,0.2383 0.074,0.2522 0.13645,0.0139 0.2856,0.26879 0.33189,0.56623 0.0463,0.29751 -0.0165,0.46361 -0.13944,0.36914 -0.12297,-0.0945 -0.23633,-0.046 -0.25195,0.10731 -0.0156,0.15332 -0.25555,0.25548 -0.53324,0.22719 -0.90574,-0.0923 -1.36908,0.17162 -1.43488,0.81735 -0.0549,0.53905 -0.17952,0.61343 -0.83029,0.49663 -0.42102,-0.0756 -0.70687,-0.21398 -0.63513,-0.3073 0.15929,-0.20731 -0.61961,-0.67165 -1.08123,-0.6447 -0.24076,0.0141 -0.3242,-0.17035 -0.26966,-0.4529 z",
3.89459, 2.76207,
"M 2.11989 -2.76207  h 1.36873 l -0.35549,1.28958 c -0.29269,1.05994 -0.25675,1.36166 0.20266,1.6976 0.30734,0.22472 0.5588,0.76635 0.5588,1.20359 0,0.73869 -0.10906,0.78002 -1.53862,0.58406 -1.12744,-0.15454 -1.65256,-0.0734 -1.96439,0.30227 -0.42093,0.50722 -1.08901,0.5684 -2.78519,0.25655 -1.43006,-0.26295 -1.74284,-0.69806 -1.33735,-1.86135 0.19689,-0.56461 0.55188,-1.02688 0.78878,-1.02688 0.23697,0 0.64286,-0.3974 0.90268,-0.88292 0.48391,-0.90417 1.76118,-1.45844 2.42085,-1.05075 0.20303,0.12556 0.36914,0.0615 0.36914,-0.14193 0,-0.20347 0.61632,-0.36982 1.3694,-0.36982 z",
2.94033, 4.04841,
"M 0.955924 -4.04819  c 0.27815,0.007 0.71396,0.18533 1.21285,0.53472 0.89248,0.62514 1.02887,1.65306 0.31401,2.36795 -0.76145,0.76144 -1.35601,2.05174 -1.12098,2.432 0.1187,0.19201 0.0454,0.34882 -0.16219,0.34882 -0.20765,0 -0.37588,0.28073 -0.37181,0.62384 0.004,0.53339 0.0558,0.54652 0.35298,0.0891 0.26617,-0.40941 0.32982,-0.30503 0.27295,0.4456 -0.0609,0.80483 -0.24384,0.98032 -1.02141,0.98032 -0.52094,0 -1.4289,0.0959 -2.01772,0.21375 -0.62778,0.12557 -1.17184,0.0511 -1.31523,-0.18103 -0.13461,-0.2177 0.0843,-0.81432 0.48665,-1.32566 0.40219,-0.51129 0.6373,-1.08236 0.52222,-1.26856 -0.11506,-0.18621 0.0922,-0.41733 0.46092,-0.51383 0.46375,-0.12126 0.65661,-0.47079 0.62453,-1.1342 -0.0293,-0.6002 0.27891,-1.26441 0.82367,-1.77476 0.4875,-0.45671 0.78245,-1.04493 0.67046,-1.33678 -0.13082,-0.34108 -0.0108,-0.50804 0.2681,-0.5013 z",
2.50487, 5.03272,
"M -2.16669 -3.17022  1.86046,-0.51083 c 1.02338,-0.28104 1.94395,-0.73695 2.04586,-1.01295 0.10187,-0.27602 0.34634,-0.41088 0.54358,-0.29979 0.19726,0.11107 0.27852,0.40249 0.17931,0.64757 -0.0993,0.24506 -0.39051,1.29581 -0.64811,2.33521 -0.319,1.28729 -0.645,1.9083 -1.02188,1.94754 -0.38441,0.0401 -0.38656,0.0854 -0.007,0.14812 0.29981,0.0496 0.43437,0.35011 0.29918,0.66799 -0.23474,0.55145 -0.25461,0.55221 -0.43484,0.0127 -0.12461,-0.37299 -0.49795,0.24186 -1.09772,1.80788 -0.49978,1.30521 -1.09181,2.41091 -1.31488,2.45715 -0.22322,0.0463 -0.0643,-0.59646 0.35215,-1.4284 0.47376,-0.94633 0.58524,-1.53429 0.29799,-1.5706 -0.9072,-0.11464 -1.60579,-1.99425 -1.33221,-3.5826 z",
4.0742, 4.47677,
"M -1.73269 -4.47592  v 5.3e-4 c -0.11923,-0.0214 -0.34907,0.20596 -0.62117,0.66663 -0.42148,0.71353 -0.27757,1.66663 0.30591,2.02726 0.18712,0.11561 0.15346,0.28882 -0.11213,0.5824 -0.75139,0.83028 -2.02901,3.17126 -1.90582,3.49229 0.21982,0.57286 0.9162,0.20501 1.12292,-0.59273 0.10928,-0.42187 0.30009,-0.76688 0.42323,-0.76688 0.46329,0 0.96755,1.15024 0.89659,2.04432 -0.0619,0.78083 -0.006,0.90333 0.4005,0.88005 0.25974,-0.0148 0.64453,-0.10023 0.85474,-0.19017 0.26972,-0.11518 0.42074,-0.0149 0.51416,0.34262 0.22016,0.8419 0.94316,0.48133 1.16377,-0.58033 0.10563,-0.50859 0.28131,-1.03924 0.39066,-1.17926 0.10925,-0.14 0.27067,-0.55414 0.35812,-0.91984 0.11033,-0.4615 0.40817,-0.76905 0.97308,-1.0051 0.52028,-0.21739 0.86677,-0.55204 0.96118,-0.92811 0.23675,-0.94326 -0.0311,-1.2976 -0.98084,-1.2976 -0.49064,0 -1.15258,-0.21143 -1.55027,-0.49454 -0.69181,-0.49257 -0.89485,-0.48303 -1.68363,0.0796 -0.35702,0.25464 -0.49911,0.1962 -1.0015,-0.41186 -0.38391,-0.46468 -0.54345,-0.8907 -0.46455,-1.2392 0.0738,-0.32572 0.0477,-0.49342 -0.045,-0.51005 z",
4.05861, 3.1259,
"M 1.4673 -3.12553  c 0.2083,-0.0104 0.30102,0.19858 0.4976,0.72986 l 0.28366,0.76589 0.85394,-0.10578 c 0.75247,-0.0929 0.83822,-0.0672 0.72384,0.2093 -0.0771,0.18457 -0.0519,0.26844 0.0579,0.20316 0.12017,-0.07 0.17818,0.0104 0.17417,0.18171 v -7.9e-4 c 0,0.17241 -0.0697,0.43828 -0.19554,0.74443 -0.13945,0.33975 -0.31333,0.45866 -0.61256,0.41861 -0.29078,-0.0393 -0.443,0.0567 -0.49759,0.31661 -0.043,0.20532 -0.16706,0.68434 -0.27521,1.06489 l -0.19702,0.6923 -1.07262,-0.0537 c -1.68416,-0.0835 -2.45547,0.0134 -2.35518,0.29672 0.0623,0.17504 -0.11423,0.24332 -0.61168,0.22847 -0.63289,-0.0166 -0.67621,0.0119 -0.45086,0.30435 0.21928,0.2847 0.20904,0.30719 -0.0905,0.18553 -0.18708,-0.0764 -0.48128,-0.1495 -0.65307,-0.1495 -0.27046,-0.0104 -0.27432,-0.0337 -0.0267,-0.18475 0.25384,-0.15875 0.17922,-0.25993 -0.10237,-0.14406 -0.0593,0.0242 -0.19242,-0.11646 -0.29508,-0.31204 -0.13945,-0.26497 -0.12017,-0.39565 0.0742,-0.51137 0.20429,-0.12255 0.18916,-0.15151 -0.0682,-0.16636 -0.22417,-0.007 -0.32431,-0.16022 -0.31586,-0.48071 0,-0.25783 0.0979,-0.46381 0.20014,-0.46381 0.10385,0.003 0.18975,-0.0786 0.19242,-0.18248 0,-0.10355 -0.13204,-0.19547 -0.29983,-0.19547 -0.24183,-0.006 -0.25607,-0.0623 -0.0757,-0.26911 0.17714,-0.20219 0.17032,-0.29634 -0.0282,-0.42243 -0.48024,-0.30624 -0.47979,-0.7686 0,-1.16533 0.25591,-0.21095 0.46895,-0.51184 0.47296,-0.66929 0,-0.20785 0.12759,-0.25156 0.45234,-0.16253 0.24509,0.0682 0.50427,0.0347 0.57652,-0.0751 0.0727,-0.11053 0.38795,-0.19961 0.70232,-0.18477 0.42357,0.0104 0.54225,0.0849 0.45917,0.2852 -0.0905,0.21809 -0.0208,0.23926 0.36348,0.1043 0.35042,-0.12269 0.43854,-0.10875 0.33425,0.0499 -0.0994,0.15132 -0.0326,0.17466 0.22238,0.0782 0.1994,-0.0751 0.63898,-0.20767 0.97664,-0.2967 0.74565,-0.19402 0.8271,-0.41151 0.10089,-0.2691 l -0.51896,0.10192 0.47608,-0.26372 c 0.25518,-0.1408 0.42148,-0.2323 0.54655,-0.2323 z",
4.06584, 3.46902,
"M 0.0899704 -3.46718  c -0.17768,-0.01 -0.28056,0.0213 -0.24431,0.083 0.0487,0.0831 -0.11034,0.13842 -0.35322,0.13842 -0.4581,-0.0115 -0.94486,0.37267 -0.74196,0.58647 0.0645,0.0674 0.0315,0.11778 -0.0731,0.11778 -0.1046,-0.003 -0.23714,0.15712 -0.29461,0.35544 -0.0745,0.25648 -0.19559,0.32418 -0.41912,0.23103 -0.25478,-0.1056 -0.29634,-0.0555 -0.22139,0.26509 0.0588,0.25046 0.0143,0.3895 -0.1347,0.3895 -0.12466,-0.003 -0.36123,0.13641 -0.52717,0.30953 -0.27842,0.29188 -0.27469,0.33282 0.0487,0.54646 0.19201,0.1271 0.34547,0.35797 0.34132,0.51242 0,0.15476 0.076,0.23203 0.17696,0.17327 0.10175,-0.059 0.18084,0.0149 0.17697,0.16438 0,0.14962 -0.0946,0.32273 -0.19989,0.38431 -0.13183,0.0768 -0.12753,0.19543 0.0143,0.37395 0.32527,0.41241 -0.0745,0.29467 -0.72864,-0.21473 -0.71245,-0.55498 -1.05936,-0.39598 -0.95892,0.43985 0.0416,0.33988 0.12324,0.48366 0.20807,0.35468 0.0888,-0.13655 0.20691,-0.004 0.32584,0.36877 0.10174,0.32039 0.28573,0.55263 0.40724,0.51536 0.34705,-0.10575 0.38459,-0.42635 0.0874,-0.73974 -0.43074,-0.4534 -0.31081,-0.73099 0.15547,-0.35986 0.25348,0.20188 0.38044,0.45716 0.32427,0.65161 -0.0502,0.17568 2.9e-4,0.32435 0.12036,0.32433 0.1175,0.003 0.16537,-0.076 0.10747,-0.17623 -0.0602,-0.1046 0.0143,-0.13327 0.17182,-0.0674 0.15345,0.0636 0.28113,0.0207 0.28428,-0.0956 0,-0.12179 0.0774,-0.1647 0.16508,-0.10732 0.0874,0.058 0.79011,0.13068 1.5603,0.16217 0.87538,0.0363 1.30439,0.11506 1.14548,0.20955 -0.1943,0.1165 -0.16808,0.15227 0.1089,0.1666 0.29763,0.0129 0.31238,0.0459 0.086,0.1792 -0.19932,0.1188 -0.22081,0.21924 -0.0803,0.36729 0.0602,0.0639 0.31711,0.0949 0.49235,0.14587 -0.0258,-0.0297 -0.0559,-0.0606 -0.0545,-0.13326 0,-0.1003 0.0903,-0.17623 0.19631,-0.17623 0.0788,0.003 0.1175,0.0486 0.1046,0.11334 0,0.0217 -0.0143,0.0455 -0.0301,0.0703 -0.0516,0.0791 -0.11177,0.12094 -0.1592,0.14286 0.55597,0.14794 1.3051,0.23898 1.40841,0.0807 0.0602,-0.0917 0.0272,-0.21988 -0.0731,-0.28583 -0.33674,-0.22037 0.1784,-0.8133 0.69525,-0.79898 0.23472,0.006 0.34648,-0.01 0.35695,-0.0652 v 0 c 0,-0.0331 -0.0229,-0.0801 -0.0846,-0.14512 -0.11607,-0.12266 -0.20835,-0.36821 -0.20362,-0.54575 0,-0.19462 -0.17309,-0.37354 -0.4495,-0.45093 -0.25205,-0.0703 -0.41684,-0.16677 -0.36583,-0.21549 0.0516,-0.0483 -0.0229,-0.21043 -0.16507,-0.35988 -0.14186,-0.14945 -0.21293,-0.34046 -0.15848,-0.42356 0.0545,-0.0832 0,-0.21354 -0.11033,-0.28804 -0.14187,-0.0937 -0.16236,-0.0546 -0.0616,0.12222 0.12466,0.22596 0.10173,0.22495 -0.16135,0 -0.24246,-0.20762 -0.25721,-0.29549 -0.0702,-0.40726 0.21766,-0.12996 0.0803,-0.22733 -0.2585,-0.1829 -0.0745,0.01 -0.13327,-0.10374 -0.1304,-0.25324 0,-0.15017 -0.0702,-0.27694 -0.16593,-0.27694 -0.0946,-0.003 -0.37127,-0.21307 -0.61386,-0.46872 -0.2426,-0.2556 -0.39377,-0.46206 -0.33617,-0.46206 0.0573,0.002 0.0143,-0.11506 -0.1003,-0.25843 -0.14042,-0.17871 -0.1443,-0.29802 -0.0143,-0.37395 0.37657,-0.21937 0.1976,-0.59398 -0.32656,-0.68568 -0.0731,-0.01 -0.14043,-0.0222 -0.19918,-0.0236 z",
3.04742, 4.1149,
"M 0.969015 -4.11451  c -0.1123,0.003 -0.4062,0.23381 -0.6532,0.51211 -0.247,0.27833 -0.4917,0.4659 -0.5436,0.41704 -0.052,-0.0489 -0.092,-0.0185 -0.089,0.0677 0,0.0862 -0.1232,0.13629 -0.2795,0.11111 -0.1564,-0.0252 -0.2829,0.0298 -0.2812,0.12247 0,0.10163 -0.054,0.0859 -0.141,-0.0393 -0.1244,-0.18039 -0.1424,-0.17678 -0.1385,0.0248 0,0.12772 -0.089,0.27161 -0.2042,0.31987 -0.1152,0.0483 -0.2111,0.0261 -0.2134,-0.0491 0,-0.0753 -0.1476,-0.13195 -0.3225,-0.12505 -0.2601,0.009 -0.2789,0.0366 -0.1017,0.14882 0.1624,0.10279 0.1233,0.11844 -0.1572,0.0625 -0.3368,-0.0673 -0.3789,-0.0286 -0.4299,0.39739 -0.081,0.67937 0.181,2.0534 0.4382,2.29547 0.2153,0.2027 0.2296,0.52597 0.052,1.16531 -0.07,0.25284 -0.1989,0.3266 -0.6031,0.34623 -0.4968,0.0241 -0.4903,0.0295 0.199,0.1819 1.3697,0.30279 2.4788,0.44216 2.7978,0.35191 0.1769,-0.05 0.674,-0.0115 1.1048,0.0858 l 0.7829,0.17673 -0.011,0.54881 c -0.051,2.51094 0.8952,0.2636 0.8723,-1.92856 h -5e-4 l -0.017,-1.60197 -0.6072,0.0429 c -0.3341,0.0235 -0.6756,0.0384 -0.7586,0.0336 -0.1011,-0.006 -0.1467,-0.54478 -0.138,-1.63452 0.013,-1.66481 -0.091,-2.04807 -0.5566,-2.03399 z",
5.34994, 2.90409,
"M 0.205271 -2.88963  c -0.13829,-0.0359 -0.26164,-0.004 -0.39597,0.0846 -0.4055,0.27123 -0.57988,1.88599 -0.20283,1.87455 0.14783,-0.004 0.19695,0.23609 0.12558,0.61444 -0.22207,1.18362 -0.24909,1.20421 -0.83216,0.65553 -1.004,-0.9451 -2.68326,-0.94665 -3.80823,-0.007 -0.588,0.49193 -0.58959,1.01822 0,1.10322 0.48466,0.0703 2.31733,0.0446 3.47298,-0.0477 1.30444,-0.10428 2.47502,0.0916 2.48901,0.41565 0.0207,0.47846 0.40456,0.83765 0.9455,0.88472 0.2971,0.0257 1.17457,0.10237 1.94935,0.17084 0.97395,0.0862 1.40585,0.0494 1.4014,-0.1208 v -7.4e-4 c 0,-0.13544 -0.24559,-0.23879 -0.53554,-0.23 -0.51615,0.0159 -0.52728,-0.021 -0.47228,-1.65113 l 0.0572,-1.66673 -0.87063,0.0172 c -1.71646,0.0362 -1.88385,-0.0251 -1.9042,-0.69824 -0.0159,-0.34413 -0.10173,-0.62322 -0.20124,-0.6202 -0.10015,0.003 -0.42332,-0.20574 -0.71803,-0.46412 -0.20188,-0.17696 -0.35766,-0.27952 -0.49691,-0.31544 z",
4.12566, 4.16038,
"M -4.12566 1.26026  c 0.3983,1.33704 0.5968,1.95916 0.8005,2.60191 0.1973,-0.15081 0.2728,-0.16272 0.4356,0.0408 0.1118,0.13999 0.2564,0.25733 0.3214,0.25733 0.2494,0.007 0.5568,-0.47474 0.585,-0.91673 0.016,-0.25389 0.1112,-0.52612 0.2113,-0.60513 0.1002,-0.0788 0.3032,-0.36546 0.4507,-0.63717 0.2314,-0.41949 0.2393,-0.48358 0.058,-0.41963 -0.11549,0.041 -0.176,0.0232 -0.1349,-0.0398 0.041,-0.063 0.2731,-0.0781 0.51471,-0.032 0.37159,0.0703 0.50419,0.0242 0.85939,-0.29872 0.2311,-0.2101 0.5599,-0.49394 0.7307,-0.63045 0.1707,-0.13658 0.2623,-0.25061 0.2036,-0.25061 -0.059,-0.002 0.01,-0.0802 0.14211,-0.17517 0.13679,-0.0944 0.2916,-0.14405 0.34419,-0.10907 0.053,0.034 0.2139,0.0285 0.3576,-0.0134 0.211,-0.0609 0.24771,-0.14503 0.1917,-0.43562 -0.059,-0.30721 0.028,-0.46208 0.5891,-1.05009 0.3622,-0.37938 0.6599,-0.741607 0.6615,-0.804607 0,-0.063 0.043,-0.21795 0.092,-0.34417 0.072,-0.18478 0.016,-0.16969 -0.2863,0.077 l -0.3762,0.30645 0.3152,-0.33952 c 0.1734,-0.18614 0.3417,-0.31097 0.3746,-0.27596 0.033,0.035 0.1383,0.002 0.2336,-0.0729 0.2257,-0.1773 0.2371,-0.4254 0.013,-0.28317 -0.092,0.0587 -0.077,0.0236 0.034,-0.078 0.11061,-0.10141 0.2745,-0.33917 0.3649,-0.52813 l 0.16431,-0.34367 -0.6878,-0.0176 c -0.57011,-0.0152 -0.7141,0.0264 -0.8377,0.24342 -0.082,0.14406 -0.1236,0.37092 -0.093,0.50382 0.036,0.15298 -0.01,0.2372 -0.1214,0.2372 -0.2107,-0.006 -0.2346,0.14356 -0.053,0.33435 0.071,0.0753 0.1477,0.31057 0.169,0.52298 0.045,0.453997 -0.308,0.827227 -0.7953,0.840237 -0.1728,0.005 -0.3814,0.0724 -0.4636,0.15039 -0.2358,0.22381 -0.5701,0.15658 -0.8759,-0.17674 -0.2915,-0.31779 -0.8445,-0.54271 -0.8495,-0.3457 0,0.0616 -0.2271,0.21883 -0.5003,0.34932 -0.5806,0.27771 -0.6869,0.47366 -0.6015,1.1152 0.075,0.56536 0.01,0.62764 -1.1157,1.05936 -0.4885,0.18751 -0.9565,0.38981 -1.4252,0.58394 z",
3.86911, 4.42196,
"M -3.86911 -0.871814  1.5374,2.2846 c 0.2008,-0.0776 0.3166,-0.10729 0.602,-0.2408 0.5019,-0.2349 0.9751,-0.3859 1.0516,-0.33592 0.082,0.0531 0.1067,0.2639 0.061,0.50541 -0.056,0.29721 -0.026,0.43696 0.1075,0.49197 0.1027,0.0425 0.1471,0.13787 0.098,0.21188 -0.061,0.093 -0.015,0.0986 0.1499,0.0196 0.3172,-0.15314 1.3733,-0.36761 1.7591,-0.35761 0.1701,0.005 0.4193,0.13343 0.5544,0.28628 0.1977,0.2236 0.2285,0.38401 0.1566,0.82322 -0.071,0.43418 -0.048,0.56343 0.1142,0.63045 0.1121,0.0465 0.1708,0.13486 0.1308,0.19584 -0.04,0.0607 0.016,0.11679 0.124,0.11679 0.1083,0.003 0.1966,0.0477 0.1953,0.0998 0,0.12689 0.7177,0.55224 0.9426,0.56224 0.1639,0.004 0.1918,-0.43416 0.1075,-1.68103 0,-0.057 -0.064,-0.0689 -0.1354,-0.0279 -0.071,0.0416 -0.1333,0.0151 -0.1406,-0.0589 -0.01,-0.0739 -0.025,-0.1915 -0.04,-0.26148 -0.016,-0.07 -0.034,-0.18423 -0.041,-0.25424 -0.01,-0.0699 -0.1177,-0.1328 -0.2459,-0.1328 -0.5387,-0.0141 -0.9,-1.15501 -0.4197,-1.325 0.3033,-0.1074 0.1118,-0.30189 -0.2402,-0.2439 -0.2282,0.0375 -0.3614,-0.0331 -0.5065,-0.26768 -0.2727,-0.44103 -0.3275,-0.86317 -0.1121,-0.8532 0.096,0.002 0.2993,-0.16322 0.4532,-0.36793 0.1542,-0.2046 0.2449,-0.37412 0.2015,-0.37412 -0.044,-0.001 0.084,-0.13883 0.2837,-0.30541 0.069,-0.0573 0.084,-0.0527 0.1308,-0.0884 -0.2037,-0.64275 -0.4022,-1.26486 -0.8005,-2.60191 -0.3301,0.13674 -0.6584,0.26051 -1.0041,0.41082 -0.2463,0.1071 -0.4953,0.2654 -0.5529,0.35139 -0.058,0.0861 -0.1658,0.15298 -0.2403,0.15298 -0.082,-0.002 -0.064,-0.0868 0.043,-0.2124 0.1554,-0.18108 0.1516,-0.19894 -0.029,-0.13693 -0.1142,0.039 -0.26049,0.0158 -0.325,-0.0522 -0.1486,-0.15661 -0.8354,0.43305 -1.0434,0.89556 -0.115,0.25451 -0.1159,0.351 0,0.40101 0.083,0.037 0.048,0.0698 -0.077,0.0698 -0.232,0.004 -0.2745,0.0305 -0.9012,0.55758 -0.19939,0.16759 -0.4385,0.30075 -0.5318,0.30075 -0.093,-0.002 -0.2048,0.0796 -0.24749,0.18293 -0.043,0.10311 -0.20891,0.18034 -0.36951,0.18034 -0.16049,-0.004 -0.3865,0.0966 -0.50179,0.22429 -0.09,0.0994 -0.2028,0.17031 -0.29871,0.22429 z",
5.21206, 3.61562,
"M 5.21167 3.31641  -2.87571,-2.99268 c -0.10415,0.0389 -0.20387,0.0802 -0.36033,0.0775 -0.39269,-0.007 -0.9471,0.19838 -1.34636,0.49674 -0.14729,-0.0882 -0.27752,-0.24828 -0.39406,-0.40426 0.0978,-0.0826 0.13979,-0.19254 0.0398,-0.3869 -0.11869,0.002 -0.19394,-0.0372 -0.23982,-0.0956 0.0252,-0.0258 0.0176,-0.0405 -0.0283,-0.0478 -0.013,-0.0295 -0.0214,-0.0615 -0.0245,-0.096 0.13076,-0.0385 0.17321,-0.0758 0.0928,-0.0834 -0.026,-0.002 -0.0566,-0.0159 -0.0886,-0.0336 0.0543,-0.31133 0.38549,-0.64795 0.53791,-0.19087 0.30229,-0.0939 0.77511,-0.0703 1.09423,-0.25331 0.7238,0.2051 0.28056,-0.77685 0.56901,-1.06934 0.11112,-0.50295 0.0275,-1.03897 0.0306,-1.55424 0.0222,-0.81487 -0.50294,0.28945 -0.82428,0.37028 -0.24188,-0.2226 -0.5545,-0.11178 -0.57697,0.24343 0.14866,0.27116 -0.28257,0.71309 -0.32723,0.29163 -0.41601,-0.68112 -0.72709,0.45649 -1.14325,0.5169 -0.44399,-0.24164 -0.97952,-0.15946 -1.35193,0.20668 -0.44324,-0.0768 -0.82988,0.4394 -1.36296,0.32681 -0.47543,-0.38381 -0.86283,0.53807 -1.33412,0.18179 -0.16273,0.26228 -0.58769,0.64528 -0.49715,1.05906 0.23477,0.14434 0.20747,0.26981 0.0897,0.31932 l 1.3417,2.68325 0.0924,0.16202 c 0.0405,-0.0728 0.12901,-0.19265 0.24502,-0.311 0.1846,-0.18774 0.29739,-0.23065 0.48407,-0.17861 0.13382,0.0374 0.26642,0.10711 0.29441,0.15529 0.088,0.15111 0.30436,0.10201 0.60459,-0.13753 0.15959,-0.12775 0.31636,-0.20397 0.34817,-0.17032 0.0321,0.0336 0.31919,0.0603 0.63861,0.068 0.3195,0.003 0.57545,0.0368 0.56948,0.0735 -0.0321,0.18681 0.0122,0.40741 0.0814,0.40741 0.0428,7.6e-4 0.16479,-0.0719 0.27032,-0.16241 0.17956,-0.15424 0.94963,-0.38544 1.25629,-0.37779 0.0757,0.002 0.16234,0.0919 0.19361,0.19996 0.0314,0.10799 0.0976,0.17291 0.14737,0.14384 0.0497,-0.0285 0.16563,-0.006 0.2577,0.051 0.14438,0.0896 0.13169,0.0996 -0.0917,0.0996 -0.14246,-0.002 -0.32385,0.0583 -0.40307,0.13318 -0.0793,0.0783 -0.1687,0.11036 -0.19921,0.0782 -0.0306,-0.0321 -0.16794,0.011 -0.30543,0.096 -0.13819,0.0859 -0.39911,0.14129 -0.58333,0.12369 -0.34235,-0.032 -0.52482,0.0456 -0.40461,0.1723 0.0375,0.0391 0.10439,0.0155 0.14898,-0.0525 0.0627,-0.0962 0.095,-0.0855 0.14185,0.0466 0.0459,0.13 0.088,0.14508 0.17627,0.0613 0.0635,-0.0608 0.23759,-0.11077 0.38648,-0.10313 0.14882,0.004 0.27185,-0.0325 0.27308,-0.0814 0,-0.0528 0.15555,-0.0724 0.34259,-0.0525 0.18705,0.0199 0.36279,6e-4 0.39085,-0.0423 0.0283,-0.0428 0.11096,-0.0528 0.18415,-0.0229 0.0734,0.0306 0.21779,0.0125 0.32163,-0.0395 0.13466,-0.0674 0.22391,-0.06 0.3114,0.0253 0.096,0.0969 0.10843,0.0961 0.0574,0 -0.0775,-0.14622 0.21305,-0.26384 0.8457,-0.34183 0.36163,-0.0446 0.45791,-0.0239 0.53033,0.1138 0.0688,0.13016 0.18399,0.16163 0.51496,0.13793 0.30955,-0.0222 0.4563,0.0117 0.53698,0.12369 0.092,0.12747 0.14476,0.13361 0.31147,0.0379 0,-0.002 0.008,-7.7e-4 0.01,-0.002 z",
4.53803, 3.30484,
"M 4.22266 2.80997  0.31537,-1.12559 c -0.009,-0.02 -0.007,-0.0204 -0.0166,-0.0433 -0.0891,-0.18672 -0.1596,-0.43114 -0.15657,-0.54337 0.007,-0.26307 -0.3226,-0.50877 -0.59978,-0.44658 -0.15585,0.0349 -0.21211,0.009 -0.20929,-0.0975 0.007,-0.14506 -0.37221,-0.30892 -0.51604,-0.22496 -0.20972,0.12206 -0.56233,-0.52094 -0.54975,-1.00228 0.007,-0.20306 0.69387,-0.85441 0.89802,-0.84718 0.0246,5.8e-4 0.22013,-0.0893 0.33034,-0.13193 l -1.11178,-1.65212 c -0.0195,0.0111 -0.0463,0.0411 -0.0615,0.0459 -0.0992,0.0317 -0.0915,0.0811 0.0376,0.23393 0.25,0.29716 0.14659,0.5097 -0.3124,0.64203 -0.308,0.0886 -0.41177,0.0914 -0.56956,0.006 -0.0485,-0.28545 -0.10942,-0.55919 -0.33331,-0.56055 l -3.8e-4,0.004 c -0.026,-3.8e-4 -0.055,-4e-5 -0.0856,0.0109 -0.0383,0.12372 -0.0535,0.2567 -0.0629,0.39275 -0.44908,-0.0654 -1.05986,-0.0415 -1.14461,0.0878 -0.0434,0.0662 -0.21861,0.14678 -0.38942,0.17789 -0.21261,0.0392 -0.30785,0.10585 -0.30192,0.21152 0,0.0285 -0.007,0.0534 -0.0145,0.0759 -0.0275,0.0216 -0.055,0.0453 -0.0822,0.0702 -0.007,0.001 -0.013,0.005 -0.0195,0.005 -0.0333,-7.2e-4 -0.12923,0.0237 -0.25528,0.0624 -0.08,-0.4784 -0.33186,-1.13708 -0.5564,-0.91446 0,0.16306 -0.0847,0.21734 -0.17609,0.21377 -3.8e-4,-3e-5 0,3.8e-4 0,3.8e-4 0.0499,0.24365 0.10392,0.55738 0.16712,0.8031 -0.14354,0.0227 -0.30849,0.0898 -0.40511,0.14461 -0.33424,0.33619 -0.74507,0.32102 -1.16899,0.1805 -0.26236,-0.0809 -0.35492,0.0333 -0.33894,0.29672 -0.15324,0.19325 -0.41676,0.31324 -0.60167,0.50786 -0.007,-0.0163 -0.013,-0.0322 -0.0181,-0.0486 0,0.0296 -0.007,0.0486 -0.008,0.0785 0.008,-0.01 0.0166,-0.0192 0.0253,-0.0284 0.0398,0.11691 0.0952,0.23149 0.15439,0.34417 -0.11751,0.0925 -0.227,0.23553 -0.25939,0.35127 -0.0586,0.21246 -0.18947,0.31968 -0.36325,0.38455 l 2.71944,2.83005 c 0.10689,-0.0587 0.32051,-0.1065 0.48394,-0.0994 0.16654,0.004 0.35001,-0.037 0.40736,-0.0912 0.0571,-0.0545 0.38558,-0.0909 0.72945,-0.0814 0.45349,0.0129 0.67232,-0.0226 0.79822,-0.12894 0.16481,-0.13918 0.99615,-0.19618 2.20563,-0.15135 0.19706,0.007 0.40331,-0.0138 0.45849,-0.0463 0.17471,-0.10371 0.36931,-0.0972 0.43389,0.0138 0.0217,0.0374 0.19258,0.0665 0.52581,0.0904 z",
1.92692, 5.24979,
"M -1.16568 5.24449  1.57665,-0.6191 c -0.0474,-0.4867 0.0954,-1.5337 0.25114,-1.6836 0.17039,-0.164 0.21291,-0.1563 0.37621,0.073 0.1012,0.1417 0.23035,0.3447 0.28834,0.4506 0.0858,0.1573 0.13611,0.1636 0.2744,0.034 0.16722,-0.1573 0.0631,-0.3063 -0.31419,-0.4501 -0.062,-0.024 0.01,-0.2644 0.15968,-0.5354 0.14994,-0.271 0.33213,-0.7281 0.40412,-1.0154 0.12608,-0.5015 0.1184,-0.5257 -0.1969,-0.5974 -0.23368,-0.053 -0.38971,-0.2226 -0.53949,-0.5876 -0.27201,-0.6626 -0.33335,-2.3657 -0.093,-2.5786 0.1421,-0.1258 0.13903,-0.1774 -0.0181,-0.2987 -0.2079,-0.1604 -0.2341,-1.2411 -0.045,-1.8552 0.10281,-0.3339 0.0903,-0.3456 -0.38601,-0.3581 -0.27072,-0.01 -0.61438,-0.1034 -0.76379,-0.214 -0.23149,-0.1708 -1.32257,-0.6546 -1.34407,0.41 -0.0512,2.5087 -0.0881,2.9436 -0.26511,3.1115 -0.141,0.1334 -0.16333,0.2237 -0.0692,0.2847 0.089,0.058 0.11213,0.5288 0.0651,1.3498 -0.0634,1.1134 -0.0451,1.2803 0.15967,1.4319 0.1278,0.094 0.22429,0.3077 0.22429,0.4734 -0.005,0.1661 0.0515,0.3404 0.12455,0.3886 0.073,0.048 0.0983,0.1683 0.0573,0.2682 -0.042,0.1012 -0.0193,0.1489 0.0517,0.1075 0.22709,-0.1322 0.12486,0.2738 -0.14264,0.5664 -0.28461,0.3116 -0.3362,0.4982 -0.13951,0.5033 0.0694,0 0.12142,0.091 0.12142,0.1964 -0.003,0.1065 0.0736,0.2267 0.17055,0.2666 0.27159,0.1124 0.32136,0.5471 0.078,0.6806 -0.118,0.064 -0.15684,0.1214 -0.0868,0.125 0.01,0 0.0111,0.067 0.0207,0.077 z",
3.13742, 6.38177,
"M 2.09259 -6.37924  c -0.0333,0.0351 -0.0639,0.10651 -0.0665,0.15881 0,0.0523 -0.18168,0.28228 -0.39724,0.51128 -0.21557,0.22899 -0.38735,0.49799 -0.38192,0.5977 0.009,0.0998 -0.12845,0.28291 -0.29726,0.40676 -0.16872,0.12384 -0.25759,0.22444 -0.19726,0.22261 0.0604,-0.002 -0.0906,0.167 -0.33569,0.37418 -0.24507,0.20716 -0.43708,0.42009 -0.42666,0.47372 0.0105,0.0536 -0.0482,0.12244 -0.12993,0.15248 -0.0977,0.0359 -0.11558,-0.005 -0.0508,-0.11945 0.0753,-0.13452 0.0578,-0.14289 -0.077,-0.0371 -0.42105,0.33004 -1.55709,-0.0183 -1.22158,-0.37463 0.0735,-0.078 0.0797,-0.24976 0.0166,-0.41989 -0.0587,-0.15829 -0.0988,-0.34453 -0.0883,-0.41354 0.0105,-0.069 -0.1864,-0.18257 -0.43752,-0.25293 -0.41046,-0.11495 -0.47884,-0.10036 -0.67549,0.14524 -0.16671,0.20808 -0.20217,0.22125 -0.14981,0.0547 0.0473,-0.14918 0.009,-0.21718 -0.13028,-0.2131 -0.15393,0.004 -0.19438,0.1355 -0.18098,0.57824 0.0149,0.48242 0.0447,0.54807 0.19,0.41943 0.13588,-0.11986 0.24384,-0.10913 0.50537,0.0511 0.28184,0.17272 0.32921,0.27874 0.3082,0.69317 l -0.0245,0.48911 0.74029,0.20541 c 0.40731,0.11309 0.788,0.17473 0.84561,0.13664 0.0578,-0.0382 0.1837,-0.008 0.2801,0.0674 0.0963,0.0751 0.31283,0.14691 0.48138,0.15972 0.35154,0.0267 0.74431,0.3456 0.61989,0.50313 -0.0464,0.0583 0.0114,0.0287 0.12757,-0.066 0.19438,-0.1583 0.21329,-0.12545 0.2294,0.40676 0.0131,0.42739 0.0569,0.5552 0.16925,0.4873 0.11986,-0.0727 0.12074,-0.0569 0.009,0.0756 -0.15786,0.18104 -1.51069,0.30581 -1.51577,0.13981 0,-0.14178 -1.40641,-0.0172 -1.51069,0.13438 -0.0517,0.0751 -0.0904,0.0544 -0.0946,-0.0516 -0.009,-0.12779 -0.0569,-0.10875 -0.18194,0.0674 -0.12678,0.17864 -0.17695,0.19682 -0.18369,0.0665 -0.009,-0.13638 -0.0429,-0.14397 -0.14885,-0.0313 -0.0867,0.0919 -0.12888,0.63215 -0.11172,1.42118 0.0149,0.70002 -0.0201,1.3785 -0.0771,1.50759 -0.0948,0.21303 -0.0569,0.22862 0.40722,0.17012 0.28115,-0.0354 0.527,-0.0233 0.54609,0.0267 0.0552,0.14325 2.4345,0.49083 2.60208,0.38007 0.0832,-0.0547 0.19542,-0.0735 0.25067,-0.0416 0.14753,0.0851 0.14473,0.78506 0,0.78953 -0.0709,0.003 -0.0902,0.13157 -0.0429,0.29772 0.0718,0.25436 0.0228,0.31646 -0.36283,0.45699 -0.3814,0.13879 -0.4806,0.13359 -0.68005,-0.0362 -0.12845,-0.10932 -0.43358,-0.20333 -0.67865,-0.20858 -0.24507,-0.005 -0.44758,-0.0599 -0.44977,-0.12127 0,-0.0615 -0.11129,0.02 -0.24253,0.18099 -0.40469,0.49616 -0.24227,0.75036 0.46562,0.72891 0.22712,-0.007 0.30163,0.0579 0.30942,0.26921 0.009,0.22501 0.0315,0.24213 0.12442,0.0891 0.0974,-0.16124 0.1161,-0.15665 0.1231,0.0299 0.009,0.17543 0.0753,0.20061 0.34611,0.12804 0.18571,-0.0497 0.38919,-0.14057 0.45205,-0.20179 0.063,-0.0612 0.2062,-0.11426 0.31801,-0.11764 0.2244,-0.007 0.65264,0.38173 0.61088,0.55426 -0.014,0.0601 0.0201,0.23309 0.0762,0.38459 0.21836,0.58332 0.33254,0.22078 0.29725,-0.94474 -0.0263,-0.87541 0.0123,-1.28197 0.13799,-1.45917 0.0959,-0.1351 0.17196,-0.37168 0.16968,-0.52576 0,-0.15407 0.0552,-0.3837 0.12941,-0.51037 0.16216,-0.27806 0.0175,-1.04903 -0.2434,-1.29449 -0.15253,-0.14353 -0.15612,-0.19583 -0.021,-0.31536 0.0892,-0.0787 0.25321,-0.1112 0.36424,-0.0724 0.2413,0.0843 0.44854,-0.40882 0.41895,-0.99812 -0.0105,-0.21248 0.0245,-0.69433 0.0779,-1.07052 0.0534,-0.37619 0.10261,-1.28473 0.10909,-2.01932 h -0.009 c 0,-0.31321 0,-0.55466 -4.3e-4,-0.7452 -0.0131,-0.57155 -0.0744,-0.68608 -0.28106,-0.91216 -0.24901,-0.27261 -0.30434,-0.29013 -0.36511,-0.11627 -0.0394,0.1127 -0.13728,0.18217 -0.21757,0.15428 -0.0975,-0.0339 -0.14036,-0.32845 -0.12897,-0.88591 0.01,-0.45938 -0.0105,-0.80654 -0.0438,-0.77144 z",
2.66052, 3.77103,
"M -0.39949 -3.77089  c -0.24655,0.002 -0.48424,0.0415 -0.7069,0.14289 -0.091,0 -0.14007,0.16428 -0.23247,0.15604 -0.56186,0.39932 -1.98716,1.04365 -0.95829,1.755351 0.37012,0.004 1.89317,-0.681921 1.58174,0.0144 -0.92118,0.19316 -0.78062,1.25644 -0.57652,1.87004 -0.5507,0.73969 -0.54645,2.07518 0.50073,2.40836 0.72023,0.13329 1.71345,0.023 1.86698,1.06161 0.14079,0.10466 0.256,0.14178 0.34647,0.13161 0.39231,-0.0442 0.32501,-0.98344 -0.0873,-1.24084 -0.12491,-0.92548 1.37777,-1.34914 1.32417,-2.16898 -0.0121,-0.11715 -0.0461,-0.24263 -0.12721,-0.37852 -0.83872,-0.59956 -0.43161,-1.77015 -0.097,-2.539971 0.45562,-0.35901 -0.33799,-0.65515 -0.53141,-0.9275 -0.74291,0.0304 -1.56344,-0.29219 -2.30307,-0.2845 z",
3.00843, 4.63,
"M -3.00308 -3.2356  c 0.0216,0.023 0.10979,0.055 0.26872,0.1013 0.56889,0.1649 0.8117,0.2144 1.07177,0.2196 0.22155,0 0.65876,0.8395 0.44286,0.846 -0.0655,0 -0.0523,0.2079 0.0295,0.4583 0.0816,0.2503 0.16335,0.662 0.1819,0.9142 0.0185,0.2523 0.0943,0.4564 0.16795,0.4542 0.0735,0 0.11453,0.2196 0.091,0.4925 -0.0236,0.273 5.3e-4,0.4946 0.0532,0.493 0.0529,0 0.10741,0.2474 0.12093,0.5529 0.0301,0.6823 0.35271,2.2028 0.57102,2.6913 0.0887,0.1985 0.29547,0.4232 0.4594,0.4992 0.4735,0.2194 0.77786,0.184 0.76378,-0.089 -0.0797,-1.554 0.36543,-3.8118 0.78032,-3.9574 0.13544,-0.047 0.34304,-0.3075 0.46147,-0.5778 0.11837,-0.2703 0.26596,-0.4621 0.32814,-0.4263 0.0622,0.036 0.15067,-0.042 0.19637,-0.1726 0.10948,-0.3133 -0.1781,-0.4655 -0.66611,-0.3525 -0.21603,0.05 -0.39776,0.035 -0.40411,-0.033 -0.0656,-0.6804 -0.18153,-1.2185 -0.31109,-1.4439 -0.19053,-0.3315 0.13535,-0.7608 0.48473,-0.6387 0.17021,0.059 0.20863,0 0.16433,-0.2589 -0.0455,-0.2639 -0.2012,-0.3747 -0.72399,-0.5147 -0.42937,-0.115 -0.69455,-0.2732 -0.74517,-0.4454 -0.0417,-0.1415 -0.0928,-0.2038 -0.2341,-0.2067 h -5.3e-4 c -0.0848,0 -0.20163,0.018 -0.36845,0.054 -0.28858,0.063 -0.6874,0.2067 -0.88573,0.3179 -0.1983,0.1112 -0.45557,0.2061 -0.57206,0.2113 -0.18831,0.01 -0.18809,0.024 0.004,0.1401 0.16267,0.098 0.094,0.1364 -0.27802,0.1524 -0.27189,0.012 -0.61486,0.1186 -0.76223,0.2372 -0.14743,0.1187 -0.39883,0.2252 -0.55862,0.2367 -0.10787,0.01 -0.15286,0.021 -0.13126,0.044 z",
4.45538, 5.00079,
"M 3.67326 3.58666  c 0.27079,-0.66346 0.32332,-0.78121 0.57421,-1.39679 -0.20034,0.0973 -0.44394,0.2056 -0.86662,0.24908 l -1.01753,0.10469 -0.32442,-0.67827 c -0.17835,-0.37327 -0.32777,-0.8685 -0.33148,-1.10044 -0.0123,-0.7472 0.88739,-1.58524 1.72954,-1.61074 0.5103,-0.0154 0.92804,-0.83692 1.00146,-1.96776 0.0482,-0.75168 0.021,-0.81363 -0.45222,-0.97588 -0.57495,-0.19732 -0.73908,-0.88336 -0.22222,-0.92927 0.16771,-0.0148 0.0408,-0.1032 -0.2829,-0.19608 -0.44802,-0.12851 -0.61919,-0.11519 -0.71732,0.0549 -0.0705,0.12285 -0.198,0.23615 -0.2829,0.25164 -0.0853,0.0156 -0.12804,0.13016 -0.0952,0.25547 0.0321,0.12524 -0.17735,0.31878 -0.46742,0.42983 -0.29007,0.11098 -0.52514,0.27295 -0.52254,0.35957 0,0.0866 -0.11865,0.16151 -0.26881,0.16606 -0.19491,0.006 -0.23866,0.10876 -0.1514,0.35702 0.0667,0.19175 0.0408,0.62287 -0.0593,0.95864 -0.17315,0.58372 -0.19898,0.59981 -0.58372,0.36405 -0.22135,-0.13498 -0.45358,-0.19271 -0.51599,-0.12646 -0.0618,0.0664 -0.11741,-0.01 -0.12235,-0.16988 -0.0124,-0.36216 -0.41947,-0.36654 -0.69941,-0.007 -0.16709,0.21426 -0.16598,0.26711 0.0124,0.26185 0.17006,-0.005 0.17254,0.0494 0.0136,0.25292 -0.11123,0.14297 -0.14918,0.41073 -0.0853,0.59461 0.0643,0.1839 0.0297,0.54243 -0.0766,0.79706 -0.208,0.49594 -0.48287,0.62505 -0.49485,0.23248 0,-0.14554 -0.19528,-0.23117 -0.48732,-0.22034 -1.11751,0.0426 -1.1384,0.0298 -1.03656,-0.62143 0.0828,-0.52592 0.0321,-0.64517 -0.37881,-0.8967 -0.45308,-0.27763 -0.48287,-0.27552 -0.70373,0.0466 -0.5061,0.73788 -0.0284,1.65526 0.58113,1.11704 0.16054,-0.14168 0.25509,-0.15557 0.25868,-0.0377 0,0.10208 -0.0692,0.23597 -0.16154,0.29699 -0.10505,0.0697 -0.0853,0.23703 0.0569,0.45089 0.18144,0.27371 0.18045,0.36979 -0.0124,0.49434 -0.28512,0.18973 -0.063,0.9574 0.23248,0.80537 0.11,-0.0566 0.33641,-0.15535 0.50264,-0.21971 0.16636,-0.0644 0.29909,-0.22545 0.29514,-0.35829 -0.0124,-0.28263 0.95709,-0.35933 1.62287,-0.12901 0.43418,0.15018 0.81187,1.06462 0.49239,1.19176 -0.0865,0.0347 -0.15412,0.20558 -0.1488,0.37938 0,0.17383 -0.10011,0.36244 -0.23569,0.41897 -0.55653,0.23312 -0.27128,1.40722 0.48027,1.97542 0.39685,0.30003 0.52477,0.25768 0.6936,-0.23056 0.12915,-0.37324 0.23049,-0.42519 0.68407,-0.34872 0.40588,0.0685 0.53095,0.0271 0.52502,-0.175 -0.0124,-0.32797 0.45877,-0.54947 0.56394,-0.26633 0.0569,0.15333 0.14905,0.14191 0.36015,-0.0434 0.19972,-0.17521 0.52736,-0.21946 1.12468,-0.16094 z",
4.04617, 5.42426,
"M -2.94195 4.57672  5.547,0.84754 c -0.0131,-0.11956 -0.002,-0.25067 -0.022,-0.34369 l -0.11879,-0.55384 0.67428,-0.21593 c 0.37071,-0.11919 0.75479,-0.27013 0.85339,-0.33717 0.28775,-0.19498 -0.63141,-0.34362 -1.29568,-0.20935 -0.56865,0.11499 -1.08906,-0.007 -1.15769,-0.27139 -0.0188,-0.0737 -0.18848,-0.31709 -0.37655,-0.54178 -0.18811,-0.22467 -0.3416,-0.53824 -0.3413,-0.69665 9.2e-4,-0.50293 -0.54357,-1.30997 -1.04569,-1.54977 -0.35267,-0.16848 -0.52754,-0.36992 -0.6645,-0.76621 -0.10221,-0.29559 -0.14563,-0.52528 -0.0964,-0.51011 0.17354,0.0539 -0.66737,-2.76228 -0.96072,-3.21725 -0.16345,-0.25346 -0.27119,-0.54462 -0.23941,-0.64722 0.036,-0.1166 0.008,-0.13435 -0.0749,-0.0476 -0.11577,0.12068 -0.2106,-0.0208 -0.25406,-0.37838 -0.006,-0.0499 -0.10171,-0.11849 -0.21226,-0.15262 -0.027,-0.008 -0.0774,-0.16167 -0.11391,-0.2073 -0.53067,-0.0874 -0.63302,-0.10779 -1.20503,-0.20227 10e-4,0.01 0.003,0.01 0.004,0.0175 0.0527,0.43032 0.14887,0.71581 0.25112,0.74742 0.10484,0.0326 0.0998,0.202 -0.0141,0.46214 -0.1944,0.44418 -0.0765,0.73425 0.34048,0.83519 0.23019,0.0556 0.28611,0.20689 0.44455,1.20419 0.20536,1.29285 0.19184,1.41099 -0.17315,1.50581 -0.32706,0.0846 -0.58936,0.79809 -0.36878,1.00225 0.20564,0.19033 0.36751,0.75346 0.20212,0.70237 -0.20848,-0.0645 -0.0864,0.17277 0.22532,0.43844 0.1607,0.13697 0.2717,0.3158 0.24664,0.39698 -0.0253,0.0806 0.0317,0.17074 0.12496,0.19997 0.11335,0.0347 0.11536,0.1239 0.004,0.26449 -0.11539,0.14655 -0.10247,0.49598 0.0428,1.1318 l 0.21014,0.91939 z",
5.49752, 3.12991,
"M 1.37015 -3.12969  v 0.001 c -0.1107,0 -0.2247,0.002 -0.3421,0.005 -0.273,0.007 -0.4227,0.0758 -0.4227,0.19482 0,0.26786 -0.2484,0.3097 -0.772,0.12971 -0.3784,-0.1301 -0.526,-0.12457 -0.8269,0.031 -0.5169,0.26723 -1.5029,0.35961 -2.096,0.19586 -0.4659,-0.12864 -0.5236,-0.11468 -0.7643,0.1912 -0.2181,0.27728 -0.3304,0.31769 -0.696,0.24908 -0.3971,-0.0745 -0.4341,-0.0551 -0.4114,0.21911 0.014,0.1656 -0.1182,0.49771 -0.2935,0.73742 -0.3107,0.42518 -0.3126,0.43928 -0.066,0.58239 0.1392,0.0807 0.5196,0.38777 0.8454,0.68265 0.3259,0.29485 0.7242,0.57704 0.8852,0.62684 0.1609,0.0498 0.2925,0.14516 0.2925,0.21187 0,0.15407 1.2613,0.16018 1.3565,0.006 0.039,-0.0632 0.3437,-0.24332 0.6769,-0.39997 0.425,-0.19982 0.6332,-0.24126 0.6967,-0.1385 0.05,0.0805 0.1662,0.12856 0.2589,0.10697 0.093,-0.0217 0.2127,0.11828 0.2671,0.31058 0.087,0.30723 0.054,0.36567 -0.2775,0.48111 -0.4545,0.1584 -0.4945,0.26092 -0.07,0.17983 0.1982,-0.0379 0.3343,0.0134 0.3845,0.14418 0.064,0.16602 0,0.18801 -0.3395,0.1235 l -0.4165,-0.0786 0.2599,0.26149 c 0.1431,0.14385 0.3029,0.23712 0.3551,0.20722 0.052,-0.0299 0.2166,0.0106 0.3658,0.0904 0.2162,0.11576 0.2544,0.22258 0.1876,0.52659 l -0.084,0.38085 h 0.8407 c 0.752,0 0.8633,-0.0341 1.0527,-0.32297 0.1164,-0.17773 0.372,-0.36144 0.5679,-0.40825 0.4395,-0.10496 0.7018,-0.3012 0.6072,-0.45424 -0.039,-0.0635 -0.01,-0.11523 0.066,-0.11523 0.076,7e-5 0.1379,0.0441 0.1379,0.0977 0,0.23852 1.0762,0.16168 1.5126,-0.10801 0.2339,-0.14457 0.366,-0.29703 0.386,-0.40307 v 0 c 0.02,-0.10602 -0.072,-0.16583 -0.28679,-0.12403 -0.13541,0.0264 -0.67191,0.0638 -1.19271,0.0827 -0.9133,0.0331 -0.9439,0.0241 -0.8619,-0.24339 0.047,-0.15257 0.1949,-0.38208 0.3292,-0.51006 0.297,-0.28318 0.3099,-0.52576 0.041,-0.74878 -0.462,-0.38339 -0.6379,-1.28709 -0.3137,-1.61127 0.069,-0.0691 0.1261,-0.22671 0.1261,-0.35038 0,-0.12366 0.089,-0.31478 0.1984,-0.42426 0.3806,-0.38055 -0.5051,-0.62598 -2.1652,-0.61598 z",
3.63043, 5.15692,
"M -3.62726 -3.36795  c -0.0468,0.32054 0.43759,0.66685 0.52881,0.98692 0.28647,0.32487 -0.28269,0.47336 -0.316,0.71696 0.10063,0.38021 0.34235,0.83542 0.13607,1.20634 -0.51118,0.31783 0.37465,0.71304 0.29987,1.10283 0.36141,0.35332 -0.44509,0.42029 -0.41731,0.69873 -0.13664,0.46798 -0.26085,0.93998 -0.21981,1.41789 0.0582,0.12446 -0.11212,0.37614 0.0969,0.41685 0.88284,-0.0585 0.30904,0.58298 0.69294,0.85291 -0.34214,0.33774 0.0244,0.78826 0.73335,0.72836 0.43959,0.29919 1.07278,0.24331 1.63813,0.35228 l 0.0736,0.0448 c 0.38778,-0.10404 0.0245,-0.72353 0.20718,-1.01432 0.51323,-0.34366 -0.20283,-0.45485 -0.61814,-0.53103 0.0274,-0.33275 0.26259,-0.74304 0.21784,-1.13747 0.23923,-0.2064 0.75099,-0.0848 1.11881,-0.12006 l -0.0149,0.53094 c 0.18821,0.46487 0.40821,0.92446 0.12828,1.42682 -0.43537,0.34603 -0.20149,0.86785 0.62785,0.73084 0.32264,-0.28321 0.35317,-0.75168 -0.0804,-1.04175 0.56867,-0.53073 1.11997,0.32514 1.3861,0.60116 -0.0255,0.3657 1.32511,0.75896 0.98294,0.21235 -0.28172,-0.41656 -0.56593,-0.8647 -1.16271,-1.16373 0.0169,-0.28998 -0.30591,-0.44892 -0.51256,-0.68143 -0.27438,-0.51775 0.92277,-0.55614 1.32949,-0.29182 0.37686,0.30057 0.6089,-0.11077 0.11594,-0.20101 0.11778,-0.45323 -0.55455,-0.70288 -0.75334,-1.04554 -0.24517,-0.39883 0.35093,-0.72371 0.47053,-1.00707 -0.74221,-0.0978 -0.87992,-0.65893 -1.14431,-1.03193 -0.0424,0.27079 -0.61393,0.63737 -0.40714,0.94708 -0.39146,0.39072 -0.2133,0.90337 -0.19158,1.3428 -0.38873,0.18762 -0.1672,0.38452 -0.0853,0.62107 -0.61435,0.26479 -0.31061,-0.34619 -0.56846,-0.50332 -0.0127,-0.48721 -0.72194,-0.78357 -0.99598,-1.2129 0.12473,-0.33506 -0.33466,-0.51389 -0.36161,-0.82807 0.1997,-0.42706 0.3237,-0.84131 0.26038,-1.28998 0.68927,-0.44553 -0.67313,-0.6391 -0.21405,-1.09468 0.22086,-0.37425 0.42402,-0.90652 -0.18371,-1.18138 -0.43025,-0.35212 -0.0683,-0.76225 -0.19938,-1.15243 -0.35117,-0.0477 -0.85403,-0.20224 -1.27169,-0.19685 -0.25089,0.003 -0.47094,0.0655 -0.60781,0.24375 -0.36841,0.38217 -0.14434,0.8521 -0.46111,1.23525 -0.16785,0.0992 -0.24225,0.20322 -0.2578,0.31006 z",
2.88035, 4.06242,
"M 1.25865 -4.0624  c -0.26985,0.16593 -0.42018,0.52485 -0.45144,0.83652 -0.0532,0.14065 -0.26733,-0.18208 -0.24919,0.11733 -0.28494,0.18546 0.14687,0.48052 0.0223,0.77037 0.036,0.36289 0.3756,0.58792 0.53905,0.88077 -0.13707,-0.006 -0.45978,0.15371 -0.4101,-0.0997 -0.0108,-0.31743 -0.1796,-0.59765 -0.41569,-0.80388 -0.20831,-0.2224 -0.50043,-0.35247 -0.56789,-0.66266 -0.28582,-0.18946 -0.72788,-0.02 -0.85216,-0.40249 -0.34073,-0.28483 -0.60874,0.19553 -0.79607,0.42709 -0.15294,0.28679 -0.28387,0.62305 -0.36898,0.94692 0.0288,0.41155 0.59654,0.31526 0.78223,0.64991 0.21197,0.12691 0.79092,0.32794 0.42931,0.61615 -0.22783,0.17371 -0.51269,0.28054 -0.73081,0.21635 -0.3129,0.0885 -0.69087,0.0242 -0.91787,0.32227 -0.0986,0.0888 -0.18264,0.18327 -0.14058,0.32426 0.0153,0.33863 0,0.68711 0.12159,1.00883 -0.17136,0.30467 0.34515,0.28606 0.24024,0.61504 0.0458,0.34026 0.22841,0.60042 0.53816,0.73395 0.11715,0.34628 0.48027,0.43439 0.68656,0.10861 0.22092,-0.24425 0.13892,-0.64131 0.4966,-0.81417 0.22425,-0.28822 0.71346,-0.14432 0.87741,-0.50084 0.18991,-0.18243 0.5743,-0.40112 0.46598,0.0664 0.0489,0.38641 -0.2614,0.6663 -0.34171,1.01509 0.0213,0.33786 0.19593,0.66602 0.38507,0.93374 0.16136,0.28358 0.42527,0.5179 0.5243,0.8088 0.11306,0.0261 0.24555,0.005 0.31311,-0.099 0.17679,-0.18075 0.42474,-0.24465 0.67359,-0.22126 0.24753,-0.26223 0.54092,-0.49076 0.73684,-0.79584 0.0703,-0.37759 -0.42896,-0.49317 -0.57771,-0.73752 -0.0631,-0.34107 -0.18323,-0.61203 -0.31803,-0.91518 0.0979,-0.36538 -0.30776,-0.49131 -0.24159,-0.85374 -0.0624,-0.37632 -0.49732,-0.49981 -0.55581,-0.87137 -0.0398,-0.29907 -0.0805,-0.80571 0.373,-0.62533 0.25606,0.0228 0.19023,-0.34809 0.48675,-0.27109 0.24961,-0.0512 0.49133,-0.32109 0.67696,-0.0456 0.3421,-0.12553 0.19087,-0.82493 -0.23511,-0.63717 -0.36514,-0.0659 0.34354,-0.40056 0.18392,-0.66355 -0.01,-0.40846 -0.52884,-0.49985 -0.47624,-0.91853 -0.12166,-0.36074 -0.58136,-0.42643 -0.90604,-0.4595 z",
3.57088, 4.77385,
"M -1.37842 -4.77385  c -0.32004,0.38562 -0.90978,0.44109 -1.06475,0.98212 -0.33196,0.58155 -0.34433,1.24636 -0.0431,1.86247 0.63654,0.61117 -0.46462,0.56209 -0.81103,0.5768 -0.71739,0.43468 0.23093,1.04171 -0.0105,1.64117 0.14314,0.6155 0.45858,1.33418 1.13619,1.48914 0.64762,-0.0825 1.4885,0.94379 0.87455,1.42229 -0.20223,0.48484 -0.89031,0.9814 -0.1661,1.42189 0.23434,0.25918 0.59181,0.15737 0.78734,-0.0868 0.49244,-0.4069 1.22385,-0.34712 1.79188,-0.53942 0.40959,-0.0676 0.43978,-0.38024 0.52823,-0.71925 0.4924,0.34324 1.17879,0.64169 1.71133,0.1823 0.40615,-0.60741 0.1742,-1.49048 -0.17359,-2.078 -0.01,-0.55871 -0.45537,-1.1785 -1.0577,-0.79483 -0.47109,0.82821 -1.0013,-0.36652 -0.57886,-0.79192 0.0974,-0.61938 -0.40188,-1.41863 -1.12248,-1.2263 -0.615,0.15415 -0.83256,-0.44564 -0.6852,-0.84049 -0.60947,-0.41783 0.57646,-1.01268 -0.19227,-1.48999 -0.1719,-0.39383 -0.74183,-0.82494 -0.92398,-1.01118 z",
3.57088, 3.49495,
"M -2.51747 -3.49406  c -0.44274,0.0168 -0.80575,0.24519 -0.65317,0.74056 0.77847,0.68723 0.006,1.14574 -0.40024,1.60094 0.45273,0.85169 -0.057,2.25152 1.13658,2.51484 0.58996,0.19997 -0.83942,0.82749 0.19445,1.21761 0,0.4092 0.84626,1.20435 1.14506,0.8068 0.96341,-0.49284 2.10194,-0.48523 3.16453,-0.62008 1.00752,-0.34932 1.59197,-1.40776 1.48958,-2.45213 -0.18766,-1.01556 -0.54571,-2.14247 -1.52062,-2.68748 -0.56572,-0.6585 -2.34805,-1.70658 -2.30696,-0.17264 1.19275,1.31973 -1.71147,1.34573 -1.07034,-0.0774 0.0918,-0.5639 -0.60964,-0.89245 -1.17887,-0.87092 z",
2.19827, 3.72855,
"M 1.79578 -3.16055  c 0.29134,0.9127 -0.8513,1.88012 -1.60506,1.23145 -0.61184,0.26144 -1.4819,1.48613 -0.34626,1.65621 0.20292,0.5854 -0.87432,1.11229 -0.86456,1.82626 -0.34284,0.66001 -1.30935,-0.58756 -1.16324,0.49712 0.015,0.16218 -0.0225,0.37679 0.0868,0.50798 0.35875,0.66438 1.22487,1.01087 1.95492,0.77514 0.59338,-0.43345 1.13653,0.4314 1.81488,0.39374 0.23014,-0.39462 0.28884,-0.97791 0.31677,-1.46193 0.0963,-1.02031 -0.0218,-2.07603 0.20824,-3.07268 -0.0106,-0.662 -0.26526,-1.55587 -0.0532,-2.3394 0.10891,-0.75342 -0.47035,-0.79317 -0.34929,-0.0139 z",
3.57085, 3.77374,
"M -1.00965 -3.77187  c -0.50773,-0.0455 -0.59851,0.75754 -1.18315,0.62357 -0.2759,0.006 -0.43238,-0.20174 -0.54732,-0.35854 -0.42974,-0.101 -0.98298,0.0786 -0.79185,0.62433 0.0918,0.27667 0.63885,0.21922 0.50031,0.65458 -0.008,0.28654 0.415,0.4124 0.12652,0.69333 -0.0968,0.21542 0.3362,0.39395 0.19527,0.57658 -0.37534,0.22876 0.12223,0.32999 0.18878,0.60483 0.26355,0.34866 -0.14544,0.74459 -0.0203,1.14239 0.22773,0.37599 0.42716,0.78503 0.71585,1.1349 0.20089,0.31178 0.59558,0.80482 0.97387,0.45305 0.53952,-0.18877 -0.0114,0.38539 -0.26603,0.32129 -0.005,0.24853 0.24187,0.42998 0.13101,0.70434 0.12782,0.27427 0.66178,0.50327 0.90612,0.28229 -0.19362,-0.28539 -0.38278,-0.53996 -0.23003,-0.90337 0.0118,-0.18256 0.14722,-0.40585 -0.0853,-0.48031 0.11501,-0.13614 0.31125,-0.45293 0.29528,-0.0705 0.13823,0.35583 0.51741,0.5319 0.87087,0.59257 0.22815,-0.32159 0.48104,-0.66059 0.8381,-0.85811 0.36776,-0.32583 0.73175,-0.66601 1.02963,-1.05813 0.31551,-0.18006 0.54504,-0.50482 0.93287,-0.49481 -0.0706,-0.38435 -0.81564,-0.26725 -0.51007,-0.7296 -0.004,-0.20098 0.28139,-0.50601 0.11752,-0.63332 -0.38126,0.008 -0.77281,0.0549 -1.14464,-0.0458 0.22089,-0.13301 0.47335,-0.47944 0.54081,-0.80185 0.14915,-0.29117 0.0598,-0.94753 -0.38205,-0.76085 -0.38265,-0.0285 -0.7182,-0.62895 -1.06013,-0.13102 -0.15706,0.19112 -0.33864,-0.13478 -0.29078,-0.28203 0.22718,-0.13574 0.44603,-0.63936 0.0285,-0.63858 -0.11854,0.35338 -0.58257,0.49089 -0.85134,0.14477 -0.29235,-0.20085 -0.66289,-0.31351 -1.02838,-0.30604 z",
3.57089, 5.62272,
"M -1.42379 -5.622  c -0.43389,0.0339 0.14764,1.23134 -0.27034,1.52136 -0.32134,0.64017 -0.3091,1.54516 -1.19918,1.59638 -0.5274,0.17075 -1.08366,0.90166 -0.25472,1.09547 0.18208,0.59783 0.78062,1.0544 0.80096,1.66886 -0.42354,0.53578 -1.0075,1.1037 -0.51402,1.81991 -0.22824,0.69684 -0.48371,1.55859 -0.0453,2.29815 0.0422,0.68578 0.8206,0.79979 1.16895,1.23743 0.21794,0.0348 0.4294,-0.063 0.61318,-0.16915 0.77997,-0.43845 0.0476,-1.6467 0.62273,-2.13958 0.98799,-0.075 0.80763,-1.11489 1.07434,-1.82644 0.15326,-0.73964 1.0527,-0.44914 1.60996,-0.48531 0.28374,0.01 0.59117,0.0806 0.75616,-0.21396 0.52814,-0.37956 0.96391,-1.07548 0.28243,-1.5707 -0.29818,-0.54672 -1.30735,-0.15148 -0.43951,0.146 0.46422,0.42259 -0.74074,0.6148 -1.00937,0.33025 -0.36553,0.13004 -0.87668,0.44845 -0.74154,-0.26028 0.0857,-0.69248 0.10805,-1.45821 -0.47727,-1.77811 -0.56593,-0.56035 0.55259,-0.94649 0.6444,-1.55057 0.2319,-0.78726 -0.94437,-0.99379 -1.45289,-1.28073 -0.40803,0.36815 -0.78127,0.16543 -0.91525,-0.37606 -0.10931,-0.0486 -0.19174,-0.0678 -0.25373,-0.0629 z",
3.57089, 3.9589,
"M -0.662738 -3.95715  c -0.22732,0.1856 -0.1616,0.62065 -0.53242,0.4242 -0.47422,0.18596 -1.1086,-0.11391 -1.44717,0.37963 -0.0317,0.5894 -0.93935,0.14582 -0.92847,0.72408 0.18671,0.31101 0.54772,0.37476 0.774,0.25063 0.73663,0.38029 -0.26627,0.73927 -0.27541,1.18012 -0.16873,0.24378 -0.22553,0.49255 -0.0791,0.74385 -0.35807,0.49099 0.49544,0.79295 0.89329,0.76898 0.7408,0.0465 -0.0421,0.63402 -0.3488,0.72743 -0.19071,0.30445 -0.16117,0.71804 -0.63596,0.65942 -0.0994,0.45223 -0.3579,0.95834 0.0972,1.33927 0.29409,0.53974 0.96448,0.4094 1.42438,0.6058 0.25557,0.31566 0.59274,-0.11505 0.78137,-0.28179 0.43541,-0.29546 0.89088,-1.01334 1.42439,-0.89697 0.37087,0.20728 0.82182,0.38467 1.17308,0.10018 0.30848,0.23171 0.74947,0.19845 1.07289,0.43056 0.23849,0.006 0.56519,-0.37111 0.53108,-0.64333 0.0427,-0.41341 0.67275,-0.95516 0.002,-1.22669 -0.51332,0.33701 -1.04337,-0.18935 -1.52993,-0.35215 -0.31537,-0.10025 -1.01146,-0.23443 -0.48651,-0.62826 0.0794,-0.34626 0.44388,-0.30417 0.61988,-0.4711 -0.5035,-0.3571 -0.4592,-1.16755 -1.13992,-1.30979 0.084,-0.42636 0.24714,-0.98594 0.42956,-1.45855 -0.12383,-0.36592 -0.23396,-1.20431 -0.77535,-0.74854 -0.42722,0.24522 -0.7282,-0.3551 -1.04407,-0.31698 z",
4.35997, 4.86053,
"M 4.12802 -4.85771  c -0.10456,-0.0226 -0.24399,0.0881 -0.41693,0.42159 -0.2125,0.98079 -0.39661,2.00485 -1.27137,2.64164 -0.55259,0.76964 -0.84457,2.28577 -2.02939,2.16135 -0.53939,-0.50526 -1.4269,0.18538 -2.12311,0.20849 -0.55867,0.36525 -1.1619,-0.53149 -1.67378,0.28501 -0.92928,0.12919 -0.84737,1.30207 -0.21641,1.57615 -0.23706,0.6136 -1.50686,1.78461 -0.11541,1.86052 0.32445,0.23972 0.55181,0.97448 0.94742,0.25664 0.58659,-0.89664 1.44208,-1.65942 1.52995,-2.76437 0.89443,-0.40943 1.08676,1.66279 2.17983,1.45475 1.174,0.49699 -0.92964,1.24862 -0.11145,1.5821 0.52277,-1.10045 2.1059,-0.15978 2.383,-1.305 0.26784,-0.82205 1.53393,-1.83991 0.82737,-2.6377 -0.22347,-0.70652 -1.39017,-0.1269 -1.3248,-0.76992 -0.23832,-0.99719 0.89102,-1.4275 1.08066,-2.29858 -0.19297,-0.68688 0,-1.06432 0.50602,-1.52073 0.12078,-0.40486 0.0586,-1.10219 -0.17153,-1.15194 z",
2.04794, 5.74865,
"M 1.10481 -5.74852  c -0.2228,0.001 -0.42374,0.10801 -0.27902,0.38066 -0.22773,0.32416 -1.04651,0.01 -0.86657,0.6221 0.26376,0.4125 0.0107,0.98862 -0.32022,1.27218 0.2249,0.43579 0.30114,0.9337 0.36381,1.39786 -0.18081,0.1263 -0.35082,0.59103 -0.47477,0.5752 0.13163,-0.4456 -0.6191,-0.80374 -0.54723,-0.18253 0.0335,0.4939 0.30706,0.8021 0.65488,1.16755 0.0573,0.18913 0.50349,0.27841 0.054,0.30188 -0.32151,0.0134 -0.70298,0.15083 -0.96036,0.26819 -0.004,0.5051 -0.55904,0.90049 -0.23512,1.41321 -0.58144,0.13334 -0.38504,0.76784 -0.14975,1.12213 0.1297,0.3844 0.11185,0.7896 -0.32623,0.90896 -0.31629,0.49376 0.60887,-0.10276 0.51957,0.45041 -0.50415,0.22985 -0.21881,0.79349 -0.0156,1.13687 0.0134,0.25246 0.20574,0.68606 0.45883,0.6615 0.0963,-0.10895 0.22038,-0.21662 0.23453,-0.37013 -0.0641,-0.49914 0.3279,-0.12706 0.46184,0.0763 0.39469,0.33331 0.98388,0.17352 1.28451,-0.20657 0.25784,-0.20591 0.90771,-0.24636 0.52498,-0.65428 0.40076,-0.52758 -0.33316,-0.93582 -0.19243,-1.46581 0.0606,-0.56794 -0.0204,-1.15758 0.12958,-1.71117 0.1498,-0.39418 -0.30538,0.12251 -0.36231,-0.28414 -0.0417,-0.31502 0.45507,-0.37624 0.54784,-0.65519 0.36583,-0.13394 0.0559,-0.67183 0.39269,-0.91255 0.2087,-0.31728 -0.37429,-0.20735 -0.2216,-0.60377 -0.11082,-0.46181 -0.47329,-0.89905 -0.93241,-1.04485 -0.14753,0.0856 -0.34858,0.34434 -0.37436,-0.04 0.25058,-0.4456 -0.0889,-0.88442 0.0608,-1.31458 -0.11306,-0.4522 0.68191,0.0944 0.52617,-0.54182 0.2472,-0.31727 0.41213,-0.71248 0.436,-1.11462 -0.0466,-0.2419 0.0663,-0.32026 0.29827,-0.3073 0.43849,-0.20873 -0.33546,-0.1678 -0.47237,-0.3142 -0.0671,-0.0208 -0.14373,-0.032 -0.21799,-0.0316 z",
3.57088, 2.80202,
"M -0.301059 -2.80177  c -0.17254,0.005 -0.33301,0.071 -0.43996,0.26942 0.0409,0.2666 0.3491,0.47076 0.41858,0.69966 -0.36962,0.1128 -0.73032,0.3493 -1.08495,0.0913 -0.38249,0.0613 -0.25611,0.72381 0.036,0.88696 0.2585,0.30716 -0.38849,0.35691 -0.58937,0.39089 -0.44316,0.0903 -0.23306,0.50499 -0.19046,0.77618 -0.15188,0.28342 -0.52426,0.566 -0.8221,0.58572 0.13061,-0.25419 0.13318,-0.51381 -0.18414,-0.64646 -0.2523,-0.20486 -0.5502,0.37635 -0.34497,0.58256 0.0492,0.32508 -0.0329,0.76492 0.43389,0.8515 0.34683,0.12336 0.70559,0.1523 1.03297,0.318 0.37589,0.0865 0.73238,-0.13092 1.03003,-0.32214 0.24664,0.20939 0.57748,0.44613 0.93799,0.5041 0.31649,0.14136 0.22886,-0.80563 0.61876,-0.47421 0.16196,0.33427 0.30384,0.70629 0.62118,0.93166 0.24438,-0.0315 0.44964,0.28662 0.6275,0.0294 0.27443,-0.16145 0.69426,0.0631 1.03638,0.0219 0.26986,0.26652 1.04494,-0.002 0.59931,-0.38117 -0.10676,-0.38309 -0.23137,-0.76018 -0.4295,-1.11046 0.008,-0.29153 -0.0528,-0.52528 -0.27889,-0.7016 0.0236,-0.29755 0.67185,-0.5939 0.17224,-0.85368 -0.25747,-0.317 0.0128,-0.57784 0.20261,-0.8357 -0.15225,-0.23193 -0.67612,-0.3641 -0.45235,-0.74897 0.0316,-0.51263 -0.46331,-0.52312 -0.81311,-0.58111 -0.28658,-0.15328 -0.55917,0.57451 -0.89206,0.23687 -0.31681,-0.0816 -0.31441,-0.59906 -0.73342,-0.41882 -0.15487,-0.0489 -0.33957,-0.10634 -0.51211,-0.1018 z",
3.16396, 6.28632,
"M -0.0681559 -6.28552  c -0.27397,0.0139 -0.48549,0.2094 -0.46561,0.68988 -0.13806,0.64257 -0.98468,1.02766 -0.7705,1.89187 -0.48185,0.70273 0.94628,1.12941 0.80874,1.464 -0.56632,0.088 -1.05088,0.56396 -1.53894,0.75344 0.69345,0.93338 -1.41697,0.2104 -1.09603,1.20509 0.35057,0.59033 0.4561,1.79086 1.47845,1.37459 0.45409,-0.19336 0.97791,0.0161 0.31471,0.32402 -0.53264,0.2974 -0.72491,1.05029 0.10387,0.95704 0.54052,0.2765 1.10715,0.49524 1.67639,0.62219 -0.14509,0.73633 -1.35981,0.56002 -1.27744,1.47226 -0.11997,0.30673 -0.90705,0.8281 -0.69143,0.97203 0.64811,0.19804 1.34276,0.11448 1.99522,0.29869 0.71434,-0.16557 1.26673,0.36676 1.79782,0.54674 -0.065,-0.01 0.33212,-0.10842 0.24134,-0.11937 0.49432,-0.10903 0.93445,-0.26785 0.43665,-0.76378 -0.0961,-0.57586 -0.91548,-1.29645 -0.12038,-1.6831 -0.4048,-0.51862 0.47421,-1.14462 -0.38966,-1.35703 -0.75428,-0.50253 -0.338,-1.22127 0.38603,-1.49396 0.40886,-0.84209 -0.29071,-1.68516 -0.11525,-2.56832 -0.038,-0.76264 0.0716,-1.5506 -0.0966,-2.2965 -0.33673,-0.72185 -1.00957,-1.22882 -1.66863,-1.66295 -0.20068,-0.36268 -0.65647,-0.64475 -1.00871,-0.62683 z",
    )
scratches=(
1.21098, 12.0742,
"M 0.000565603 -12.0742  c -0.13128,0.38351 -0.75628,0.6648 -1.19561,1.82211 -0.1414,0.37239 0.70808,0.91154 0.5936,1.35961 -0.4133,1.61789 0.0395,2.85731 0.44366,3.80288 0.14158,0.26984 -0.43294,2.12241 -0.31668,2.13113 0.97645,1.62823 0.21172,1.22883 0.13349,3.04219 0.57549,2.43115 0.2224,3.6139 0.12544,6.25854 0.33054,2.36608 -0.73036,1.37715 0.43984,5.73195 0.69244,-3.11326 0.13428,-4.81509 0.0535,-6.02599 0.14464,0.36647 0.22405,-0.58222 0.79938,-1.15807 -0.0141,-0.45337 -0.0246,-0.90699 -0.0494,-1.35961 -0.12372,-0.71424 -0.19704,-1.42486 -0.13769,-2.16524 -0.26034,-0.76585 0.18753,-1.53555 -0.15299,-2.29857 -0.018,-0.66796 0.41747,-1.89258 0.47387,-2.52801 -0.26969,-0.62391 -0.19911,-1.3802 -0.23483,-1.23662 -0.16028,-0.23971 -0.65345,-1.05872 -0.69112,-1.99936 0.0472,-0.699 0.26798,-1.33045 0.18587,-2.0593 0.0935,-0.5642 -0.25882,-1.6984 0.35341,-1.78749 0.55821,-0.1124 -6e-5,-0.993 -0.21304,-0.42272 -0.52661,-0.0264 0.002,-0.62728 -0.20154,-0.94723 -0.17458,-0.0432 -0.29981,0.0337 -0.38438,0.17519 -0.007,-0.1122 -0.0131,-0.22491 -0.0249,-0.33538 z m 0.0394,1.76526 c 0.16024,0.34312 0.14013,0.70048 -0.0104,0.99271 0,-0.34768 0.005,-0.67207 0.0104,-0.99271 z",
1.21091, 19.8304,
"M 0.362522 -19.83  c -0.0425,-0.0104 -0.0565,0.20572 -0.0739,0.68316 -0.17118,4.69021 -0.36552,8.35313 -0.44801,8.44393 -0.0551,0.0606 -0.0981,0.35879 -0.0961,0.66198 0.003,0.45283 0.0497,0.48001 0.26043,0.15399 0.14105,-0.2183 0.25847,-0.63823 0.26046,-0.93327 0.002,-0.29501 0.19851,-0.81366 0.43616,-1.1529 0.17211,-0.24575 0.26287,-0.41037 0.27647,-0.53486 0.0136,-0.12454 -0.0499,-0.20942 -0.18603,-0.29352 -0.25998,-0.1606 -0.30745,-0.89742 -0.19897,-3.07991 0.0784,-1.57793 0.0496,-3.16673 -0.0641,-3.53054 -0.0829,-0.26526 -0.1333,-0.40997 -0.1664,-0.41806 z m -0.52554,11.09286 c -0.0402,-0.006 -0.0675,0.0911 -0.0755,0.28371 -0.0115,0.27886 0.0511,0.43465 0.13901,0.34675 0.0438,-0.044 0.0678,-0.12308 0.0718,-0.21498 h 5.3e-4 c 0.004,-0.0919 -0.013,-0.19595 -0.0512,-0.29145 -0.0317,-0.0792 -0.0606,-0.12035 -0.0848,-0.12403 z m 0.13952,1.125 c -0.007,-9.6e-4 -0.0143,0.001 -0.0217,0.006 -0.11779,0.0728 -0.22286,1.22667 -0.23307,2.56419 -0.004,0.48057 -0.0157,0.84616 -0.032,1.15393 -0.11018,0.1207 -0.22863,0.25041 -0.36277,0.42995 -0.0519,-0.0854 -0.0727,-0.0405 -0.0842,0.12247 -0.15026,0.2185 -0.30279,0.48437 -0.4377,0.83975 -0.11105,0.29241 0.38748,0.68769 0.55189,1.06091 -0.003,0.13173 0.002,0.29377 0.006,0.45631 -0.24328,1.08473 -0.099,1.99333 0.14417,2.75642 0.002,0.0868 0.0117,0.25962 0.0109,0.30282 -0.005,0.2518 0.103,0.52662 0.23926,0.61082 0.0341,0.0211 0.067,0.0269 0.0966,0.0191 0.0622,0.28054 -0.21254,1.2988 -0.31004,1.80195 -0.0228,-0.0699 -0.058,-0.1111 -0.1044,-0.1111 -0.12534,0 -0.22789,0.29714 -0.22789,0.66094 0,0.36375 0.0455,0.66198 0.10128,0.66198 0.0558,0 0.1583,-0.29823 0.22789,-0.66198 0.0132,-0.0692 0.0186,-0.13206 0.0238,-0.19482 0.90374,1.53576 0.17148,1.18929 0.0946,2.97243 0.57549,2.43115 0.22254,3.6139 0.12557,6.25853 0.33054,2.36608 -0.73044,1.37714 0.43976,5.73195 0.69244,-3.11326 0.13401,-4.81509 0.0532,-6.02599 0.14464,0.36647 0.2241,-0.58222 0.79943,-1.15807 -0.0141,-0.45337 -0.0243,-0.90698 -0.0491,-1.3596 -0.12372,-0.71424 -0.19733,-1.42486 -0.13798,-2.16525 -0.26033,-0.76585 0.18756,-1.53554 -0.15296,-2.29856 -0.018,-0.66796 0.41749,-1.89257 0.47387,-2.52802 -0.26969,-0.6239 -0.19889,-1.38019 -0.23461,-1.23661 -0.16028,-0.23971 -0.65375,-1.05872 -0.69143,-1.99936 0.0472,-0.699 0.26815,-1.33046 0.18605,-2.05931 0.0935,-0.5642 -0.25876,-1.6984 0.35346,-1.78749 0.55822,-0.1124 8e-5,-0.99299 -0.21291,-0.42271 -0.10411,-0.005 -0.1661,-0.0335 -0.20153,-0.077 10e-4,-0.13439 10e-4,-0.26153 -10e-4,-0.37982 0.0406,-0.17118 0.0872,-0.35502 10e-4,-0.49041 -0.0109,-0.003 -0.0199,-8.1e-4 -0.0305,-0.003 -0.0297,-0.26499 -0.0801,-0.43408 -0.1633,-0.43408 -0.17193,0 -0.23233,-0.58849 -0.16124,-1.572 0.0586,-0.81059 0.0233,-1.43103 -0.0786,-1.44539 z m -0.66869,13.69167 c -0.0402,-0.006 -0.068,0.091 -0.076,0.2837 -0.0115,0.27886 0.0511,0.43465 0.13901,0.34675 0.0439,-0.044 0.0684,-0.12256 0.0723,-0.21446 0.004,-0.0919 -0.013,-0.19647 -0.0512,-0.29197 -0.0317,-0.0792 -0.0601,-0.12034 -0.0842,-0.12402 z m 0,1.05833 c -0.0402,-0.006 -0.068,0.0911 -0.076,0.28371 -0.0115,0.27886 0.0511,0.43465 0.13901,0.34675 0.0439,-0.044 0.0684,-0.12308 0.0723,-0.21498 0.004,-0.0919 -0.013,-0.19595 -0.0512,-0.29145 -0.0317,-0.0792 -0.0601,-0.12035 -0.0842,-0.12403 z",
1.49495, 18.3425,
"M -0.389083 -18.3412  c -0.0137,-0.002 -0.0282,-0.002 -0.0419,0.002 -0.10038,0.0279 -0.19454,0.22976 -0.24185,0.70177 -0.26567,1.67026 -0.34483,3.24287 -0.40257,4.9351 -0.13546,0.59246 -0.11445,1.16489 -0.0517,1.73632 -0.0585,0.4088 -0.0804,0.92918 0.002,1.32964 -0.12052,1.75209 -0.60079,3.29057 -0.2346,5.20847 0.16549,1.86259 -0.34446,4.07682 0.14933,5.25032 0.0936,0.18527 0.19288,-0.31455 0.29921,-0.15916 0.13603,-0.0189 0.22781,-0.1762 0.28991,-0.41806 0.0507,0.92595 0.15436,1.84189 0.32038,2.69544 0.53221,-0.0551 0.41124,0.9134 0.20775,1.79317 0.51184,0.22894 0.34367,2.25374 0.002,2.89388 -0.0368,0.48543 -0.0891,1.00555 -0.13539,1.53634 0.0369,-1.13406 0.0173,-2.27104 -0.57621,-2.24896 -0.21034,3.7268 -0.29522,7.48925 -0.0279,11.21121 0.34301,0.69517 0.62318,-0.48978 0.76737,-0.70229 -0.15983,-1.72878 -0.22984,-3.51098 -0.2041,-5.28185 0.0306,0.26264 0.0729,0.51868 0.13383,0.76275 0.76367,0.6551 0.19733,-2.26198 0.72554,-2.8298 0.19862,-1.54766 0.11464,-3.48741 0.30178,-5.14232 -0.27355,-1.32725 -0.3366,-3.24034 0.42067,-3.55069 0.7031,-1.9288 -0.89779,-1.8191 -0.56743,-3.27577 0.0274,-1.9927 -0.44524,-3.31209 -0.79632,-4.81779 -0.27045,1.00189 -0.45267,2.49784 -0.54364,3.82767 -0.0214,-0.31841 -0.0436,-0.64831 -0.0615,-0.91622 -0.0602,-1.06935 0.3048,-2.00573 0.18499,-3.34967 0.15293,-0.0682 0.31221,-0.28396 0.43977,-0.72089 0.4395,-1.55568 -0.22239,-3.13405 -0.0129,-4.73201 -0.24661,-1.51874 -0.005,-3.03839 -0.11162,-4.46122 0.17729,-0.52717 -0.0266,-1.24147 -0.23461,-1.27745 z m 0.33951,11.62978 c 2.4e-4,-9.5e-4 2.7e-4,-0.002 5.1e-4,-0.003 h -0.002 c 2.4e-4,0.001 7.7e-4,0.002 10e-4,0.003 z",
1.53299, 18.3425,
"M -0.351212 -18.3412  c -0.0137,-0.002 -0.0281,-0.002 -0.0418,0.002 -0.10036,0.0279 -0.19452,0.22975 -0.24183,0.70176 -0.26567,1.67026 -0.34483,3.24287 -0.40256,4.9351 -0.13547,0.59246 -0.11444,1.1649 -0.0517,1.73633 -0.0484,0.33819 -0.0679,0.74913 -0.0279,1.10898 -0.19071,1.02501 -0.32009,2.2804 -0.38034,3.36517 -0.0761,1.42435 -0.0331,2.95284 0.15555,4.37751 -0.0636,1.14999 -0.12615,2.23412 0.16949,2.93677 0.0697,0.13803 0.14251,-0.10384 0.21913,-0.17208 -0.005,0.0233 -0.01,0.047 -0.015,0.0703 0.51186,0.22893 0.34366,2.25375 0.002,2.89388 -0.1252,1.65317 -0.43757,3.69316 -0.0424,5.2772 0.0544,0.0467 0.10184,0.0733 0.14364,0.0858 -0.12761,3.049 -0.14573,6.11235 0.0724,9.1488 0.343,0.69517 0.62267,-0.48977 0.76686,-0.70228 -0.15983,-1.72878 -0.22934,-3.51098 -0.20359,-5.28185 0.0306,0.26264 0.0724,0.51867 0.13332,0.76274 0.76367,0.6551 0.19733,-2.26198 0.72554,-2.8298 0.19862,-1.54766 0.11465,-3.48741 0.30179,-5.14232 -0.27356,-1.32725 -0.3361,-3.24034 0.42116,-3.55069 0.7031,-1.9288 -0.89829,-1.8191 -0.56793,-3.27577 0.0274,-1.9927 -0.44524,-3.31209 -0.79634,-4.81779 -0.0429,0.15891 -0.0825,0.33356 -0.12092,0.5147 -0.005,-0.41675 -0.0275,-0.80863 -0.0698,-1.17513 0.0749,-0.11708 0.14656,-0.28021 0.21035,-0.49867 0.4395,-1.55568 -0.22191,-3.13405 -0.0124,-4.73201 -0.24662,-1.51874 -0.005,-3.0384 -0.11163,-4.46123 0.1773,-0.52717 -0.0266,-1.24146 -0.23461,-1.27744 z m 0.33952,11.62978 c 2.6e-4,-9.5e-4 2.9e-4,-0.002 5.3e-4,-0.003 h -0.002 c 2.4e-4,0.001 8e-4,0.002 10e-4,0.003 z m -0.0956,9.67279 c 0.34693,0.1444 0.2326,0.99256 0.0522,1.77251 0.51183,0.22894 0.34374,2.25373 0.002,2.89388 -0.0368,0.48543 -0.0891,1.00555 -0.13539,1.53634 0.0309,-0.95055 0.0204,-1.90052 -0.33279,-2.17093 0.0287,-0.40839 0.0871,-0.76874 0.23823,-0.93121 0.1206,-0.9398 0.13944,-2.0231 0.17568,-3.10059 z",
1.20633, 23.4716,
"M 1.20633 -23.4715  c -0.5053,0.33607 -0.23791,1.76733 -0.95755,1.18391 l -0.14986,-0.19483 c -0.008,-0.0416 -0.016,-0.007 -0.0238,0.0713 v 5.2e-4 c 0.18164,0.13239 0.32232,0.3276 0.39534,0.62063 0.0427,0.088 0.061,0.19852 0.061,0.32143 -0.002,0.4222 -0.235,0.98918 -0.51157,1.29759 -0.021,1.53032 -0.0353,3.79606 -0.046,6.48384 0.0137,0.012 0.0215,0.0282 0.0155,0.0491 -0.001,0.008 -0.0116,0.0176 -0.0155,0.0264 -0.003,0.66229 -0.002,1.56265 -0.005,2.27222 -0.26217,1.91668 -0.35258,3.7582 -0.4196,5.72264 -0.4378,1.91459 0.36031,3.66604 0.19897,5.64617 -0.22442,0.63928 -0.0753,1.29189 0.201,1.54461 -0.003,11.37386 0.0137,22.45681 -0.006,21.87567 0.14362,-0.69488 -0.32321,-2.78058 0.72811,-2.35232 -0.12814,0.70466 0.19505,1.16669 0.36795,1.67432 -0.0668,-2.83822 -0.14226,-12.31522 -0.15967,-21.95783 0.0124,-0.0381 0.0257,-0.0691 0.0377,-0.11007 0.42336,-1.49849 0.0674,-3.0137 -0.0413,-4.54288 2.1e-4,-0.48242 10e-4,-0.94352 0.002,-1.42058 0.007,-0.0756 0.0123,-0.15121 0.0222,-0.22686 -0.008,-0.051 -0.0143,-0.102 -0.0217,-0.15296 0.008,-5.2851 0.0362,-10.20918 0.10488,-13.52062 -0.0288,0.0211 -0.0623,0.0406 -0.10282,0.0573 -0.22574,-0.68079 -0.0267,-0.7215 0.11576,-0.56947 0.0507,-2.17256 0.11869,-3.59653 0.20978,-3.79925 z m -2.21742,27.16423 c -0.21037,3.7268 -0.29472,7.48924 -0.0274,11.2112 0.343,0.69518 0.62317,-0.48977 0.76737,-0.70228 -0.19468,-2.10548 -0.25887,-4.29063 -0.17518,-6.43733 -0.14325,-1.11457 0.39523,-4.10729 -0.5643,-4.07159 z",
1.19276, 25.078,
"M -0.421376 -3.88054  c -0.27631,0.25272 -0.42545,0.90533 -0.20103,1.54461 0,0.0137 0,0.0271 0.003,0.0408 -0.0445,-0.0116 -0.0846,-0.0199 -0.13695,-0.0377 l -0.11678,-0.0403 0.0201,0.0682 c 0.1305,0.45681 0.22437,0.8044 0.24133,0.92553 0,0.015 0,0.03 -0.003,0.045 -0.003,8e-4 -0.005,0.002 -0.0109,-0.005 0.003,0.20217 -0.0377,0.38449 -0.14211,0.45011 0.0449,0.007 0.0822,0.0209 0.11626,0.0382 -0.0299,0.3032 -0.0693,0.6025 -0.11057,0.90072 -0.0932,0.0649 -0.2028,0.0993 -0.32401,0.078 -0.14565,0.51493 0.25239,1.08423 -0.0393,1.56063 0.0722,0.17488 0.13539,0.34219 0.17312,0.50901 0.003,0.14782 0.0111,0.29635 0.0254,0.44545 -0.0156,0.11938 -0.0498,0.24159 -0.11213,0.36897 0.24339,0.56784 0.14218,1.21602 0.18034,1.82263 0.0321,0.60964 -0.054,1.23666 0.0951,1.83296 0.21513,0.29946 -0.39643,0.31017 -0.11369,0.67955 0.039,0.74934 0.21111,1.48345 0.17364,2.23707 0.0111,0.43606 0.0354,1.12135 -0.22066,1.29191 0.64399,0.51467 -0.31361,0.91939 -0.18965,1.4857 0.0477,0.49885 0.30371,1.02019 0.0961,1.48363 -0.0396,0.59756 -0.0501,1.20341 0.21238,1.7234 0.0946,0.62129 -0.0953,1.21664 -0.0124,1.84589 -0.0453,0.5497 0.3098,1.19228 -0.0542,1.66449 -0.19063,0.44185 0,0.91783 -0.19327,1.33791 -0.0311,0.51712 -0.0706,1.05694 -0.0429,1.57355 -0.0493,0.5227 -0.0711,1.06009 -0.003,1.58078 0.079,0.51085 0.0409,0.93085 -0.0816,1.42058 0.0312,0.0244 0.0439,0.0586 0.0692,0.0863 l 1.44333,-3.35897 c -0.0392,-0.12324 -0.10192,-0.24394 -0.21963,-0.35346 -0.0796,-0.4752 0.0952,-0.97383 0.23461,-1.4056 0.16936,0.27841 0.1652,0.98193 0.50075,1.20509 -0.0911,-0.20272 -0.15912,-1.62669 -0.20982,-3.79925 -0.14245,0.15203 -0.34149,0.11132 -0.11575,-0.56947 0.0405,0.0167 0.074,0.0363 0.10284,0.0574 -0.006,-0.28889 -0.0101,-0.65271 -0.0156,-0.96532 0.0822,-0.14175 0.16769,-0.28152 0.24908,-0.42891 -0.0214,-0.20087 -0.14751,-0.3117 -0.26252,-0.4191 -0.005,-0.33733 -0.0116,-0.6269 -0.0161,-0.98547 0.0805,-0.35753 0.10441,-0.72698 0.19428,-1.08055 0.0366,-0.47155 -0.0944,-0.90339 -0.21703,-1.34462 -0.0209,-2.51731 -0.0327,-5.33076 -0.0372,-8.29665 0.007,-0.051 0.0138,-0.10197 0.0217,-0.15297 -0.01,-0.0757 -0.0153,-0.15126 -0.0222,-0.22685 0,-0.47706 -0.003,-0.93817 -0.003,-1.42059 0.0257,-0.36017 0.0648,-0.71939 0.10441,-1.07797 0.10797,-0.15575 0.16115,-0.31721 0.0486,-0.4532 0.0116,-0.11133 0.0217,-0.22266 0.032,-0.33383 0.56079,-1.63193 0.75962,-1.19491 0.0356,-1.64228 -0.0274,-0.34604 -0.082,-0.69129 -0.1793,-1.0356 -0.0119,-0.041 -0.0254,-0.072 -0.0377,-0.11007 0.0175,-9.64261 0.0929,-19.11961 0.15967,-21.95783 -0.1729,0.50763 -0.49607,0.96966 -0.36793,1.67432 -0.80568,3.15608 0.55147,3.77736 -0.72101,19.52334 z",
0.670938, 8.19849,
"M -0.407338 -8.00763  c -0.0754,0.25511 -0.0986,0.76103 -0.002,1.48952 -0.15641,1.5163 0.005,3.60812 -0.2616,5.1103 0.11078,2.14938 0.16104,4.46253 0.31355,6.6328 0.0244,0.8744 0.0171,4.68192 0.38457,2.0748 0.22082,-1.3347 0.65056,-2.59474 0.58926,-4.39985 0.27685,-1.75982 -0.59717,-2.80885 -0.47885,-4.72632 -0.0715,-1.81309 0.41368,-3.66693 -0.0322,-5.23362 -0.11395,-1.10094 -0.38606,-1.37269 -0.51194,-0.94764 z",
0.670938, 12.9232,
"M -0.407338 -12.6223  c -0.0754,0.40213 -0.0986,1.19961 -0.002,2.34792 -0.15641,2.39013 0.005,5.68743 -0.2616,8.05531 0.11078,3.38804 0.16104,7.03423 0.31355,10.4552 0.0244,1.37832 0.0171,7.38006 0.38457,3.27049 0.22082,-2.10387 0.65056,-4.09006 0.58926,-6.93544 0.27685,-2.77398 -0.59717,-4.42755 -0.47885,-7.45005 -0.0715,-2.85794 0.41368,-5.78013 -0.0322,-8.24969 -0.11395,-1.7354 -0.38606,-2.16376 -0.51194,-1.49375 z",
0.753237, 9.39521,
"M -0.337287 -9.3947  c -0.0586,-0.007 -0.11261,0.0579 -0.15195,0.19068 h -5.1e-4 c -0.0754,0.25511 -0.0987,0.76083 -0.002,1.48932 -0.15642,1.51629 0.005,3.6081 -0.26149,5.11028 0.10925,2.11959 0.16037,4.39716 0.308,6.5412 -0.63693,1.04513 0.47877,2.68047 -0.28112,3.96927 0.31912,0.41893 0.25879,2.08373 1.06919,1.26659 0.80684,-0.22766 0.17536,-1.16368 0.17414,-2.66186 -0.30014,-0.12043 -0.46132,-0.32959 -0.54001,-0.58446 0.22741,-1.27036 0.61566,-2.49792 0.55705,-4.22299 0.27686,-1.75983 -0.59735,-2.80886 -0.47902,-4.72633 -0.0715,-1.81309 0.41383,-3.6671 -0.032,-5.23379 -0.0783,-0.7569 -0.23124,-1.12184 -0.36017,-1.13791 z",
1.73753, 21.1382,
"M 0.27454 -21.1377  c -0.0586,-0.007 -0.11258,0.0573 -0.15192,0.19016 l -5e-4,5.2e-4 c -0.0754,0.25511 -0.0987,0.76083 -0.002,1.48932 -0.15452,1.498 -4.5e-4,3.55549 -0.25371,5.05343 -0.15516,0.069 -0.19791,0.18561 -0.17881,0.37363 0.093,0.91418 -0.0185,1.41725 -0.27855,1.25677 -0.1455,-0.0898 -0.26453,-0.51355 -0.26459,-0.94155 3e-5,-0.49146 -0.137,-0.77825 -0.37155,-0.77825 -0.2589,0 -0.40127,0.42059 -0.47077,1.38907 -0.0191,0.26627 -0.0317,0.52238 -0.0367,0.76739 -0.0365,1.71506 0.26421,2.89312 0.85627,3.30781 0.5721,0.40073 0.5816,0.49454 0.39946,3.88917 -0.10265,1.91329 -0.16684,4.25278 -0.14263,5.19865 0.0283,1.10976 -0.0992,1.98889 -0.35968,2.47788 -0.50411,0.94668 -0.52832,1.39496 -0.0542,1.00149 0.28816,-0.23923 0.33626,1.20389 0.27647,8.26771 -0.0584,6.89886 -0.002,8.65365 0.28887,9.05165 0.33266,0.45496 0.37624,0.41244 0.55449,-0.53795 0.10638,-0.56717 0.0821,-1.77797 -0.0543,-2.6908 -0.13639,-0.91282 -0.17248,-1.61256 -0.0796,-1.55546 0.0928,0.0573 0.18032,-0.51603 0.19431,-1.27434 0.0252,-1.36423 0.11504,-1.96676 0.60204,-4.02456 0.13776,-0.58206 0.29511,-3.3592 0.34933,-6.1712 0.0542,-2.81205 0.20923,-5.22336 0.34467,-5.35884 0.13546,-0.13545 0.26493,-0.75165 0.28734,-1.36943 0.0309,-0.85348 -0.0157,-0.99638 -0.19431,-0.59428 -0.17402,0.392 -0.23993,0.42036 -0.25374,0.11007 -0.0102,-0.23043 0.0657,-0.50284 0.16846,-0.60565 0.1623,-0.16235 0.15642,-3.36764 -0.0145,-8.03723 -0.0279,-0.76399 0.0292,-1.38905 0.12663,-1.38906 0.0973,-1e-5 0.17671,-0.41671 0.17671,-0.92604 0,-0.90243 -0.0261,-0.92604 -1.04799,-0.92605 -0.0114,0 -0.0199,4.9e-4 -0.031,5.2e-4 0.001,-0.0895 0.002,-0.17939 0.008,-0.27285 -0.0715,-1.81309 0.41386,-3.6671 -0.032,-5.23379 -0.0784,-0.7569 -0.23125,-1.12184 -0.36021,-1.13791 z",
0.90975, 7.69929,
"M 0.5118 4.3328  0.0773,-1.89734 c -0.78022,-0.65432 -0.7026,-1.85535 0.0991,-2.36604 l 0.22155,-5.30816 c -0.0671,-0.11321 -0.12273,-0.0764 -0.193,-0.25561 -0.39953,-0.40971 -0.50958,-1.87041 -0.78623,-2.16685 l -0.14591,-0.0381 c -0.0432,0.0195 -0.10273,0.16373 -0.15446,0.23376 l -0.5399,11.53632 c 0.09,-0.0554 0.16763,-0.042 0.26343,-0.13249 0.2235,1.06334 -0.41852,2.98144 0.13712,3.761 1.10534,-0.76211 0.3114,-2.4459 1.02162,-3.36601 z",
0.909985, 21.0802,
"M 0.069295 -21.0802  c -1.60221,0.95118 0.0978,2.94503 -0.79529,4.45968 0.31909,0.41893 0.25829,2.08373 1.06866,1.26659 0.80687,-0.22766 0.17589,-1.16368 0.17467,-2.66185 -1.23089,-0.49394 -0.14943,-2.46581 -0.44751,-3.06442 z m -0.28475,6.45077 c -0.0432,0.0195 -0.10276,0.16407 -0.15449,0.2341 l -0.54004,11.53624 c 0.09,-0.0554 0.16777,-0.0423 0.26355,-0.13281 0.22352,1.06334 -0.41867,2.98145 0.13695,3.76101 0.11266,-0.0777 0.20122,-0.16697 0.27803,-0.262 -0.02,0.26698 -0.062,0.40312 -0.11938,0.34572 -0.0852,-0.0852 -0.1323,0.67721 -0.1323,2.13269 0,0.59617 -0.0104,1.04983 -0.0263,1.41593 -0.043,0.0648 -0.0864,0.12977 -0.12454,0.18552 -0.0367,0.0537 -0.0978,0.0547 -0.11316,0.11265 -0.0217,0.0815 0.0922,0.078 0.0744,0.15503 -0.0144,0.0623 -0.10165,-0.009 -0.1116,0.0537 -0.0128,0.0803 0.18314,0.1196 0.17,0.20206 -0.007,0.0452 -0.0103,0.091 -0.009,0.13746 -9.3e-4,1.7e-4 -0.002,0.003 -0.003,0.003 h -5.1e-4 c -0.0783,0 -0.1189,0.16869 -0.091,0.37517 0.0279,0.20649 0.12144,0.35145 0.20825,0.32194 0.003,-9.4e-4 0.005,0.001 0.008,0.001 0.0106,0.0622 0.0188,0.12677 0.0248,0.19224 10e-4,0.0292 0.002,0.0591 0.004,0.0884 0.002,0.0585 0.005,0.11669 0,0.17931 8.2e-4,0.0518 -0.0608,0.12352 -0.0594,0.17312 0.001,0.0496 0.0533,0.0771 0.0553,0.12454 0.002,0.0475 -0.046,0.11479 -0.0434,0.16019 0.002,0.0335 0.0533,0.17164 0.0863,0.27079 0.0126,0.33511 0.0159,0.66968 -0.007,1.00149 -0.0129,0.0186 -0.027,0.0231 -0.0424,0.008 -0.0679,-0.068 -0.10681,0.27415 -0.11419,0.92501 -0.066,0.23273 -0.1523,0.46252 -0.27183,0.68678 0.1062,0.21699 0.19947,0.37376 0.28527,0.49919 0.009,0.30425 0.0207,0.49363 0.0408,0.46354 -0.008,0.28451 -0.0302,0.75484 -0.0398,1.26349 -0.10361,0.44818 -0.1356,1.31544 -0.0109,1.44022 0.14782,0.14777 0.17589,1.00278 0.0796,2.44481 -0.0191,0.28562 -0.0175,0.6403 -0.003,0.9586 -0.0617,0.0399 -0.12502,0.089 -0.19225,0.1571 0.0518,0.38782 0.074,0.81499 0.0935,1.24953 -0.004,0.003 -0.009,0.003 -0.0124,0.007 -0.0863,0.0863 -0.11387,0.41392 -0.0672,0.79169 0.0438,0.35436 0.0791,0.9009 0.0791,1.21439 0,0.46794 0.0655,0.57783 0.36483,0.61443 0.0158,0.002 0.023,0.001 0.0372,0.003 0.0344,0.0746 0.0714,0.14651 0.11266,0.21445 0.15124,-0.0321 0.23686,-0.14678 0.29507,-0.28525 0.0212,0.0538 0.0649,0.0638 0.11575,0.0129 0.0515,-0.0515 0.0774,-0.33668 0.0873,-0.67231 0.29014,-0.88165 0.123,-3.135 -0.53848,-3.37654 l -5e-4,5.2e-4 c -0.0427,-0.0173 -0.0878,-0.0231 -0.13489,-0.0207 0.002,-0.84663 0.0224,-1.14112 0.0972,-1.12552 0.002,4.9e-4 0.005,0.004 0.008,0.005 0.1579,0.11986 0.38113,0.13515 0.67231,-0.0475 -0.022,-0.60336 -0.25202,-1.93024 -0.50488,-2.00556 -0.0135,-0.004 -0.027,-0.005 -0.0408,-10e-4 0.0589,-0.0887 0.12261,-0.19962 0.19431,-0.34623 -0.0435,-0.19088 -0.0688,-0.3762 -0.0847,-0.56017 0.0114,0.004 0.0236,0.008 0.0362,0.016 0.24339,0.15111 0.38658,-0.25127 0.37258,-1.04593 -0.006,-0.35248 0.0815,-0.86868 0.19431,-1.14722 0.12518,-0.30915 0.16568,-0.65349 0.10388,-0.88367 -0.0556,-0.20738 -0.10724,-0.52146 -0.11422,-0.69814 -0.007,-0.17668 -0.10248,-0.32443 -0.21188,-0.32763 -0.0382,-9.7e-4 -0.0689,-0.0206 -0.0946,-0.0522 0.007,-0.0407 0.0134,-0.0917 0.0212,-0.12815 0.0171,-0.0261 0.0359,-0.0485 0.0511,-0.0729 l 5.3e-4,-0.002 -0.047,-3.21014 c 0.0994,-0.6389 0.22457,-1.57045 0.1824,-1.57045 -0.0641,0 -0.11626,-0.10246 -0.11626,-0.22738 0,-0.12493 0.0594,-0.16741 0.13229,-0.0946 0.0729,0.0729 0.1328,-0.0128 0.1328,-0.18966 0,-0.17699 -0.0574,-0.3783 -0.12711,-0.44803 -0.0698,-0.0697 -0.0933,-0.26863 -0.0522,-0.44184 0.0465,-0.19637 -0.005,-0.3147 -0.13798,-0.3147 -0.11716,0 -0.21341,-0.13706 -0.21341,-0.30438 0,-0.16734 -0.0396,-0.47302 -0.0884,-0.67903 -0.0605,-0.2553 -0.0396,-0.32487 0.0661,-0.21911 0.10713,0.10714 0.15502,-0.0773 0.15502,-0.59582 0,-0.65769 0.18227,-0.88497 0.0692,-0.89452 -0.0382,-0.003 -0.10946,0.0186 -0.23151,0.0563 0.0311,-0.6426 0.004,-1.30921 0.36587,-1.77818 l -5e-4,-5.2e-4 0.077,-1.89756 c -0.78023,-0.65432 -0.70247,-1.85506 0.0992,-2.36574 l 0.22169,-5.30821 c -0.0671,-0.11321 -0.123,-0.0766 -0.19328,-0.2558 -0.39952,-0.40971 -0.50935,-1.87035 -0.78599,-2.16679 z m 0.0336,28.36158 c -0.0146,0.0334 -0.0293,0.0897 -0.0439,0.13384 -0.006,-0.0264 -0.0126,-0.0554 -0.0196,-0.0832 0.0197,-0.011 0.0416,-0.0302 0.0636,-0.0506 z",
0.70928, 12.466,
"M 0.18321 -12.4659  c -0.10784,-0.006 -0.2246,0.65736 -0.33539,2.3642 -0.2559,1.61556 0.0378,4.44866 0.86146,3.05511 -0.022,-1.62742 -0.25194,-5.20773 -0.50488,-5.40898 l -0.002,-0.002 c 4.3e-4,3.1e-4 6.4e-4,0.002 0.001,0.002 -0.007,-0.006 -0.0136,-0.01 -0.0207,-0.0103 z m -0.37155,7.36079 c -0.0907,0.007 -0.12097,0.50508 -0.0698,1.9544 -0.26241,4.289499 0.38667,9.178069 -0.45114,13.416749 1.02915,5.67045 1.03722,-1.11191 1.24077,-3.68918 0.0171,-0.0704 0.0353,-0.13161 0.0506,-0.1974 l 0.001,-0.005 -0.0532,-9.785969 c -0.32449,-0.66216 -0.58576,-1.7036 -0.71829,-1.69395 z",
0.791801, 10.8742,
"M 0.37094 -10.8739  c -0.0383,-0.003 -0.10962,0.0185 -0.23202,0.0563 -0.32348,0.10008 -0.3638,0.20621 -0.3638,0.95808 0,0.50923 -0.053,0.793 -0.13282,0.71313 -0.0852,-0.0852 -0.13229,0.67721 -0.13229,2.13269 0,0.59618 -0.0104,1.04983 -0.0263,1.41593 -0.043,0.0648 -0.0864,0.12977 -0.12454,0.18552 -0.0367,0.0537 -0.0978,0.0547 -0.11316,0.11266 -0.0217,0.0815 0.0922,0.078 0.0744,0.15503 -0.0144,0.0623 -0.10166,-0.009 -0.11161,0.0537 -0.0128,0.0803 0.18315,0.11959 0.17,0.20205 -0.007,0.0452 -0.01,0.091 -0.009,0.13746 -9.3e-4,1.7e-4 -0.002,0.003 -0.003,0.003 h -5.3e-4 c -0.0783,0 -0.11885,0.16869 -0.0909,0.37517 0.0279,0.20648 0.12145,0.35145 0.20826,0.32195 0.003,-9.4e-4 0.005,0.001 0.008,0.001 0.0245,0.14353 0.0418,0.2935 0.0284,0.45992 8e-4,0.0518 -0.0608,0.12351 -0.0594,0.17311 10e-4,0.0496 0.0533,0.0771 0.0553,0.12454 0.002,0.0475 -0.046,0.11479 -0.0434,0.1602 0.003,0.0454 0.11377,0.32348 0.11679,0.3669 0.0104,0.14968 0.0239,0.28357 0.0383,0.40928 -0.0199,0.3693 -0.0606,0.56165 -0.11834,0.50384 -0.0843,-0.0844 -0.12634,0.43821 -0.11526,1.43558 0.009,0.84768 0.0233,1.18693 0.0553,1.13894 -0.008,0.28451 -0.0302,0.75484 -0.0398,1.26349 -0.10361,0.44818 -0.13563,1.31545 -0.0109,1.44023 0.1478,0.14777 0.1759,1.00278 0.0796,2.44481 -0.0316,0.47191 -0.0123,1.14721 0.0434,1.5012 0.0959,0.60757 0.19674,0.59567 0.19274,-0.21446 -0.006,-1.40297 0.0164,-1.64026 0.14005,-1.5167 0.0647,0.0647 0.19682,-0.0363 0.29353,-0.22479 0.14898,-0.29044 0.13581,-0.38968 -0.0848,-0.65061 -0.14306,-0.1693 -0.24148,-0.32752 -0.21857,-0.35192 0.0229,-0.0244 -0.002,-0.23205 -0.0569,-0.46095 -0.0157,-0.0663 -0.0256,-0.11223 -0.0398,-0.16847 0.10073,-0.0561 0.22942,-0.23081 0.39328,-0.56585 -0.0435,-0.19088 -0.0689,-0.37621 -0.0848,-0.56018 0.0114,0.004 0.0236,0.008 0.0362,0.016 0.24339,0.15111 0.38658,-0.25127 0.37259,-1.04593 -0.006,-0.35248 0.0815,-0.86867 0.19428,-1.14721 0.12517,-0.30915 0.16568,-0.65349 0.10387,-0.88367 -0.0556,-0.20738 -0.1072,-0.52147 -0.11419,-0.69815 -0.007,-0.17668 -0.10247,-0.32443 -0.21188,-0.32763 -0.0396,-0.001 -0.0716,-0.0212 -0.0977,-0.0548 -0.006,-0.51267 -0.0453,-1.07976 -0.14057,-1.72237 -0.0104,-0.12615 -0.027,-0.22448 -0.0444,-0.32195 0.004,-0.009 0.006,-0.0184 0.0104,-0.0274 0.0792,-0.15448 0.12295,-0.41373 0.0971,-0.57619 -0.0258,-0.16259 -0.0261,-0.2425 -5.3e-4,-0.17828 0.0256,0.0642 0.35351,-2.15491 0.2894,-2.15491 -0.0641,0 -0.11628,-0.10246 -0.11628,-0.22738 0,-0.12493 0.0594,-0.16745 0.13229,-0.0946 0.0729,0.0729 0.13282,-0.0128 0.13282,-0.18965 0,-0.17699 -0.0574,-0.3783 -0.12713,-0.44803 -0.0698,-0.0697 -0.0933,-0.26863 -0.0522,-0.44184 0.0465,-0.19637 -0.005,-0.31471 -0.13798,-0.31471 -0.11715,0 -0.21344,-0.13705 -0.21344,-0.30437 0,-0.16734 -0.0396,-0.47302 -0.0884,-0.67903 -0.0605,-0.2553 -0.0396,-0.32487 0.0661,-0.21911 0.10713,0.10713 0.15504,-0.0773 0.15504,-0.59583 0,-0.6594 0.18355,-0.88659 0.0687,-0.89503 z m -0.77409,10.7518 c 0.008,0.0177 0.0132,0.12011 0.0176,0.27698 -0.03,0.0588 -0.0609,0.11849 -0.0967,0.18449 0.0414,-0.31691 0.0656,-0.49083 0.0791,-0.46147 z m -0.14833,8.34212 c -0.0352,-0.005 -0.064,0.002 -0.0842,0.0222 -0.0863,0.0863 -0.1139,0.41391 -0.0672,0.79168 0.0438,0.35437 0.079,0.90091 0.079,1.2144 0,0.46794 0.0655,0.57783 0.36484,0.61443 0.27919,0.0342 0.31816,-0.004 0.16586,-0.16433 -0.17018,-0.17883 -0.16531,-0.21039 0.0331,-0.21239 0.12763,-0.002 0.23204,0.0988 0.23204,0.22376 0,0.1249 0.0583,0.16907 0.1297,0.0977 0.14147,-0.14143 0.0967,-2.02722 0.0274,-1.95802 -0.0692,0.0692 -0.28162,-0.0722 -0.47181,-0.31419 -0.14263,-0.18155 -0.30326,-0.30098 -0.40875,-0.31523 z",
1.07687, 9.8574,
"M -0.468434 -9.85737  c 0.35117,1.50705 0.82497,2.82661 0.79752,4.82073 -0.33038,1.45668 1.2704,1.34706 0.56728,3.27586 -0.75726,0.31035 -0.69421,2.22344 -0.42064,3.55069 -0.18715,1.65491 -0.10345,3.59501 -0.30207,5.14267 -0.52822,0.56782 0.0385,3.48463 -0.72515,2.82953 -0.39519,-1.58404 -0.0832,-3.62405 0.042,-5.27722 0.3416,-0.64014 0.50986,-2.66493 -0.002,-2.89386 0.20349,-0.87978 0.32479,-1.84821 -0.20741,-1.79313 -0.33006,-1.69697 -0.41673,-3.64057 -0.32176,-5.41903 0.0788,-1.41878 0.27531,-3.13331 0.5735,-4.23627 z",
2.2155, 17.1585,
"M 0.126795 -17.1552  c -0.0862,0.0295 -0.10147,0.26481 0.15555,0.17053 -0.0201,-0.0566 -0.0417,-0.0964 -0.0625,-0.12402 -0.0181,-0.0239 -0.0356,-0.0379 -0.0522,-0.045 h -5.3e-4 c -0.0144,-0.006 -0.0281,-0.006 -0.0403,-0.002 z m -0.11575,1.07797 c -0.2067,8.5e-4 -0.36484,0.15284 -0.2589,0.56792 -0.78081,0.4633 -0.40071,1.63009 -0.15092,2.37195 -0.26111,0.38499 -0.0968,0.69767 0.0336,0.94413 0.003,0.15192 0.005,0.3124 0.008,0.46612 -0.0424,0.0281 -0.0954,0.0546 -0.17159,0.078 -0.38372,0.36544 -0.96946,0.62763 -0.47024,1.2516 0.61513,-0.37333 -0.39214,1.24462 0.37827,1.11983 0.58034,0.63299 -0.16417,1.58188 -0.29199,0.99788 -0.36309,0.82279 -0.74943,2.75375 0.30697,2.90059 0.65154,0.38277 -1.22994,0.47827 -0.32247,1.03353 0.12882,0.11405 0.2826,0.54247 0.4527,0.34778 0.0882,-0.10334 0.15261,-0.21193 0.20463,-0.32349 0.004,0.47543 0.008,0.94591 0.0124,1.42162 -0.1546,-0.15317 -0.30245,-0.264 -0.41445,-0.27802 -0.19592,-0.0245 -0.2839,0.23429 -0.11266,1.02733 0.39989,1.176989 -1.12924,1.251499 -0.25992,2.377629 0.20293,1.10216 -0.30896,2.46141 -0.11473,3.72122 -0.26596,1.39175 1.38674,0.67395 0.53589,2.24534 -0.60793,1.13701 0.0733,2.34518 0.5674,3.05614 -0.24691,0.6918 -0.4676,-0.15486 -0.46559,-0.52399 -0.52125,-0.0811 -0.56888,0.96424 -0.10543,1.27589 0.44555,0.85109 -0.8023,1.42473 0.10078,2.29133 0.72273,0.54138 -0.004,1.44593 -0.46149,0.78445 -0.46868,0.32104 0.99774,0.86268 0.33591,1.56889 -0.12715,0.56519 0.112,1.93607 0.42477,0.65009 0.31115,-0.77077 0.98546,-1.04884 1.25005,-1.8526 -0.41074,-0.17674 -0.96338,-0.29486 -0.21032,-0.51986 0.87294,-0.30163 0.89675,-2.52384 0.20465,-3.22254 0.0979,-0.4361 -0.17264,-0.70085 -0.44082,-0.96635 0.0414,-0.83119 0.0704,-1.69667 0.11163,-2.32389 0.0512,-0.6609 0.0902,-1.54857 0.11988,-2.55954 0.17206,-0.23191 0.19942,-0.56928 0.0243,-0.8847 0.002,-0.0714 0.002,-0.15084 0.004,-0.22324 0.21306,-0.76122 0.28736,-1.5621 0.0243,-2.17454 0.002,-0.29119 0.003,-0.5884 0.004,-0.88677 0.22603,-0.40201 0.27916,-1.31894 0.0305,-1.5658 -0.008,-0.008 -0.0188,-0.0119 -0.0279,-0.0186 -0.0212,-4.586229 -0.18235,-9.480359 -0.49194,-11.419969 0.21301,-0.40925 0.4383,-0.83985 0.42579,-1.43867 0.27437,-0.6706 -0.68641,-0.14121 -0.21238,-0.89969 -0.0211,-0.10544 -0.0977,-0.20384 -0.19741,-0.27854 l -0.001,-10e-4 c -0.11229,-0.0842 -0.25379,-0.13798 -0.37774,-0.13746 z m -1.07128,3.59616 c -0.046,0.0125 -0.11609,0.15283 -0.19843,0.50126 -0.0447,0.11946 -0.12068,0.55421 0.0775,0.36484 0.19362,-0.34691 0.21455,-0.81723 0.14004,-0.86403 h -5e-4 c -0.006,-0.003 -0.012,-0.004 -0.0186,-0.002 z m -0.6356,1.5627 c -0.079,-0.0128 -0.19479,0.0402 -0.34883,0.19792 -0.19704,0.0386 -0.26882,0.95532 0.0331,0.63458 0.39754,-0.0448 0.5438,-0.64657 0.38243,-0.79943 -0.0183,-0.0177 -0.0403,-0.0276 -0.0667,-0.0326 z m 3.53157,0.28783 c -0.0145,-0.003 -0.0279,-0.003 -0.0398,0.002 -0.0792,0.0285 -0.0914,0.22185 0.078,0.65371 0.0425,0.16251 0.28297,0.43936 0.34157,0.13023 -0.006,-0.29077 -0.11668,-0.53703 -0.22944,-0.67335 h -5.1e-4 c -0.0529,-0.0638 -0.10628,-0.10342 -0.14986,-0.11214 z m -3.06699,0.71417 c -0.11473,0.005 -0.36497,0.34787 -0.21913,0.6656 l 0.11472,-0.12816 c 0.15291,-0.30965 0.17669,-0.46521 0.14161,-0.51832 0,7.2e-4 0.002,-2.3e-4 0.002,5.2e-4 l -0.002,-0.002 v 5.2e-4 c -0.009,-0.012 -0.0222,-0.0194 -0.0372,-0.0186 z m 0.91829,1.10691 c 0,1.7e-4 0,3.5e-4 0,5.2e-4 -5.3e-4,1.2e-4 -0.001,-1.2e-4 -0.002,0 5.5e-4,-1.3e-4 9.8e-4,-3.8e-4 0.002,-5.2e-4 z m 1.47484,23.342759 c -0.0965,-0.0137 -0.20217,0.35222 0.0749,0.14676 -0.0109,-0.0548 -0.0222,-0.0916 -0.0357,-0.11472 -0.0122,-0.0202 -0.0259,-0.03 -0.0393,-0.032 z m -0.98237,0.9338 c -0.14174,0.0139 -0.29345,0.21991 -0.24496,0.71003 -0.0263,0.31998 0.15116,0.22384 0.22998,0.0662 0.29102,-0.34448 0.26099,-0.62222 0.14571,-0.72864 v -5.1e-4 c -0.037,-0.0349 -0.0834,-0.052 -0.13073,-0.047 z m 0.4005,1.03714 c -0.1714,-0.0358 -0.43241,0.24697 -0.11526,0.65423 l 0.1447,-0.10853 c 0.1255,-0.28076 0.10745,-0.44152 0.0388,-0.50952 l -0.002,-0.001 c -0.0189,-0.0187 -0.0416,-0.0301 -0.0661,-0.0351 z",
0.780383, 20.4192,
"M -0.105503 -20.4188  c -0.0405,-0.002 -0.0956,0.004 -0.16536,0.017 -0.29935,0.0549 -0.36433,0.21983 -0.36433,0.92115 0,0.46985 -0.0358,1.28906 -0.0796,1.82015 -0.0467,0.5662 -0.0191,1.05683 0.0672,1.18617 0.081,0.1214 0.30279,-0.076 0.49299,-0.43885 0.19016,-0.36274 0.40254,-0.575 0.47181,-0.47122 0.0693,0.10377 0.11406,-2.72234 -0.0274,-2.93432 -0.0714,-0.10706 -0.1297,-0.0411 -0.1297,0.14614 0,0.18729 -0.1044,0.33898 -0.23201,0.33532 -0.19841,-0.003 -0.20326,-0.0502 -0.0331,-0.31828 0.1142,-0.18005 0.12105,-0.25755 -5.3e-4,-0.2633 z m -0.26458,4.60917 c -0.0347,0.0205 -0.0779,0.24114 -0.11988,0.63953 -0.0557,0.53054 -0.075,1.54319 -0.0434,2.25048 0.0963,2.16125 0.0682,3.44269 -0.0796,3.66416 -0.16271,0.24387 -0.0597,2.39029 0.12248,2.55214 0.0663,0.0589 0.11298,0.78233 0.10335,1.60754 -0.0172,1.47849 -0.0189,1.48377 -0.11523,0.37494 -0.0781,-0.90009 -0.10126,-0.6552 -0.11523,1.22451 -0.0111,1.494811 0.0309,2.278081 0.11523,2.151631 0.0606,-0.0909 0.10337,0.21775 0.12195,0.82742 -0.11997,0.0834 -0.25326,0.46905 -0.39997,1.18617 0.0168,0.60699 0.0279,1.24832 0.0393,1.8896 -0.002,0.0692 -7.2e-4,0.13106 0.004,0.18789 0.10261,6.050652 0.0186,13.334762 0.56481,17.672782 0.34438,-0.35219 0.35478,-2.7314 0.4532,-3.9671 0.15868,-1.69005 0.20889,-5.0472 0.15865,-8.334692 0.08,-0.0549 0.10705,-0.21387 0.0698,-0.4495 -0.0411,-0.25961 -0.0181,-0.55719 0.0517,-0.66168 0.0697,-0.10451 0.12713,-0.40665 0.12713,-0.67191 0,-0.26509 -0.0599,-0.39341 -0.13282,-0.28418 -0.0729,0.10924 -0.13229,0.0454 -0.13229,-0.14188 0,-0.18721 0.0523,-0.34043 0.11629,-0.34043 0.0641,0 -0.26379,-3.32582 -0.28938,-3.22958 -0.0256,0.0962 -0.0253,-0.0239 5e-4,-0.26757 0.0258,-0.24348 -0.0179,-0.63167 -0.0971,-0.8632 -0.10261,-0.30004 -0.10358,-0.60608 -0.002,-1.06389 0.078,-0.35332 0.12729,-0.97353 0.10903,-1.37832 -0.0212,-0.47251 0.0377,-0.73954 0.16537,-0.74562 0.1094,-0.005 0.20484,-0.225601 0.21187,-0.490401 0.007,-0.2648 0.0586,-0.73603 0.1142,-1.04684 0.0618,-0.34499 0.0213,-0.86088 -0.10385,-1.32421 -0.11285,-0.41748 -0.19997,-1.19131 -0.19381,-1.71961 0.014,-1.19099 -0.12967,-1.79354 -0.37309,-1.56707 -0.1185,0.11022 -0.18552,-0.0461 -0.18552,-0.4333 0,-0.33342 -0.0706,-0.66331 -0.15658,-0.73284 -0.0967,-0.0782 -0.0238,-0.365 0.0372,-0.75073 0.0545,-0.34306 0.0797,-0.6532 0.0569,-0.6898 -0.0229,-0.0366 0.0755,-0.27457 0.21859,-0.52832 0.22053,-0.39106 0.23374,-0.53911 0.0847,-0.97441 -0.0967,-0.28243 -0.22886,-0.43441 -0.29353,-0.33744 -0.12361,0.18519 -0.1457,-0.17035 -0.14004,-2.27306 0.002,-0.68298 -0.0283,-0.98758 -0.0729,-0.96121 z",
2.50105, 20.7737,
"M 1.99533 -20.1853  c -0.0375,-0.25789 -0.33261,-0.42551 -0.56905,-0.58839 -0.23996,0.23872 -0.47489,0.49163 -0.6009,0.8121 -0.0812,0.48893 -0.29007,0.81482 -0.38877,1.29015 0.002,0.54267 -0.41169,0.99749 -0.40079,1.54481 -0.29754,0.24097 -0.55536,0.37238 -0.50367,0.84202 0.0614,0.57416 -0.22723,1.12925 -0.28672,1.7031 -0.14255,0.60343 -0.28189,1.20471 -0.38066,1.8095 -0.0293,0.59441 -0.2814,1.18246 -0.12046,1.75323 -0.2131,-0.15862 -0.45105,-0.0488 -0.16056,0.12001 -0.0693,0.38384 -0.39321,0.81136 -0.33681,1.29405 0.0267,0.64265 -0.0113,1.27603 -0.14224,1.89832 -0.003,0.60064 -0.10155,1.21537 -0.24239,1.79332 -0.22967,0.32796 0.15308,0.6786 0.11208,1.04197 0.2095,-0.3435 0.24518,-0.79906 0.5964,-1.0637 0.58133,0.18835 0.24969,0.75808 0.31901,1.19987 -0.26469,0.52559 -0.0489,0.91024 0.47559,1.11661 0.25093,-0.0203 0.43656,0.83787 0.43804,0.4442 -0.0618,-0.51559 0.4207,-0.87267 0.41211,-1.37583 0.013,-0.55963 0.17412,-1.11826 0.20939,-1.67553 -0.40638,-0.35369 -0.56249,-0.94788 -0.16108,-1.40288 -0.26584,-0.54166 0.46688,-1.0424 -0.027,-1.55948 -0.72899,-0.49884 0.91668,-0.76908 0.36771,-1.29742 0.0275,0.33925 -0.48808,0.48735 -0.42907,0.0774 0.21303,-0.33662 0.68355,-1.28644 0.0603,-1.29291 -0.1772,-0.18811 -0.2848,-1.18628 0.34402,-1.22453 0.56288,-0.22657 -0.41458,0.0439 -0.23023,-0.39889 0.21834,-0.27814 0.12847,-0.9857 0.5386,-0.9766 0.24676,0.4666 0.0743,-0.15601 -0.0145,-0.31345 0.32655,-0.0856 0.16625,-0.26335 0.21161,-0.564 -0.17635,-0.4556 0.40637,-0.92508 0.0292,-1.47058 -0.0333,-0.47745 0.11583,-1.16532 0.30655,-1.6979 0.30437,-0.47718 0.2963,-1.07386 0.51455,-1.54766 0.0565,-0.10923 0.0727,-0.20496 0.0598,-0.29086 z m -0.67527,5.18673 c -0.005,-0.0207 -0.0401,-0.0511 -0.1331,-0.0891 -0.12685,0.13115 0.14583,0.15153 0.1331,0.0891 z m -0.51447,2.32592 c -0.01,-0.026 -0.0378,-0.0564 -0.0928,-0.089 -0.1546,0.23186 0.13611,0.20162 0.0928,0.089 z m -0.55676,9.92556 c -0.004,-0.0178 -0.0117,-0.0421 -0.0244,-0.0749 l -3.8e-4,8e-5 c -0.10421,-0.01 0.0523,0.1987 0.0247,0.0748 z m -0.40593,0.62749 c -0.029,-0.13397 -0.065,-0.26569 -0.10661,-0.39404 -0.0954,0.0982 -0.57871,0.21425 -0.47507,0.55849 0.0692,0.55607 -0.16309,0.96182 -0.64341,1.21137 -0.634,0.49658 -0.78164,-0.34647 -0.88631,-0.80154 -0.48235,0.13084 -0.0802,0.8574 0.0167,1.20292 -0.36114,0.39753 0.0177,1.11042 -0.11524,1.68891 -0.23801,0.62967 0.24476,1.16486 0.31513,1.76431 0.0279,0.59829 0.024,1.19727 0.091,1.79736 l 3.7e-4,-8e-5 c -5.9e-4,0.16055 0.01,0.32283 0.0418,0.48106 0.14355,0.59006 0.21051,0.10446 0.26051,-0.22378 -0.0205,-0.63973 -0.27186,-1.28964 0.006,-1.90725 0.12037,-0.14839 0.19349,-0.5856 0.41752,-0.32353 0.46234,0.3705 0.3394,1.09108 0.53985,1.60971 0.27126,0.36763 0.39796,0.77484 0.51709,1.26206 -0.002,-0.30581 -0.0844,-1.00791 0.24199,-1.29206 0.11117,-0.32729 -0.38379,-0.91151 -0.20475,-1.38978 0.34489,-0.30119 -0.033,-0.25259 -0.0944,-0.49538 -0.24459,-0.22385 -0.04,-0.56816 -0.16501,-0.76622 -0.1494,-0.4034 -0.40711,-0.71884 -0.52267,-1.12154 -0.24067,-0.7641 0.5871,-1.03146 0.79857,-1.63 0.0764,-0.40778 0.0532,-0.8291 -0.0337,-1.23099 z m 0.89502,9.36585 c -0.0536,-0.20202 -0.20583,-0.37279 -0.33916,-0.2913 0.0256,0.46608 -0.22812,0.1812 -0.20758,-0.11481 0.50238,-0.20886 -0.136,-0.53619 -0.35884,-0.67825 -0.19176,0.37421 -0.0305,0.92091 -0.0919,1.3655 0.20082,0.60747 -0.0474,1.25588 0.23051,1.83712 0.46931,0.45584 0.1681,1.16914 0.30673,1.73368 -0.0209,1.03972 0.15888,2.08076 0.41848,3.08564 0.2179,0.9083 0.43884,1.83468 0.46756,2.76969 l -3.8e-4,8e-5 c -0.0559,0.1059 -0.0189,0.23303 -0.009,0.34529 -0.0466,0.49251 0.47733,1.10683 0.86726,0.99131 0.0869,-0.81136 -0.1849,-1.61553 -0.23186,-2.42464 -0.0153,-0.79202 -0.18322,-1.5676 -0.33686,-2.34287 -0.2662,-0.49125 0.35995,-0.65771 -0.11227,-1.06692 -0.29724,-0.73686 -0.33568,-1.55313 -0.49201,-2.32897 -0.13547,-0.6156 -0.29699,-1.09551 -0.32023,-1.74958 -0.42463,-0.16399 -0.2469,-0.50188 0.0726,-0.6905 0.13034,-0.0698 0.16881,-0.21103 0.15303,-0.35262 -0.003,-0.0295 -0.009,-0.059 -0.0165,-0.0878 z m -2.57871,-11.32017 c -0.2708,-0.55639 -0.38057,-0.0872 -0.31795,0.29402 l -0.0116,0.0821 c -0.0994,0.78435 0.40245,-0.0619 0.32958,-0.37609 z m 4.33185,23.83573 c -0.0258,-0.0834 -0.10025,-0.11446 -0.21568,0.0379 -0.0385,0.0658 -0.10012,0.12716 -0.0971,0.2101 0.26257,0.17623 0.35541,-0.10828 0.31288,-0.24765 z m -0.18954,-0.48941 c -0.002,-0.0411 -0.0436,-0.0877 -0.15571,-0.12611 -0.17795,0.32518 0.16312,0.24936 0.15571,0.12611 z m -3.39268,-11.78328 c -0.007,-0.0222 -0.015,-0.0442 -0.0268,-0.0646 -0.39951,-0.3938 -0.32809,-1.07152 -0.63904,-1.49675 0.0493,0.59576 0.19414,1.18256 0.32136,1.76612 0.004,0.62023 0.44185,0.12916 0.34453,-0.20475 z m 2.27835,12.58899 c -0.0101,-0.0525 -0.027,-0.0984 -0.0521,-0.13269 -0.007,-0.44623 -0.38635,-0.40063 -0.30696,0.0345 -6.4e-4,0.18442 0.0187,0.36863 0.0509,0.55019 0.11903,0.61469 0.37941,-0.0844 0.30817,-0.45198 z m -3.29846,-15.55671 c -0.007,-0.0317 -0.0473,-0.0578 -0.14608,-0.0599 -0.11114,0.30062 0.16683,0.15497 0.14608,0.0599 z",
1.65654, 15.5833,
"M 1.65648 7.45006  c -0.014,-0.2553 -0.50256,-0.33488 -0.51056,-0.80616 -0.15816,0.27247 -0.44145,0.24588 -0.34003,-0.16846 0.52996,-0.0216 0.43279,-0.95853 0.27957,-1.38131 -0.46566,0.32298 -1.49686,0.71091 -1.96216,0.28887 0.47799,-0.50839 -0.10776,-0.98406 0.0393,-1.53376 0.24102,-1.11239 1.05865,-0.30535 1.5689,0.0594 0.29417,-0.54317 -0.42213,-1.43185 -0.2713,-2.12699 0.053,-0.33438 -0.14256,-0.76674 0.18035,-0.99478 0.74093,-0.4137 0.071,-1.20457 0.48731,-1.77662 0.16189,-0.87009 0.12472,-1.775858 0.0253,-2.653088 -0.37764,-0.45524 0.11408,-0.79274 0.0103,-1.24229 -0.0546,-1.40989 0.11322,-2.83197 0.009,-4.24522 -0.18099,-0.66609 0.36361,-1.5512 -0.2403,-2.13217 -0.76022,-0.19291 -1.74581,-0.46318 -1.47846,-1.48984 0.23135,-0.3261 0.005,-1.03309 -0.27595,-0.47181 -0.1149,1.29353 -0.0485,2.59721 -0.0899,3.89486 0.0166,1.91664 -0.15304,3.83585 -0.0129,5.74952 -0.10524,0.84254 -0.15444,1.608928 -0.11524,2.481498 -0.1778,1.05201 -0.0675,2.11953 -0.11111,3.18223 -0.0119,0.29604 0.0774,0.64029 -0.25631,0.78393 h -5.2e-4 c -0.29571,0.18497 -0.24881,0.55168 -0.24081,0.84905 0.12861,0.90109 0.33978,-0.31956 0.34778,-0.59272 0.25303,0.5402 0.0916,1.52043 0.1478,2.22414 -0.0982,0.37068 0.22385,1.646 -0.31678,1.42214 0.003,0.32776 -0.0945,1.3313 0.0145,1.937875 -0.20501,0.86879 0.24098,1.79626 -0.0889,2.66081 0.58049,0.22895 -0.1934,0.54647 0.0367,1.02682 0.20168,0.73761 0.51285,1.49881 0.25838,2.27219 -0.43599,0.48255 0.11253,1.05024 0.3359,0.29043 0.1027,-1.16376 -0.0313,-2.33825 0.0682,-3.50416 0.18831,-1.09985 0.14716,0.68392 0.80718,0.4754 0.30346,0.13933 1.74688,-0.14938 0.9679,-0.23408 -0.58672,-0.4472 0.61164,-1.04584 0.54674,-1.60403 -0.11252,-0.73311 -0.62171,-1.66272 0,-2.307345 0.13633,-0.14516 0.18483,-0.24977 0.17983,-0.33489 z m -0.94413,-22.308708 c -0.13146,-0.53592 -0.80296,-1.00793 -1.25935,-0.52142 -0.21226,0.79762 0.82396,0.58812 1.25935,0.52142 z m -0.25166,28.192583 c 5.1e-4,-0.0774 -0.0451,-0.16422 -0.17674,-0.2346 -0.24185,0.61669 0.17474,0.46696 0.17674,0.2346 z m -0.27799,2.11821 c 0.006,-0.0684 -0.0265,-0.1592 -0.1323,-0.26303 -0.24658,0.46943 0.11469,0.4681 0.1323,0.26303 z m -1.26453,-22.438413 c -10e-4,-0.053 -0.009,-0.11049 -0.0264,-0.17208 0.003,-0.25223 -0.003,-1.72141 -0.21342,-1.21079 -0.0657,0.43773 -0.0695,0.88514 -0.0537,1.32707 -0.0378,0.55499 0.30152,0.42662 0.29352,0.0558 z m -0.0667,3.28145 c -0.004,-0.0288 -0.0281,-0.0608 -0.0847,-0.092 -0.26244,0.28511 0.10135,0.21687 0.0847,0.092 z m -0.2961,17.933793 c -0.005,-0.0297 -0.0175,-0.0677 -0.0413,-0.11629 -0.28174,0.10517 0.0752,0.32372 0.0413,0.11629 z",
1.55663, 17.8159,
"M 0.508125 -17.8159  c 1.02192,1e-5 1.04851,0.0236 1.04851,0.92603 0,0.50934 -0.0799,0.92605 -0.17722,0.92606 -0.0974,1e-5 -0.15454,0.62507 -0.1266,1.38906 0.17088,4.66959 0.17679,7.8749 0.0145,8.03725 -0.10277,0.10281 -0.17864,0.3752 -0.16849,0.60563 0.0138,0.31029 0.0797,0.28188 0.25371,-0.11012 0.17857,-0.4021 0.22525,-0.2592 0.19434,0.59428 -0.0224,0.61778 -0.15187,1.23396 -0.28734,1.36941 -0.13546,0.13548 -0.29046,2.54681 -0.34467,5.35886 -0.0542,2.812 -0.21156,5.58913 -0.34933,6.17119 -0.48698,2.0578 -0.57684,2.66036 -0.60202,4.02459 -0.014,0.75831 -0.10149,1.33161 -0.19432,1.27433 -0.0929,-0.0571 -0.0568,0.64262 0.0796,1.55544 0.13639,0.91283 0.16059,2.12362 0.0542,2.69079 -0.17825,0.95039 -0.22183,0.99291 -0.55449,0.53795 -0.29134,-0.398 -0.34675,-2.15278 -0.28836,-9.05164 0.0598,-7.06382 0.0117,-8.50695 -0.27647,-8.26772 -0.47407,0.39347 -0.44992,-0.0548 0.0542,-1.00148 0.26051,-0.48899 0.38793,-1.36812 0.35966,-2.47788 -0.0242,-0.94587 0.04,-3.28537 0.14265,-5.19866 0.18216,-3.39463 0.32775,-9.3534 1.16794,-9.35337 z",
1.17205, 11.7473,
"M 1.02268 -10.861  c 0.1091,0.1814 0.17719,0.37632 0.13837,0.50034 -0.0351,0.11193 -0.27619,0.1724 -0.53581,0.13446 -0.37479,-0.0547 -0.43967,0.0318 -0.3153,0.41964 0.0861,0.2686 0.0411,0.54794 -0.0997,0.62149 -0.25308,0.13194 -0.18099,1.81839 0.10401,2.43261 0.0809,0.17447 0.095,0.4829 0.0316,0.68524 -0.0634,0.20236 -0.0497,0.46476 0.0303,0.58334 0.25808,0.3824 -0.27881,17.31188 -0.58526,17.230934 -0.163,-0.04269 -0.48752,-0.232899 -0.72224,-0.421235 -0.33692,-0.271053 -0.32403,-0.457558 0.0633,-0.889384 0.26957,-0.300525 0.40319,-0.70571 0.29707,-0.900197 -0.34087,0.04626 0.15589,-9.963114 -0.40834,-15.903028 0.12837,-0.35962 0.16399,-0.9136 0.0792,-1.23105 -0.0848,-0.31748 -0.12098,-0.89469 -0.0803,-1.28251 0.0407,-0.38778 -0.003,-1.16267 -0.0972,-1.72233 -0.12675,-0.75336 -0.0633,-1.00192 0.24376,-0.95706 0.22809,0.0334 0.44313,-0.031 0.47827,-0.14277 0.0351,-0.11192 0.19636,-0.007 0.35828,0.23297 0.16257,0.24082 0.3604,0.33132 0.44178,0.2019 0.10143,-0.16052 0.39628,0.10439 0.57817,0.40666 z",
1.95058, 13.6269,
"M 0.471013 -13.6265  c -0.0226,0.002 -0.0393,0.019 -0.0481,0.0517 -0.0351,0.13069 -0.24992,0.20591 -0.47799,0.16691 -0.30708,-0.0524 -0.37068,0.23818 -0.24392,1.1188 0.0942,0.65419 0.13785,1.56002 0.0971,2.01331 -0.0407,0.45333 -0.004,1.12802 0.0806,1.49913 0.0848,0.37107 0.0488,1.01882 -0.0796,1.43919 -0.20937,0.68561 -0.0718,1.6412 0.37155,2.62827 -0.11168,0.0113 -0.23463,0.0283 -0.38137,0.0563 -0.6455,0.12341 -0.72607,0.25452 -0.72607,1.18184 -1.6e-4,0.62786 -0.10541,0.97834 -0.26458,0.87954 -0.16997,-0.10511 -0.26459,0.835241 -0.26459,2.630332 0,1.81333 -0.0999,2.79364 -0.28421,2.79363 h -5e-4 c -0.15627,-2e-5 -0.23813,0.20835 -0.18243,0.46302 0.0557,0.25469 0.2427,0.43297 0.416,0.39687 0.21384,-0.0449 0.3152,0.40637 0.31522,1.40457 0,0.91314 -0.10041,1.40916 -0.26458,1.30741 -0.16825,-0.10405 -0.25207,0.54034 -0.22998,1.77042 0.0279,1.54684 0.074,1.74841 0.22998,1.0077 0.19206,-0.91245 0.19513,-0.90816 0.22945,0.3085 0.0192,0.67909 -0.0733,1.27442 -0.20569,1.32292 -0.36356,0.1332 -0.56917,1.89947 -0.24442,2.10013 0.29488,0.18225 0.69575,0.93239 0.58706,0.65009 -0.12179,-0.3174 -0.0778,-0.55313 0.11522,-0.61753 0.17206,-0.0571 0.3121,-0.3287 0.31213,-0.60307 3.4e-4,-0.31822 0.13407,-0.44675 0.37052,-0.35605 0.4857,0.18639 0.77203,-0.30974 0.74414,-1.28984 -0.41298,-4.18519 -0.41174,-10.39178 -0.21031,-12.921172 0.13216,0.0905 0.25572,0.15931 0.33795,0.18449 0.30644,0.0946 0.84307,-3.81941 0.58497,-4.26641 -0.08,-0.13861 -0.0934,-0.44506 -0.0299,-0.68161 0.0634,-0.23652 0.0494,-0.59705 -0.0315,-0.80099 -0.28499,-0.71796 -0.35746,-2.68952 -0.10438,-2.84375 0.14078,-0.086 0.18584,-0.4126 0.0997,-0.72657 -0.12435,-0.45333 -0.0596,-0.55441 0.31522,-0.49041 0.25964,0.0444 0.50081,-0.0263 0.53589,-0.15709 0.0388,-0.14498 -0.0294,-0.37294 -0.13848,-0.58498 -0.1819,-0.35333 -0.4763,-0.66307 -0.57775,-0.47543 -0.0814,0.15128 -0.27926,0.0453 -0.44182,-0.23616 -0.12145,-0.21024 -0.24284,-0.33152 -0.31057,-0.32401 z",
0.746818, 13.5457,
"M 0.651624 -11.1239  c 0.0695,0.49565 0.1129,1.02826 0.0882,1.36713 -0.0224,0.30582 -0.17599,0.47106 -0.34142,0.36739 -0.23882,-0.1496 -0.28016,0.087 -0.20091,1.14663 0.0549,0.73391 0.0262,1.49718 -0.0635,1.69814 -0.16126,0.36052 -0.11532,4.968579 0.0663,6.646849 0.0516,0.47672 0.0605,1.31946 0.0201,1.87234 -0.0404,0.55294 -0.0317,1.2699 0.0193,1.59391 0.16445,1.04487 -0.17766,10.194453 -0.37293,9.973273 -0.10386,-0.11664 -0.31064,-0.63636 -0.46021,-1.15097 -0.21468,-0.74062 -0.20647,-1.25023 0.0403,-2.43014 0.17177,-0.82115 0.25691,-1.928273 0.18929,-2.459683 -0.30256,-2.37873 -0.39749,-4.69547 -0.26019,-6.34493 0.0818,-0.98262 0.10449,-2.49631 0.0505,-3.363699 -0.054,-0.86748 -0.0771,-2.44465 -0.0512,-3.50432 0.0259,-1.05958 -0.002,-3.17689 -0.0619,-4.70608 -0.0808,-2.05847 -0.0403,-2.73764 0.15533,-2.61506 0.14533,0.0912 0.28236,-0.0846 0.30475,-0.39011 0.0224,-0.30582 0.12512,-0.0187 0.22829,0.63658 0.10359,0.65802 0.22965,0.90528 0.2815,0.55166 0.0646,-0.43862 0.25251,0.28522 0.36841,1.11114 z",
1.38957, 16.4755,
"M 0.138886 -16.4755  -1.35549,1.84795 c 0.27683,0.42497 0.50699,1.05938 0.54983,1.48879 0.0601,0.48725 0.0332,0.97552 -0.0119,1.46658 -0.0892,-0.49054 -0.17108,-0.69618 -0.19068,-0.4284 -0.0224,0.30551 -0.15905,0.48136 -0.30438,0.39016 -0.19563,-0.12258 -0.23635,0.55687 -0.15555,2.61535 0.0599,1.52918 0.0879,3.64658 0.062,4.70617 -0.0259,1.05967 -0.003,2.63669 0.0512,3.50417 0.054,0.86739 0.0311,2.381 -0.0506,3.36362 -0.13732,1.64946 -0.0421,3.96611 0.26043,6.34484 0.0676,0.5314 -0.0179,1.63864 -0.18966,2.45979 -0.24675,1.17991 -0.25497,1.68972 -0.0403,2.43035 0.12094,0.41612 0.27699,0.82396 0.38809,1.03094 -0.25501,0.55068 -0.28125,1.13805 -0.22479,1.73064 l 1.4609,-0.003 c 0.0168,-0.42614 0.0252,-0.85476 0.0677,-1.27434 0.25389,-1.34995 0.14287,-2.66726 0.36689,-4.02611 0.031,-1.4516 0.11288,-2.8398 0.19175,-4.25762 0.12581,-1.2814 -0.13108,-2.63661 0.16224,-3.91397 -0.10988,-1.37381 -0.11983,-2.74916 -0.14417,-4.11603 0.0786,-1.41425 -0.2367,-2.81377 -0.12764,-4.19458 -0.0646,-1.50176 -0.23391,-3.00258 -0.21238,-4.4483 0.78798,-0.10067 0.97116,-1.25283 0.22117,-1.50379 0.42751,0.71673 -0.5987,0.75403 -0.39428,0.0444 0.49067,-0.38166 0.47429,-0.58582 0.0196,-0.82733 0.15126,-1.10944 -0.28977,-2.18413 -0.17312,-3.35639 -0.008,-0.38118 -0.13687,-0.71882 -0.22635,-1.07229 z m -0.46355,8.37158 c 0.0189,0.005 0.0414,0.0166 0.0641,0.0295 0.11988,0.62091 -0.0121,1.26337 -0.16021,1.91254 5.9e-4,-0.24158 -0.005,-0.49741 -0.0243,-0.75809 -0.0694,-0.92718 -0.0464,-1.22436 0.12041,-1.18391 z m -0.25374,3.57756 c 0.0898,0.93296 0.42876,1.87606 0.6594,2.79259 0.23042,1.45009 0.26061,2.8781 -0.008,4.3031 -0.24439,1.27673 -0.30588,2.48657 0.0294,3.77031 -0.13909,1.17698 -0.45156,2.32446 -0.53951,3.48144 0.0761,-2.22329 0.11075,-4.44633 0.0346,-4.92993 -0.051,-0.32401 -0.06,-1.04076 -0.0196,-1.5937 0.0404,-0.55288 0.0314,-1.39552 -0.0202,-1.87224 -0.1455,-1.34446 -0.20132,-4.53663 -0.13592,-5.95157 z m 0.0796,15.91375 c 0.14541,0.99939 0.5629,2.03294 -0.10957,2.94194 -0.0303,0.0418 -0.0541,0.0847 -0.0811,0.12713 0.0646,-0.57923 0.13222,-1.74103 0.19069,-3.06907 z",
2.34195, 11.5159,
"M 2.34146 -4.20638  c 0.002,0.0553 -0.002,0.11609 -0.0145,0.17985 -0.0644,0.3366 -0.42979,0.50248 -1.33841,0.60719 -0.68743,0.0792 -1.29195,0.0792 -1.34307,0 -0.0514,-0.079 -0.0507,0.0201 4.4e-4,0.22012 0.0516,0.20034 -0.0357,0.52005 -0.19353,0.71065 -0.20476,0.24688 -0.20659,0.49864 -0.005,0.87541 0.15568,0.29074 0.2535,0.80116 0.21707,1.13429 -0.0424,0.38883 0.0761,0.60836 0.33071,0.6134 0.21831,0.004 0.40817,0.18566 0.42215,0.40362 0.014,0.21788 0.11684,0.60569 0.22773,0.8615 0.12326,0.2839 0.0426,0.70855 -0.20724,1.08985 -0.22511,0.34352 -0.39934,0.98017 -0.38738,1.41481 0.0279,0.9801 -0.25842,1.47623 -0.74413,1.28984 -0.23645,-0.0907 -0.37001,0.038 -0.37035,0.35622 -2e-5,0.27437 -0.14008,0.54597 -0.31215,0.60306 -0.19306,0.0644 -0.23703,0.30013 -0.11526,0.61753 0.10871,0.2823 0.15952,0.5378 0.11365,0.56793 -0.0459,0.0299 0.15058,0.22581 0.4361,0.43468 0.44013,0.32178 0.46571,0.44377 0.1685,0.80197 -0.19293,0.23238 -0.45647,0.35733 -0.58549,0.27753 -0.2465,-0.15271 -0.29031,0.14035 -0.27906,1.87067 0.007,0.99916 -0.004,1.01393 -0.19528,0.26464 -0.11079,-0.43669 -0.15026,-1.27006 -0.0873,-1.8521 0.19222,-1.77852 0.13676,-2.83306 -0.15812,-3.01531 -0.32475,-0.20066 -0.11915,-1.96693 0.24443,-2.10013 0.13238,-0.0485 0.2249,-0.64383 0.20567,-1.32292 -0.0343,-1.21666 -0.0374,-1.22096 -0.22946,-0.30851 -0.15593,0.74072 -0.20203,0.53914 -0.22995,-1.0077 -0.0221,-1.23009 0.0617,-1.87448 0.22995,-1.77043 0.16417,0.10175 0.26459,-0.39427 0.2646,-1.30741 -2e-5,-0.99819 -0.10145,-1.44949 -0.31528,-1.40457 -0.17329,0.0361 -0.36031,-0.1422 -0.41598,-0.39689 -0.0557,-0.25466 0.0262,-0.46301 0.18245,-0.463 h 5.2e-4 c 0.18433,1e-5 0.28421,-0.9803 0.28422,-2.79364 0,-1.79509 0.0946,-2.73543 0.26457,-2.63032 0.15915,0.0988 0.2644,-0.25166 0.26456,-0.87952 -1e-5,-0.92732 0.0806,-1.05844 0.72609,-1.18185 0.97706,-0.18674 1.12602,-0.0503 1.126,1.03409 10e-6,0.63956 -0.0957,0.86695 -0.30949,0.73485 -0.21128,-0.13012 -0.25315,-0.0441 -0.13229,0.27076 0.0974,0.25407 0.17731,0.63077 0.17721,0.83716 -2e-5,0.20637 0.19145,0.37518 0.42527,0.37521 0.26515,2e-5 0.36883,0.14645 0.27599,0.38857 -0.082,0.21364 -0.0353,0.45868 0.10391,0.54466 0.13913,0.086 0.2532,0.33466 0.2532,0.55296 -7e-5,0.21815 -0.11907,0.32343 -0.26457,0.23359 -0.14556,-0.0898 -0.26459,-0.0373 -0.2646,0.11674 -8e-5,0.15409 0.10469,0.28009 0.23249,0.28014 0.12785,-1e-5 0.38035,0.35716 0.56117,0.79376 0.18081,0.4366 0.44232,0.79375 0.58084,0.79377 0.10393,0 0.17072,0.11718 0.1778,0.28317 z",
1.43162, 19.3805,
"M -0.137187 -19.38  c -0.0842,-0.005 -0.19127,0.031 -0.32713,0.0946 -0.35901,0.16796 -0.4036,0.34656 -0.4036,1.60868 -7e-5,0.85455 -0.0588,1.3313 -0.14726,1.19683 -0.0946,-0.14306 -0.14727,1.13695 -0.14727,3.58014 -3e-5,2.46804 -0.0551,3.80237 -0.15761,3.80236 h -5.3e-4 c -0.0869,-10e-6 -0.13229,0.28333 -0.10128,0.62993 0.031,0.34665 0.13459,0.58964 0.23098,0.54054 0.11895,-0.0611 0.17571,0.55293 0.17571,1.91151 0,1.24282 -0.056,1.9177 -0.14727,1.77922 -0.0936,-0.14162 -0.14047,0.73546 -0.12816,2.409662 0.0155,2.10531 0.0415,2.37964 0.12816,1.37149 0.10681,-1.24188 0.10853,-1.23579 0.12764,0.42013 0.0107,0.92427 -0.0411,1.73441 -0.11473,1.80041 -0.20225,0.1813 -0.31652,2.58511 -0.13591,2.85822 0.16401,0.24805 0.19476,1.6835 0.0879,4.104145 -0.035,0.792181 -0.013,1.926426 0.0485,2.520775 0.10663,1.019821 0.11306,0.999718 0.10907,-0.360182 -0.006,-2.355041 0.0179,-2.753938 0.15502,-2.546099 0.0718,0.108619 0.21825,-0.06148 0.32557,-0.377759 0.16531,-0.48752 0.15123,-0.65344 -0.0936,-1.0914 -0.15883,-0.28428 -0.26837,-0.551 -0.24286,-0.5917 0.0255,-0.041 -0.003,-0.38885 -0.063,-0.77308 -0.0677,-0.43199 -0.0433,-0.75256 0.0641,-0.84026 0.0957,-0.0777 0.17362,-0.44719 0.17362,-0.82062 1.8e-4,-0.43313 0.0747,-0.60869 0.20619,-0.48524 0.27014,0.25368 0.42944,-0.42149 0.41394,-1.75545 -0.007,-0.59156 0.0903,-1.45792 0.21548,-1.92546 0.13896,-0.51897 0.18375,-1.09672 0.11525,-1.48312 -0.0617,-0.348152 -0.11914,-0.875982 -0.12713,-1.172522 -0.008,-0.29666 -0.11319,-0.54484 -0.23461,-0.54984 -0.14163,-0.007 -0.20757,-0.30536 -0.18396,-0.83457 0.0203,-0.45341 -0.0338,-1.14839 -0.12042,-1.5441 -0.11189,-0.5128 -0.11128,-0.85512 0.003,-1.19114 0.0877,-0.25941 0.13668,-0.69471 0.108,-0.96738 -0.0285,-0.2722 -0.0291,-0.40724 -5.3e-4,-0.29972 0.025,0.0947 0.29102,0.10153 0.61341,0.03 0.0405,0.52236 0.0906,1.04394 0.0915,1.57561 -0.0608,0.60189 -0.078,1.11737 -0.0465,1.74821 0.0696,0.63278 0.297,1.09496 0.0998,1.78077 -0.058,0.73385 0.0101,1.474622 0.002,2.207102 -0.0197,0.81022 0.0418,1.60891 0.004,2.4262 -0.0875,0.63698 -0.0763,1.43498 -0.22429,1.9482 0.12394,0.22305 0.34293,0.21881 0.24702,0.84078 -0.0607,1.1452 -0.0473,2.30122 -0.0796,3.45147 0.0458,0.825198 -0.0683,1.62066 -0.093,2.435508 0.0204,1.1895 -0.1012,2.349915 -0.17311,3.516066 -0.0662,0.696058 -0.0537,1.411333 -0.0966,2.110981 -0.0148,0.25309 -0.0926,0.598914 0.0625,0.599964 0.29631,0.0457 0.33951,0.87489 0.30128,1.493449 -0.10774,0.477189 0.0255,1.113514 0.0103,1.577165 l 0.14004,-0.146246 c 0.0399,-0.520959 0.0745,-1.044482 0.11576,-1.564762 0.18883,-2.023379 0.28009,-4.091098 0.39534,-6.147427 0.0855,-1.171151 0.0493,-2.370113 0.13795,-3.538798 -0.0101,-0.73951 0.0498,-1.47317 0.032,-2.21279 0.0113,-0.74373 0.0533,-1.47966 0.0351,-2.2226 0.0413,-0.71668 -0.0126,-1.40059 -0.001,-2.13166 0.0365,-0.74313 0.0171,-1.48879 -0.0109,-2.22673 0.07,-0.4474 -0.0966,-1.347812 0.0646,-1.455202 -0.015,0.783202 0.13674,0.20012 0.1297,-0.24288 -0.002,-0.28913 -0.0311,-1.39426 -0.15089,-0.66404 0.14613,0.14266 -0.0403,0.66546 -0.10491,0.6289 0.1066,-0.45472 0.0111,-0.96117 0.0522,-1.43971 0.15606,-0.65158 0.0842,-1.40153 -0.077,-1.97559 -0.0905,-0.69785 -0.11025,-1.42475 -0.13901,-2.14043 -0.009,-0.18957 -0.0222,-0.37781 -0.0341,-0.56689 0.11245,-0.13013 0.17013,-0.29776 0.18757,-0.5209 0.007,-0.0868 0.009,-0.16965 0.008,-0.24495 h -5.3e-4 c -0.004,-0.22592 -0.0414,-0.3855 -0.0992,-0.3855 -0.0459,-2e-5 -0.11597,-0.18397 -0.18653,-0.44959 -10e-4,-0.0362 -0.004,-0.0729 -0.005,-0.10904 0.003,-0.57553 -0.0511,-1.12705 -0.10387,-1.6769 h -0.001 l -0.26149,0.19689 c -0.0296,-0.0724 -0.0569,-0.12196 -0.0775,-0.12196 -0.0711,-7e-5 -0.12922,-0.17164 -0.12916,-0.38137 0,-0.20968 0.0663,-0.28087 0.14726,-0.15864 0.0809,0.12226 0.14724,-0.0209 0.14727,-0.31781 0,-0.29713 -0.0637,-0.63588 -0.14108,-0.75293 -0.0774,-0.11702 -0.10347,-0.45026 -0.0579,-0.74104 0.0516,-0.32953 -0.006,-0.52914 -0.15349,-0.52917 -0.13004,-4e-5 -0.23669,-0.22969 -0.23667,-0.51056 0,-0.28091 -0.044,-0.79366 -0.0982,-1.13947 -0.0672,-0.42852 -0.0441,-0.54555 0.0734,-0.36845 0.11893,0.17979 0.17208,-0.12947 0.17208,-0.99994 3e-5,-1.10695 -0.0466,-1.48805 -0.29919,-1.50223 z",
1.43294, 9.17485,
"M 0.959443 6.24537  c 0.0279,0.9801 -0.25842,1.47623 -0.74413,1.28984 -0.23645,-0.0907 -0.37001,0.038 -0.37035,0.35622 -2e-5,0.27437 -0.14008,0.54597 -0.31215,0.60306 -0.19306,0.0644 -0.23703,0.30013 -0.11526,0.61753 0.10871,0.2823 -0.29212,-0.46774 -0.587,-0.64999 -0.32475,-0.20066 -0.11915,-1.96693 0.24443,-2.10013 0.13238,-0.0485 0.2249,-0.64383 0.20567,-1.32292 -0.0343,-1.21666 -0.0374,-1.22096 -0.22946,-0.30851 -0.15593,0.74072 -0.20203,0.53914 -0.22995,-1.0077 -0.0221,-1.23009 0.0617,-1.87448 0.22995,-1.77043 0.16417,0.10175 0.26459,-0.39427 0.2646,-1.30741 -2e-5,-0.99819 -0.10145,-1.44949 -0.31528,-1.40457 -0.17329,0.0361 -0.36031,-0.1422 -0.41598,-0.39689 -0.0557,-0.25466 0.0262,-0.46301 0.18245,-0.463 h 5.2e-4 c 0.18433,1e-5 0.28421,-0.980298 0.28422,-2.793628 0,-1.79509 0.0946,-2.73543 0.26457,-2.63032 0.15915,0.0988 0.2644,-0.25166 0.26456,-0.87952 -1e-5,-0.92732 0.0806,-1.05844 0.72609,-1.18185 0.97706,-0.18674 1.12602,-0.0503 1.126,1.03409 10e-6,0.63956 -0.0957,0.86695 -0.30949,0.73485 -0.49107,-2.81613 -0.76298,7.511348 -0.16401,13.581278 z",
0.865975, 16.9234,
"M 0.579895 11.5198  c 0.0168,1.80783 -0.15617,2.72296 -0.4497,2.37916 -0.14289,-0.1673 -0.22361,0.0701 -0.22381,0.65706 -10e-6,0.50609 -0.0846,1.00706 -0.18865,1.11237 -0.11667,0.11879 -0.14325,0.5536 -0.0697,1.13906 0.0657,0.52071 -0.17654,-0.86277 -0.35474,-1.19894 -0.19626,-0.37012 -0.072,-3.62807 0.14772,-3.87377 0.08,-0.0895 0.13591,-1.18757 0.12429,-2.44017 -0.0208,-2.24418 -0.0226,-2.25211 -0.13867,-0.56906 -0.0942,1.36629 -0.1221,0.99446 -0.13897,-1.85874 -0.0133,-2.26895 0.0373,-3.45755 0.13897,-3.26563 0.0992,0.18768 0.1599,-0.72725 0.15991,-2.41157 -10e-6,-1.8412 -0.0613,-2.67364 -0.19055,-2.59078 -0.10472,0.0666 -0.21773,-0.26229 -0.25139,-0.73208 -0.0337,-0.46973 0.0158,-0.85404 0.11023,-0.85402 h 3.2e-4 c 0.11135,2e-5 0.17176,-1.8082 0.17176,-5.15295 0,-3.31112 0.0571,-5.04561 0.1599,-4.85173 0.0962,0.18224 0.15978,-0.4642 0.15987,-1.62232 0,-1.71048 0.0487,-1.95233 0.4388,-2.17997 0.59049,-0.34444 0.6805,-0.0928 0.68049,1.90742 10e-6,1.17971 -0.0578,1.59914 -0.18703,1.35547 -0.29677,-5.19447 -0.46111,13.85497 -0.0991,25.05119 z",
0.920438, 15.73,
"M 0.600428 -15.73  -0.20928,0.13642 c 0.0116,0.10797 -0.008,0.20908 -0.10025,0.29611 -0.24135,0.42805 0.0459,1.00331 -0.24961,1.42213 0.2354,0.42752 0.12838,0.94985 0.13438,1.42472 0.10674,0.60433 -0.83865,0.28316 -0.87696,0.59066 0.2572,0.43226 -0.0173,0.8381 -0.0377,1.27641 0.21408,0.45321 -0.16298,0.88129 0.0558,1.32705 -0.23974,0.30068 0.19899,0.60454 -0.0589,0.92553 -0.006,0.92527 -0.0266,1.85081 -0.0517,2.77606 0.0152,0.4801 -0.0752,1.02655 0.1478,1.45779 -0.25858,-0.0769 -0.30935,0.24479 -0.16174,0.47594 -0.075,0.46382 0.1289,0.91188 -0.016,1.36787 -0.0107,1.507429 0.0413,3.015669 0.10387,4.521689 0.1733,0.47524 0.0581,0.99741 0.19534,1.44487 -0.36306,0.3697 -0.20534,0.96671 -0.25064,1.43919 -0.0318,0.49346 0.11663,1.03192 -0.0951,1.49087 0.14563,0.84649 0.0783,1.712415 0.1111,2.567805 -0.0665,0.45222 0.24575,0.922311 -0.0563,1.326021 -0.16113,0.30199 -0.14399,0.87954 0.10178,1.04644 -0.008,0.50476 0.33634,0.95241 0.1287,1.45211 0.19711,0.12306 0.23505,0.16337 0.093,0.39998 -0.25935,0.34252 0.27749,0.87321 -0.10438,1.09347 -0.51512,0.10675 0.14734,0.092 0.17105,0.3359 -0.11327,0.32055 -0.0946,1.20549 0.39066,0.72502 0.77099,0.005 -0.24847,-0.55676 0.12041,-0.89814 0.0876,-0.7571 0.0667,-1.52622 0.0171,-2.28565 -0.17201,-0.6034 -0.11922,-1.24776 -0.0837,-1.86965 0.24858,-0.51407 -0.0498,-1.029341 0.016,-1.553911 -0.006,-0.45575 0.10213,-0.879465 -0.0367,-1.321375 -0.0154,-0.45173 -0.18732,-0.86783 0.13748,-1.26039 0.17298,-0.38763 0.16687,0.10802 0.19378,0.28939 0.13417,0.39563 0.14967,0.90483 0.22738,1.25987 0.1107,-0.96718 0.0622,-1.94664 0.0801,-2.9192 0.0259,-4.03603 -0.0233,-8.072429 0.0181,-12.108299 0.23979,-0.22266 0.005,0.79598 0.22736,0.78083 0.0587,-0.30736 0.032,-0.62753 0.03,-0.93948 0,-0.1962 -0.0468,-0.69798 -0.0713,-0.75137 -0.0348,0.24268 0.0825,0.63699 -0.16899,0.74 -0.058,-0.72244 -0.0291,-1.46041 -0.0476,-2.18901 -0.006,-2.09788 -0.0295,-4.19582 -0.0243,-6.29367 z",
1.1248, 15.7337,
"M 0.154165 -15.7331  c -0.024,0.006 -0.0442,0.058 -0.0594,0.16072 -0.0119,0.0804 -0.0224,0.198 -0.032,0.33693 -0.18859,0.42108 0.0568,0.96324 -0.22582,1.36374 0.2354,0.42752 0.12835,0.94985 0.13436,1.42472 0.10673,0.60433 -0.83863,0.28316 -0.87694,0.59066 0.2572,0.43226 -0.0173,0.8381 -0.0377,1.27641 0.21408,0.45321 -0.16298,0.88129 0.0558,1.32705 -0.23974,0.30068 0.199,0.60453 -0.0589,0.92553 -0.006,0.92527 -0.0266,1.8508 -0.0517,2.77605 0.0152,0.4801 -0.0752,1.02656 0.14777,1.4578 -0.25857,-0.0769 -0.30935,0.24479 -0.16174,0.47594 -0.075,0.46382 0.12888,0.91188 -0.016,1.36787 -0.0107,1.507433 0.0413,3.015673 0.10384,4.521693 0.17331,0.47524 0.0581,0.99741 0.19535,1.44487 -0.36307,0.3697 -0.20532,0.96671 -0.25062,1.43919 -0.0318,0.49346 0.11663,1.03192 -0.0951,1.49086 0.14563,0.84649 0.0783,1.71241 0.1111,2.5678 -0.0665,0.45222 0.24572,0.9223 -0.0563,1.32601 -0.16113,0.30199 -0.14396,0.87955 0.10181,1.04645 -0.008,0.50476 0.33631,0.95241 0.12867,1.45211 0.19711,0.12306 0.23508,0.16336 0.093,0.39998 -0.25935,0.34252 0.27746,0.87321 -0.10441,1.09347 -0.51512,0.10675 0.14735,0.092 0.17105,0.3359 -0.11326,0.32055 -0.0946,1.20548 0.39069,0.72501 0.77099,0.005 -0.24847,-0.55675 0.12041,-0.89813 0.0876,-0.7571 0.0666,-1.52622 0.017,-2.28565 -0.17201,-0.6034 -0.11922,-1.24776 -0.0837,-1.86965 0.24857,-0.51407 -0.0498,-1.02934 0.016,-1.55391 -0.006,-0.45575 0.10213,-0.87945 -0.0367,-1.32136 -0.0154,-0.45173 -0.18735,-0.86783 0.13745,-1.26039 0.17299,-0.38763 0.16688,0.10802 0.19378,0.28939 0.13417,0.39563 0.14968,0.90483 0.22739,1.25987 0.1107,-0.96718 0.0622,-1.94664 0.0801,-2.9192 0.006,-0.9718 0.007,-1.94324 0.007,-2.91507 0.0104,0.0106 0.0205,0.0219 0.0315,0.031 0.27921,0.23143 0.31816,-0.0277 0.16589,-1.11104 -0.17018,-1.20977 -0.16534,-1.42171 0.0331,-1.43661 0.12761,-0.0166 0.23202,0.66842 0.23202,1.51361 0,0.84503 0.0583,1.14304 0.12972,0.65991 0.14145,-0.95658 0.0967,-13.710923 0.0274,-13.242613 -0.0693,0.46832 -0.28163,-0.48851 -0.47181,-2.12545 -0.062,-0.53393 -0.12687,-0.97738 -0.18965,-1.3286 -1.9e-4,-0.26869 -0.003,-0.53746 -0.002,-0.80616 l -0.16123,0.10542 c -0.0298,-0.0783 -0.0571,-0.11812 -0.0806,-0.11213 z",
1.19528, 10.0475,
"M 0.114948 -10.0474  -0.50148,0.14618 c 0.0736,0.18661 0.16739,0.36967 0.261,0.54185 -0.0538,0.4307 0.22483,0.82056 0.0968,1.2648 -0.0525,0.36807 -0.1816,0.11366 -0.21076,-0.0961 -0.007,-0.23246 -0.37153,-0.97536 -0.51525,-0.43993 -0.11169,0.66702 0.20719,1.30602 0.21001,1.96991 -0.10058,0.38477 -0.12871,0.71438 -0.0766,1.11766 0.11511,0.40452 0.49088,0.70014 0.16472,1.13856 -0.0959,0.46912 0.0158,0.94258 0.003,1.41083 -0.0326,0.51795 0.0689,1.02842 0.006,1.55089 -0.14471,0.4072 -0.12605,0.91736 -0.37066,1.24545 0.20492,0.14259 0.56731,0.13997 0.4087,0.53758 -0.10038,0.73209 -0.0783,1.47103 -0.13167,2.20635 0.0757,0.52752 -0.1133,1.03601 -0.15412,1.55692 0.0337,0.76041 -0.16675,1.50234 -0.28566,2.24782 -0.1094,0.44497 -0.0894,0.90202 -0.16034,1.34929 -0.0245,0.16179 -0.1532,0.38291 0.10343,0.38358 0.48997,0.0292 0.56196,0.55944 0.49872,0.95487 -0.17814,0.30505 0.0416,0.71193 0.0159,1.00833 l 0.2316,-0.0937 c 0.066,-0.33303 0.12356,-0.66776 0.19171,-1.00036 0.31225,-1.29348 0.46274,-2.61533 0.65329,-3.92988 0.14143,-0.74868 0.0819,-1.51508 0.22844,-2.26219 -0.0167,-0.47274 0.0824,-0.94169 0.053,-1.41451 0.0186,-0.47544 0.0882,-0.94598 0.058,-1.42092 0.0683,-0.45815 -0.0209,-0.89535 -0.002,-1.36269 0.0604,-0.47506 0.0288,-0.95181 -0.0173,-1.42355 0.1158,-0.28601 -0.15986,-0.86145 0.10674,-0.9301 -0.0248,0.50068 0.22634,0.12776 0.21472,-0.15544 -0.004,-0.18483 -0.0516,-0.89113 -0.24975,-0.42432 0.24166,0.0912 -0.0666,0.42542 -0.17345,0.40205 0.17628,-0.29069 0.0179,-0.61457 0.0859,-0.92048 0.25808,-0.41654 0.13936,-0.89587 -0.12721,-1.26285 -0.14964,-0.44611 -0.18222,-0.91074 -0.22982,-1.36825 -0.0568,-0.4894 -0.17777,-0.9772 -0.21376,-1.45567 0.005,-0.36792 -0.0846,-0.72055 -0.17196,-1.07205 z",
1.30323, 16.4753,
"M 0.0525373 -16.4753  -1.35577,1.84766 c 0.27683,0.42497 0.50698,1.05949 0.54982,1.4889 0.16442,1.33355 -0.31785,2.67593 -0.0632,4.05717 0.89557,1.12078 0.40959,2.34393 0.16413,3.57614 -0.17728,1.24541 0.33502,2.53287 0.6468,3.771819 0.23041,1.45009 0.26088,2.878319 -0.008,4.303319 -0.24439,1.27673 -0.30622,2.48615 0.0291,3.76989 -0.15519,1.31328 -0.52854,2.58989 -0.56215,3.88389 -0.40545,1.32231 0.79113,2.83626 -0.1482,4.10603 -0.48269,0.66505 -0.53542,1.40168 -0.46451,2.14587 l 1.46058,-0.003 c 0.0168,-0.42614 0.0252,-0.85474 0.0677,-1.27432 0.2539,-1.34995 0.14317,-2.66731 0.3672,-4.02616 0.031,-1.4516 0.11297,-2.83964 0.19182,-4.25746 0.12583,-1.2814 -0.13142,-2.63659 0.16191,-3.91395 -0.1099,-1.37381 -0.1194,-2.74933 -0.14373,-4.116199 0.0786,-1.414249 -0.23712,-2.813419 -0.12807,-4.194219 -0.0646,-1.50176 -0.23348,-3.00275 -0.21195,-4.44847 0.788,-0.10067 0.97118,-1.25313 0.22118,-1.50409 0.42751,0.71673 -0.59912,0.75408 -0.39471,0.0445 0.49068,-0.38166 0.47426,-0.5855 0.0196,-0.82701 0.15128,-1.10943 -0.28974,-2.18426 -0.17309,-3.35652 -0.008,-0.38118 -0.13645,-0.71905 -0.22593,-1.07252 z",
3.07777, 16.9417,
"M -0.540315 -15.4793  c -0.24634,0.11349 -1.09658,-1.59624 -1.32736,-1.45407 -0.46023,-0.0191 -0.87738,0.005 -0.4728,0.85065 -0.38438,0.23736 0.33291,1.49353 -0.0462,1.73799 0.11651,0.36917 0.22846,-0.0337 0.42244,7.7e-4 0.15386,0.31092 0.11637,0.69337 0.1657,1.02389 0.34684,0.15366 0.10848,0.64269 0.28358,0.76644 0.41195,-0.0452 0.55085,0.74543 0.0419,0.5194 -0.27864,0.26975 0.19724,0.25988 0.15384,0.53117 0.1548,0.28493 0.18472,-0.0741 0.12234,-0.20191 0.13378,-0.5353 0.43906,0.1542 0.21787,0.35432 -0.44319,-0.0388 -0.0774,0.48499 0.15909,0.23185 0.10925,-0.0917 0.16042,0.22403 0.24283,0.28936 0.16253,0.3502 -0.0983,-0.0308 -0.20977,-0.0917 0.0884,0.0915 -0.32447,0.49774 -0.0756,0.36635 0.504,-0.0226 -0.0992,0.43767 0.20795,0.63982 -0.12496,0.26315 0.26137,0.29982 0.17242,0.57031 0.17152,0.21667 0.0867,0.45418 -0.10428,0.6239 0.17103,0.10816 0.33957,0.15221 0.1982,0.38978 -0.16129,0.25419 0.1763,0.20125 0.0976,0.39005 0.32354,0.0622 -0.0735,0.53313 0.26064,0.44952 0.16282,-0.294 0.37219,0.22828 0.27145,0.4162 -0.47933,0.1604 0.29042,0.60248 -0.0488,0.85553 -0.3968,0.13941 0.16426,0.23706 -0.0494,0.45862 0.28882,0.23742 -0.26764,0.54324 0.13284,0.74483 -0.10693,0.19204 0.083,0.29149 0.17713,0.48105 -0.15407,0.2054 0.003,0.54361 0.2659,0.31837 0.41731,-0.0408 -0.0181,0.41236 -0.166,0.51107 -0.33606,0.19565 -0.0219,0.65064 -0.0739,0.87903 0.16745,0.29274 -0.13216,0.58183 -0.1062,0.84082 0.34137,-0.0301 0.0236,0.11514 0.051,0.28009 0.0409,0.34986 0.13526,0.72795 0.0252,1.04597 0.15197,0.3677 0.14533,0.69256 0.16035,1.08996 -0.036,0.32787 -0.12398,0.67271 -0.0792,1.00699 0.0442,0.35265 -0.0843,0.70567 0.0129,1.05246 -0.16927,0.31514 0.0256,0.61454 -0.0918,0.92499 0.16825,0.22895 -0.33833,0.64208 -0.0292,0.80537 0.15925,-0.0153 0.26898,-0.23555 0.49496,-0.0945 0.26864,0.26814 -0.20509,0.57525 -0.34709,0.79008 -0.22991,0.0506 -0.30996,0.1628 -0.20241,0.38893 -0.17522,0.31665 -0.083,0.71442 -0.19792,1.05749 -0.13603,0.29173 -0.0706,0.64159 -0.18877,0.95263 -0.0571,0.35637 -0.19909,0.70354 -0.20502,1.0641 0.0598,0.35902 -0.37999,0.56428 -0.25411,0.85446 -0.17361,0.0964 -0.0151,0.24287 -0.10293,0.35328 -0.12389,0.31997 -0.23871,0.64343 -0.31422,0.9698 0.37628,0.16039 -0.17227,0.39286 -0.16858,0.63597 0.30078,0.23058 -0.13834,0.45129 -0.031,0.70849 -0.25057,0.0583 -0.25681,0.38137 -0.29542,0.62486 -0.16162,0.27054 -0.36499,0.58731 -0.41785,0.94511 -0.20866,0.32369 -0.2598,0.71855 -0.47571,1.0374 -0.0588,0.22792 -0.45417,0.58871 -0.21593,0.77317 0.17559,0.16477 -0.45739,0.3482 -0.45852,0.6172 -0.19443,0.35075 0.20244,0.16374 0.23716,-0.0487 0.19866,0.0213 -0.14873,0.46087 -0.16338,0.61248 0.061,0.23503 -0.25086,0.084 -0.18389,0.22621 0.0489,0.15635 0.11616,0.29085 0.27212,0.14598 0.22858,-0.16237 0.2946,-0.0576 0.29863,0.10731 l 0.58122,-0.23538 c 0.0821,-0.26109 0.13335,-0.52666 0.32626,-0.73129 0.29731,-0.21914 0.4927,-0.5536 0.42053,-0.90933 0.28474,-0.14849 -0.14372,-0.26041 0.14824,-0.39219 0.0582,-0.25514 0.25636,-0.49292 0.21565,-0.73236 0.20454,-0.25284 0.2954,-0.60737 0.43347,-0.91621 0.15348,-0.36215 0.22129,-0.75885 0.40247,-1.10584 -0.28553,-0.13508 0.0396,-0.0926 0.0685,-0.27653 0.002,-0.34291 0.29517,-0.66136 0.262,-0.97518 -0.081,-0.33699 0.1328,-0.65558 0.15942,-0.96449 0.0684,-0.31442 0.27457,-0.63665 0.23586,-0.96492 0.0625,-0.50054 0.23178,-0.97825 0.42326,-1.44136 0.18554,-0.19438 0.0484,-0.42979 0.14205,-0.62178 0.0484,-0.35206 0.10613,-0.7143 0.13996,-1.07237 -0.0486,-0.32538 0.0402,-0.66028 0.0568,-0.98284 -0.1766,-0.19954 0.22096,-0.45145 -0.005,-0.58272 0.24328,-0.0558 0.0135,-0.13491 -0.0604,-0.28181 0.19276,-0.31297 0.27144,-0.70626 0.22753,-1.05971 -0.0252,-0.37093 0.27933,-0.65074 0.23872,-1.02209 0.13549,-0.19061 0.0276,-0.13206 -0.15026,-0.14387 0.20177,-0.27604 0.0426,-0.68124 0.0698,-1.00127 -0.25174,-0.36563 0.5493,-0.54553 0.3428,-0.81705 -0.25356,-0.23622 0.32006,-0.0315 0.44898,-0.15681 0.26237,0.0525 0.16508,-0.0532 0.004,-0.17008 -0.44135,-0.0992 -0.0848,-0.46999 0.19506,-0.38124 0.28648,0.19195 0.22981,-0.48563 0.0384,-0.17107 0.12844,-0.16201 -0.19188,-0.2907 -0.11936,-0.0706 -0.26147,0.0464 0.31544,-0.3974 0.12276,-0.52606 -0.5222,-0.16455 0.47293,-0.4875 0.0622,-0.80128 -0.15487,-0.35815 -0.30698,0.0345 -0.12834,0.21111 0.11054,0.25141 -0.0432,0.17819 -0.13073,0.002 -0.25855,-0.2018 0.0244,-0.36195 -0.0498,-0.57233 -0.0965,-0.35513 0.12392,-0.0329 0.17524,0.11028 0.25543,0.19264 0.35139,-0.0251 0.16432,-0.22037 -0.004,-0.33007 -0.26981,-0.63496 -0.12873,-0.93904 -0.26093,-0.11267 -0.0552,-0.27303 0.0915,-0.39209 -0.21433,-0.10496 0.0965,-0.35929 -0.11769,-0.54934 -0.25618,-0.32729 0.28716,-0.15825 0.25115,-0.39389 0.11279,-0.26574 -0.36057,-0.39294 -0.28675,-0.7419 0.005,-0.33252 -0.16188,-0.73649 0.0676,-0.99258 -0.30161,0.0812 -0.28705,-0.41089 -0.53868,-0.52123 -0.15675,0.11505 -0.4179,0.16922 -0.34564,-0.13387 0.14195,-0.3866 0.5009,0.36188 0.45085,-0.0604 0.23245,-0.0749 0.17085,-0.38571 -0.1051,-0.20675 -0.4024,0.21999 -0.32985,-0.45314 -0.03,-0.54873 0.23973,-0.30604 -0.0514,-0.68237 0.0219,-1.00559 -0.25351,0.14792 -0.18752,-0.10625 -0.0293,-0.19294 -0.17696,-0.18581 0.13556,-0.36147 -0.0298,-0.49713 0.0619,-0.35137 -0.28931,-0.20494 -0.41784,0.002 -0.25619,0.36645 -0.46783,-0.021 -0.55009,-0.28332 -0.23712,-0.21349 0.077,-0.44237 -0.0703,-0.63757 -0.0231,-0.364 -0.38759,-0.5027 -0.41947,-0.85291 -0.40308,-0.16283 -0.33097,-0.5917 -0.39057,-0.93046 0.066,-0.25081 0.0521,-0.49286 -0.16035,-0.69864 -0.0175,-0.32952 -0.20434,-0.65743 -0.31229,-0.96069 -0.27709,-0.23307 -0.26297,-0.67784 -0.45519,-0.90665 z",
1.20026, 9.02829,
"M -0.647852 8.69897  c 0.27123,0.89699 0.28293,-0.25188 1.17294,-0.91121 -0.019,-0.45337 -0.0333,-0.90699 -0.0667,-1.35961 -0.16717,-0.71424 -0.26623,-1.42486 -0.18603,-2.16524 -0.35177,-0.76585 0.25336,-1.53555 -0.20672,-2.29857 -0.0243,-0.66796 0.56421,-1.8924 0.64041,-2.52784 -0.36439,-0.62391 0.76929,-1.52704 0.43002,-2.6141 -0.11243,0.31694 -0.2569,0.21823 -0.27552,0.42189 -0.0451,0.49293 -0.42354,0.81201 -0.47179,0.95559 -0.21656,-0.23971 -0.88287,-1.05873 -0.93376,-1.99936 0.0638,-0.699 0.36156,-1.33045 0.25064,-2.05931 0.12634,-0.5642 -0.3497,-1.6984 0.4775,-1.78749 0.75422,-0.1124 -8e-5,-0.99299 -0.28784,-0.42271 -0.71152,-0.0264 0.003,-0.62728 -0.27184,-0.94723 -0.76967,-0.14096 -0.83801,0.98907 -0.55758,1.51309 0.38695,0.50259 0.21479,1.05455 -0.26614,1.39216 0.0501,0.97361 0.11051,1.94686 0.15451,2.92075 0.0822,1.46354 0,2.94343 0.2372,4.39147 0.007,0.81629 0.20886,1.59379 0.30335,2.39622 0.37107,0.77953 -0.48689,1.40959 -0.29922,2.23398 0.25837,0.71917 0.26924,1.49082 0.23463,2.2629 0.0409,0.22802 -0.003,0.42297 -0.078,0.60462 z",
1.28909, 16.4888,
"M -0.601805 -16.4887  c -0.64066,0.01 -0.68406,1.0337 -0.42117,1.52497 0.38696,0.50259 0.21482,1.05455 -0.26611,1.39216 0.0501,0.97361 0.11051,1.94686 0.15451,2.92075 0.0822,1.46354 -2e-5,2.94343 0.23718,4.39147 0.007,0.81629 0.20886,1.5938 0.30334,2.39623 0.37108,0.77953 -0.48686,1.40959 -0.29919,2.23397 0.25837,0.71917 0.26921,1.49084 0.23461,2.26291 0.0409,0.22802 -0.003,0.42297 -0.078,0.60462 0.2036,0.67334 0.26077,0.19399 0.65318,-0.36587 0.0969,0.52347 -0.0537,1.09567 -0.19016,1.50792 0.0168,0.88995 0.65551,0.74475 0.31729,1.64124 -0.0177,0.60701 -0.0656,1.18107 -0.18346,1.53996 0.25987,0.69342 -0.0198,1.56597 -0.17106,2.204 0.0508,0.45325 0.55251,1.01924 0.18243,1.5999 -0.21179,0.38644 0.0962,0.7179 0.22789,0.6749 0.10991,0.92985 -0.3225,-0.36708 -0.2527,0.61391 0.31975,-0.39811 0.20545,1.13303 0.061,1.51257 -0.1973,0.59553 0.0532,1.92311 -0.32557,1.88361 -0.17997,0.49675 0.006,0.49018 0.12301,0.36483 0.33612,0.4428 0.44119,1.18723 0.49659,2.04277 0.35883,0.26987 0.25639,-1.31676 0.30748,-1.8836 0.23582,-0.55477 -0.0262,-1.73369 0.22997,-2.05673 0.17235,-0.37447 0.2776,-0.92052 0.4041,-1.32911 0.0863,-0.79196 -0.53517,-1.19078 -0.13745,-1.73065 0.23614,0.1059 0.15208,-0.41602 0.14108,-0.81338 -0.096,-0.43366 0.0207,-0.65455 0.1421,-0.69195 l -0.0377,-1.03611 c -0.11308,0.12238 -0.18047,0.14259 -0.0718,-0.41652 0.0189,-0.52013 -0.27541,-0.60437 -0.0723,-1.15703 -0.14343,-0.6791 -0.24172,-1.29806 -0.2718,-2.10478 -0.0691,-0.7367 -0.28104,-1.22923 -0.22945,-1.98438 -0.19344,-0.65924 0.37674,-0.26839 0.13023,-1.00149 -0.20709,-0.36744 0.15814,-0.68667 -0.0765,-0.90382 0.23915,-0.85322 -0.17302,-1.75788 0.092,-2.56418 h 5e-4 c 0.18233,-0.97684 -0.41499,-1.56326 -0.47749,-2.1239 -0.0678,0.0482 -0.15229,0.007 -0.23513,-0.0357 -0.0187,-0.0404 -0.0387,-0.0806 -0.063,-0.12092 -0.0243,-0.66796 0.56457,-1.89257 0.64077,-2.52801 -0.36438,-0.62391 0.76922,-1.52674 0.42995,-2.6138 -0.11242,0.31694 -0.25681,0.21802 -0.27543,0.42168 -0.0451,0.49293 -0.42355,0.81192 -0.47181,0.9555 -0.21656,-0.23971 -0.88289,-1.05873 -0.93379,-1.99936 0.0638,-0.699 0.36155,-1.33045 0.25064,-2.05931 0.12634,-0.5642 -0.3497,-1.6984 0.47749,-1.78748 0.75422,-0.1124 -8e-5,-0.993 -0.28784,-0.42272 -0.71152,-0.0264 0.003,-0.62728 -0.27183,-0.94723 -0.0481,-0.009 -0.0937,-0.0125 -0.13642,-0.0119 z",
0.891012, 10.9638,
"M 0.353692 -8.74076  c -0.26499,0.8063 0.14695,1.71094 -0.0922,2.56416 0.23462,0.21715 -0.13037,0.53651 0.0767,0.90395 0.24652,0.7331 -0.32371,0.3421 -0.13029,1.00134 -0.0516,0.75515 0.16007,1.24804 0.22912,1.98474 0.0301,0.80672 0.12886,1.42571 0.27229,2.10481 -0.20308,0.55266 0.0911,0.63666 0.0722,1.1568 -0.10866,0.55911 -0.0412,0.53899 0.0719,0.4166 l 0.0376,1.03625 c -0.12144,0.0374 -0.23824,0.25801 -0.14224,0.69167 0.011,0.39736 0.0951,0.91925 -0.14103,0.81335 -0.39773,0.53987 0.224,0.93857 0.13772,1.73053 -0.12651,0.40859 -0.23189,0.95495 -0.40425,1.32942 -0.25615,0.32304 0.006,1.5018 -0.22982,2.05657 -0.0511,0.56684 0.0512,2.15362 -0.30762,1.88375 -0.0554,-0.85554 -0.16078,-1.60008 -0.49692,-2.04288 -0.11732,0.12535 -0.30284,0.13188 -0.12285,-0.36487 0.37871,0.0395 0.12845,-1.28817 0.32574,-1.8837 0.14445,-0.37954 0.25865,-1.91074 -0.0611,-1.51263 -0.0698,-0.98099 0.36247,0.31593 0.25256,-0.61392 -0.13168,0.043 -0.43956,-0.28828 -0.22776,-0.67472 0.37007,-0.58066 -0.13156,-1.14676 -0.18233,-1.60001 0.15123,-0.63803 0.43113,-1.51051 0.17126,-2.20393 0.11781,-0.35889 0.16536,-0.93265 0.18299,-1.53966 0.33822,-0.89649 -0.30017,-0.75144 -0.31703,-1.64139 0.16345,-0.49382 0.35255,-1.21826 0.1047,-1.81199 -0.36685,-0.1049 0.17326,-0.62553 0.0859,-1.18473 0.11267,-0.55844 0.1666,-0.62712 -0.10721,-0.68 0.21516,-0.68305 -0.17153,-1.14907 -0.0833,-1.906 0.13977,-0.58259 0.0436,-0.99951 -0.009,-1.65408 -0.0315,-1.04621 0.34527,-0.33266 0.55695,-0.48326 0.0625,0.56064 0.66002,1.14699 0.47769,2.12383 z",
1.40858, 15.4784,
"M -0.215316 -14.2449  c -0.0573,-0.39663 -0.09,-0.9845 -0.25167,-1.23351 l -0.008,0.0889 c -0.008,0.51419 -0.0135,1.82829 -0.0164,3.62614 -0.16272,-0.0566 -0.15885,-0.0396 -0.42542,-0.0862 0.12658,0.42973 0.13679,0.99178 0.0761,0.91046 0.003,0.20217 -0.0381,0.38449 -0.14261,0.45011 0.57288,0.0852 0.23548,1.11425 -0.31834,1.01699 -0.14563,0.51493 0.25241,1.08423 -0.0393,1.56063 0.18624,0.45092 0.31719,0.8511 0.0863,1.32291 0.24339,0.56784 0.14216,1.21654 0.18034,1.82315 0.0321,0.60964 -0.054,1.23667 0.0951,1.83296 0.21513,0.29947 -0.39592,0.30966 -0.11316,0.67903 0.039,0.74934 0.21058,1.48397 0.17309,2.23759 0.0111,0.43606 0.0354,1.12083 -0.22064,1.29139 0.64397,0.51467 -0.3131,0.91939 -0.18915,1.4857 0.0477,0.49886 0.30321,1.02072 0.0956,1.48415 -0.0396,0.59756 -0.0501,1.20341 0.21238,1.72341 0.0946,0.62128 -0.0948,1.21611 -0.0119,1.84536 -0.0453,0.5497 0.30977,1.1923 -0.0543,1.66451 -0.19063,0.44184 -8e-5,0.91834 -0.19378,1.33842 -0.0311,0.51712 -0.0706,1.05695 -0.0429,1.57355 -0.0493,0.5227 -0.0706,1.05957 -0.003,1.58026 0.079,0.51085 0.0409,0.93086 -0.0816,1.42059 0.0312,0.0244 0.0439,0.0591 0.0692,0.0868 l 1.44333,-3.35947 c -0.0392,-0.12325 -0.10192,-0.24343 -0.21964,-0.35295 -0.0919,-0.54896 0.15828,-1.13571 0.29766,-1.60611 -0.0725,-0.11045 -0.11144,-0.22339 -0.13435,-0.33744 l 1.00769,-1.16841 c -2.7e-4,-0.0874 0.0233,-0.25803 -0.051,-0.4113 -0.0574,-0.11829 -0.21368,-0.21272 -0.23231,-0.33456 -0.0501,-0.32747 0.096,-0.64257 0.0949,-1.07853 0,-0.22596 -0.12986,-0.49448 -0.12229,-0.73317 0.0391,-1.23285 0.2989,-2.52798 0.29792,-3.1837 0.0119,-0.61935 -0.007,-12.50878 -0.007,-13.10462 0.0514,-0.4966 -0.0418,0.0395 0.1633,-0.14005 0.0463,0.3542 0.0498,-0.47899 0.0491,-0.11317 0.0143,-1.07594 -0.0857,-2.58572 -0.28713,-2.87271 -0.1198,-0.17068 -0.28631,0.41503 -0.45638,0.89253 -0.0151,-0.0313 -0.61574,-1.82898 -0.72041,-1.81959 z m 0.94094,4.32328 c -0.007,2.38506 -0.13253,2.52318 -0.13645,1.38389 -0.004,-0.49656 -0.0196,-0.97179 0.13645,-1.38389 z m -0.7084,1.35338 c 0,7e-5 0,-9e-5 0.003,0 0.0662,0.009 0.17023,0.0887 0.1819,0.15865 0.008,0.0466 -0.0254,0.0889 -0.13798,0.10335 -0.0214,-0.0326 -0.0326,-0.0554 -0.0465,-0.0811 2.1e-4,-0.0603 0,-0.12059 0,-0.18087 z m 0.48411,-0.0599 c 0.0272,0.008 0.063,0.0884 0.0682,0.18345 0.003,0.0634 -0.007,0.13334 -0.0419,0.19224 -0.0347,-0.0107 -0.0299,-0.0884 -0.0419,-0.12299 -0.0298,-0.19578 -0.0116,-0.2606 0.0156,-0.2527 z m -0.55491,2.5006 c 0.0892,0.006 0.19317,0.0297 0.24858,0.0346 0.11311,-0.0172 0.16436,0.0169 0.17415,0.0708 0.003,0.0101 0.003,0.0207 0,0.032 -0.0114,0.14338 -0.23088,0.36319 -0.37465,0.25063 -0.0662,-0.13041 -0.10052,-0.25938 -0.1106,-0.38706 0.0193,-0.001 0.0391,-0.002 0.0615,-0.001 z m 0.13229,1.002 c 0.0214,-0.01 0.0347,0.0183 0.0383,0.0615 0.007,0.0865 -0.0249,0.23269 -0.11009,0.25321 -0.0212,-0.0395 -0.0345,-0.10075 -0.014,-0.14831 0.0356,-0.10904 0.0645,-0.15666 0.0858,-0.1664 z m 0,0.52504 c 0.006,-0.002 0.0143,2.2e-4 0.0233,0.005 0.036,0.0197 0.0874,0.0851 0.0889,0.12505 0,0.0264 -0.0206,0.0421 -0.0842,0.0248 -0.0461,-0.10744 -0.046,-0.14933 -0.0279,-0.15502 z m 0.0201,4.1553 v 0 c 0.0656,-0.0183 0.009,0.15149 0.0156,0.29559 0.0532,0.0409 0.0778,0.0874 0.0837,0.13229 0.003,0.0122 0.003,0.0243 0,0.0362 -0.0151,0.17868 -0.30001,0.30917 -0.21085,-0.0398 -0.005,-0.11149 -0.0596,-0.23076 -0.003,-0.33538 0.056,-0.0558 0.0913,-0.0828 0.11319,-0.0889 z m 1.20973,2.00814 c -0.003,-2.6e-4 -0.005,0.004 -0.007,0.0114 v 0 c -0.008,0.0314 -0.0132,0.12504 0,0.25528 l 0.005,0.0145 0.006,-0.03 c 0.003,-0.0485 0.005,-0.0893 0.005,-0.12299 0.003,-0.0882 -0.003,-0.12732 -0.01,-0.12815 z m -0.0517,0.48576 c -0.007,-0.004 -0.0151,-0.003 -0.0228,0.005 -0.0201,0.0212 -0.0409,0.0895 -0.0594,0.22738 -0.11983,0.55066 -0.2176,3.41162 0.003,3.71346 0.36878,-0.48993 0.18709,-2.28477 0.17984,-2.57762 -0.007,-0.29285 -0.0406,-0.59455 -0.0305,-0.87643 0.0206,-0.16536 -0.0193,-0.46288 -0.0703,-0.49196 z",
1.40858, 15.4784,
"M -0.215316 -14.2449  c -0.0573,-0.39663 -0.09,-0.9845 -0.25167,-1.23351 l -0.008,0.0889 c -0.008,0.51419 -0.0135,1.82828 -0.0164,3.62614 -0.16272,-0.0566 -0.15885,-0.0396 -0.42542,-0.0862 0.12658,0.42973 0.13679,0.99178 0.0761,0.91046 0.003,0.20216 -0.0381,0.38449 -0.14261,0.4501 0.57288,0.0852 0.23548,1.11425 -0.31834,1.017 -0.14563,0.51492 0.25241,1.08422 -0.0393,1.56063 0.18624,0.45092 0.31719,0.8511 0.0863,1.32291 0.24339,0.56784 0.14216,1.21653 0.18034,1.82315 0.0321,0.60964 -0.054,1.23666 0.0951,1.83296 0.21513,0.29947 -0.39592,0.30966 -0.11316,0.67902 0.039,0.74935 0.21058,1.48397 0.17309,2.23759 0.0111,0.43606 0.0354,1.12083 -0.22064,1.2914 0.64397,0.51466 -0.3131,0.91938 -0.18915,1.48569 0.0477,0.49885 0.30321,1.02071 0.0956,1.48415 -0.0396,0.59756 -0.0501,1.20342 0.21238,1.72341 0.0946,0.62127 -0.0948,1.21611 -0.0119,1.84536 -0.0453,0.5497 0.30977,1.19229 -0.0543,1.6645 -0.19063,0.44184 -8e-5,0.91834 -0.19378,1.33842 -0.0311,0.51712 -0.0706,1.05695 -0.0429,1.57355 -0.0493,0.5227 -0.0706,1.05957 -0.003,1.58026 0.079,0.51085 0.0409,0.93086 -0.0816,1.42059 0.0312,0.0244 0.0439,0.0591 0.0692,0.0868 l 1.44333,-3.35948 c -0.0392,-0.12324 -0.10192,-0.24343 -0.21964,-0.35295 -0.0919,-0.54896 0.15828,-1.13571 0.29766,-1.60611 -0.0725,-0.11045 -0.11144,-0.22339 -0.13435,-0.33744 l 1.00769,-1.16841 c -2.7e-4,-0.0874 0.0233,-0.25803 -0.051,-0.4113 -0.0574,-0.11829 -0.21368,-0.21272 -0.23231,-0.33456 -0.0501,-0.32747 0.096,-0.64257 0.0949,-1.07853 0,-0.22595 -0.12986,-0.49447 -0.12229,-0.73317 0.0391,-1.23284 0.2989,-2.52797 0.29792,-3.18369 0.0119,-0.61935 -0.007,-12.50878 -0.007,-13.10462 0.0514,-0.4966 -0.0418,0.0395 0.1633,-0.14005 0.0463,0.3542 0.0498,-0.47899 0.0491,-0.11317 0.0143,-1.07594 -0.0857,-2.58572 -0.28713,-2.87271 -0.1198,-0.17068 -0.28631,0.41503 -0.45638,0.89253 -0.0151,-0.0313 -0.61574,-1.82898 -0.72041,-1.81959 z m 0.94094,4.32327 c -0.007,2.38507 -0.13253,2.52318 -0.13645,1.3839 -0.004,-0.49657 -0.0196,-0.97179 0.13645,-1.3839 z m -0.7084,1.35339 c 0,7e-5 0,-9e-5 0.003,0 0.0662,0.009 0.17023,0.0887 0.1819,0.15865 0.008,0.0466 -0.0254,0.0889 -0.13798,0.10335 -0.0214,-0.0326 -0.0326,-0.0554 -0.0465,-0.0811 2.1e-4,-0.0603 0,-0.12058 0,-0.18087 z m 0.48411,-0.0599 c 0.0272,0.008 0.063,0.0884 0.0682,0.18345 0.003,0.0634 -0.007,0.13333 -0.0419,0.19224 -0.0347,-0.0107 -0.0299,-0.0884 -0.0419,-0.123 -0.0298,-0.19577 -0.0116,-0.26059 0.0156,-0.25269 z m -0.55491,2.5006 c 0.0892,0.006 0.19317,0.0297 0.24858,0.0346 0.11311,-0.0172 0.16436,0.0169 0.17415,0.0708 0.003,0.0101 0.003,0.0207 0,0.032 -0.0114,0.14339 -0.23088,0.36319 -0.37465,0.25063 -0.0662,-0.13041 -0.10052,-0.25937 -0.1106,-0.38705 0.0193,-0.001 0.0391,-0.002 0.0615,-0.001 z m 0.13229,1.002 c 0.0214,-0.01 0.0347,0.0183 0.0383,0.0615 0.007,0.0865 -0.0249,0.23269 -0.11009,0.25321 -0.0212,-0.0395 -0.0345,-0.10075 -0.014,-0.14832 0.0356,-0.10903 0.0645,-0.15665 0.0858,-0.16639 z m 0,0.52504 c 0.006,-0.002 0.0143,2.1e-4 0.0233,0.005 0.036,0.0197 0.0874,0.0851 0.0889,0.12505 0,0.0264 -0.0206,0.0421 -0.0842,0.0248 -0.0461,-0.10744 -0.046,-0.14934 -0.0279,-0.15502 z m 0.0201,4.1553 v 0 c 0.0656,-0.0183 0.009,0.15149 0.0156,0.29558 0.0532,0.0409 0.0778,0.0874 0.0837,0.1323 0.003,0.0122 0.003,0.0243 0,0.0362 -0.0151,0.17868 -0.30001,0.30918 -0.21085,-0.0398 -0.005,-0.1115 -0.0596,-0.23077 -0.003,-0.33539 0.056,-0.0558 0.0913,-0.0828 0.11319,-0.0889 z m 1.20973,2.00814 c -0.003,-2.6e-4 -0.005,0.004 -0.007,0.0114 v 0 c -0.008,0.0314 -0.0132,0.12505 0,0.25528 l 0.005,0.0145 0.006,-0.03 c 0.003,-0.0485 0.005,-0.0893 0.005,-0.12298 0.003,-0.0882 -0.003,-0.12732 -0.01,-0.12816 z m -0.0517,0.48576 c -0.007,-0.004 -0.0151,-0.003 -0.0228,0.005 -0.0201,0.0212 -0.0409,0.0895 -0.0594,0.22737 -0.11983,0.55067 -0.2176,3.41162 0.003,3.71346 0.36878,-0.48992 0.18709,-2.28477 0.17984,-2.57762 -0.007,-0.29285 -0.0406,-0.59455 -0.0305,-0.87643 0.0206,-0.16535 -0.0193,-0.46288 -0.0703,-0.49196 z",
1.19541, 13.7255,
"M -0.876944 -13.7255  0.0201,0.0681 c 0.1772,0.62029 0.2896,1.04659 0.2289,0.96528 0.002,0.20217 -0.038,0.38439 -0.14247,0.45001 0.57288,0.0852 0.23535,1.11439 -0.31847,1.01713 -0.14564,0.51493 0.25248,1.08419 -0.0392,1.56059 0.18623,0.45093 0.31728,0.85144 0.0864,1.32325 0.24339,0.56784 0.14204,1.21638 0.18021,1.82299 0.0321,0.60964 -0.0539,1.23657 0.0952,1.83287 0.21514,0.29946 -0.3963,0.30988 -0.11356,0.67925 0.039,0.74934 0.21098,1.48377 0.17351,2.23739 0.0112,0.43606 0.0353,1.12094 -0.22077,1.2915 0.64399,0.51467 -0.31336,0.91952 -0.1894,1.48583 0.0477,0.49885 0.3035,1.02028 0.0959,1.48372 -0.0396,0.59756 -0.0502,1.20353 0.21231,1.72353 0.0946,0.62128 -0.0949,1.21631 -0.012,1.84556 -0.0453,0.5497 0.30974,1.19243 -0.0543,1.66464 -0.19062,0.44184 9e-5,0.91798 -0.19363,1.33806 -0.0311,0.51712 -0.0704,1.05705 -0.0427,1.57365 -0.0493,0.5227 -0.071,1.05966 -0.003,1.58035 0.079,0.51085 0.041,0.93086 -0.0815,1.42059 0.0312,0.0244 0.0438,0.059 0.0691,0.0867 l 1.4435,-3.35926 c -0.0392,-0.12324 -0.102,-0.24386 -0.21971,-0.35338 -0.0919,-0.54896 0.15862,-1.13534 0.29801,-1.60574 -0.33671,-0.51282 -0.0478,-1.07406 0.0233,-1.5934 0.1794,-0.51138 -0.42493,-0.9927 -0.1319,-1.57117 0.10052,-0.42731 0.34806,-0.7585 0.55653,-1.13598 -0.0485,-0.45487 -0.63618,-0.4481 -0.40805,-0.98503 0.20565,-0.4798 0.19847,-1.00573 0.32411,-1.49994 0.0473,-0.60976 -0.18707,-1.15221 -0.31882,-1.73588 -0.20363,-0.49064 -0.48402,-0.98562 -0.62777,-1.483 -0.0236,-0.51779 0.25969,-0.98172 0.1594,-1.47137 0.1783,-0.63379 0.0603,-1.30668 0.0592,-1.95596 -0.13465,-0.3989 -0.16238,-0.76032 0.0586,-1.11763 -0.46748,-0.37921 0.27905,-0.83492 0.14106,-1.29856 -0.31768,-0.55003 0.008,-1.06122 -0.008,-1.63184 0.006,-0.47253 0.005,-0.94519 0.008,-1.41776 0.24498,-0.23512 0.71548,-0.65116 0.36114,-0.92225 0.73256,-2.2984 0.9711,-1.21193 -0.28683,-2.21043 -0.41602,-0.19937 -0.35419,0.1806 -1.06564,-0.0622 l -0.11675,-0.0402 z",
0.884301, 13.8117,
"M -0.123121 -2.30275  c -0.004,0.0919 -0.0285,0.17097 -0.0723,0.21497 -0.0879,0.0879 -0.15049,-0.0684 -0.13901,-0.34726 0.0127,-0.30824 0.0757,-0.37079 0.1602,-0.15968 0.0382,0.0955 0.0552,0.20005 0.0512,0.29197 z m 1.00562,-4.18682 c -0.0136,0.12449 -0.10434,0.28911 -0.27647,0.53486 -0.23763,0.33924 -0.43417,0.85838 -0.43615,1.15339 -0.002,0.29504 -0.1194,0.71499 -0.26045,0.93329 -0.21074,0.32602 -0.25744,0.29832 -0.26045,-0.15451 -0.002,-0.30319 0.041,-0.60087 0.0961,-0.66146 0.0825,-0.0908 0.27686,-3.75423 0.44804,-8.44444 0.031,-0.84879 0.0508,-0.8709 0.24029,-0.26459 0.11369,0.36381 0.14246,1.9526 0.0641,3.53053 -0.10848,2.18249 -0.061,2.9193 0.19896,3.0799 0.13612,0.0841 0.19963,0.16849 0.18603,0.29303 z m -0.97307,14.8394 c 0,0.28238 -0.11129,0.44439 -0.24753,0.36017 -0.13626,-0.0842 -0.24426,-0.35901 -0.23926,-0.61081 0.005,-0.25181 -0.0503,-1.31485 -0.12247,-2.36263 -0.0894,-1.29767 -0.0341,-1.94495 0.17311,-2.02985 0.21135,-0.0867 0.20302,-0.13124 -0.0269,-0.14574 -0.1819,-0.0114 -0.32231,-0.28855 -0.31212,-0.61597 0.0178,-0.56801 0.029,-0.57121 0.24598,-0.0661 0.15007,0.34917 0.23368,-0.29792 0.24597,-1.90273 0.0102,-1.33752 0.11528,-2.49142 0.23307,-2.56418 0.11788,-0.0728 0.16275,0.57507 0.10025,1.4397 -0.0711,0.98351 -0.0107,1.572 0.16123,1.572 0.54297,0 -0.21133,6.64377 -0.21133,6.92614 z m -0.43305,2.47427 c -0.003,0.0791 -0.0141,0.16582 -0.0315,0.25683 -0.0696,0.36375 -0.17208,0.66146 -0.22789,0.66146 -0.0558,0 -0.10129,-0.29771 -0.10129,-0.66146 0,-0.3638 0.10256,-0.66145 0.2279,-0.66145 0.094,0 0.14338,0.16742 0.1328,0.40462 z m -0.12867,1.68982 c -0.004,0.0919 -0.0285,0.17097 -0.0724,0.21497 -0.0879,0.0879 -0.1505,-0.0684 -0.13901,-0.34726 0.0127,-0.30824 0.0757,-0.37077 0.1602,-0.15968 0.0382,0.0955 0.0552,0.20005 0.0512,0.29197 z m 0,1.05833 c -0.004,0.0919 -0.0285,0.17097 -0.0724,0.21497 -0.0879,0.0879 -0.1505,-0.0684 -0.13901,-0.34726 0.0127,-0.30824 0.0757,-0.37079 0.1602,-0.15968 0.0382,0.0955 0.0552,0.20005 0.0512,0.29197 z",
0.762531, 16.9731,
"M 0.232564 -5.04002  c -0.0256,-0.11674 -0.0255,0.0289 2.6e-4,0.32444 0.0258,0.29532 -0.0179,0.76648 -0.0971,1.0473 -0.1026,0.36391 -0.10352,0.73496 -0.002,1.29023 0.078,0.42853 0.12701,1.18085 0.10877,1.67181 -0.0212,0.57309 0.0381,0.8967 0.16574,0.90407 0.1094,0.006 0.20455,0.27367 0.21158,0.59484 0.007,0.32117 0.0586,0.89271 0.11421,1.26969 0.0618,0.41843 0.0213,1.04432 -0.10385,1.60629 -0.11282,0.50634 -0.20014,1.44467 -0.19397,2.08543 0.014,1.44453 -0.12951,2.17576 -0.37292,1.90107 -0.1185,-0.13368 -0.18543,0.0559 -0.18543,0.52551 0,0.40439 -0.0704,0.80453 -0.15642,0.88886 -0.0967,0.0949 -0.0239,0.44232 0.0371,0.91016 0.0545,0.41609 0.0799,0.79268 0.057,0.83706 -0.0229,0.0444 0.0755,0.33281 0.21858,0.64057 0.22056,0.47431 0.2334,0.65408 0.0844,1.18204 -0.0967,0.34255 -0.22876,0.52665 -0.29342,0.40904 -0.12361,-0.22462 -0.1455,0.20684 -0.13985,2.75717 0.004,1.47266 -0.0968,1.49441 -0.19275,0.38996 -0.0557,-0.64348 -0.0753,-1.8719 -0.0437,-2.72975 0.0963,-2.62133 0.0685,-4.17561 -0.0793,-4.44423 -0.16273,-0.29578 -0.0597,-2.89904 0.12249,-3.09534 0.0663,-0.0715 0.1127,-0.94893 0.10307,-1.94981 -0.0172,-1.79322 -0.0187,-1.79958 -0.11498,-0.45471 -0.0781,1.0917 -0.10126,0.79463 -0.11525,-1.48522 -0.0111,-1.81302 0.0309,-2.76277 0.11525,-2.6094 0.0823,0.14963 0.1326,-0.58109 0.1326,-1.92698 0,-1.47122 -0.0508,-2.13633 -0.15798,-2.07016 -0.0868,0.0537 -0.18056,-0.20961 -0.20847,-0.58495 -0.0279,-0.37534 0.0131,-0.68244 0.0914,-0.68244 h 2.6e-4 c 0.0924,0 0.14244,-1.44486 0.14244,-4.11748 0,-2.64579 0.0474,-4.03168 0.13259,-3.87679 0.0798,0.14518 0.13259,-0.37062 0.13259,-1.29631 0,-1.36676 0.0404,-1.55998 0.36387,-1.7419 0.48964,-0.27527 0.1634,-0.0741 0.1634,1.52406 0,0.94267 -0.048,1.2778 -0.15513,1.08305 -0.10578,-0.19225 -0.12686,-0.065 -0.0663,0.39909 0.0488,0.37448 0.0888,0.92971 0.0888,1.23389 0,0.30416 0.096,0.55297 0.21314,0.55297 0.13289,0 0.18485,0.21576 0.13829,0.57273 -0.0411,0.31487 -0.0177,0.67604 0.0521,0.80278 0.0697,0.12676 0.1269,0.49327 0.1269,0.815 0,0.32153 -0.0597,0.47673 -0.1326,0.34425 -0.0729,-0.1325 -0.1326,-0.055 -0.1326,0.17214 0,0.22706 0.0525,0.41282 0.11654,0.41282 0.0641,0 -0.26373,4.03388 -0.28935,3.91715 z",
0.560077, 8.98122,
"M 0.467182 8.52902  c -0.0714,0.48313 -0.12974,0.18491 -0.12974,-0.66012 0,-0.84519 -0.10442,-1.5299 -0.23204,-1.51334 -0.1984,0.0149 -0.20328,0.22696 -0.0331,1.43673 0.15228,1.08338 0.11346,1.34238 -0.16574,1.11095 -0.29936,-0.24779 -0.36464,-0.99245 -0.36464,-4.15731 0,-2.12029 -0.0357,-5.81576 -0.0795,-8.212422 -0.0467,-2.55505 -0.019,-4.76947 0.0673,-5.35317 0.081,-0.54783 0.30289,0.34342 0.49308,1.9808 0.19018,1.63694 0.4026,2.59378 0.47186,2.12546 0.0693,-0.46831 0.11396,12.285842 -0.0275,13.242422 z",
0.762531, 19.7675,
"M 0.232564 -7.83435  c -0.0256,-0.11674 -0.0255,0.0289 2.6e-4,0.32444 0.0258,0.29532 -0.0179,0.76648 -0.0971,1.0473 -0.1026,0.36391 -0.10352,0.73496 -0.002,1.29023 0.078,0.42853 0.12701,1.18085 0.10877,1.67181 -0.0212,0.57309 0.0381,0.8967 0.16574,0.90407 0.1094,0.006 0.20455,0.27367 0.21158,0.59484 0.007,0.32117 0.0586,0.89271 0.11421,1.26969 0.0618,0.41843 0.0213,1.04432 -0.10385,1.60629 -0.11282,0.50634 -0.20014,1.44467 -0.19397,2.08543 0.014,1.44453 -0.12951,2.17576 -0.37292,1.90107 -0.1185,-0.13368 -0.18543,0.0559 -0.18543,0.52551 0,0.40439 -0.0704,0.80453 -0.15642,0.88886 -0.0967,0.0949 -0.0239,0.44232 0.0371,0.91016 0.0545,0.41609 0.0799,0.79268 0.057,0.83706 -0.0229,0.0444 0.0755,0.33281 0.21858,0.64057 0.22056,0.47431 0.2334,0.65408 0.0844,1.18204 -0.0967,0.34255 -0.22876,0.52665 -0.29342,0.40904 -0.12361,-0.22462 -0.1455,0.20684 -0.13985,2.75717 0.004,1.47266 -0.0968,1.49441 -0.19275,0.38996 -0.0557,-0.64348 -0.0753,-1.8719 -0.0437,-2.72975 0.0963,-2.62133 0.0685,-4.17561 -0.0793,-4.44423 -0.16273,-0.29578 -0.0597,-2.89904 0.12249,-3.09534 0.0663,-0.0715 0.1127,-0.94893 0.10307,-1.94981 -0.0172,-1.79322 -0.0187,-1.79958 -0.11498,-0.45471 -0.0781,1.0917 -0.10126,0.79463 -0.11525,-1.48522 -0.0111,-1.81302 0.0309,-2.76277 0.11525,-2.6094 0.0823,0.14963 0.1326,-0.58109 0.1326,-1.92698 0,-1.47122 -0.0508,-2.13633 -0.15798,-2.07016 -0.0868,0.0537 -0.18056,-0.20961 -0.20847,-0.58495 -0.0279,-0.37534 0.0131,-0.68244 0.0914,-0.68244 h 2.6e-4 c 0.0924,0 0.14244,-1.44486 0.14244,-4.11748 0,-2.64579 0.0474,-4.03168 0.13259,-3.87679 0.0798,0.14518 0.13259,-0.37062 0.13259,-1.29631 0,-1.36676 0.0404,-1.55998 0.36387,-1.7419 0.48964,-0.27527 0.1634,-0.0741 0.1634,1.52406 0,0.94267 -0.048,1.2778 -0.15513,1.08305 -0.10578,-0.19225 -0.12686,-0.065 -0.0663,0.39909 0.0488,0.37448 0.0888,0.92971 0.0888,1.23389 0,0.30416 0.096,0.55297 0.21314,0.55297 0.13289,0 0.18485,0.21576 0.13829,0.57273 -0.0411,0.31487 -0.0177,0.67604 0.0521,0.80278 0.0697,0.12676 0.1269,0.49327 0.1269,0.815 0,0.32153 -0.0597,0.47673 -0.1326,0.34425 -0.0729,-0.1325 -0.1326,-0.055 -0.1326,0.17214 0,0.22706 0.0525,0.41282 0.11654,0.41282 0.0641,0 -0.26373,4.03388 -0.28935,3.91715 z m 0.0396,27.48027 c -0.0714,0.12985 -0.12974,0.0497 -0.12974,-0.17742 0,-0.22716 -0.10442,-0.41119 -0.23204,-0.40674 -0.1984,0.004 -0.20328,0.061 -0.0331,0.38615 0.15228,0.29118 0.11346,0.36079 -0.16574,0.29859 -0.29936,-0.0666 -0.36464,-0.26674 -0.36464,-1.11736 0,-0.56987 -0.0357,-1.5631 -0.0795,-2.20725 -0.0467,-0.68672 -0.019,-1.28189 0.0673,-1.43877 0.081,-0.14724 0.30289,0.0923 0.49308,0.53238 0.19018,0.43996 0.4026,0.69713 0.47186,0.57126 0.0693,-0.12587 0.11396,3.30206 -0.0275,3.55916 z",
1.269, 19.5416,
"M 0.14825 -19.5416  c -0.32661,0.39468 -0.3994,1.66107 -0.7598,2.04169 -0.41626,0.2009 -0.37637,1.71843 -0.58357,2.35306 0.0458,1.94112 -0.13983,3.86046 -0.048,5.80025 0.0294,1.34326 0.10658,2.68769 0.0855,4.03529 -0.0762,1.10649 0.0737,2.4593 -0.0252,3.70953 -0.0204,0.90305 0.056,1.55369 -0.0307,2.40511 0.0317,2.24936 0.21841,4.51036 0.0946,6.75626 -0.0753,1.08892 0.0479,1.9506 0.41334,1.69121 0.43294,0.76903 0.0677,2.22794 -0.0667,3.18578 -0.19001,0.81644 0.11544,1.40278 0.30591,1.39693 -0.0451,-0.42285 -0.28028,-2.17954 0.0281,-1.08788 0.33743,0.48853 -0.12596,2.7085 -0.38629,1.41622 -0.28035,0.0578 0.0439,1.62405 -0.26723,2.03911 -0.13348,0.55093 0.1996,1.81699 0.24795,0.69045 -0.11076,1.53067 0.61772,1.10071 0.51013,2.65016 0.0573,-0.0207 0.22667,-1.00323 0.23225,-1.34788 0.08,-0.9498 0.53243,-2.15075 0.25784,-3.07732 -0.30065,-0.79082 0.41871,-1.52126 0.11979,-2.62364 -0.0371,-0.56266 -0.17904,-1.49854 -0.12697,-0.30689 0.003,0.94503 -0.34219,0.63414 -0.29234,1.70477 -0.30776,-0.58284 0.30053,-2.415 -0.32326,-2.26952 0.11911,-0.39881 0.21227,-1.78449 0.37582,-0.58638 0.38472,0.39819 0.28972,-1.2848 0.30849,-2.1732 0.01,-1.30431 0.11678,-2.2807 -0.0501,-3.51089 -0.22551,-0.80379 0.19385,-2.07867 -0.24626,-2.47433 -0.35322,-0.76698 0.20938,-2.5558 0.24625,-0.9447 0.20473,1.61631 0.33115,-1.44215 -0.0187,-1.00535 -0.20738,-0.73783 0.50146,-0.8757 0.0966,-1.34105 -0.23718,0.80488 -0.81457,0.0175 -0.51184,-1.058 0.23378,-0.45163 0.79437,0.53657 0.57233,-1.14201 -0.13148,-0.58581 -0.34247,-1.16952 -0.45141,-1.63641 0.0722,-0.37607 0.21467,-1.5057 0.37321,-0.66014 0.20019,0.95972 0.14835,-1.38787 0.16689,-1.8434 -0.0266,-0.64918 0.1111,-1.97392 -0.17902,-1.90736 -0.13214,-0.95422 0.003,-2.15651 0.24077,-2.14256 -0.0855,-1.27561 0.30739,-2.14491 0.59873,-1.54848 0.46782,-0.57364 0.0341,-2.11974 -0.0515,-3.1238 -0.0488,-1.29164 -0.35659,-0.91037 -0.47231,-0.43792 -0.20293,-0.6018 -0.22527,-1.26026 -0.1283,-2.00164 -0.13967,-0.42976 -0.33758,0.56271 -0.34577,-0.61191 0.10408,-0.26168 0.33651,-0.79224 0.0907,-1.01316 z",
2.21549, 17.1584,
"M -0.167255 -17.1539  c -0.0167,0.007 -0.0346,0.0209 -0.0528,0.0449 -0.0209,0.0276 -0.0424,0.0677 -0.0625,0.1243 0.29371,0.10775 0.23216,-0.21534 0.1151,-0.1692 z m 0.15604,1.07656 c -0.12396,-5.2e-4 -0.2654,0.0531 -0.37769,0.13732 l -0.001,8.7e-4 c -0.0997,0.0747 -0.17632,0.1734 -0.19745,0.27884 0.47404,0.75848 -0.48688,0.22899 -0.21252,0.89958 -0.0284,1.36166 1.17753,1.85975 0.65585,3.19253 1.11667,0.51287 -0.4379,2.54995 0.4584,2.75912 -0.46709,-0.10692 -0.70481,0.87074 -0.3325,1.21901 0.65675,1.18822 -0.35631,2.59721 0.49336,3.59223 0.1701,0.19469 0.32366,-0.23386 0.45247,-0.34791 0.90751,-0.55526 -0.97382,-0.65062 -0.32228,-1.03339 1.05641,-0.14684 0.66978,-2.07793 0.30667,-2.90072 -0.1278,0.58401 -0.87195,-0.365 -0.2916,-0.99799 0.7704,0.12479 -0.23691,-1.49304 0.37823,-1.11971 0.49922,-0.62397 -0.0865,-0.88636 -0.47023,-1.2518 -0.78012,-0.24 0.39106,-0.72068 -0.12964,-1.48835 0.24984,-0.74186 0.6296,-1.90821 -0.15119,-2.37151 0.10592,-0.41508 -0.0521,-0.5672 -0.25878,-0.56805 z m 1.05236,3.59827 c -0.0745,0.0468 -0.0538,0.51726 0.13983,0.86417 0.1982,0.18937 0.12229,-0.24571 0.0775,-0.36517 -0.0941,-0.39821 -0.17239,-0.52458 -0.21681,-0.499 z m 0.65476,1.56087 c -0.0264,0.005 -0.0484,0.015 -0.0667,0.0327 -0.16138,0.15286 -0.015,0.75465 0.38253,0.79942 0.3019,0.32074 0.22976,-0.59592 0.0327,-0.63453 -0.15403,-0.15771 -0.26966,-0.21053 -0.34864,-0.1977 z m -3.49173,0.28921 c -0.0476,-0.0172 -0.11945,0.0254 -0.18993,0.11043 h -5.4e-4 c -0.11277,0.13632 -0.22294,0.3826 -0.22919,0.67337 0.0586,0.30913 0.29916,0.0322 0.34164,-0.13031 0.16947,-0.43186 0.1572,-0.62497 0.078,-0.65352 z m 3.02689,0.71223 c -0.015,-7.5e-4 -0.0279,0.007 -0.0371,0.019 -0.0354,0.0528 -0.0117,0.20856 0.1415,0.51884 l 0.1146,0.1278 c 0.14585,-0.31772 -0.10424,-0.66014 -0.21896,-0.6656 z m -0.0371,0.019 c 0,0 0,-8.9e-4 0,-8.9e-4 l -0.002,0.002 c 5e-4,-7.5e-4 0.002,-1.6e-4 0.002,-8.8e-4 z m -0.52135,6.72 c -0.3266,0.0409 -0.95395,0.87087 -1.18526,1.33727 -0.16786,-0.13281 -0.29294,-0.12997 -0.38091,-0.0431 -0.26269,0.26077 -0.19142,1.27654 0.0678,1.63511 0.24643,-0.25664 0.56346,-0.67538 0.39223,0.23741 -0.75859,0.63588 -0.69093,1.82753 -0.36102,2.9059 -0.34879,0.52437 -0.0947,1.14594 0.44171,1.23281 0.99104,0.75169 0.10007,2.18367 0.0608,3.29439 0.68087,1.18413 -0.94373,1.30332 -0.72471,2.27914 -0.69212,0.6987 -0.66851,2.92111 0.20443,3.22274 0.75306,0.225 0.20039,0.34297 -0.21036,0.51971 0.26459,0.80376 0.93921,1.0819 1.25036,1.85267 0.31277,1.28598 0.55164,-0.0849 0.42449,-0.65009 -0.66183,-0.70621 0.80496,-1.24759 0.33626,-1.56863 -0.45715,0.66148 -1.18436,-0.24337 -0.46161,-0.78475 0.90308,-0.8666 -0.34499,-1.44013 0.10057,-2.29122 0.46346,-0.31165 0.41583,-1.35705 -0.10541,-1.27598 0.002,0.36914 -0.21848,1.21584 -0.4654,0.52404 0.49413,-0.71096 1.17556,-1.9191 0.56762,-3.05611 -0.85085,-1.57139 0.80183,-0.85373 0.53587,-2.24548 0.19424,-1.25981 -0.31804,-2.6187 -0.11511,-3.72086 0.8693,-1.12613 -0.65975,-1.20057 -0.25986,-2.37756 0.17124,-0.79304 0.0835,-1.05187 -0.11243,-1.02734 z m -1.83464,17.71081 c -0.0134,0.002 -0.0271,0.0117 -0.0393,0.0319 -0.0134,0.0231 -0.0247,0.06 -0.0356,0.11477 0.2771,0.20545 0.1713,-0.16054 0.0748,-0.14684 z m 0.98242,0.93411 c -0.0473,-0.005 -0.0936,0.0117 -0.13065,0.0466 v 8.8e-4 c -0.11527,0.10642 -0.14517,0.38416 0.14585,0.72864 0.0788,0.1577 0.25607,0.25348 0.22974,-0.0665 0.0485,-0.49012 -0.10308,-0.69576 -0.24481,-0.70964 z m -0.40028,1.03683 c -0.0246,0.005 -0.0473,0.0167 -0.0662,0.0354 l -0.002,8.9e-4 c -0.0687,0.068 -0.0867,0.22859 0.0388,0.50935 l 0.14467,0.10876 c 0.31715,-0.40726 0.0558,-0.69018 -0.1156,-0.65438 z",
1.67889, 12.4576,
"M 0.170368 -12.417  c -0.11243,0.0479 -0.25614,0.13683 -0.43964,0.27748 -1.27389,0.78909 -1.21189,2.51347 -0.18945,3.31231 -0.13431,0.55548 -0.36031,1.04129 -0.87302,1.22619 0.20361,1.23934 0.72541,2.15584 0.32094,3.54244 0.008,0.02 0.015,0.0396 0.024,0.0595 -0.27904,1.22039 -1.13252,2.77766 -0.40612,3.53991 0.14735,-0.35853 0.2824,-0.44551 0.41469,-0.40806 -0.0972,0.57648 -0.15854,1.15046 -0.0182,1.72729 -0.06,2.38124 0.71903,4.78601 0.0209,7.13418 0.14601,1.27941 0.088,2.90648 1.11768,4.46334 0.20931,-0.32402 -0.26141,-4.04795 -0.1288,-4.40228 -0.87044,-5.44038 0.51261,-4.61027 0.27264,-7.20329 -0.0762,-1.65323 0.16439,0.11009 -0.0256,-1.53491 0.0358,-0.0631 0.0717,-0.12629 0.10943,-0.208 0.33307,-0.84154 1.06509,-1.4845 1.30907,-2.19339 -0.67902,-0.22846 -0.0546,-1.46994 -1.00086,-1.45768 -0.0423,-0.98007 1.23341,-1.30442 0.84312,-2.52886 -0.10207,-0.60431 -0.005,-1.4265 -0.08,-2.10218 -0.0192,-0.64914 -0.14685,-1.29953 -0.44026,-1.94485 -0.24035,-0.076 -0.43253,-0.0872 -0.58683,-0.0473 -0.003,7.9e-4 -0.007,0.003 -0.01,0.005 0.1136,-0.70342 0.31863,-1.48809 -0.23437,-1.25593 z",
0.606945, 19.5532,
"M 0.535288 -19.101  c -0.12312,-0.64981 -0.32031,-0.73191 -0.52216,0.6492 0.13498,2.03239 -0.19035,3.48484 -0.30626,5.15433 -0.22468,2.50128 -0.19194,5.27546 -0.28382,8.06552 0.14367,2.92828 -0.0892,5.74081 -0.015,8.6394 0.14902,2.95697 -0.015,5.88803 0.0117,8.84688 -0.15971,3.34873 0.87967,3.73489 0.85929,7.01865 -0.015,-0.18654 0.22435,0.61848 0.19054,0.10792 0.3206,-2.15319 -0.0271,-5.6381 0.0215,-8.31354 -0.0332,-2.93964 0.13264,-5.81334 0.0481,-8.77509 0.005,-0.0175 0.005,0.002 0.01,-0.0134 l -10e-4,-1.53853 c -0.0167,-0.17135 -0.0117,-0.39165 -4.9e-4,-0.67826 l -0.007,-8.52476 c -0.0251,-0.17877 -0.0216,-0.38728 -0.0573,-0.55526 -0.008,-0.46963 0.0383,-0.97844 0.056,-1.46172 z",
2.75846, 24.0847,
"M -0.53697 -24.0496  c -0.0684,0.29866 -0.0297,0.66596 0.13574,0.88282 0.0473,-0.31373 0.004,-0.65245 -0.12095,-0.91797 z m 0.36555,2.05274 c -0.1428,0.26785 -0.15703,0.69044 0.006,0.94726 -0.0394,0.35884 -0.44025,0.2613 -0.4462,0.64649 -0.13341,0.55376 -0.0618,1.26397 0.2782,1.6289 0.38647,-0.11637 0.44767,-0.80593 0.38665,-1.27862 -0.042,-0.65364 -0.0767,-1.31482 -0.20924,-1.94599 z m -0.0685,3.66015 c 0.0544,0.007 0.0464,-0.065 0,0 z m 0.13171,0.71485 c 0.0476,-0.0128 0.0369,-0.0718 0,0 z m 0.22847,2.58984 c -0.2894,0.77897 -0.28573,1.72489 -0.63166,2.4668 -0.47716,0.40261 -0.61905,1.44325 -0.27551,2.07422 0.26278,1.07486 -0.0427,2.34757 -0.66794,3.01562 -0.17172,0.2786 -0.0476,0.77973 0.20981,0.82791 0.38219,0.22562 0.82919,-0.30745 0.78616,-0.89087 0.15146,-1.09007 0.68275,-2.05871 0.59978,-3.21392 -0.003,-1.4276 0.18264,-2.88329 -0.0126,-4.29148 z m -0.97436,8.98632 c -0.0867,0.087 0.0488,0.007 0,0 z m -0.15187,0.90625 c -0.52835,0.40878 -0.4693,1.70374 0.0726,2.04297 0.51682,0.6066 0.26815,1.98415 -0.36821,2.19969 -0.32196,0.17741 -0.71983,0.44113 -0.75527,1.00981 -0.0704,0.4972 0.0592,1.03177 0.30905,1.38034 0.22493,0.75424 -0.1245,1.49918 -0.25804,2.21485 -0.161,0.74022 -0.28768,1.49458 -0.42066,2.24609 -0.42887,0.46517 -0.40727,1.40504 -0.13373,2.00508 0.29613,1.28961 0.23854,2.67245 0.42671,3.99298 0.18501,0.44913 0.6129,0.71487 0.60881,1.30273 -0.0974,1.02173 -0.7334,1.84226 -0.61687,2.92188 0.12358,1.36478 0.23492,2.73359 0.422,4.08398 -0.24493,1.20409 -0.0575,2.46873 0.1344,3.6582 0.13244,0.3425 0.51859,0.10392 0.49139,-0.2639 0.55319,-3.33295 0.74881,-6.77676 0.71199,-10.19928 -0.0341,-0.83387 -0.0795,-1.67038 -0.0382,-2.50557 0.0769,-0.26194 0.38773,-0.21698 0.42872,0.0606 0.15576,0.33179 0.55489,0.0325 0.49055,-0.3418 0.06,-0.507 0.13468,-1.1082 0.46904,-1.38867 0.49047,-0.0745 0.76751,0.85571 1.26869,0.67578 0.19449,-0.0457 0.51816,-0.0809 0.52146,-0.46093 -0.0813,-0.81436 -0.69657,-1.24081 -1.21561,-1.19272 -0.33451,-7.4e-4 -0.67174,-0.17488 -0.89675,-0.54267 -0.25083,-0.25457 -0.59572,0.13269 -0.82418,-0.2197 -0.16795,-0.3508 -0.22547,-0.80084 -0.2634,-1.22044 -0.006,-0.32789 -0.0804,-0.74575 0.14222,-0.96155 0.15244,-0.17908 0.31154,-0.47312 0.18536,-0.76146 -0.21856,-0.80287 -0.55515,-1.67299 -0.34299,-2.55633 0.20639,-1.08152 0.37362,-2.21388 0.29836,-3.33932 -0.28596,-1.22431 -0.25129,-2.59774 -0.71303,-3.73317 -0.026,-0.0821 -0.07,-0.13612 -0.1344,-0.10742 z m -0.89104,25.33205 c 0.0499,0.0937 0.0486,9.1e-4 0,0 z",
1.8822, 22.4677,
"M 1.25124 -21.5684  c -0.28418,-2.48971 -0.62962,0.79472 -0.69653,2.55163 -0.0134,0.33593 -0.0346,0.5327 -0.047,0.43714 -0.02,-0.15681 -0.0495,-0.15546 -0.3007,0.0153 -0.15327,0.10424 0.0424,3.35182 -0.0498,3.34843 -0.21601,-0.007 -0.41768,0.37741 -0.36668,0.70163 0.0286,0.18236 0.006,0.24818 -0.0802,0.2448 -0.11908,-0.004 -0.25364,0.43529 -0.18471,0.6037 0.71933,5.30949 -0.84428,1.36714 -0.18964,5.27515 l -0.17336,0.21594 c 0.0917,0.13776 0.14905,0.31886 0.14544,0.60895 -0.002,0.12917 0.013,0.27497 0.0381,0.42665 l -0.0218,0.006 c 0.0168,0.0338 0.0268,0.0634 0.039,0.094 0.0143,0.0729 0.0307,0.14633 0.05,0.22032 0.005,0.11878 -0.013,0.27501 -0.054,0.53201 -0.0541,0.33939 -0.13657,0.63375 -0.18319,0.65397 -0.11671,0.0506 -0.47174,-0.31269 -0.41853,-0.4284 0.0236,-0.0514 -0.008,-0.20294 -0.0696,-0.33617 -0.10666,-0.22943 -0.11316,-0.23096 -0.12398,-0.0315 -0.009,0.17127 -0.0213,0.16013 -0.0663,-0.059 -0.0305,-0.14804 -0.06,-0.19604 -0.0661,-0.10622 -0.005,0.0898 -0.0795,0.23039 -0.16355,0.31212 -0.18138,0.17617 -0.20093,0.37374 -0.046,0.46512 0.061,0.0363 0.11066,0.11279 0.10955,0.17048 -8.5e-4,0.0577 0.0954,0.32276 0.21449,0.58884 0.119,0.26609 0.21567,0.53536 0.21448,0.59889 -8.5e-4,0.0635 0.0919,0.18952 0.20682,0.27978 0.33391,0.262 0.4046,0.46417 0.37893,1.08499 -0.0255,0.61623 0.034,0.7011 0.50815,0.72217 0.35806,0.0162 0.65582,0.76507 0.53973,1.35777 l -0.0519,0.2636 -0.14759,-0.22163 c -0.0812,-0.122 -0.18117,-0.19419 -0.22184,-0.16043 -0.0407,0.034 -0.0928,-0.0143 -0.11598,-0.10711 -0.0668,-0.26845 -0.30372,-0.21195 -0.36208,0.0865 -0.0286,0.14608 -0.10177,0.26395 -0.16293,0.26142 -0.0612,-0.002 -0.11204,0.0439 -0.11323,0.10273 -8.4e-4,0.0588 -0.0557,0.10482 -0.1212,0.10228 -0.0995,-0.003 -0.11067,0.0715 -0.0685,0.45245 0.0278,0.2512 0.0474,0.70532 0.0445,1.00937 -0.005,0.46262 0.0255,0.59462 0.19024,0.8096 0.10822,0.14128 0.21652,0.38557 0.24057,0.54249 0.0239,0.15696 0.10666,0.38658 0.18349,0.51059 0.0826,0.13306 0.13642,0.3788 0.13286,0.59976 -0.005,0.29947 -0.0602,0.43199 -0.27492,0.66315 -0.29456,0.31713 -0.32923,0.45939 -0.16264,0.66709 0.0663,0.0827 0.10722,0.31839 0.10188,0.59277 -0.006,0.36907 0.035,0.51904 0.22,0.79079 0.17112,0.25109 0.20751,0.35709 0.18442,0.59584 l -0.67781,0.21289 c 0.0524,0.58516 -0.0123,0.94307 -0.20099,0.93593 -0.1353,-0.005 -0.20989,0.31179 -0.2378,1.00981 -0.0241,0.60164 -0.11859,0.82734 -0.43141,1.02948 -0.21966,0.14195 -0.22403,0.19455 -0.112,1.34903 0.0368,0.38004 0.68935,2.90966 0.7272,2.97271 0.0949,0.15747 0.0814,0.40225 -0.0366,0.66257 -0.11787,0.2598 -0.13449,0.64857 -0.0423,0.98759 0.0667,0.24519 0.0456,0.44952 -0.0548,0.532 -0.0228,0.0188 -0.0711,0.63594 0.0835,1.35624 0.0409,0.19057 0.50838,1.19206 0.55879,1.39985 0.35151,1.44898 0.46272,2.37383 0.46272,2.37383 0.79566,1.73874 0.72766,1.04386 0.63346,0.41525 2.3e-4,-0.0113 0.27534,-0.54652 0.19995,-0.70081 -0.15301,-0.31339 -0.17759,-0.66368 -0.083,-1.18261 0.0379,-0.20818 0.0431,-0.45443 0.016,-0.78147 -0.0498,-0.60249 -0.0427,-0.9744 0.0205,-1.07732 0.0273,-0.0442 0.0455,-0.20317 0.0407,-0.35337 -0.005,-0.15019 0.008,-0.40879 0.0283,-0.57478 0.0361,-0.29302 -0.0429,-0.6788 -0.13963,-0.68256 -0.0551,-0.002 -0.036,-0.53013 0.0208,-0.57728 0.0735,-0.061 0.0636,-0.34385 -0.0156,-0.44254 -0.0385,-0.048 -0.0692,-0.13755 -0.0682,-0.19891 8.5e-4,-0.0614 -0.0387,-0.22756 -0.0887,-0.36892 -0.0547,-0.15488 -0.0881,-0.38704 -0.0846,-0.58367 0.004,-0.17953 0.25499,-2.55145 0.21196,-2.61624 -0.10744,-0.16134 -0.0424,-0.84634 0.0767,-0.80785 0.0511,0.0165 0.0952,-0.062 0.097,-0.17486 0.005,-0.31876 -0.0724,-0.5529 -0.16081,-0.47955 -0.0451,0.0375 -0.0572,0.0165 -0.0276,-0.0476 0.006,-0.0155 0.0119,-0.0321 0.0145,-0.049 l 0.1838,-0.0577 c -0.0115,-0.0716 -0.0185,-0.14175 -0.0282,-0.21332 0.20698,0.0786 0.41395,-0.23158 0.42283,-0.71561 0.005,-0.26614 0.0327,-0.33495 0.1166,-0.28546 0.12218,0.0721 0.27032,-0.32611 0.27768,-0.7484 0.002,-0.0908 -0.0206,-0.17365 -0.0559,-0.23737 5.9e-4,-0.0146 0.002,-0.0265 0.004,-0.0424 0.005,-0.34543 -0.25055,-1.04474 -0.35226,-0.9604 -0.11645,0.0967 -0.30456,-0.0735 -0.301,-0.2719 0.002,-0.10386 -0.0615,-0.41419 -0.14115,-0.68982 -0.11849,-0.41032 -0.13308,-0.61892 -0.0811,-1.1497 0.0697,-0.71055 0.11872,-0.77415 0.58883,-0.76544 0.238,0.004 0.28272,-0.0309 0.28627,-0.22645 0.002,-0.12743 0.0471,-0.25037 0.0991,-0.27321 0.0739,-0.0324 0.0727,-0.0806 -0.005,-0.2212 -0.0565,-0.10309 -0.0965,-0.4463 -0.0923,-0.80434 0.005,-0.50518 0.0383,-0.66811 0.17397,-0.85156 0.2095,-0.28327 0.22253,-0.69385 0.0313,-0.98096 -0.13165,-0.19763 -0.12715,-0.26243 0.016,-0.22994 0.027,0.006 0.0267,-0.12053 -3e-4,-0.28152 -0.0271,-0.16095 -0.0753,-0.26956 -0.10744,-0.24131 -0.0321,0.0282 -0.10477,-0.11781 -0.16139,-0.32479 -0.08,-0.29191 -0.11861,-0.34197 -0.17307,-0.22339 -0.11294,0.2457 -0.17081,-0.12246 -0.0727,-0.46293 0.0682,-0.23598 0.0579,-0.30034 -0.0653,-0.39999 -0.10565,-0.0857 -0.14006,-0.21238 -0.12273,-0.45333 0.0132,-0.18386 -0.0139,-0.35393 -0.0604,-0.37813 -0.0466,-0.0241 -0.17138,-0.16327 -0.27646,-0.30993 -0.14602,-0.2042 -0.17876,-0.32905 -0.13838,-0.53551 0.0291,-0.14806 0.0832,-0.29471 0.12058,-0.32567 0.11733,-0.0974 0.0792,-0.33666 -0.10677,-0.66884 -0.0991,-0.17651 -0.15862,-0.34993 -0.13256,-0.38512 0.0261,-0.0352 0.0492,-0.35308 0.0528,-0.70643 0.004,-0.35346 0.0287,-0.6367 0.0573,-0.62993 0.14508,0.0352 0.31458,-0.0129 0.31636,-0.0896 4.8e-4,-0.0395 -0.042,-0.13954 -0.0951,-0.23868 l 0.11112,-0.0301 c -0.0111,-0.0958 -0.0399,-0.22407 -0.0386,-0.28983 0.002,-0.0834 0.0713,-0.1796 0.15527,-0.2142 0.0838,-0.0346 0.13399,-0.10798 0.11167,-0.16306 -0.0224,-0.055 -0.005,-0.18319 0.0394,-0.28502 0.0692,-0.15973 0.0802,-0.14567 0.0772,0.10142 -0.002,0.15765 0.0451,0.35827 0.10343,0.44589 0.0579,0.0877 0.0857,0.18709 0.0608,0.22076 -0.004,0.004 0.0112,0.0354 0.01,0.0433 l 0.1746,-0.0472 c -0.008,-0.0688 -0.0286,-0.0867 -0.0208,-0.19104 0.0154,-0.20696 6e-4,-0.4013 -0.0325,-0.43233 -0.0332,-0.0311 -0.041,-0.12276 -0.0174,-0.20371 0.0236,-0.0809 0.009,-0.24097 -0.031,-0.35584 -0.0524,-0.14895 -0.0466,-0.24767 0.0196,-0.34447 0.0735,-0.10616 0.0727,-0.11992 -0.002,-0.0634 -0.11355,0.0858 -0.20488,-0.23611 -0.26541,-0.93593 -0.0346,-0.40002 -0.0183,-0.50647 0.0887,-0.58665 0.10165,-0.076 0.1255,-0.20309 0.1071,-0.56917 -0.0173,-0.34244 0.009,-0.5044 0.0958,-0.59364 0.0992,-0.10173 0.1041,-0.161 0.0298,-0.34535 -0.0495,-0.12234 -0.0867,-0.38173 -0.0832,-0.57659 0.004,-0.19488 -0.0302,-0.40942 -0.0748,-0.47649 -0.0448,-0.0671 -0.0747,-0.45108 -0.0669,-0.85331 0.0111,-0.53051 -0.0185,-0.78692 -0.10588,-0.93418 -0.11429,-0.19274 -0.16109,-1.16523 -0.12426,-2.58004 0.005,-0.23051 -0.0111,-0.47175 -0.0377,-0.53638 -0.085,-0.20429 -0.0795,-0.43155 0.0115,-0.50709 0.0628,-0.0521 0.0982,-0.7782 0.12335,-2.54026 0.0111,-0.7367 0.0229,-1.05157 0.0353,-1.42947 z",
2.42744, 13.3932,
"M -1.21246 -13.3117  c -0.52065,0.12572 -0.51196,0.12241 -1.03612,0.24908 -0.001,0.0734 0.0127,0.13564 0.008,0.20774 l -0.0413,0.60203 0.49247,-0.0656 c 0.3575,-0.0477 0.5449,-0.0103 0.68368,0.13592 0.10525,0.11049 0.21523,0.17879 0.24443,0.15089 0.0553,-0.0525 0.18585,0.42322 0.50591,1.84847 0.40153,1.78589 0.46949,2.16266 0.39429,2.16266 -0.043,2.5e-4 0.005,0.0582 0.10801,0.12919 0.16623,0.1151 0.1658,0.14317 -0.006,0.25114 -0.1387,0.0873 -0.15256,0.14904 -0.0501,0.22015 0.21167,0.14601 0.22839,0.42645 0.0548,0.91828 -0.0852,0.24126 -0.19974,0.751 -0.25424,1.13224 -0.0545,0.38124 -0.13282,0.79864 -0.17312,0.9281 -0.0476,0.15154 -0.0232,0.20647 0.0687,0.15297 0.0797,-0.0464 0.13994,5.4e-4 0.13694,0.10697 -0.003,0.10433 -0.0625,0.22286 -0.13332,0.26406 -0.0707,0.0411 -0.13559,0.30272 -0.1416,0.58085 -0.009,0.34992 0.0322,0.47879 0.13488,0.41909 0.0929,-0.0543 0.11283,-0.0321 0.0532,0.0589 -0.22213,0.33938 -0.21073,0.73976 0.0279,0.96739 0.27232,0.25969 0.31048,0.5796 0.0935,0.78548 -0.1029,0.0973 -0.10442,0.20891 -0.007,0.40359 0.0743,0.14815 0.11655,0.465619 0.094,0.705379 -0.0314,0.33474 0.0303,0.49527 0.26613,0.69092 0.39591,0.32862 0.4506,0.97704 0.0971,1.15445 -0.19818,0.0995 -0.25884,0.08 -0.30024,-0.0971 -0.0413,-0.17754 -0.0559,-0.18235 -0.0718,-0.0238 -0.0147,0.14713 -0.11818,0.18202 -0.39791,0.13332 -0.33766,-0.0586 -0.37754,-0.0306 -0.38654,0.26458 -0.005,0.1817 -0.0655,0.33614 -0.1354,0.34314 -0.0697,0.007 -0.18114,0.0178 -0.24804,0.0248 -0.0668,0.007 -0.20301,0.11829 -0.30231,0.24701 -0.0994,0.12871 -0.25695,0.2064 -0.34985,0.1726 -0.19915,-0.0722 -0.63789,0.30833 -0.64389,0.55914 -0.008,0.28898 0.32739,0.49631 0.64751,0.40101 0.16115,-0.0479 0.37848,-0.0986 0.48317,-0.11265 0.6821,-0.0915 0.75775,-0.16898 0.69815,-0.71314 -0.0515,-0.47073 -0.0192,-0.55211 0.31006,-0.78135 0.25416,-0.17695 0.44432,-0.22349 0.61702,-0.15089 0.24881,0.10484 0.24866,0.10527 -0.004,0.12247 -0.1394,0.009 -0.29578,0.009 -0.34778,0 -0.0519,-0.01 -0.0957,0.043 -0.0977,0.11731 -0.002,0.0774 0.10491,0.10318 0.25115,0.061 0.26076,-0.0754 0.35205,0.12076 0.11937,0.25631 -0.071,0.0413 -0.0973,0.21478 -0.0594,0.38551 0.0649,0.2906 -0.0353,0.37589 -0.36948,0.31419 -0.0517,-0.01 -0.0962,0.0455 -0.0982,0.12247 -0.002,0.0766 0.0821,0.10935 0.18655,0.0724 0.10479,-0.0372 0.18115,-0.0268 0.17415,0.0227 -0.009,0.0495 -0.004,0.14704 0.0109,0.21704 0.0872,0.41425 -0.088,1.66625 -0.25579,1.82779 -0.17035,0.164 -0.21288,0.15606 -0.37621,-0.0729 -0.1012,-0.14177 -0.23055,-0.34474 -0.28835,-0.45062 -0.0858,-0.15729 -0.1361,-0.16357 -0.2744,-0.0336 -0.16719,0.15723 -0.063,0.30632 0.31419,0.4501 0.062,0.0236 -0.01,0.26437 -0.15968,0.53537 -0.14984,0.27102 -0.33181,0.72807 -0.40411,1.01544 -0.12606,0.50143 -0.11796,0.52568 0.1974,0.59738 0.23372,0.0532 0.38925,0.22256 0.53899,0.58757 0.27181,0.66258 0.33329,2.3657 0.093,2.57865 -0.14193,0.12572 -0.13896,0.17735 0.0181,0.29869 0.20771,0.16041 0.23365,1.24108 0.0444,1.85518 -0.10282,0.33391 -0.0903,0.34552 0.38603,0.35812 0.0498,0.001 0.0875,0.0324 0.14056,0.0393 l 2.02778,-1.91874 c 0.0391,-0.84014 0.0901,-1.31251 0.20567,-1.42214 0.1408,-0.13348 0.16273,-0.22373 0.0687,-0.28474 -0.0888,-0.0584 -0.11201,-0.52874 -0.0651,-1.34978 0.0634,-1.11341 0.0451,-1.28034 -0.15968,-1.43196 -0.12777,-0.0943 -0.22728,-0.30706 -0.22428,-0.47284 0.005,-0.16601 -0.0515,-0.3404 -0.12454,-0.3886 -0.0731,-0.048 -0.0993,-0.16882 -0.0579,-0.26872 0.0422,-0.10162 0.0192,-0.14879 -0.0517,-0.10749 -0.22711,0.13224 -0.12436,-0.27378 0.14314,-0.56637 0.2846,-0.31158 0.33629,-0.4973 0.13953,-0.50229 -0.0693,-0.002 -0.12444,-0.0905 -0.12144,-0.19689 0.003,-0.10651 -0.074,-0.22624 -0.17053,-0.26614 -0.27152,-0.11237 -0.32136,-0.54803 -0.078,-0.68161 0.1171,-0.0642 0.15641,-0.12105 0.0868,-0.12506 -0.0694,-0.004 -0.13414,-0.24406 -0.14314,-0.5333 -0.009,-0.28926 -0.11377,-0.63887 -0.23255,-0.77721 -0.18088,-0.21019 -0.21428,-0.54858 -0.20928,-2.06292 0.006,-1.693699 0.0211,-1.808719 0.23564,-1.769919 0.21803,0.0394 0.41737,-0.1166 0.42736,-0.33435 0.002,-0.055 -0.0668,-0.0927 -0.15399,-0.0837 -0.56087,0.0579 -0.60799,-0.005 -0.56689,-0.7519 0.0221,-0.39949 -0.0155,-0.87281 -0.0832,-1.05213 -0.14126,-0.3744 0.0208,-0.75214 0.31936,-0.74414 0.10486,0.003 0.15977,-0.048 0.12247,-0.11265 -0.0376,-0.0646 0.004,-0.13992 0.092,-0.16692 0.12652,-0.039 0.12829,-0.0886 0.008,-0.24236 -0.0831,-0.10642 -0.15413,-0.46452 -0.15813,-0.79582 -0.004,-0.33891 -0.0742,-0.62959 -0.1602,-0.66559 -0.0981,-0.0407 -0.14427,-0.3485 -0.13177,-0.85886 0.0114,-0.43746 -0.0148,-0.97902 -0.0579,-1.20355 -0.0579,-0.30105 -0.0202,-0.47 0.14418,-0.64233 0.17836,-0.18698 0.20786,-0.35499 0.14676,-0.8413 -0.0419,-0.33433 -0.0924,-0.83108 -0.11214,-1.10381 -0.0199,-0.27268 -0.0595,-0.52067 -0.0879,-0.55087 -0.0287,-0.0302 0.0105,-0.22077 0.0873,-0.42374 0.17376,-0.45944 -0.0535,-1.17751 -0.34055,-1.07591 -0.10107,0.0357 -0.20737,0.0633 -0.23667,0.0631 -0.0292,-7.8e-4 -0.0549,0.0807 -0.0579,0.18087 -0.003,0.10015 -0.0715,0.22122 -0.15296,0.26871 -0.0924,0.0533 -0.11257,0.0315 -0.0543,-0.0574 0.052,-0.0791 0.0375,-0.14476 -0.032,-0.14676 -0.0696,-0.002 -0.16379,0.0535 -0.20929,0.12299 -0.0455,0.0695 -0.14895,0.0835 -0.22945,0.0305 -0.0807,-0.0528 -0.34584,-0.10465 -0.58911,-0.11265 -0.2954,-0.011 -0.39714,-0.0935 -0.48317,-0.27751 z",
    )
specks=(
0.19582, 0.378634,
"M -0.161959 -0.371071  c -0.1979,0.116362 0.54642,1.511327 0.31137,0.17337 -0.16941,-0.158525 -0.26567,-0.200234 -0.31137,-0.17337 z",
0.0857523, 0.142767,
"M -0.0570877 -0.130051  c -0.079,0.08213 0.0173,0.518112 0.14284,0.08157 -0.0697,-0.09649 -0.11628,-0.108996 -0.14284,-0.08157 z",
0.25372, 0.210386,
"M 0.0480998 -0.205499  c -0.42327,0.07051 -0.4419,0.885499 0.20562,0.02193 -0.0763,-0.02736 -0.14491,-0.032 -0.20562,-0.02193 z",
0.140526, 0.148695,
"M -0.13122 -0.139672  c -0.0745,0.07221 0.32456,0.53906 0.26582,0.107512 -0.16476,-0.111376 -0.2411,-0.131561 -0.26582,-0.107512 z",
0.249924, 0.342306,
"M -0.114966 -0.341551  c -0.34386,0.03623 0.04,1.35907 0.36489,0.226387 -0.16717,-0.174662 -0.28557,-0.234744 -0.36489,-0.226387 z",
0.825503, 0.732431,
"M -0.326263 0.732431  c 0.0466,-0.194767 0.0978,-0.389834 0.33522,-0.575911 0.67157,-0.184934 1.35761,-1.225527 0.16941,-0.77828 -0.39118,0.182831 -0.72402,0.18703 -1.00387,0.107874 z",
0.16647, 0.33039,
"M -0.07457 -0.329965  c -0.15388,0.01622 -0.13629,0.783005 0.22124,0.643371 l 0.0198,-0.08888 -0.0371,-0.25956 c -0.0823,-0.222794 -0.15267,-0.300475 -0.20355,-0.295097 z",
0.262126, 0.290338,
"M -0.233294 -0.271662  c -0.17079,0.144103 0.47197,1.053092 0.49542,0.204111 -0.29067,-0.215468 -0.43843,-0.252153 -0.49542,-0.204111 z",
0.240004, 0.127086,
"M -0.124553 -0.0975163  c -0.46137,0.111203 0.63127,0.460222 0.30142,-0.02957 -0.13801,0.0028 -0.23555,0.01328 -0.30142,0.02957 z",
0.221255, 0.286652,
"M -0.173286 -0.268838  c -0.24785,0.147346 0.55483,1.130889 0.36527,0.06706 -0.19413,-0.09043 -0.30799,-0.101042 -0.36527,-0.06706 z",
0.187681, 0.221884,
"M -0.00844687 -0.221884  c -0.63303,0.150003 0.66682,0.918773 0,0 z",
0.186594, 0.274126,
"M -0.108183 -0.271572  c -0.33666,0.06033 0.53913,1.16614 0.22586,0.04727 -0.10437,-0.04356 -0.17776,-0.05583 -0.22586,-0.04727 z",
0.143206, 0.14809,
"M -0.134411 -0.138893  c -0.0733,0.07232 0.33765,0.536966 0.27008,0.106063 -0.16769,-0.110755 -0.24568,-0.130215 -0.27008,-0.106063 z",
0.205727, 0.193789,
"M -0.128883 -0.184279  c -0.27335,0.08926 0.26505,0.768387 0.33461,0.05638 -0.16492,-0.06656 -0.27156,-0.07703 -0.33461,-0.05638 z",
0.305162, 0.456411,
"M 0.128243 -0.44135  c -0.29988,0.132509 -0.79724,1.080982 -0.0121,0.866257 l 0.0851,-0.147363 0.073,-0.208094 c 0.0833,-0.462398 -0.0138,-0.57096 -0.1475,-0.510616 z",
0.218898, 0.391658,
"M -0.131822 -0.379923  c -0.19943,0.127506 -0.0454,1.240958 0.35072,0.543718 l -0.0293,-0.2389 c -0.14923,-0.280285 -0.25533,-0.347357 -0.32181,-0.304858 z",
0.314969, 0.187616,
"M -0.110991 -0.158343  c -0.32093,0.126299 -0.30492,0.600436 0.35146,0.16703 l 0.0607,-0.06813 0.0138,-0.06262 c -0.17113,-0.08176 -0.31579,-0.07863 -0.42277,-0.03643 z",
0.144384, 0.265972,
"M -0.0991 -0.265684  c -0.18575,-0.02027 0.26116,1.038075 0.24294,0.222501 -0.1237,-0.15592 -0.19996,-0.217867 -0.24294,-0.222501 z",
0.249929, 0.219643,
"M -0.0945313 -0.210573  c -0.25801,0.08907 -0.22795,0.792505 0.34446,0.186141 -0.1406,-0.177396 -0.25858,-0.215823 -0.34446,-0.186141 z",
0.311515, 0.537401,
"M -0.140345 -0.534025  c -0.2896,0.082 -0.25686,1.549308 0.45186,0.911044 l -0.009,-0.160716 -0.0654,-0.241334 c -0.1551,-0.409456 -0.28705,-0.536186 -0.38358,-0.508808 z",
0.177607, 0.337109,
"M -0.0901732 -0.335988  c -0.24366,0.04365 0.0797,1.339997 0.26778,0.215464 -0.12508,-0.168462 -0.21151,-0.225565 -0.26778,-0.215464 z",
0.174549, 0.307061,
"M -0.0785809 -0.300444  c -0.14612,0.07522 -0.16717,0.806477 0.23933,0.555164 l 0.0138,-0.09759 v 0 c -0.0894,-0.391069 -0.18135,-0.492186 -0.24739,-0.457993 z",
0.103904, 0.181685,
"M -0.0611958 -0.180411  c -0.10023,0.0313 -0.0155,0.631126 0.1651,0.216629 -0.0781,-0.174417 -0.13129,-0.227114 -0.1651,-0.216629 z",
0.300084, 0.415141,
"M -0.145146 -0.407237  c -0.287,0.106735 -0.19552,1.250324 0.44523,0.648168 l -0.0387,-0.306045 c -0.17379,-0.297695 -0.31118,-0.377743 -0.40684,-0.342148 z",
0.871496, 0.707663,
"M 0.738024 -0.703243  c 0.1632,-0.03202 0.20538,0.105011 -0.038,0.573046 -0.007,0.328165 0.23952,1.188452 -0.39899,0.681635 -0.51284,-0.241795 -1.94836,0.149951 -0.63082,-0.4331 0.18654,-0.245067 0.7955,-0.768223 1.06753,-0.82157 z",
0.197296, 0.279834,
"M -0.119824 -0.279832  c -0.264,-0.0013 0.22383,1.101454 0.31712,0.214591 -0.15509,-0.155747 -0.25613,-0.214387 -0.31712,-0.214591 z",
0.362209, 0.376951,
"M -0.0190514 -0.364089  c 0.31129,-0.08445 0.64101,0.259813 0.068,0.656187 -0.0781,-0.02833 -0.31369,0.170722 -0.40954,0.03985 -0.0204,-0.440504 0.15544,-0.645356 0.34191,-0.696058 z",
0.322117, 0.43818,
"M 0.123244 -0.437076  c 0.21552,-0.01825 0.34427,0.190347 -0.0901,0.577559 0.0262,0.147674 -0.23676,0.452306 -0.33939,0.202753 -0.0887,-0.518231 0.21366,-0.762308 0.42909,-0.780276 z",
0.245771, 0.389516,
"M -0.127069 -0.385828  c -0.24435,0.07803 -0.10731,1.36138 0.37284,0.449505 -0.16786,-0.367879 -0.29138,-0.475521 -0.37284,-0.449505 z",
0.230177, 0.205962,
"M 0.0753834 -0.190211  c 0.31684,-0.138599 0.11369,0.694489 -0.30556,0.278851 0.13249,-0.162699 0.23246,-0.246844 0.30556,-0.278851 z",
0.195686, 0.378332,
"M -0.161823 -0.370693  c -0.19787,0.116276 0.54614,1.510428 0.31108,0.172533 -0.16941,-0.158542 -0.26538,-0.199394 -0.31108,-0.172533 z",
0.156073, 0.178711,
"M -0.123092 -0.173272  c 0.12455,-0.06557 0.54867,0.486539 0.0207,0.320803 -0.073,-0.208356 -0.063,-0.298951 -0.0207,-0.320803 z",
0.292868, 0.408748,
"M -0.0994831 -0.377205  c 0.2675,-0.127455 0.66049,0.135735 0.12818,0.70988 -0.0423,0.149106 -0.21965,0.04812 -0.24627,-0.04082 -0.15751,-0.374986 -0.0435,-0.592568 0.11835,-0.669165 z",
0.23406, 0.207114,
"M 0.15688 -0.205786  c 0.21509,-0.03505 -0.053,0.63824 -0.39094,0.33352 0.19441,-0.231554 0.31926,-0.321849 0.39094,-0.33352 z",
0.201117, 0.166598,
"M -0.0170272 -0.150178  c 0.25106,-0.12761 0.39144,0.539989 -0.18409,0.23713 0.0628,-0.135856 0.1261,-0.207647 0.18409,-0.23713 z",
0.114132, 0.18984,
"M -0.0125222 -0.189323  c 0.12162,-0.02091 0.24963,0.600837 -0.10161,0.293838 0.0204,-0.208237 0.0609,-0.286747 0.10161,-0.293838 z",
0.0915151, 0.173045,
"M -0.00111513 -0.171143  c 0.1168,-0.04285 0.15872,0.6575 -0.0904,0.172206 0.0312,-0.113136 0.0631,-0.162096 0.0904,-0.172206 z",
0.197077, 0.0630351,
"M -0.0264672 -0.0279184  c 0.34231,-0.128835 0.29613,0.142843 -0.17061,0.08191 0.0647,-0.03707 0.12128,-0.06352 0.17061,-0.08191 z",
0.32193, 0.326561,
"M -0.0297401 -0.311587  c 0.33824,-0.13351 0.67663,0.670587 -0.23729,0.637132 l -0.0549,-0.125695 c 0.0671,-0.318517 0.17898,-0.466864 0.29172,-0.511371 z",
0.15739, 0.161874,
"M -0.112614 -0.152979  c 0.14163,-0.07868 0.53585,0.392922 -0.0121,0.303528 -0.0537,-0.189612 -0.0352,-0.27729 0.0121,-0.303528 z",
0.492892, 0.413638,
"M -0.474932 -0.402985  c 0.083,-0.07235 0.43945,0.238304 0.76375,0.494301 0.64017,0.214078 -0.41841,0.52196 -0.55019,0.144016 -0.21378,-0.413279 -0.26311,-0.594893 -0.21413,-0.638157 z",
0.280378, 0.178761,
"M -0.23433 -0.147759  c 0.20564,-0.136598 1.00237,0.215878 0.0761,0.32652 -0.14871,-0.18142 -0.14543,-0.28092 -0.0761,-0.32652 z",
0.162268, 0.210341,
"M -0.102042 -0.210249  c -0.15268,-0.0093 0.003,0.684278 0.26431,0.309297 -0.12766,-0.225447 -0.21324,-0.306126 -0.26431,-0.309297 z",
0.173218, 0.312533,
"M 0.0960123 -0.312183  c 0.1701,0.02182 0.0498,1.05689 -0.26923,0.415217 0.12369,-0.317554 0.21268,-0.422492 0.26923,-0.415217 z",
0.225385, 0.163164,
"M -0.170322 -0.147418  c 0.19286,-0.101284 0.78196,0.319518 0.005,0.310437 -0.0894,-0.185115 -0.0687,-0.276769 -0.005,-0.310437 z",
0.256583, 0.29513,
"M 0.153207 -0.27621  c -0.24739,0.13332 -0.78467,0.91045 0.003,0.399077 0.15923,-0.386903 0.10662,-0.459705 -0.003,-0.399077 z",
0.142144, 0.160053,
"M -0.0769161 -0.159666  c -0.19378,0.01932 0.0939,0.633237 0.21906,0.107305 -0.10454,-0.08281 -0.17446,-0.111446 -0.21906,-0.107305 z",
0.257763, 0.339205,
"M -0.152337 -0.339123  c -0.34676,0.01225 0.26021,1.339727 0.4101,0.24746 -0.19896,-0.18325 -0.33009,-0.250231 -0.4101,-0.24746 z",
0.358938, 0.665171,
"M -0.0631218 -0.64564  c 0.1073,-0.03088 0.24635,-0.02708 0.42206,0.02536 -0.16562,0.294692 0.0819,1.345849 -0.37241,1.282725 -0.38085,-0.375619 -0.51502,-1.172797 -0.0474,-1.308222 z",
0.514964, 0.316052,
"M 0.0658635 -0.289668  c -0.47326,0.143016 -1.01805,0.807806 -0.02,0.544771 0.16837,-0.03778 0.41057,-0.08108 0.4691,-0.279942 -0.034,-0.28679 -0.23428,-0.329746 -0.44939,-0.264863 z",
0.28151, 0.425671,
"M -0.238974 -0.406955  c 0.18858,-0.163182 0.91161,0.793301 0.2413,0.832626 l -0.18057,-0.220986 c -0.12698,-0.378745 -0.12387,-0.557216 -0.0604,-0.611607 z",
0.146439, 0.286498,
"M -0.0593209 -0.282103  c -0.20198,0.07761 -0.019,1.143862 0.20576,0.141757 -0.0918,-0.124884 -0.15889,-0.159664 -0.20576,-0.141757 z",
0.224774, 0.345217,
"M -0.143886 -0.34214  c -0.21102,0.06745 0.0186,1.204903 0.36866,0.40122 -0.1801,-0.327159 -0.29833,-0.423613 -0.36866,-0.40122 z",
0.364393, 0.448279,
"M 0.118202 -0.416096  c -0.35164,0.168272 -0.86793,0.937323 -0.0295,0.85876 l 0.1563,-0.199134 c 0.2359,-0.645352 0.0842,-0.760655 -0.1268,-0.659676 z",
0.178445, 0.406469,
"M 0.0463255 -0.4059  c 0.18761,0.03174 0.22558,1.377763 -0.22477,0.5358 0.0833,-0.411175 0.16234,-0.546375 0.22477,-0.5358 z",
0.304022, 0.447803,
"M -0.00200764 -0.442434  c -0.3188,0.09523 -0.5662,1.415475 0.23893,0.655491 l 0.0671,-0.219199 c -0.0863,-0.363197 -0.19964,-0.467949 -0.30603,-0.436344 z",
0.738368, 0.337605,
"M -0.320121 -0.185236  c 0.66788,-0.249967 1.7692,-0.202284 0.41521,0.216789 -0.27722,0.04994 -0.4964,0.346511 -0.78619,0.3014 -0.14647,-0.229779 0.0671,-0.404455 0.37101,-0.518134 z",
0.375081, 0.882719,
"M -0.0815792 -0.881282  c -0.0944,0.0069 -0.19344,0.08954 -0.2912,0.259464 0.43267,0.359195 -0.52183,2.120841 0.4626,1.279309 0.52762,-0.392181 0.23605,-1.588716 -0.17148,-1.538687 z",
0.352805, 1.09079,
"M -0.163519 -1.08216  c 0.0236,-0.0085 0.05,-0.01121 0.0787,-0.0059 0.20435,0.405278 0.95305,2.696139 -0.15267,2.071683 -0.0376,-0.500541 -0.27879,-1.943774 0.0742,-2.06522 z",
0.41009, 1.14354,
"M 0.143813 -1.1422  c 0.118,-0.0138 0.21959,0.07711 0.26221,0.355831 0.0248,0.57158 -0.0509,3.125281 -0.59151,1.256692 -0.51863,-0.436258 -0.024,-1.551867 0.32926,-1.612506 z",
0.140552, 0.154723,
"M 0.0177994 -0.154723  c -0.5212,0.106098 0.44996,0.639827 0,0 z",
0.533274, 0.437204,
"M -0.502236 -0.433749  c -0.0649,0.02284 -0.0361,0.171895 0.16389,0.539065 0.1023,0.205429 0.86901,0.563972 0.87162,0.118795 -0.2874,-0.156058 -0.89227,-0.709351 -1.03553,-0.657957 z",
0.200361, 0.314278,
"M -0.124469 -0.30535  c -0.17097,0.09682 -0.049,0.951731 0.32483,0.477136 l -0.0547,-0.244774 c -0.12404,-0.21357 -0.21339,-0.264707 -0.27044,-0.232426 z",
0.274543, 0.539917,
"M -0.0453373 -0.535301  c -0.54472,0.105752 0.0193,1.964059 0.27489,0.540906 0.031,-0.148363 0.0654,-0.312167 0.03,-0.464016 -0.12663,-0.07108 -0.22779,-0.09212 -0.30563,-0.07677 z",
0.171092, 0.292515,
"M -0.137189 -0.285982  c 0.12732,-0.0772 0.5456,0.554711 0.12559,0.578497 l -0.11369,-0.216331 c -0.0635,-0.232403 -0.0552,-0.336408 -0.0103,-0.362199 z",
0.270716, 0.41871,
"M -0.224043 -0.39102  c -0.27325,0.221011 0.75659,1.649373 0.43008,0.09126 -0.23446,-0.129041 -0.36699,-0.142153 -0.43008,-0.09126 z",
0.173802, 0.188374,
"M 0.0173279 -0.186489  c 0.18489,-0.03985 0.2935,0.568112 -0.19113,0.311443 0.0597,-0.212936 0.13008,-0.298228 0.19113,-0.311443 z",
0.23479, 0.730233,
"M -0.0770773 -0.699585  c -0.11437,0.07401 -0.20017,0.281408 -0.13542,0.641983 0.0405,0.983203 0.0543,1.109157 0.35222,0.01811 0.23794,-0.534592 -0.0224,-0.783153 -0.21742,-0.659545 z",
0.439088, 0.707771,
"M 0.246767 -0.701228  c -0.10299,0.02191 -0.25293,0.102474 -0.45902,0.264588 -0.16872,0.17726 -0.43644,1.694364 0.0511,0.931154 0.43805,-0.288567 0.85607,-1.292506 0.40786,-1.195742 z",
0.139899, 0.344117,
"M -0.0663814 -0.336297  c -0.14129,0.105407 -0.0823,1.226019 0.20628,0.344756 -0.0904,-0.302541 -0.15923,-0.379811 -0.20628,-0.344756 z",
0.102613, 0.185851,
"M 0.0541974 -0.185523  c 0.10196,0.01725 0.0397,0.631684 -0.15681,0.242402 0.0711,-0.187404 0.12283,-0.247802 0.15681,-0.242402 z",
0.302008, 0.560567,
"M -0.0270707 -0.55462  c -0.18178,0.03761 -0.35011,0.264001 -0.23973,0.713802 0.0228,0.671208 0.15302,0.377879 0.44061,0.0069 0.28805,-0.504213 0.0331,-0.769335 -0.20105,-0.721031 z",
0.085711, 0.142485,
"M -0.056959 -0.129834  c -0.0792,0.08246 0.0173,0.517289 0.14267,0.08091 -0.0697,-0.09661 -0.11541,-0.107822 -0.14267,-0.08091 z",
0.180055, 0.427331,
"M -0.0363549 -0.425781  c 0.19187,0.05348 0.43026,1.467351 -0.1437,0.536886 0.0224,-0.422025 0.08,-0.554673 0.1437,-0.536886 z",
0.253994, 0.158651,
"M -0.157396 -0.151199  c -0.33829,0.07177 0.31285,0.629548 0.41139,0.04756 -0.20239,-0.05529 -0.33338,-0.06404 -0.41139,-0.04756 z",
0.142239, 0.160356,
"M -0.077041 -0.160032  c -0.19381,0.01932 0.0942,0.634082 0.21928,0.108167 -0.10472,-0.08281 -0.17481,-0.111962 -0.21928,-0.108167 z",
0.204532, 0.23177,
"M -0.125112 -0.231668  c 0.18825,-0.0086 0.62843,0.52902 0.0155,0.456753 -0.14112,-0.336734 -0.10644,-0.453147 -0.0155,-0.456753 z",
0.302086, 0.355289,
"M -0.150514 -0.321921  c -0.41485,0.206156 0.12059,1.269337 0.4281,0.224788 l 0.0245,-0.194374 c -0.21141,-0.07953 -0.35705,-0.07849 -0.45276,-0.03088 z",
0.319431, 0.249425,
"M -0.132339 -0.243618  c -0.4405,0.08143 -0.0266,0.995486 0.45177,0.108513 -0.20224,-0.10282 -0.34995,-0.127317 -0.45177,-0.108513 z",
0.318818, 0.231988,
"M 0.0318226 -0.193927  c -0.30533,0.13422 -0.649,0.55842 0.0989,0.38482 l 0.16407,-0.11973 c 0.0881,-0.31984 -0.0811,-0.34462 -0.26387,-0.2645 z",
0.196202, 0.386572,
"M -0.0913577 -0.381538  c -0.40284,0.05141 0.50388,1.52278 0.23778,0.253201 l -0.14026,-0.243385 c -0.0383,-0.01208 -0.0716,-0.01725 -0.0973,-0.0138 z",
0.252923, 0.25611,
"M -0.0991266 -0.255205  c -0.34893,0.0343 -0.0492,1.01793 0.35205,0.16303 -0.15613,-0.12749 -0.27152,-0.17097 -0.35205,-0.16303 z",
0.227743, 0.364507,
"M -0.129097 -0.362268  c -0.18019,0.05106 -0.14232,0.970885 0.35684,0.664393 -0.14854,-0.531815 -0.27503,-0.687804 -0.35684,-0.664393 z",
0.164352, 0.175428,
"M -0.045238 -0.173643  c -0.22431,0.03864 -0.11921,0.700104 0.20959,0.09557 -0.0887,-0.08074 -0.15768,-0.104545 -0.20959,-0.09557 z",
0.39131, 0.330262,
"M -0.0448403 -0.29959  c -0.36646,0.172171 -0.65831,0.979631 0.30262,0.455235 l 0.13353,-0.110583 c -0.087,-0.369322 -0.26893,-0.422388 -0.43553,-0.344152 z",
0.08851, 0.144333,
"M -0.06158 -0.131595  c -0.0797,0.0825 0.0316,0.52359 0.15009,0.0835 -0.0751,-0.0988 -0.12318,-0.11093 -0.15009,-0.0835 z",
0.142269, 0.160385,
"M -0.0770807 -0.160034  c -0.19379,0.01932 0.0942,0.634151 0.21935,0.108167 -0.10471,-0.08281 -0.17483,-0.112135 -0.21935,-0.108167 z",
0.141781, 0.160158,
"M -0.076419 -0.159773  c -0.19379,0.01932 0.093,0.633582 0.2182,0.107477 -0.10472,-0.08281 -0.17386,-0.111617 -0.2182,-0.107477 z",
0.209206, 0.30405,
"M -0.116284 -0.296837  c -0.20484,0.0951 -0.0599,1.08425 0.32549,0.30178 -0.15043,-0.26622 -0.25717,-0.33337 -0.32549,-0.30178 z",
0.201742, 0.3747,
"M -0.146133 -0.374084  c -0.24908,0.0343 0.42422,1.48646 0.34064,0.25287 -0.17607,-0.19356 -0.28321,-0.26077 -0.34064,-0.25287 z",
0.322667, 0.259474,
"M -0.114613 -0.226479  c -0.33039,0.161993 -0.30744,0.843431 0.36278,0.242972 l 0.0745,-0.165098 c -0.17776,-0.127834 -0.3269,-0.132147 -0.4369,-0.07815 z",
0.182682, 0.131912,
"M -0.0706049 -0.115145  c -0.38616,0.0825 0.35908,0.51914 0.2403,-0.0121 -0.10644,-0.0121 -0.18509,0.002 -0.2403,0.0121 z",
0.226747, 0.401068,
"M -0.210404 -0.386184  c -0.12456,0.15492 0.50616,1.44357 0.43089,0.3531 -0.26357,-0.33026 -0.38924,-0.40472 -0.43089,-0.3531 z",
0.231976, 0.335615,
"M -0.105621 -0.325515  c 0.23153,-0.10299 0.64145,0.61459 -0.029,0.66113 l -0.0958,-0.22943 c -0.0121,-0.27202 0.048,-0.39727 0.1249,-0.4317 z",
0.335965, 0.335139,
"M -0.0942353 -0.30707  c -0.34582,0.173534 -0.39751,1.086591 0.3636,0.373859 l 0.0511,-0.10489 0.0155,-0.10903 c -0.16113,-0.19389 -0.30903,-0.218301 -0.42429,-0.160612 z",
0.0697059, 0.240406,
"M -0.0569386 -0.237373  c 0.0566,0.055 0.24727,0.84545 0.0121,0.26736 -0.0321,-0.22226 -0.0297,-0.28589 -0.0121,-0.26736 z",
0.242776, 0.518825,
"M -0.137028 -0.512424  c -0.43441,0.07384 0.62996,1.783124 0.32347,0.636705 -0.0242,-0.208434 -0.0183,-0.497208 -0.21632,-0.628062 -0.0433,-0.0138 -0.0788,-0.01725 -0.10696,-0.0138 z",
0.0911062, 0.261377,
"M -0.00830619 -0.25616  c 0.11369,-0.08798 0.1788,0.975612 -0.0828,0.280459 0.0274,-0.178605 0.0571,-0.260274 0.0828,-0.280459 z",
0.175081, 0.18501,
"M -0.0388005 -0.179991  c 0.20821,-0.07297 0.40631,0.681507 -0.13628,0.207658 0.0405,-0.129905 0.0885,-0.190855 0.13628,-0.207658 z",
0.118897, 0.176509,
"M -0.0742332 -0.173289  c -0.15768,0.0526 0.1494,0.70495 0.19313,0.0823 -0.0947,-0.0745 -0.15647,-0.0942 -0.19313,-0.0823 z",
    )

inxhead = """<?xml version="1.0" encoding="UTF-8"?>
<inkscape-extension xmlns="http://www.inkscape.org/namespace/inkscape/extension">
  <_name>ChipScratches</_name>
  <id>ca.sfu.AT.kurn.ChipScratches</id>
  <dependency type="executable" location="extensions">chipScratches.py</dependency>
  <dependency type="executable" location="extensions">inkex.py</dependency>
"""

inxfoot = """  <effect>
    <object-type>all</object-type>
    <effects-menu>
       <submenu _name="Generate">
         <submenu _name="Noise"/>
       </submenu>
    </effects-menu>
  </effect>
  <script>
    <command reldir="extensions" interpreter="python">chipScratches.py</command>
  </script>
</inkscape-extension>
"""

inxtype = {
    "i" : '<param  name="%s" type="int" %s _gui-text="%s">%s</param>\n',
    "f" : '<param  name="%s" type="float" %s _gui-text="%s">%s</param>\n',
    "s" : '<param  name="%s" type="string" %s _gui-text="%s">%s</param>\n',
    "b" : '<param  name="%s" type="boolean" %s _gui-text="%s">%s</param>\n',
    "n" : '<!-- Dummy line for index. %s %s %s %s  -->\n',
    }

inxbk = {
    "tabs" : '<param type="notebook" name="Nmain">\n<page name="%(s)s" _gui-text="%(s)s">\n',
    "tab"  : '</page>\n<page name="%(s)s" _gui-text="%(s)s">\n',
    "tabe"  : '</page>\n</param>\n',
    "lit" : '<_param name="%(s)s"   type="description" >%(s)s</_param>\n',

    }

def mt( s ) :
    return s.isspace() or len(s) == 0

def addparms( p ) :
    s = ''
    for parm in p.split(',') :
        if mt(parm) :
            continue
        s += parm + ' '
    return s

if inxGen :
    print inxhead

    for parm in parms.split('\n') :
        if mt(parm) :
            continue
        w = parm.split('#')
        if len(w) == 3 :
            print inxbk[w[1]] % {"s" : w[2]}, 
            continue
        if len(w) > 1 :
            continue
        w = parm.split(None, 4)
        print inxtype[w[0]] % (w[1], addparms(w[3]), w[4], w[2]),
    
    print inxfoot
else :
# Create effect instance and apply it.
    effect = noiseEffect()
    effect.affect()
    sys.exit( 0 )

